@@ -1,3 +1,3 @@
-(define-package "agda2-mode" "2.6.2.2"
+(define-package "agda2-mode" "2.6.3"
   "interactive development for Agda, a dependently typed functional programming language"
   '((emacs "24.3") (annotation "1.0") (eri "1.0")))
