@@ -28,7 +28,7 @@
 
 ;;; Code:
 
-(defvar agda2-version "2.6.2.2"
+(defvar agda2-version "2.6.3"
   "The version of the Agda mode.
 Note that the same version of the Agda executable must be used.")
 
@@ -108,10 +108,6 @@ argument, and does not need to be listed here."
   :type 'string
   :group 'agda2)
 
-(defcustom agda2-toplevel-module "Agda.Interaction.GhciTop"
-  "The name of the Agda toplevel module."
-  :type 'string :group 'agda2)
-
 (defcustom agda2-information-window-max-height
   0.35
   "The maximum height of the information window.
@@ -1306,7 +1302,8 @@ top-level scope."
 
 (agda2-maybe-normalised
  agda2-elaborate-give
- "Elaborate check the given expression against the hole's type and fill in hole with the elaborated term"
+ "Elaborate check the given expression against the hole's type and fill in the
+ hole with the elaborated term"
  "Cmd_elaborate_give"
  "expression to elaborate and give")
 
@@ -1324,7 +1321,8 @@ top-level scope."
 
 (agda2-maybe-normalised
  agda2-goal-and-context-and-checked
- "Shows the context, the goal and check the given expression's against the hole's type"
+ "Shows the context, the goal and check the given expression's against
+ the hole's type"
  "Cmd_goal_type_context_check"
  "expression to type")
 
@@ -1976,7 +1974,10 @@ the argument is a positive number, otherwise turn it off."
 This command assumes that the agda and agda-mode executables for
 Agda version VERSION are called agda-VERSION and
 agda-mode-VERSION, and that they are located on the PATH. (If
-VERSION is empty, then agda and agda-mode are used instead.)"
+VERSION is empty, then agda and agda-mode are used instead.)
+
+An attempt is made to preserve the default value of
+`agda2-mode-hook'."
   (interactive
    (list (completing-read "Version: " (agda2-get-agda-program-versions))))
 
@@ -1988,6 +1989,8 @@ VERSION is empty, then agda and agda-mode are used instead.)"
                       (list buf))))
                 (buffer-list)))
 
+       (default-hook (default-value 'agda2-mode-hook))
+
        (version-suffix (if (or (equal version "")
                                (equal version nil))
                            ""
@@ -2042,6 +2045,10 @@ VERSION is empty, then agda and agda-mode are used instead.)"
     (require 'agda2-mode)
     (setq agda2-program-name (concat "agda" version-suffix))
 
+    ;; Restore the Agda mode's default hook (if any).
+    (when default-hook
+      (set-default 'agda2-mode-hook default-hook))
+
     ;; Restart the Agda mode in all former Agda mode buffers.
     (mapc (lambda (buf)
             (with-current-buffer buf
