@@ -5,7 +5,7 @@ module Agda.Auto.Auto
       , AutoProgress(..)
       ) where
 
-import Prelude hiding (null)
+import Prelude hiding ((!!), null)
 
 import Control.Monad          ( filterM, forM, guard, join, when )
 import Control.Monad.Except
@@ -54,6 +54,7 @@ import Agda.Auto.CaseSplit
 import Agda.Utils.Functor
 import Agda.Utils.Impossible
 import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Null
 import Agda.Utils.Pretty ( prettyShow )
@@ -196,7 +197,7 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
         let getsols :: [I.Term] -> TCM [(MetaId, A.Expr)]
             getsols sol = do
              exprs <- forM (zip (Map.keys tccons) sol) $ \ (mi, e) -> do
-               mv   <- lookupMeta mi
+               mv   <- lookupLocalMetaAuto mi
                e    <- etaContract e
                expr <- modifyAbstractExpr <$> do withMetaInfo (getMetaInfo mv) $ reify e
                return (mi, expr)
@@ -234,7 +235,7 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
                  else do
                   aexprss <- mapM getsols rsols
                   cexprss <- forM aexprss $ mapM $ \(mi, e) -> do
-                    mv <- lookupMeta mi
+                    mv <- lookupLocalMetaAuto mi
                     withMetaInfo (getMetaInfo mv) $ do
                       (mi,) <$> abstractToConcrete_ e
                   let ss = dropWhile (== ' ') . dropWhile (/= ' ') . prettyShow
@@ -290,7 +291,7 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
               -- cexprss <- mapM (mapM (\(mi, e) -> lookupMeta mi >>= \mv -> withMetaInfo (getMetaInfo mv) $ abstractToConcrete_ e >>= \e' -> return (mi, e'))) aexprss
               cexprss <- forM aexprss $ do
                 mapM $ \ (mi, e) -> do
-                  mv <- lookupMeta mi
+                  mv <- lookupLocalMetaAuto mi
                   withMetaInfo (getMetaInfo mv) $ do
                     e' <- abstractToConcrete_ e
                     return (mi, e')
@@ -337,7 +338,7 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
                            -- (const retry)
                            -- (\_ -> return (Nothing, Just ("Failed to give expr for side solution of " ++ show mi)))
                          Just ii' -> do ae <- give WithoutForce ii' Nothing expr
-                                        mv <- lookupMeta mi
+                                        mv <- lookupLocalMetaAuto mi
                                         let scope = getMetaScope mv
                                         ce <- abstractToConcreteScope scope ae
                                         let cmnt = if ii' == ii then agsyinfo ticks else ""
@@ -381,12 +382,12 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
             case cls' of
              Left{} -> stopWithMsg "No solution found"
              Right cls' -> do
-              cls'' <- forM cls' $ \ (I.Clause _ _ tel ps body t catchall exact recursive reachable ell) -> do
+              cls'' <- forM cls' $ \ (I.Clause _ _ tel ps body t catchall exact recursive reachable ell wm) -> do
                 withCurrentModule (AN.qnameModule def) $ do
                  -- Normalise the dot patterns
                  ps <- addContext tel $ normalise ps
                  body <- etaContract body
-                 fmap modifyAbstractClause $ inTopContext $ reify $ AN.QNamed def $ I.Clause noRange noRange tel ps body t catchall exact recursive reachable ell
+                 fmap modifyAbstractClause $ inTopContext $ reify $ AN.QNamed def $ I.Clause noRange noRange tel ps body t catchall exact recursive reachable ell wm
               moduleTel <- lookupSection (AN.qnameModule def)
               pcs <- withInteractionId ii $ inTopContext $ addContext moduleTel $ mapM prettyA cls''
               ticks <- liftIO $ readIORef ticks
@@ -400,7 +401,7 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
        _ -> stopWithMsg "Metavariable is not at top level of clause RHS"
 
      MRefine listmode -> do
-      mv <- lookupMeta mi
+      mv <- lookupLocalMetaAuto mi
       let tt = jMetaType $ mvJudgement mv
           minfo = getMetaInfo mv
       targettyp <- withMetaInfo minfo $ do
@@ -421,13 +422,14 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
             cn <- withMetaInfo minfo $ runAbsToCon $ toConcrete n
             if C.isInScope cn == C.NotInScope then
               return Nothing
-            else do
-              c <- getConstInfo n
-              ctyp <- normalise $ defType c
-              cdfv <- withMetaInfo minfo $ getDefFreeVars n
-              return $ case matchType cdfv tctx ctyp targettyp of
-               Nothing -> Nothing
-               Just score -> Just (prettyShow cn, score)
+            else getConstInfo' n >>= \case
+              Left{} -> return Nothing
+              Right c -> do
+                ctyp <- normalise $ defType c
+                cdfv <- withMetaInfo minfo $ getDefFreeVars n
+                return $ case matchType cdfv tctx ctyp targettyp of
+                  Nothing -> Nothing
+                  Just score -> Just (prettyShow cn, score)
          ) alldefs
        else do
         let scopeinfo = clScope (getMetaInfo mv)
@@ -437,13 +439,14 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
             modnames = case thisdefinfo of
                         Just (def, _, _) -> filter (\(_, n) -> n /= def) qnames
                         Nothing -> qnames
-        catMaybes <$> mapM (\(cn, n) -> do
-          c <- getConstInfo n
-          ctyp <- normalise $ defType c
-          cdfv <- withMetaInfo minfo $ getDefFreeVars n
-          return $ case matchType cdfv tctx ctyp targettyp of
-           Nothing -> Nothing
-           Just score -> Just (prettyShow cn, score)
+        catMaybes <$> mapM (\(cn, n) -> getConstInfo' n >>= \case
+          Left{} -> return Nothing
+          Right c -> do
+            ctyp <- normalise $ defType c
+            cdfv <- withMetaInfo minfo $ getDefFreeVars n
+            return $ case matchType cdfv tctx ctyp targettyp of
+              Nothing -> Nothing
+              Just score -> Just (prettyShow cn, score)
          ) modnames
 
       let sorthits = List.sortBy (\(_, (pa1, pb1)) (_, (pa2, pb2)) -> case compare pa2 pa1 of {EQ -> compare pb1 pb2; o -> o}) hits
@@ -466,14 +469,14 @@ auto ii rng argstr = liftTCM $ locallyTC eMakeCase (const True) $ do
 -- Get the functions and axioms defined in the same module as @def@.
 autohints :: AutoHintMode -> I.MetaId -> Maybe AN.QName -> TCM [Hint]
 autohints AHMModule mi (Just def) = do
-  scope <- clScope . getMetaInfo <$> lookupMeta mi
+  scope <- clScope . getMetaInfo <$> lookupLocalMetaAuto mi
   let names     = Scope.nsNames $ Scope.everythingInScope scope
       qnames    = map (Scope.anameName . head) $ Map.elems names
       modnames  = filter (\n -> AN.qnameModule n == AN.qnameModule def && n /= def) qnames
   map (Hint False) <$> do
-    (`filterM` modnames) $ \ n -> do
-      c <- getConstInfo n
-      case theDef c of
+    (`filterM` modnames) $ \ n -> getConstInfo' n >>= \case
+      Left{} -> return False
+      Right c -> case theDef c of
         Axiom{}    -> return True
         AbstractDefn{} -> return True
         Function{} -> return True
