@@ -1,6 +1,8 @@
 
 module Agda.Auto.Convert where
 
+import Prelude hiding ((!!))
+
 import Control.Monad          ( when )
 import Control.Monad.Except
 import Control.Monad.IO.Class ( MonadIO(..) )
@@ -27,7 +29,8 @@ import qualified Agda.TypeChecking.Monad.Base as MB
 import Agda.TypeChecking.Monad.Signature (getConstInfo, getDefFreeVars, ignoreAbstractMode)
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.TypeChecking.Monad.Base (mvJudgement, mvPermutation, getMetaInfo, envContext, clEnv)
-import Agda.TypeChecking.Monad.MetaVars (lookupMeta, withMetaInfo, lookupInteractionPoint)
+import Agda.TypeChecking.Monad.MetaVars
+  (lookupMeta, withMetaInfo, lookupInteractionPoint)
 import Agda.TypeChecking.Monad.Context (getContextArgs)
 import Agda.TypeChecking.Monad.Constraints (getAllConstraints)
 import Agda.TypeChecking.Substitute (applySubst, renamingR)
@@ -47,6 +50,7 @@ import Agda.Auto.CaseSplit hiding (lift)
 
 import Agda.Utils.Either
 import Agda.Utils.Lens
+import Agda.Utils.List
 import Agda.Utils.Monad       ( forMaybeMM )
 import Agda.Utils.Permutation ( Permutation(Perm), permute, takeP, compactP )
 import Agda.Utils.Pretty      ( prettyShow )
@@ -162,7 +166,7 @@ tomy imi icns typs = do
           when (Map.notMember eqi eqsm) $ do
            modify $ \s -> s {sEqs = (Map.insert eqi Nothing eqsm, eqi : eqsl)}
         ) (zip eqs [0..])
-       mv <- lift $ lookupMeta mi
+       mv <- lift $ lookupLocalMetaAuto mi
        msol <- case MB.mvInstantiation mv of
                      MB.InstV{} ->
                       lift $ withMetaInfo (getMetaInfo mv) $ do
@@ -264,9 +268,23 @@ getConst iscon name mode = do
 
 getdfv :: I.MetaId -> A.QName -> MB.TCM Cm.Nat
 getdfv mainm name = do
- mv <- lookupMeta mainm
+ mv <- lookupLocalMetaAuto mainm
  withMetaInfo (getMetaInfo mv) $ getDefFreeVars name
 
+-- | A variant of 'lookupLocalMeta' that, if applied to a remote
+-- meta-variable, raises a special error message noting that remote
+-- meta-variables are not handled by the auto command.
+
+lookupLocalMetaAuto :: I.MetaId -> MB.TCM MB.MetaVariable
+lookupLocalMetaAuto m = do
+  mv <- lookupMeta m
+  case mv of
+    Just (Right mv) -> return mv
+    Nothing         -> __IMPOSSIBLE__
+    Just Left{}     -> MB.typeError $ MB.GenericError $
+      "The auto command does not support remote meta-variables," ++
+      "consider using --no-save-metas"
+
 getMeta :: I.MetaId -> TOM (Metavar (Exp O) (RefInfo O))
 getMeta name = do
  mmap <- gets (fst . sMetas)
@@ -288,10 +306,6 @@ getEqs = forMaybeMM getAllConstraints $ \ eqc -> do
       return $ Just (tomyIneq ineq, ee, ei)
     _ -> return Nothing
 
-copatternsNotImplemented :: MB.TCM a
-copatternsNotImplemented = MB.typeError $ MB.NotImplemented $
-  "The Agda synthesizer (Agsy) does not support copatterns yet"
-
 literalsNotImplemented :: MB.TCM a
 literalsNotImplemented = MB.typeError $ MB.NotImplemented $
   "The Agda synthesizer (Agsy) does not support literals yet"
@@ -336,9 +350,9 @@ instance Conversion TOM (Cm.Arg I.Pattern) (Pat O) where
       cc    <- lift $ liftIO $ readIORef c
       let Just (npar,_) = fst $ cdorigin cc
       return $ PatConApp c (replicate npar PatExp ++ pats')
+    I.ProjP _ q -> PatProj <$> getConst True q TMAll
 
     -- UNSUPPORTED CASES
-    I.ProjP{}   -> lift copatternsNotImplemented
     I.LitP{}    -> lift literalsNotImplemented
     I.DefP{}    -> lift hitsNotImplemented
 
@@ -581,7 +595,10 @@ constructPats cmap mainm clause = do
        I.DotP _ t -> do
         (t2, _) <- runStateT (convert t) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
         return (ns, HI hid (CSPatExp t2))
-       I.ProjP{} -> copatternsNotImplemented
+       I.ProjP po c -> do
+        (c2, _) <- runStateT (getConst True c TMAll) (S {sConsts = (cmap, []), sMetas = initMapS, sEqs = initMapS, sCurMeta = Nothing, sMainMeta = mainm})
+        cc <- liftIO $ readIORef c2
+        return (ns, HI hid (CSPatProj c2))
        I.LitP{} -> literalsNotImplemented
        I.DefP{} -> hitsNotImplemented
 
@@ -651,18 +668,19 @@ frommyClause (ids, pats, mrhs) = do
           Nothing -> return $ Nothing
           Just e -> Just <$> convert e
  let cperm =  Perm nv perm
- return $ I.Clause
+ return I.Clause
    { I.clauseLHSRange  = SP.noRange
    , I.clauseFullRange = SP.noRange
    , I.clauseTel       = tel
    , I.namedClausePats = IP.numberPatVars __IMPOSSIBLE__ cperm $ applySubst (renamingR $ compactP cperm) ps
-   , I.clauseBody  = body
-   , I.clauseType  = Nothing -- TODO: compute clause type
-   , I.clauseCatchall = False
+   , I.clauseBody      = body
+   , I.clauseType      = Nothing -- TODO: compute clause type
+   , I.clauseCatchall    = False
    , I.clauseExact       = Nothing -- TODO
    , I.clauseRecursive   = Nothing -- TODO: Don't know here whether recursive or not !?
    , I.clauseUnreachable = Nothing -- TODO: Don't know here whether reachable or not !?
-   , I.clauseEllipsis = Cm.NoEllipsis
+   , I.clauseEllipsis    = Cm.NoEllipsis
+   , I.clauseWhereModule = Nothing
    }
 
 contains_constructor :: [CSPat O] -> Bool
