@@ -1,10 +1,12 @@
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 module Agda.Auto.NarrowingSearch where
 
+import Control.Monad       ( foldM, when )
+import Control.Monad.State ( MonadState(..), modify, StateT, evalStateT, runStateT )
+import Control.Monad.Trans ( lift )
+
 import Data.IORef hiding (writeIORef, modifyIORef)
 import qualified Data.IORef as NoUndo (writeIORef, modifyIORef)
-import Control.Monad.State
 
 import Agda.Utils.Impossible
 import Agda.Utils.Empty
@@ -12,13 +14,18 @@ import Agda.Utils.Empty
 newtype Prio = Prio { getPrio :: Int }
  deriving (Eq, Ord, Num)
 
-class Trav a blk | a -> blk where
-  trav :: Monad m => (forall b . Trav b blk => MM b blk -> m ()) -> a -> m ()
+class Trav a where
+  type Block a
+  trav :: Monad m => (forall b. TravWith b (Block a) => MM b (Block b) -> m ()) -> a -> m ()
 
-instance Trav a blk => Trav (MM a blk) blk where
+-- | Trav instance 'a' with block type 'blk'
+type TravWith a blk = (Trav a, Block a ~ blk)
+
+instance TravWith a blk => Trav (MM a blk) where
+  type Block (MM a blk) = blk
   trav f me = f me
 
-data Term blk = forall a . Trav a blk => Term a
+data Term blk = forall a. TravWith a blk => Term a
 
 -- | Result of type-checking.
 data Prop blk
@@ -240,9 +247,7 @@ mmmcase x fm f = case x of
  NotM x -> f x
  Meta m -> do
   bind <- readIORef $ mbind m
-  case bind of
-   Just x -> f x
-   Nothing -> fm
+  maybe fm f bind
 
 mmpcase :: Refinable a blk => BlkInfo blk -> MM a blk -> (a -> MetaEnv (PB blk)) -> MetaEnv (PB blk)
 mmpcase blkinfo x f = case x of
@@ -283,8 +288,7 @@ mmbpcase x fm f = do
 
 waitok :: OKHandle blk -> MetaEnv (MB b blk) -> MetaEnv (MB b blk)
 waitok okh f =
- mmcase okh $ \b -> case b of -- principle constraint is never present for okhandle so it will not be refined
-  OKVal -> f
+ mmcase okh $ \ OKVal -> f -- principle constraint is never present for okhandle so it will not be refined
 
 mbret :: a -> MetaEnv (MB a blk)
 mbret x = return $ NotB x
@@ -413,10 +417,11 @@ topSearch ticks nsol hsol envinfo p searchdepth depthinterval = do
          ) mcomptr
         obs <- ureadIORef (mobs m)
         res <- recalcs obs
-        case res of
-         True -> -- failed
-          return $ Left False
-         False -> lift $ search depthleft -- succeeded
+        if res
+          then
+            return $ Left False     -- failed
+          else
+            lift $ search depthleft -- succeeded
 
     doit = do
      res <- search depth
@@ -448,13 +453,14 @@ topSearch ticks nsol hsol envinfo p searchdepth depthinterval = do
 
  runUndo $ do
   res <- reccalc p (Just mainroot)
-  case res of
-   True -> -- failed immediately
-    return False
-   False -> do
-    Left _solFound <- lift $ searchSubProb [(mainroot, Nothing)] searchdepth
-    dr <- lift $ readIORef depthreached
-    return dr
+  if res -- failed immediately
+    then
+        return False
+    else
+      do
+        Left _solFound <- lift $ searchSubProb [(mainroot, Nothing)] searchdepth
+        lift $ readIORef depthreached
+
 
 extractblkinfos :: Metavar a blk -> IO [blk]
 extractblkinfos m = do
@@ -469,8 +475,7 @@ extractblkinfos m = do
   f ((QPDoubleBlocked{}, _) : cs) = f cs
 
 recalcs :: [(QPB a blk, Maybe (CTree blk))] -> Undo Bool
-recalcs [] = return False
-recalcs (c : cs) = seqc (recalc c) (recalcs cs)
+recalcs cs = foldr (seqc . recalc) (return False) cs
 
 seqc :: Undo Bool -> Undo Bool -> Undo Bool
 seqc x y = do
@@ -480,33 +485,33 @@ seqc x y = do
   False -> y
 
 recalc :: (QPB a blk, Maybe (CTree blk)) -> Undo Bool
-recalc (con, node) =
- case con of
-  QPBlocked _ cont -> reccalc cont node
+recalc (con, node) = case con of
+  QPBlocked       _    cont -> reccalc cont node
   QPDoubleBlocked flag cont -> do
-   fl <- ureadIORef flag
-   if fl then
-     return False
-    else do
-     uwriteIORef flag True
-     reccalc cont node
+    fl <- ureadIORef flag
+    if fl
+      then return False
+      else do
+        uwriteIORef flag True
+        reccalc cont node
 
 reccalc :: MetaEnv (PB blk) -> Maybe (CTree blk) -> Undo Bool
 reccalc cont node = do
- res <- calc cont node
- case res of
-  Nothing -> return True
-  Just pendhandles ->
-   foldM (\res1 h ->
-    case res1 of
-     True -> return res1
-     False -> do
-
-
-      uwriteIORef (mbind h) $ Just OKVal
-      obs <- ureadIORef (mobs h)
-      recalcs obs
-   ) False pendhandles
+  res <- calc cont node
+  case res of
+    Nothing -> return True
+    Just pendhandles ->
+      foldM
+        ( \res1 h ->
+            if res1
+              then return res1
+              else do
+                uwriteIORef (mbind h) $ Just OKVal
+                obs <- ureadIORef (mobs h)
+                recalcs obs
+        )
+        False
+        pendhandles
 
 calc :: forall blk . MetaEnv (PB blk) -> Maybe (CTree blk) -> Undo (Maybe [OKMeta blk])
 calc cont node = do
@@ -558,7 +563,10 @@ calc cont node = do
        storeprio node (NoPrio False) []
     PDoubleBlocked m1 m2 cont -> do
      flag <- lift $ newIORef False
-     let newobs = ((QPDoubleBlocked flag cont, node) :)
+
+     let newobs :: forall b. [(QPB b blk, Maybe (CTree blk))]
+                          -> [(QPB b blk, Maybe (CTree blk))]
+         newobs = ((QPDoubleBlocked flag cont, node) :)
      umodifyIORef (mobs m1) newobs
      umodifyIORef (mobs m2) newobs
      storeprio node (NoPrio False) []
@@ -610,32 +618,37 @@ calc cont node = do
     ConnectHandle (NotM _) _ -> __IMPOSSIBLE__
 
 choosePrioMeta :: Bool -> PrioMeta blk -> PrioMeta blk -> PrioMeta blk
-choosePrioMeta flip pm1@(PrioMeta p1 _) pm2@(PrioMeta p2 _) = if p1 > p2 then pm1 else if p2 > p1 then pm2 else if flip then pm2 else pm1
+choosePrioMeta flip pm1@(PrioMeta p1 _) pm2@(PrioMeta p2 _)
+  | p1 > p2 = pm1
+  | p2 > p1 = pm2
+  | flip = pm2
+  | otherwise = pm1
 choosePrioMeta _ pm@(PrioMeta _ _) (NoPrio _) = pm
 choosePrioMeta _ (NoPrio _) pm@(PrioMeta _ _) = pm
 choosePrioMeta _ (NoPrio d1) (NoPrio d2) = NoPrio (d1 && d2)
 
 propagatePrio :: CTree blk -> Undo [OKMeta blk]
 propagatePrio node = do
- parent <- lift $ readIORef $ ctparent node
- case parent of
-  Nothing -> return []
-  Just parent -> do
-   Just sc <- ureadIORef (ctsub parent)
-   pm1 <- ureadIORef $ ctpriometa $ scsub1 sc
-   pm2 <- ureadIORef $ ctpriometa $ scsub2 sc
-   flip <- ureadIORef $ scflip sc
-   let pm = choosePrioMeta flip pm1 pm2
-   opm <- ureadIORef (ctpriometa parent)
-   if (not (pm == opm)) then do
-     uwriteIORef (ctpriometa parent) pm
-     phs <- case pm of
-      NoPrio True -> ureadIORef (cthandles parent)
-      _ -> return []
-     phs2 <- propagatePrio parent
-     return $ phs ++ phs2
-    else
-     return []
+  parent <- lift $ readIORef $ ctparent node
+  case parent of
+    Nothing     -> return []
+    Just parent -> do
+      Just sc <- ureadIORef (ctsub parent)
+      pm1     <- ureadIORef $ ctpriometa $ scsub1 sc
+      pm2     <- ureadIORef $ ctpriometa $ scsub2 sc
+      flip    <- ureadIORef $ scflip sc
+      let pm = choosePrioMeta flip pm1 pm2
+      opm <- ureadIORef (ctpriometa parent)
+      if (pm /= opm)
+        then do
+          uwriteIORef (ctpriometa parent) pm
+          phs <- case pm of
+            NoPrio True -> ureadIORef (cthandles parent)
+            _           -> return []
+          phs2 <- propagatePrio parent
+          return $ phs ++ phs2
+        else
+          return []
 
 data Choice = LeftDisjunct | RightDisjunct
 
