@@ -1,6 +1,9 @@
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.Auto.Typecheck where
 
+import Prelude hiding ((!!))
+
 import Data.IORef
 
 import Agda.Syntax.Common (Hiding (..))
@@ -9,6 +12,8 @@ import Agda.Auto.Syntax
 import Agda.Auto.SearchControl
 
 import Agda.Utils.Impossible
+import Agda.Utils.List
+import Agda.Utils.Maybe
 
 -- ---------------------------------
 
@@ -16,20 +21,23 @@ import Agda.Utils.Impossible
 tcExp :: Bool -> Ctx o -> CExp o -> MExp o -> EE (MyPB o)
 tcExp isdep ctx typ@(TrBr typtrs ityp@(Clos _ itypexp)) trm =
   mbpcase prioTypeUnknown Nothing (hnn_checkstep ityp) $ \(hntyp, iotastepdone) ->
-  mmpcase (True, prioTypecheck isdep, Just (RIMainInfo (length ctx) hntyp iotastepdone)) trm $ \trm -> case trm of
+  mmpcase (True, prioTypecheck isdep, Just (RIMainInfo (length ctx) hntyp iotastepdone)) trm $ \case
    App _ okh elr args -> case rawValue hntyp of
     HNPi{} | isdep -> mpret $ Error "tcExp, dep terms should be eta-long"
     _ -> do
-     (ityp, sc) <- case elr of
-              Var v -> -- assuming within scope
-               return (weak (v + 1) (snd $ ctx !! v), id)
+     res <- case elr of
+              Var v ->
+                case ctx !!! v of
+                  Nothing     -> return Nothing
+                  Just (_, a) -> return $ Just (weak (v+1) a, id)
               Const c -> do
                cdef <- readIORef c
-               return (closify (cdtype cdef), \x -> mpret $ And (Just [Term args]) (noiotastep_term c args) x)
+               return $ Just (closify (cdtype cdef), mpret . And (Just [Term args]) (noiotastep_term c args))
+     caseMaybe res (mpret $ Error "tcExp, variable not in scope") $ \ (ityp, sc) -> do
 
      ndfv <- case elr of
               Var{} -> return 0
-              Const c -> readIORef c >>= \cd -> return (cddeffreevars cd)
+              Const c -> cddeffreevars <$> readIORef c
 
 
      isconstructor <- case elr of
@@ -408,6 +416,25 @@ iotastep smartcheck e = case rawValue e of
       mbfailed "dopat: wrong amount of args"
     else
      mbret $ Left (Left aa)
+ dopat (PatProj cs) a =
+  case a of
+   PENo a ->
+    if smartcheck then
+     mbcase (meta_not_constructor a) $ \notcon -> if notcon then mbret $ Left $ Right noblks else qq -- to know more often if iota step is possible
+    else
+     qq
+    where
+     qq =
+      mbcase (hnn_blks a) $ \(hna, blks) -> case rawValue hna of
+       HNApp (Const c') as ->
+        do
+         cd <- readIORef c'
+         case cdcont cd of
+          Constructor{} -> mbcase (getAllArgs as) $ \as' ->
+           mbret $ Left (Left (PEConApp a c' (map PENo as')))
+          _ -> mbret $ Left (Right (addblk hna blks))
+       _ -> mbret $ Left (Right (addblk hna blks))
+   aa@(PEConApp a c' as) -> mbret $ Left (Left aa)
  dopat PatVar{} a@(PENo a') = mbret $ Right (a, [a'])
  dopat PatVar{} a@(PEConApp a' _ _) = mbret $ Right (a, [a'])
  dopat PatExp a = mbret $ Right (a, [])
