@@ -13,9 +13,9 @@ import GHC.Generics (Generic)
 
 import System.IO.Unsafe
 
-import Agda.Syntax.Concrete.Name (TopLevelModuleName)
 import Agda.Syntax.Concrete.Pretty () --instance only
 import Agda.Syntax.Abstract.Name
+import Agda.Syntax.TopLevelModuleName (TopLevelModuleName)
 import Agda.Utils.Benchmark (MonadBench(..))
 import qualified Agda.Utils.Benchmark as B
 import Agda.Utils.Null
