@@ -5,10 +5,14 @@ import Agda.Syntax.Common (Nat)
 import Agda.Syntax.Treeless
 
 import Agda.TypeChecking.Monad
+  ( HasConstInfo
+  , getErasedConArgs
+  , getTreeless
+  )
 
 -- | Insert unsafeCoerce (in the form of 'TCoerce') everywhere it's needed in
 --   the right-hand side of a definition.
-addCoercions :: TTerm -> TCM TTerm
+addCoercions :: HasConstInfo m => TTerm -> m TTerm
 addCoercions = coerceTop
   where
     -- Don't coerce top-level lambdas.
@@ -65,12 +69,12 @@ addCoercions = coerceTop
         TLet e b       -> TLet <$> softCoerce e <*> softCoerce b
         TCase x t d bs -> TCase x t <$> coerce d <*> mapM coerceAlt bs
 
-funArity :: TTerm -> TCM Nat
+funArity :: HasConstInfo m => TTerm -> m Nat
 funArity (TDef q)  = maybe 0 (fst . tLamView) <$> getTreeless q
 funArity (TCon q)  = erasedArity q
 funArity (TPrim _) = return 3 -- max arity of any primitive
 funArity _         = return 0
 
 -- | The number of retained arguments after erasure.
-erasedArity :: QName -> TCM Nat
+erasedArity :: HasConstInfo m => QName -> m Nat
 erasedArity q = length . filter not <$> getErasedConArgs q
