@@ -54,6 +54,7 @@ import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Names (namesIn)
 import qualified Agda.Syntax.Treeless as T
 import Agda.Syntax.Literal
+import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Primitive (getBuiltinName)
@@ -70,6 +71,7 @@ import Agda.Utils.Float
 import Agda.Utils.IO.Directory
 import Agda.Utils.Lens
 import Agda.Utils.List
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Pretty (prettyShow, render)
@@ -192,7 +194,7 @@ data GHCDefinition = GHCDefinition
   , ghcDefDecls      :: [HS.Decl]
   , ghcDefDefinition :: Definition
   , ghcDefMainDef    :: Maybe MainFunctionDef
-  , ghcDefImports    :: Set ModuleName
+  , ghcDefImports    :: Set TopLevelModuleName
   }
 
 --- Top-level compilation ---
@@ -245,7 +247,7 @@ ghcPreCompile flags = do
   msub        <- getBuiltinName builtinSub
   msubin      <- getBuiltinName builtinSubIn
   mid         <- getBuiltinName builtinId
-  mconid      <- getBuiltinName builtinConId
+  mconid      <- getPrimitiveName' builtinConId
 
   istcbuiltin <- do
     builtins <- mapM getBuiltinName
@@ -267,6 +269,8 @@ ghcPreCompile flags = do
       , builtinAgdaTCMFreshName
       , builtinAgdaTCMDeclareDef
       , builtinAgdaTCMDeclarePostulate
+      , builtinAgdaTCMDeclareData
+      , builtinAgdaTCMDefineData
       , builtinAgdaTCMDefineFun
       , builtinAgdaTCMGetType
       , builtinAgdaTCMGetDefinition
@@ -275,12 +279,14 @@ ghcPreCompile flags = do
       , builtinAgdaTCMIsMacro
       , builtinAgdaTCMWithNormalisation
       , builtinAgdaTCMWithReconsParams
+      , builtinAgdaTCMFormatErrorParts
       , builtinAgdaTCMDebugPrint
       , builtinAgdaTCMOnlyReduceDefs
       , builtinAgdaTCMDontReduceDefs
       , builtinAgdaTCMNoConstraints
       , builtinAgdaTCMRunSpeculative
       , builtinAgdaTCMExec
+      , builtinAgdaTCMGetInstances
       ]
     return $
       flip HashSet.member $
@@ -320,7 +326,8 @@ ghcPreCompile flags = do
     , ghcEnvIsTCBuiltin = istcbuiltin
     }
 
-ghcPostCompile :: GHCEnv -> IsMain -> Map ModuleName GHCModule -> TCM ()
+ghcPostCompile ::
+  GHCEnv -> IsMain -> Map TopLevelModuleName GHCModule -> TCM ()
 ghcPostCompile _cenv _isMain mods = do
   -- FIXME: @curMName@ and @curIF@ are evil TCM state, but there does not appear to be
   --------- another way to retrieve the compilation root ("main" module or interaction focused).
@@ -336,7 +343,7 @@ ghcPostCompile _cenv _isMain mods = do
 ghcPreModule
   :: GHCEnv
   -> IsMain      -- ^ Are we looking at the main module?
-  -> ModuleName
+  -> TopLevelModuleName
   -> Maybe FilePath    -- ^ Path to the @.agdai@ file.
   -> TCM (Recompile GHCModuleEnv GHCModule)
                  -- ^ Could we confirm the existence of a main function?
@@ -356,7 +363,8 @@ ghcPreModule cenv isMain m mifile =
     ifileDesc = fromMaybe "(memory)" mifile
 
     noComp = do
-      reportSLn "compile.ghc" 2 . (++ " : no compilation is needed.") . prettyShow . A.mnameToConcrete =<< curMName
+      reportSLn "compile.ghc" 2 .
+        (++ " : no compilation is needed.") . prettyShow =<< curMName
       menv <- ask
       mainDefs <- ifM curIsMainModule
                          (mainFunctionDefs <$> curIF)
@@ -364,7 +372,7 @@ ghcPreModule cenv isMain m mifile =
       return . Skip $ GHCModule menv mainDefs
 
     yesComp = do
-      m   <- prettyShow . A.mnameToConcrete <$> curMName
+      m   <- prettyShow <$> curMName
       out <- curOutFile
       reportSLn "compile.ghc" 1 $ repl [m, ifileDesc, out] "Compiling <<0>> in <<1>> to <<2>>"
       asks Recompile
@@ -373,7 +381,7 @@ ghcPostModule
   :: GHCEnv
   -> GHCModuleEnv
   -> IsMain        -- ^ Are we looking at the main module?
-  -> ModuleName
+  -> TopLevelModuleName
   -> [GHCDefinition]   -- ^ Compiled module content.
   -> TCM GHCModule
 ghcPostModule _cenv menv _isMain _moduleName ghcDefs = do
@@ -421,7 +429,9 @@ ghcMayEraseType q = getHaskellPragma q <&> \case
 
 -- Compilation ------------------------------------------------------------
 
-imports :: BuiltinThings PrimFun -> Set ModuleName -> [Definition] -> [HS.ImportDecl]
+imports ::
+  BuiltinThings PrimFun -> Set TopLevelModuleName -> [Definition] ->
+  [HS.ImportDecl]
 imports builtinThings usedModules defs = hsImps ++ imps where
   hsImps :: [HS.ImportDecl]
   hsImps = [unqualRTE, decl mazRTE]
@@ -442,7 +452,7 @@ imports builtinThings usedModules defs = hsImps ++ imps where
   decl :: HS.ModuleName -> HS.ImportDecl
   decl m = HS.ImportDecl m True Nothing
 
-  mnames :: [ModuleName]
+  mnames :: [TopLevelModuleName]
   mnames = Set.elems usedModules
 
   uniq :: [HS.ModuleName] -> [HS.ModuleName]
@@ -657,13 +667,14 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
 
       -- Id x y is compiled as a pair of a boolean and whatever
       -- Path x y is compiled to.
-      Axiom{} | is ghcEnvId -> do
+      Datatype{} | is ghcEnvId -> do
         _        <- sequence_ [primInterval]
         Just int <- getBuiltinName builtinInterval
         int      <- xhqn TypeK int
+        -- re  #3733: implement reflId
         retDecls $
           [ HS.TypeDecl (unqhname TypeK q)
-              [HS.UnkindedVar (ihname A i) | i <- [0..3]]
+              [] -- [HS.UnkindedVar (ihname A i) | i <- [0..3]]
               (HS.TyApp (HS.FakeType "(,) Bool")
                  (HS.TyFun (HS.TyCon int) mazAnyType))
           , HS.FunBind
@@ -673,7 +684,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
           ]
 
       -- conid.
-      Axiom{} | is ghcEnvConId -> do
+      Primitive{} | is ghcEnvConId -> do
         strict <- optGhcStrictData <$> askGhcOpts
         let var = (if strict then HS.PBangPat else id) . HS.PVar
         retDecls $
@@ -705,7 +716,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
                    fb axiomErr
       Primitive{ primName = s } -> (mempty,) . fb <$> (liftTCM . primBody) s
 
-      PrimitiveSort{ primName = s } -> retDecls []
+      PrimitiveSort{} -> retDecls []
 
       Function{} -> function pragma $ functionViaTreeless q
 
@@ -787,7 +798,7 @@ definition def@Defn{defName = q, defType = ty, theDef = d} = do
       def <- getConstInfo q
       (argTypes0, resType) <- hsTelApproximation $ defType def
       let pars = case theDef def of
-                   Function{ funProjection = Just Projection{ projIndex = i } } | i > 0 -> i - 1
+                   Function{ funProjection = Right Projection{ projIndex = i } } | i > 0 -> i - 1
                    _ -> 0
           argTypes  = drop pars argTypes0
           argTypesS = filterUsed used argTypes
@@ -1040,6 +1051,7 @@ alt sc a = do
                       (HS.GuardedRhss [HS.GuardedRhs [HS.Qualifier g] b])
                       emptyBinds
     T.TALit { T.aLit = LitQName q } -> mkAlt (litqnamepat q)
+    T.TALit { T.aLit = LitMeta _ m } -> mkAlt (litmetapat m)
     T.TALit { T.aLit = l@LitFloat{}, T.aBody = b } -> do
       tell YesFloat
       l <- literal l
@@ -1071,6 +1083,10 @@ literal l = case l of
   LitFloat  x   -> floatExp x "Double"
   LitQName  x   -> return $ litqname x
   LitString s   -> return $ litString s
+  LitMeta _ m   ->
+    return $ HS.FakeExp "(,)" `HS.App`
+             hsTypedInt (metaId m) `HS.App`
+             (hsTypedInt (moduleNameHash $ metaModule m))
   _             -> return $ l'
   where
     l'    = HS.Lit $ hslit l
@@ -1135,6 +1151,12 @@ litqnamepat x =
   where
     NameId n (ModuleNameHash m) = nameId $ qnameName x
 
+litmetapat :: MetaId -> HS.Pat
+litmetapat (MetaId m h) =
+  HS.PApp (hsName "(,)")
+          [ HS.PLit (HS.Int $ fromIntegral m)
+          , HS.PLit (HS.Int $ fromIntegral $ moduleNameHash h) ]
+
 condecl :: QName -> Induction -> HsCompileM HS.ConDecl
 condecl q _ind = do
   opts <- askGhcOpts
@@ -1270,10 +1292,11 @@ curOutFile = snd <$> curOutFileAndDir
 
 callGHC :: ReaderT GHCModule TCM ()
 callGHC = do
-  opts    <- askGhcOpts
-  hsmod   <- prettyPrint <$> curHsMod
-  agdaMod <- curAgdaMod
-  let outputName = lastWithDefault __IMPOSSIBLE__ $ mnameToList agdaMod
+  opts     <- askGhcOpts
+  agdaOpts <- lift commandLineOptions
+  hsmod    <- prettyPrint <$> curHsMod
+  agdaMod  <- curAgdaMod
+  let outputName = Text.unpack $ List1.last $ moduleNameParts agdaMod
   (mdir, fp) <- curOutFileAndDir
   let ghcopts = optGhcFlags opts
 
@@ -1288,7 +1311,7 @@ callGHC = do
 
   let overridableArgs =
         [ "-O"] ++
-        (if isMain then ["-o", mdir </> prettyShow (nameConcrete outputName)] else []) ++
+        (if isMain then ["-o", mdir </> outputName] else []) ++
         [ "-Werror"]
       otherArgs       =
         [ "-i" ++ mdir] ++
@@ -1307,4 +1330,8 @@ callGHC = do
   -- those versions of GHC we don't print any progress information
   -- unless an error is encountered.
   let doCall = optGhcCallGhc opts
-  liftTCM $ callCompiler doCall ghcBin args (Just utf8)
+      cwd    = if optGHCiInteraction agdaOpts ||
+                  optJSONInteraction agdaOpts
+               then Just mdir
+               else Nothing
+  liftTCM $ callCompiler doCall ghcBin args cwd (Just utf8)
