@@ -26,6 +26,8 @@ import Agda.TypeChecking.Pretty
 
 import Agda.Utils.Either
 import Agda.Utils.Lens
+import Agda.Utils.List   (hasElem)
+import Agda.Utils.Maybe
 import Agda.Utils.Pretty (prettyShow)
 import qualified Agda.Utils.Haskell.Syntax as HS
 
@@ -42,17 +44,17 @@ data CheckedMainFunctionDef = CheckedMainFunctionDef
 -- TODO: Also only consider top-level definition (not buried inside a module).
 asMainFunctionDef :: Definition -> Maybe MainFunctionDef
 asMainFunctionDef d = case (theDef d) of
-    Axiom{}                             -> perhaps
-    Function{ funProjection = Nothing } -> perhaps
-    Function{ funProjection = Just{}  } -> no
-    AbstractDefn{}                      -> no
-    GeneralizableVar{}                  -> no
-    DataOrRecSig{}                      -> no
-    Datatype{}                          -> no
-    Record{}                            -> no
-    Constructor{}                       -> no
-    Primitive{}                         -> no
-    PrimitiveSort{}                     -> no
+    Axiom{}                              -> perhaps
+    Function{ funProjection = Left _ }   -> perhaps
+    Function{ funProjection = Right{}  } -> no
+    AbstractDefn{}                       -> no
+    GeneralizableVar{}                   -> no
+    DataOrRecSig{}                       -> no
+    Datatype{}                           -> no
+    Record{}                             -> no
+    Constructor{}                        -> no
+    Primitive{}                          -> no
+    PrimitiveSort{}                      -> no
   where
   isNamedMain = "main" == prettyShow (nameConcrete . qnameName . defName $ d)  -- ignores the qualification!?
   perhaps | isNamedMain = Just $ MainFunctionDef d
@@ -156,12 +158,14 @@ xForPrim :: Map String a -> BuiltinThings PrimFun -> [Definition] -> [a]
 xForPrim table builtinThings defs = catMaybes
     [ Map.lookup s table
     | (s, def) <- Map.toList builtinThings
-    , getName def `Set.member` qs
+    , maybe False elemDefs $ getName def
     ]
   where
-  qs = Set.fromList $ map defName defs
-  getName (Builtin t)            = getPrimName t
-  getName (Prim (PrimFun q _ _)) = q
+  elemDefs = hasElem $ map defName defs
+  getName = \case
+    Builtin t                 -> Just $ getPrimName t
+    Prim (PrimFun q _ _)      -> Just q
+    BuiltinRewriteRelations _ -> Nothing
 
 
 -- | Definition bodies for primitive functions
@@ -282,10 +286,13 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primQNameFixity"     |-> return "MAlonzo.RTE.qnameFixity"
   , "primQNameToWord64s"  |-> return "\\ qn -> (MAlonzo.RTE.nameId qn, MAlonzo.RTE.moduleId qn)"
   , "primQNameToWord64sInjective" |-> return mazErasedName
-  , "primMetaEquality"    |-> rel "(==)" "Integer"
-  , "primMetaLess"        |-> rel "(<)" "Integer"
-  , "primShowMeta"        |-> return "\\ x -> Data.Text.pack (\"_\" ++ show (x :: Integer))"
-  , "primMetaToNat"       |-> return "(id :: Integer -> Integer)"
+  , "primMetaEquality"    |-> rel "(==)" "(Integer, Integer)"
+  , "primMetaLess"        |-> rel "(<)" "(Integer, Integer)"
+  -- Should be kept in sync with version in `primitiveFunctions` in
+  -- Agda.TypeChecking.Primitive
+  , "primShowMeta"        |-> return "\\ (m, h) -> Data.Text.pack (\"_\" ++ show (m :: Integer) ++ \"@\" ++ show (h :: Integer))"
+  -- Should be kept in sync with `metaToNat` in Agda.TypeChecking.Primitive
+  , "primMetaToNat"       |-> return "\\ (m, h) -> (h :: Integer) * 2^64 + (m :: Integer)"
   , "primMetaToNatInjective" |-> return mazErasedName
 
   -- Seq
@@ -313,7 +320,6 @@ primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
   , "primDepIMin"     |-> return "\\i f -> if i then f () else False"
   , "primIdFace"      |-> return "\\_ _ _ _ -> fst"
   , "primIdPath"      |-> return "\\_ _ _ _ -> snd"
-  , "primIdJ"         |-> return "\\_ _ _ _ _ x _ _ -> x"
   , builtinIdElim     |-> return
                             "\\_ _ _ _ _ f x y -> f (fst y) x (snd y)"
   ]
