@@ -4,6 +4,8 @@ module Agda.Compiler.ToTreeless
   , closedTermToTreeless
   ) where
 
+import Prelude hiding ((!!))
+
 import Control.Arrow        ( first )
 import Control.Monad        ( filterM, foldM, forM, zipWithM )
 import Control.Monad.Reader ( MonadReader(..), asks, ReaderT, runReaderT )
@@ -22,6 +24,7 @@ import Agda.Syntax.Treeless (TTerm, EvaluationStrategy, ArgUsage(..))
 
 import Agda.TypeChecking.CompiledClause as CC
 import qualified Agda.TypeChecking.CompiledClause.Compile as CC
+import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.EtaContract (binAppView, BinAppView(..))
 import Agda.TypeChecking.Monad as TCM
 import Agda.TypeChecking.Pretty
@@ -57,7 +60,7 @@ getCompiledClauses :: QName -> TCM CC.CompiledClauses
 getCompiledClauses q = do
   def <- getConstInfo q
   let cs = defClauses def
-      isProj | Function{ funProjection = proj } <- theDef def = isJust (projProper =<< proj)
+      isProj | Function{ funProjection = Right x } <- theDef def = isJust (projProper x)
              | otherwise = False
       translate | isProj    = CC.DontRunRecordPatternTranslation
                 | otherwise = CC.RunRecordPatternTranslation
@@ -253,6 +256,8 @@ casetree cc = do
     CC.Case (Arg _ n) (CC.Branches True conBrs _ _ Nothing _ _) -> lambdasUpTo n $ do
       mkRecord =<< traverse casetree (CC.content <$> conBrs)
     CC.Case (Arg i n) (CC.Branches False conBrs etaBr litBrs catchAll _ lazy) -> lambdasUpTo (n + 1) $ do
+      -- re #3733 TODO: revise when compiling --cubical
+      conBrs <- fmap Map.fromList $ filterM (isConstructor . fst) (Map.toList conBrs)
                     -- We can treat eta-matches as regular matches here.
       let conBrs' = caseMaybe etaBr conBrs $ \ (c, br) -> Map.insertWith (\ new old -> old) (conName c) br conBrs
       if Map.null conBrs' && Map.null litBrs then do
