@@ -12,6 +12,7 @@ import System.Exit (exitSuccess, exitWith, ExitCode(ExitFailure))
 data AgdaError = UnknownError      -- ^ 1
                | TCMError          -- ^ 42
                | OptionError       -- ^ 71
+               | CommandError      -- ^ 113
                | ImpossibleError   -- ^ 154
                deriving (Show, Eq, Enum, Bounded)
 
@@ -19,6 +20,7 @@ agdaErrorToInt :: AgdaError -> Int
 agdaErrorToInt UnknownError    = 1
 agdaErrorToInt TCMError        = 42
 agdaErrorToInt OptionError     = 71
+agdaErrorToInt CommandError    = 113
 agdaErrorToInt ImpossibleError = 154
 
 -- ^ Return the error corresponding to an exit code from the
@@ -30,5 +32,5 @@ agdaErrorFromInt = -- We implement this in a somewhat more inefficient
                                | error  <- [minBound..maxBound]
                                ]
 
-exitAgdaWith :: AgdaError -> IO ()
+exitAgdaWith :: AgdaError -> IO a
 exitAgdaWith = exitWith . ExitFailure . agdaErrorToInt
