@@ -6,6 +6,8 @@ module Agda.Interaction.Highlighting.Emacs
   , lispifyTokenBased
   ) where
 
+import Prelude hiding (null)
+
 import Agda.Interaction.Highlighting.Common
 import Agda.Interaction.Highlighting.Precise
 import Agda.Interaction.Highlighting.Range (Range(..))
@@ -14,12 +16,14 @@ import Agda.Interaction.Response
 import Agda.TypeChecking.Monad (HighlightingMethod(..), ModuleToSource)
 import Agda.Utils.FileName (filePath)
 import Agda.Utils.IO.TempFile (writeToTempFile)
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.String (quote)
 
 import qualified Data.List as List
 import qualified Data.Map as Map
 import Data.Maybe
 
+import Agda.Utils.Null
 import Agda.Utils.Impossible
 
 ------------------------------------------------------------------------
@@ -40,9 +44,9 @@ showAspects modFile (r, m) = L $
     dropNils (
       [lispifyTokenBased (tokenBased m)]
         ++
-      [A $ maybe "nil" quote $ note m]
+      [A $ ifNull (note m) "nil" quote]
         ++
-      (maybeToList $ fmap defSite $ definitionSite m))
+      maybeToList (defSite <$> definitionSite m))
   where
   defSite (DefinitionSite m p _ _) =
     Cons (A $ quote $ filePath f) (A $ show p)
@@ -81,7 +85,7 @@ lispifyHighlightingInfo h remove method modFile =
   info = (case remove of
                 RemoveHighlighting -> A "remove"
                 KeepHighlighting   -> A "nil") :
-             map (showAspects modFile) (ranges h)
+             map (showAspects modFile) (toList h)
 
   direct :: IO (Lisp String)
   direct = return $ L (A "agda2-highlight-add-annotations" :
@@ -89,7 +93,7 @@ lispifyHighlightingInfo h remove method modFile =
 
   indirect :: IO (Lisp String)
   indirect = do
-    filepath <- writeToTempFile (show $ L info)
+    filepath <- writeToTempFile (prettyShow $ L info)
     return $ L [ A "agda2-highlight-load-and-delete-action"
                , A (quote filepath)
                ]
