@@ -64,7 +64,9 @@ import qualified Agda.Syntax.Literal as L
 import qualified Agda.Syntax.Parser as Pa
 import qualified Agda.Syntax.Parser.Tokens as T
 import qualified Agda.Syntax.Position as P
-import Agda.Syntax.Position (Range, HasRange, getRange, noRange)
+import Agda.Syntax.Position
+  (RangeFile, Range, HasRange, getRange, noRange)
+import Agda.Syntax.TopLevelModuleName
 
 import Agda.Syntax.Scope.Base     ( WithKind(..) )
 import Agda.Syntax.Abstract.Views ( KName, declaredNames )
@@ -127,19 +129,19 @@ generateAndPrintSyntaxInfo
   -> TCM ()
 generateAndPrintSyntaxInfo decl _ _ | null $ getRange decl = return ()
 generateAndPrintSyntaxInfo decl hlLevel updateState = do
-  file <- getCurrentPath
+  top <- fromMaybe __IMPOSSIBLE__ <$> currentTopLevelModule
 
-  reportSLn "import.iface.create" 15 $ concat
-    [ "Generating syntax info for "
-    , filePath file
-    , case hlLevel of
-        Full   {} -> " (final)."
-        Partial{} -> " (first approximation)."
-    ]
+  reportSDoc "import.iface.create" 15 $
+    TCM.fwords
+      ("Generating syntax info for the following declaration " ++
+       case hlLevel of
+         Full   {} -> "(final):"
+         Partial{} -> "(first approximation):")
+      TCM.$$
+    TCM.prettyA decl
 
   ignoreAbstractMode $ do
-    modMap <- sourceToModule
-    kinds  <- nameKinds hlLevel decl
+    kinds <- nameKinds hlLevel decl
 
     -- After the code has been type checked more information may be
     -- available for overloaded constructors, and
@@ -147,11 +149,11 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     -- Note, however, that highlighting for overloaded constructors is
     -- included also in @nameInfo@.
     constructorInfo <- case hlLevel of
-      Full{} -> generateConstructorInfo modMap file kinds decl
+      Full{} -> generateConstructorInfo top kinds decl
       _      -> return mempty
 
     -- Main source of scope-checker generated highlighting:
-    let nameInfo = runHighlighter modMap file kinds decl
+    let nameInfo = runHighlighter top kinds decl
 
     reportSDoc "highlighting.warning" 60 $ TCM.hcat
       [ "current path = "
@@ -184,21 +186,31 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
 
 generateTokenInfo :: AbsolutePath -> TCM HighlightingInfo
 generateTokenInfo file =
-  generateTokenInfoFromSource file . Text.unpack =<<
-    runPM (Pa.readFilePM file)
+  generateTokenInfoFromSource rf . Text.unpack =<<
+    runPM (Pa.readFilePM rf)
+  where
+  -- Note the use of Nothing here. The file might not even parse, but
+  -- it should still be possible to obtain token-based highlighting
+  -- information. The top-level module names seem to be *mostly*
+  -- unused, but one cannot use __IMPOSSIBLE__ instead of Nothing,
+  -- because the top-level module names are used by interleaveRanges,
+  -- which is used by parseLiterateWithComments, which is used by
+  -- generateTokenInfoFromSource.
+  rf = P.mkRangeFile file Nothing
 
 -- | Generate and return the syntax highlighting information for the
 -- tokens in the given file.
 
 generateTokenInfoFromSource
-  :: AbsolutePath
+  :: RangeFile
      -- ^ The module to highlight.
   -> String
      -- ^ The file contents. Note that the file is /not/ read from
      -- disk.
   -> TCM HighlightingInfo
 generateTokenInfoFromSource file input =
-  runPM $ tokenHighlighting . fst <$> Pa.parseFile Pa.tokensParser file input
+  runPM $ tokenHighlighting . fst . fst <$>
+          Pa.parseFile Pa.tokensParser file input
 
 -- | Generate and return the syntax highlighting information for the
 -- tokens in the given string, which is assumed to correspond to the
@@ -207,7 +219,8 @@ generateTokenInfoFromSource file input =
 generateTokenInfoFromString :: Range -> String -> TCM HighlightingInfo
 generateTokenInfoFromString r _ | r == noRange = return mempty
 generateTokenInfoFromString r s = do
-  runPM $ tokenHighlighting <$> Pa.parsePosString Pa.tokensParser p s
+  runPM $ tokenHighlighting . fst <$>
+          Pa.parsePosString Pa.tokensParser p s
   where
     Just p = P.rStart r
 
@@ -323,12 +336,12 @@ instance Collection KName NameKindBuilder
 -- the type checker.
 
 generateConstructorInfo
-  :: SourceToModule  -- ^ Maps source file paths to module names.
-  -> AbsolutePath    -- ^ The module to highlight.
+  :: TopLevelModuleName
+     -- ^ The module to highlight.
   -> NameKinds
   -> A.Declaration
   -> TCM HighlightingInfoBuilder
-generateConstructorInfo modMap file kinds decl = do
+generateConstructorInfo top kinds decl = do
 
   -- Get boundaries of current declaration.
   -- @noRange@ should be impossible, but in case of @noRange@
@@ -345,7 +358,7 @@ generateConstructorInfo modMap file kinds decl = do
         constrs = IntMap.elems m2
 
     -- Return suitable syntax highlighting information.
-    return $ foldMap (runHighlighter modMap file kinds) constrs
+    return $ foldMap (runHighlighter top kinds) constrs
 
 printSyntaxInfo :: Range -> TCM ()
 printSyntaxInfo r = do
@@ -460,9 +473,11 @@ warningHighlighting' b w = case tcWarning w of
   RewriteMissingRule{}       -> confluenceErrorHighlighting w
   PragmaCompileErased{}      -> deadcodeHighlighting w
   NotInScopeW{}              -> deadcodeHighlighting w
+  UnsupportedIndexedMatch{}  -> mempty
   AsPatternShadowsConstructorOrPatternSynonym{}
                              -> deadcodeHighlighting w
   RecordFieldWarning w       -> recordFieldWarningHighlighting w
+  OptionWarning w            -> mempty
   ParseWarning w             -> case w of
     Pa.UnsupportedAttribute{}     -> deadcodeHighlighting w
     Pa.MultipleAttributes{}       -> deadcodeHighlighting w
@@ -482,6 +497,12 @@ warningHighlighting' b w = case tcWarning w of
     EmptyPrivate{}                   -> deadcodeHighlighting w
     EmptyGeneralize{}                -> deadcodeHighlighting w
     EmptyField{}                     -> deadcodeHighlighting w
+    HiddenGeneralize{}               -> mempty
+      -- Andreas, 2022-03-25, issue #5850
+      -- We would like @deadcodeHighlighting w@ for the braces in
+      -- @variable {x} : A@, but these have no range, so we cannot highlight them.
+      -- Highlighting the variable instead might be misleading,
+      -- suggesting that it is not generalized over.
     UselessAbstract{}                -> deadcodeHighlighting w
     UselessInstance{}                -> deadcodeHighlighting w
     UselessPrivate{}                 -> deadcodeHighlighting w
@@ -530,7 +551,7 @@ terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
     functionDefs = foldMap (\x -> H.singleton (rToR $ bindingSite x) m) $
                    concatMap termErrFunctions termErrs
     callSites    = foldMap (\r -> H.singleton (rToR r) m) $
-                   concatMap (map callInfoRange . termErrCalls) termErrs
+                   concatMap (map getRange . termErrCalls) termErrs
     bindingSite  = A.nameBindingSite . A.qnameName
 
 -- | Generate syntax highlighting for not-strictly-positive inductive
