@@ -20,12 +20,14 @@ import Prelude hiding (log)
 import Data.Bifunctor (second)
 import Data.Char
 import Data.Maybe
-import Data.Function
+import Data.Function (on)
 import Data.Foldable (toList)
 #if !(MIN_VERSION_base(4,11,0))
 import Data.Semigroup (Semigroup(..))
 #endif
+import qualified Data.Text as T
 
+import Control.Exception.Base (IOException, try)
 import Control.Monad (forM_, mapM_, unless, when)
 import Control.Monad.Trans.Reader as R ( ReaderT(runReaderT))
 import Control.Monad.RWS.Strict
@@ -59,22 +61,22 @@ import qualified Data.List    as List
 
 import Paths_Agda
 
-import Agda.Syntax.Abstract (toTopLevelModuleName)
 import Agda.Syntax.Common
-import Agda.Syntax.Concrete (TopLevelModuleName, moduleNameParts)
 import Agda.Syntax.Parser.Literate (literateTeX, LayerRole, atomizeLayers)
 import qualified Agda.Syntax.Parser.Literate as L
-import Agda.Syntax.Position (startPos)
+import Agda.Syntax.Position (RangeFile, startPos)
+import Agda.Syntax.TopLevelModuleName
+  (TopLevelModuleName, moduleNameParts)
 
 import Agda.Interaction.Highlighting.Precise hiding (toList)
 
-import Agda.TypeChecking.Monad (Interface(..))
+import Agda.TypeChecking.Monad (Interface(..)) --, reportSLn)
 
-import Agda.Utils.FileName (AbsolutePath)
 import Agda.Utils.Function (applyWhen)
 import Agda.Utils.Functor  ((<&>))
 import Agda.Utils.List     (last1, updateHead, updateLast)
 import Agda.Utils.Maybe    (whenJust)
+import Agda.Utils.Monad
 import qualified Agda.Utils.List1 as List1
 
 import Agda.Utils.Impossible
@@ -695,7 +697,7 @@ defaultStyFile = "agda.sty"
 
 data LaTeXOptions = LaTeXOptions
   { latexOptOutDir         :: FilePath
-  , latexOptSourceFileName :: Maybe AbsolutePath
+  , latexOptSourceFileName :: Maybe RangeFile
     -- ^ The parser uses a @Position@ which includes a source filename for
     -- error reporting and such. We don't actually get the source filename
     -- with an @Interface@, and it isn't necessary to look it up.
@@ -718,28 +720,45 @@ getTextWidthEstimator _countClusters =
 -- | Create the common base output directory and check for/install the style file.
 prepareCommonAssets :: (MonadLogLaTeX m, MonadIO m) => FilePath -> m ()
 prepareCommonAssets dir = do
-  liftIO $ createDirectoryIfMissing True dir
-  (code, _, _) <-
-    liftIO $
-      readProcessWithExitCode
+  -- Make sure @dir@ will exist.
+  dirExisted <- liftIO $ doesDirectoryExist dir
+  unless dirExisted $
+    -- Create directory @dir@ and parent directories.
+    liftIO $ createDirectoryIfMissing True dir
+
+  -- Check whether TeX will find @agda.sty@.
+  texFindsSty <- liftIO $ try $
+      readProcess
         "kpsewhich"
-        ["--path=" ++ dir, defaultStyFile]
+        (applyWhen dirExisted (("--path=" ++ dir) :) [defaultStyFile])
         ""
-  when (code /= ExitSuccess) $ do
-    logLaTeX $ LogMessage FileSystem
-      (T.pack $ unwords [defaultStyFile, "was not found. Copying a default version of", defaultStyFile, "into", dir])
-      []
-    liftIO $ do
-      styFile <- getDataFileName $
-        latexDataDir </> defaultStyFile
-      copyFile styFile (dir </> defaultStyFile)
+  case texFindsSty of
+    Right _ -> return ()
+    Left (e :: IOException) -> do
+     -- -- we are lacking MonadDebug here, so no debug printing via reportSLn
+     -- reportSLn "compile.latex.sty" 70 $ unlines
+     --   [ unwords [ "Searching for", defaultStyFile, "in", dir, "returns:" ]
+     --   , show e
+     --   ]
+     let agdaSty = dir </> defaultStyFile
+     unlessM (pure dirExisted `and2M` liftIO (doesFileExist agdaSty)) $ do
+       -- It is safe now to create the default style file in @dir@ without overwriting
+       -- a possibly user-edited copy there.
+       logLaTeX $ LogMessage FileSystem
+         (T.pack $ unwords [defaultStyFile, "was not found. Copying a default version of", defaultStyFile, "into", dir])
+         []
+       liftIO $ do
+         styFile <- getDataFileName $
+           latexDataDir </> defaultStyFile
+         copyFile styFile agdaSty
 
 -- | Generates a LaTeX file for the given interface.
 generateLaTeXIO :: (MonadLogLaTeX m, MonadIO m) => LaTeXOptions -> Interface -> m ()
 generateLaTeXIO opts i = do
   let textWidthEstimator = getTextWidthEstimator (latexOptCountClusters opts)
   let baseDir = latexOptOutDir opts
-  let outPath = baseDir </> (latexOutRelativeFilePath $ toTopLevelModuleName $ iModuleName i)
+  let outPath = baseDir </>
+                latexOutRelativeFilePath (iTopLevelModuleName i)
   latex <- E.encodeUtf8 <$> toLaTeX
               (emptyEnv textWidthEstimator)
               (latexOptSourceFileName opts)
@@ -750,7 +769,10 @@ generateLaTeXIO opts i = do
     BS.writeFile outPath latex
 
 latexOutRelativeFilePath :: TopLevelModuleName -> FilePath
-latexOutRelativeFilePath m = List.intercalate [pathSeparator] (List1.toList $ moduleNameParts m) <.> "tex"
+latexOutRelativeFilePath m =
+  List.intercalate [pathSeparator]
+    (map T.unpack $ List1.toList $ moduleNameParts m) <.>
+  "tex"
 
 groupByFst :: forall a b. Eq a => [(a,b)] -> [(a,[b])]
 groupByFst =
@@ -765,7 +787,7 @@ groupByFst =
 toLaTeX
   :: MonadLogLaTeX m
   => Env
-  -> Maybe AbsolutePath
+  -> Maybe RangeFile
   -> L.Text
   -> HighlightingInfo
   -> m L.Text
