@@ -1,4 +1,6 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE RecursiveDo #-}
 
 {-| This module deals with finding imported modules and loading their
     interface files.
@@ -35,14 +37,15 @@ import qualified Control.Exception as E
 import Control.Monad.Fail (MonadFail)
 #endif
 
-import Data.Either (lefts)
-import qualified Data.Map as Map
+import Data.Either
 import qualified Data.List as List
-import Data.Set (Set)
-import qualified Data.Set as Set
 import Data.Maybe
 import Data.Map (Map)
+import qualified Data.Map as Map
 import qualified Data.HashMap.Strict as HMap
+import qualified Data.HashSet as HSet
+import Data.Set (Set)
+import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text.Lazy as TL
 
@@ -53,12 +56,14 @@ import Agda.Benchmarking
 
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Concrete as C
+import Agda.Syntax.Concrete.Attribute
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
 import Agda.Syntax.Parser
 import Agda.Syntax.Position
 import Agda.Syntax.Scope.Base
-import Agda.Syntax.Translation.ConcreteToAbstract
+import Agda.Syntax.TopLevelModuleName
+import Agda.Syntax.Translation.ConcreteToAbstract as CToA
 
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Warnings hiding (warnings)
@@ -94,8 +99,10 @@ import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.IO.Binary
 import Agda.Utils.Pretty hiding (Mode)
+import qualified Agda.Utils.ProfileOptions as Profile
 import Agda.Utils.Hash
 import qualified Agda.Utils.Trie as Trie
+import Agda.Utils.WithDefault
 
 import Agda.Utils.Impossible
 
@@ -116,19 +123,27 @@ data Source = Source
   , srcFileType    :: FileType              -- ^ Source file type
   , srcOrigin      :: SourceFile            -- ^ Source location at the time of its parsing
   , srcModule      :: C.Module              -- ^ The parsed module.
-  , srcModuleName  :: C.TopLevelModuleName  -- ^ The top-level module name.
+  , srcModuleName  :: TopLevelModuleName    -- ^ The top-level module name.
   , srcProjectLibs :: [AgdaLibFile]         -- ^ The .agda-lib file(s) of the project this file belongs to.
+  , srcCohesion    :: !CohesionAttributes
+    -- ^ Every encountered occurrence of a cohesion attribute.
   }
 
 -- | Parses a source file and prepares the 'Source' record.
 
 parseSource :: SourceFile -> TCM Source
 parseSource sourceFile@(SourceFile f) = Bench.billTo [Bench.Parsing] $ do
-  source                <- runPM $ readFilePM f
-  (parsedMod, fileType) <- runPM $
-                           parseFile moduleParser f $ TL.unpack source
-  parsedModName         <- moduleName f parsedMod
-  libs <- getAgdaLibFiles $ takeDirectory $ filePath f
+  (source, fileType, parsedMod, coh, parsedModName) <- mdo
+    -- This piece of code uses mdo because the top-level module name
+    -- (parsedModName) is obtained from the parser's result, but it is
+    -- also used by the parser.
+    let rf = mkRangeFile f (Just parsedModName)
+    source                       <- runPM $ readFilePM rf
+    ((parsedMod, coh), fileType) <- runPM $ parseFile moduleParser rf $
+                                    TL.unpack source
+    parsedModName                <- moduleName f parsedMod
+    return (source, fileType, parsedMod, coh, parsedModName)
+  libs <- getAgdaLibFiles f parsedModName
   return Source
     { srcText        = source
     , srcFileType    = fileType
@@ -136,6 +151,7 @@ parseSource sourceFile@(SourceFile f) = Bench.billTo [Bench.Parsing] $ do
     , srcModule      = parsedMod
     , srcModuleName  = parsedModName
     , srcProjectLibs = libs
+    , srcCohesion    = coh
     }
 
 srcDefaultPragmas :: Source -> [OptionsPragma]
@@ -196,10 +212,12 @@ mergeInterface :: Interface -> TCM ()
 mergeInterface i = do
     let sig     = iSignature i
         builtin = Map.toAscList $ iBuiltin i
-        prim    = [ x | (_,Prim x) <- builtin ]
-        bi      = Map.fromDistinctAscList [ (x, Builtin t) | (x, Builtin t) <- builtin ]
-                    -- Andreas, 2021-08-19: this seeming identity filters out the @Prim@s
-                    -- and converts the type.
+        primOrBi = \case
+          (_, Prim x)                     -> Left x
+          (x, Builtin t)                  -> Right (x, Builtin t)
+          (x, BuiltinRewriteRelations xs) -> Right (x, BuiltinRewriteRelations xs)
+        (prim, bi') = partitionEithers $ map primOrBi builtin
+        bi      = Map.fromDistinctAscList bi'
         warns   = iWarnings i
     bs <- getsTC stBuiltinThings
     reportSLn "import.iface.merge" 10 "Merging interface"
@@ -209,9 +227,13 @@ mergeInterface i = do
     let check b (Builtin x) (Builtin y)
               | x == y    = return ()
               | otherwise = typeError $ DuplicateBuiltinBinding b x y
+        check _ (BuiltinRewriteRelations xs) (BuiltinRewriteRelations ys) = return ()
         check _ _ _ = __IMPOSSIBLE__
     sequence_ $ Map.intersectionWithKey check bs bi
-    addImportedThings sig bi
+    addImportedThings
+      sig
+      (iMetaBindings i)
+      bi
       (iPatternSyns i)
       (iDisplayForms i)
       (iUserWarnings i)
@@ -230,6 +252,7 @@ mergeInterface i = do
 
 addImportedThings
   :: Signature
+  -> RemoteMetaStore
   -> BuiltinThings PrimFun
   -> A.PatternSynDefns
   -> DisplayForms
@@ -237,8 +260,10 @@ addImportedThings
   -> Set QName             -- ^ Name of imported definitions which are partial
   -> [TCWarning]
   -> TCM ()
-addImportedThings isig ibuiltin patsyns display userwarn partialdefs warnings = do
+addImportedThings isig metas ibuiltin patsyns display userwarn
+                  partialdefs warnings = do
   stImports              `modifyTCLens` \ imp -> unionSignatures [imp, isig]
+  stImportedMetaStore    `modifyTCLens` HMap.union metas
   stImportedBuiltins     `modifyTCLens` \ imp -> Map.union imp ibuiltin
   stImportedUserWarnings `modifyTCLens` \ imp -> Map.union imp userwarn
   stImportedPartialDefs  `modifyTCLens` \ imp -> Set.union imp partialdefs
@@ -250,16 +275,18 @@ addImportedThings isig ibuiltin patsyns display userwarn partialdefs warnings =
 -- | Scope checks the given module. A proper version of the module
 -- name (with correct definition sites) is returned.
 
-scopeCheckImport :: ModuleName -> TCM (ModuleName, Map ModuleName Scope)
-scopeCheckImport x = do
+scopeCheckImport ::
+  TopLevelModuleName -> ModuleName ->
+  TCM (ModuleName, Map ModuleName Scope)
+scopeCheckImport top x = do
     reportSLn "import.scope" 5 $ "Scope checking " ++ prettyShow x
     verboseS "import.scope" 10 $ do
       visited <- prettyShow <$> getPrettyVisitedModules
       reportSLn "import.scope" 10 $ "  visited: " ++ visited
     -- Since scopeCheckImport is called from the scope checker,
     -- we need to reimburse her account.
-    i <- Bench.billTo [] $ getNonMainInterface (toTopLevelModuleName x) Nothing
-    addImport x
+    i <- Bench.billTo [] $ getNonMainInterface top Nothing
+    addImport top
 
     -- If that interface was supposed to raise a warning on import, do so.
     whenJust (iImportWarning i) $ warning . UserWarning
@@ -273,7 +300,7 @@ scopeCheckImport x = do
 -- used to find the interface and the computed interface is stored for
 -- potential later use.
 
-alreadyVisited :: C.TopLevelModuleName ->
+alreadyVisited :: TopLevelModuleName ->
                   MainInterface ->
                   PragmaOptions ->
                   TCM ModuleInfo ->
@@ -331,7 +358,7 @@ alreadyVisited x isMain currentOptions getModule =
       _                             -> storeDecodedModule mi
 
     reportS "warning.import" 10
-      [ "module: " ++ show (C.moduleNameParts x)
+      [ "module: " ++ show (moduleNameParts x)
       , "WarningOnImport: " ++ show (iImportWarning (miInterface mi))
       ]
 
@@ -383,33 +410,42 @@ typeCheckMain
 typeCheckMain mode src = do
   -- liftIO $ putStrLn $ "This is typeCheckMain " ++ prettyShow f
   -- liftIO . putStrLn . show =<< getVerbosity
-  reportSLn "import.main" 10 "Importing the primitive modules."
-  libdirPrim <- liftIO getPrimitiveLibDir
-  reportSLn "import.main" 20 $ "Library primitive dir = " ++ show libdirPrim
-  -- Turn off import-chasing messages.
-  -- We have to modify the persistent verbosity setting, since
-  -- getInterface resets the current verbosity settings to the persistent ones.
-  bracket_ (getsTC Lens.getPersistentVerbosity) Lens.putPersistentVerbosity $ do
-    Lens.modifyPersistentVerbosity (Trie.delete [])  -- set root verbosity to 0
 
-    -- We don't want to generate highlighting information for Agda.Primitive.
-    withHighlightingLevel None $
-      forM_ (Set.map (libdirPrim </>) Lens.primitiveModules) $ \f -> do
-        primSource <- parseSource (SourceFile $ mkAbsolute f)
-        checkModuleName' (srcModuleName primSource) (srcOrigin primSource)
-        void $ getNonMainInterface (srcModuleName primSource) (Just primSource)
+  -- For the main interface, we also remember the pragmas from the file
+  setOptionsFromSourcePragmas src
+  loadPrims <- optLoadPrimitives <$> pragmaOptions
 
-  reportSLn "import.main" 10 $ "Done importing the primitive modules."
+  when loadPrims $ do
+    reportSLn "import.main" 10 "Importing the primitive modules."
+    libdirPrim <- liftIO getPrimitiveLibDir
+    reportSLn "import.main" 20 $ "Library primitive dir = " ++ show libdirPrim
+    -- Turn off import-chasing messages.
+    -- We have to modify the persistent verbosity setting, since
+    -- getInterface resets the current verbosity settings to the persistent ones.
+
+    bracket_ (getsTC Lens.getPersistentVerbosity) Lens.putPersistentVerbosity $ do
+      Lens.modifyPersistentVerbosity
+        (Strict.Just . Trie.insert [] 0 . Strict.fromMaybe Trie.empty)
+        -- set root verbosity to 0
+
+      -- We don't want to generate highlighting information for Agda.Primitive.
+      withHighlightingLevel None $
+        forM_ (Set.map (libdirPrim </>) Lens.primitiveModules) $ \f -> do
+          primSource <- parseSource (SourceFile $ mkAbsolute f)
+          checkModuleName' (srcModuleName primSource) (srcOrigin primSource)
+          void $ getNonMainInterface (srcModuleName primSource) (Just primSource)
+
+    reportSLn "import.main" 10 $ "Done importing the primitive modules."
 
   -- Now do the type checking via getInterface.
   checkModuleName' (srcModuleName src) (srcOrigin src)
 
-  -- For the main interface, we also remember the pragmas from the file
-  setOptionsFromSourcePragmas src
-
   mi <- getInterface (srcModuleName src) (MainInterface mode) (Just src)
 
-  stCurrentModule `setTCLens` Just (iModuleName (miInterface mi))
+  stCurrentModule `setTCLens'`
+    Just ( iModuleName (miInterface mi)
+         , iTopLevelModuleName (miInterface mi)
+         )
 
   return $ CheckResult' mi src
   where
@@ -428,7 +464,7 @@ typeCheckMain mode src = do
 --   Do not use this for the main file, use 'typeCheckMain' instead.
 
 getNonMainInterface
-  :: C.TopLevelModuleName
+  :: TopLevelModuleName
   -> Maybe Source
      -- ^ Optional: the source code and some information about the source code.
   -> TCM Interface
@@ -445,7 +481,7 @@ getNonMainInterface x msrc = do
 -- errors.
 
 getInterface
-  :: C.TopLevelModuleName
+  :: TopLevelModuleName
   -> MainInterface
   -> Maybe Source
      -- ^ Optional: the source code and some information about the source code.
@@ -503,25 +539,25 @@ getInterface x isMain msrc =
 -- | Check if the options used for checking an imported module are
 --   compatible with the current options. Raises Non-fatal errors if
 --   not.
-checkOptionsCompatible :: PragmaOptions -> PragmaOptions -> ModuleName -> TCM Bool
+checkOptionsCompatible ::
+  PragmaOptions -> PragmaOptions -> TopLevelModuleName -> TCM Bool
 checkOptionsCompatible current imported importedModule = flip execStateT True $ do
   reportSDoc "import.iface.options" 5 $ P.nest 2 $ "current options  =" P.<+> showOptions current
   reportSDoc "import.iface.options" 5 $ P.nest 2 $ "imported options =" P.<+> showOptions imported
-  forM_ coinfectiveOptions $ \ (opt, optName) -> do
-    unless (opt current `implies` opt imported) $ do
+  forM_ infectiveCoinfectiveOptions $ \opt -> do
+    unless (icOptionOK opt current imported) $ do
       put False
-      warning (CoInfectiveImport optName importedModule)
-  forM_ infectiveOptions $ \ (opt, optName) -> do
-    unless (opt imported `implies` opt current) $ do
-      put False
-      warning (InfectiveImport optName importedModule)
+      warning $
+        (case icOptionKind opt of
+           Infective   -> InfectiveImport
+           Coinfective -> CoInfectiveImport)
+        (icOptionWarning opt importedModule)
   where
-    implies :: Bool -> Bool -> Bool
-    p `implies` q = p <= q
-
-    showOptions opts = P.prettyList (map (\ (o, n) -> (P.text n <> ": ") P.<+> P.pretty (o opts))
-                                 (coinfectiveOptions ++ infectiveOptions))
-
+  showOptions opts =
+    P.prettyList $
+    map (\opt -> (P.text (icOptionDescription opt) <> ": ") P.<+>
+                 P.pretty (icOptionActive opt opts))
+      infectiveCoinfectiveOptions
 
 -- | Compare options and return collected warnings.
 -- | Returns `Nothing` if warning collection was skipped.
@@ -532,14 +568,15 @@ getOptionsCompatibilityWarnings isMain isPrim currentOptions i = runMaybeT $ exc
   -- They weren't logged before, but they're nice for documenting the early returns.
   when isPrim $
     throwError "Options consistency checking disabled for always-available primitive module"
-  whenM (lift $ checkOptionsCompatible currentOptions (iOptionsUsed i) (iModuleName i)) $
+  whenM (lift $ checkOptionsCompatible currentOptions (iOptionsUsed i)
+                  (iTopLevelModuleName i)) $
     throwError "No warnings to collect because options were compatible"
   lift $ getAllWarnings' isMain ErrorWarnings
 
 -- | Try to get the interface from interface file or cache.
 
 getStoredInterface
-  :: C.TopLevelModuleName
+  :: TopLevelModuleName
      -- ^ Module name of file we process.
   -> SourceFile
      -- ^ File we process.
@@ -634,7 +671,7 @@ getStoredInterface x file msrc = do
           readInterface ifile
 
         -- Ensure that the given module name matches the one in the file.
-        let topLevelName = toTopLevelModuleName $ iModuleName i
+        let topLevelName = iTopLevelModuleName i
         unless (topLevelName == x) $
           -- Andreas, 2014-03-27 This check is now done in the scope checker.
           -- checkModuleName topLevelName file
@@ -644,7 +681,9 @@ getStoredInterface x file msrc = do
 
         lift $ chaseMsg "Loading " x $ Just ifp
         -- print imported warnings
-        let ws = filter ((Strict.Just (srcFilePath file) ==) . tcWarningOrigin) (iWarnings i)
+        let ws = filter ((Strict.Just (Just x) ==) .
+                         fmap rangeFileName . tcWarningOrigin) $
+                 iWarnings i
         unless (null ws) $ reportSDoc "warning" 1 $ P.vcat $ P.prettyTCM <$> ws
 
         loadDecodedModule file $ ModuleInfo
@@ -673,23 +712,16 @@ loadDecodedModule file mi = do
   -- want the pragmas to apply to interactive commands in the UI.
   -- Jesper, 2021-04-18: Check for changed options in library files!
   -- (see #5250)
-  libOptions <- lift $ getLibraryOptions $ takeDirectory fp
+  libOptions <- lift $ getLibraryOptions
+    (srcFilePath file)
+    (iTopLevelModuleName i)
   lift $ mapM_ setOptionsFromPragma (libOptions ++ iFilePragmaOptions i)
 
   -- Check that options that matter haven't changed compared to
   -- current options (issue #2487)
   unlessM (lift $ Lens.isBuiltinModule fp) $ do
-    currentOptions <- useTC stPragmaOptions
-    let disagreements =
-          [ optName | (opt, optName) <- restartOptions,
-                      opt currentOptions /= opt (iOptionsUsed i) ]
-    unless (null disagreements) $ do
-      reportSLn "import.iface.options" 4 $ concat
-        [ "  Changes in the following options in "
-        , prettyShow fp
-        , ", re-typechecking: "
-        , prettyShow disagreements
-        ]
+    current <- useTC stPragmaOptions
+    when (recheckBecausePragmaOptionsChanged (iOptionsUsed i) current) $
       throwError "options changed"
 
   -- If any of the imports are newer we need to retype check
@@ -721,7 +753,7 @@ loadDecodedModule file mi = do
 --   in order to forget some state changes after successful type checking.
 
 createInterfaceIsolated
-  :: C.TopLevelModuleName
+  :: TopLevelModuleName
      -- ^ Module name of file we process.
   -> SourceFile
      -- ^ File we process.
@@ -739,6 +771,7 @@ createInterfaceIsolated x file msrc = do
       ds          <- getDecodedModules
       opts        <- stPersistentOptions . stPersistentState <$> getTC
       isig        <- useTC stImports
+      metas       <- useTC stImportedMetaStore
       ibuiltin    <- useTC stImportedBuiltins
       display     <- useTC stImportsDisplayForms
       userwarn    <- useTC stImportedUserWarnings
@@ -768,7 +801,8 @@ createInterfaceIsolated x file msrc = do
                setInteractionOutputCallback ho
                stModuleToSource `setTCLens` mf
                setVisitedModules vs
-               addImportedThings isig ibuiltin ipatsyns display userwarn partialdefs []
+               addImportedThings isig metas ibuiltin ipatsyns display
+                 userwarn partialdefs []
 
                r  <- createInterface x file NotMainInterface msrc
                mf' <- useTC stModuleToSource
@@ -800,7 +834,7 @@ createInterfaceIsolated x file msrc = do
 
 chaseMsg
   :: String               -- ^ The prefix, like @Checking@, @Finished@, @Loading @.
-  -> C.TopLevelModuleName -- ^ The module name.
+  -> TopLevelModuleName   -- ^ The module name.
   -> Maybe String         -- ^ Optionally: the file name.
   -> TCM ()
 chaseMsg kind x file = do
@@ -897,7 +931,7 @@ writeInterface file i = let fp = filePath file in do
 -- information.
 
 createInterface
-  :: C.TopLevelModuleName  -- ^ The expected module name.
+  :: TopLevelModuleName    -- ^ The expected module name.
   -> SourceFile            -- ^ The file to type check.
   -> MainInterface         -- ^ Are we dealing with the main module?
   -> Maybe Source      -- ^ Optional information about the source code.
@@ -912,7 +946,9 @@ createInterface mname file isMain msrc = do
        (chaseMsg checkMsg x $ Just fp)
        (const $ do ws <- getAllWarnings AllWarnings
                    let classified = classifyWarnings ws
-                   let wa' = filter ((Strict.Just (srcFilePath file) ==) . tcWarningOrigin) (tcWarnings classified)
+                   let wa' = filter ((Strict.Just (Just mname) ==) .
+                                     fmap rangeFileName . tcWarningOrigin) $
+                             tcWarnings classified
                    unless (null wa') $
                      reportSDoc "warning" 1 $ P.vcat $ P.prettyTCM <$> wa'
                    when (null (nonFatalErrors classified)) $ chaseMsg "Finished" x Nothing)
@@ -934,10 +970,16 @@ createInterface mname file isMain msrc = do
     let srcPath = srcFilePath $ srcOrigin src
 
     fileTokenInfo <- Bench.billTo [Bench.Highlighting] $
-      generateTokenInfoFromSource srcPath (TL.unpack $ srcText src)
+      generateTokenInfoFromSource
+        (let !top = srcModuleName src in
+         mkRangeFile srcPath (Just top))
+        (TL.unpack $ srcText src)
     stTokens `modifyTCLens` (fileTokenInfo <>)
 
     setOptionsFromSourcePragmas src
+    checkCohesionAttributes (srcCohesion src)
+    syntactic <- optSyntacticEquality <$> pragmaOptions
+    localTC (\env -> env { envSyntacticEqualityFuel = syntactic }) $ do
 
     verboseS "import.iface.create" 15 $ do
       nestingLevel      <- asksTC (pred . length . envImportPath)
@@ -1006,9 +1048,9 @@ createInterface mname file isMain msrc = do
     unfreezeMetas
 
     -- Profiling: Count number of metas.
-    verboseS "profile.metas" 10 $ do
-      MetaId n <- fresh
-      tickN "metas" (fromIntegral n)
+    whenProfile Profile.Metas $ do
+      m <- fresh
+      tickN "metas" (fromIntegral (metaId m))
 
     -- Highlighting from type checker.
     reportSLn "import.iface.create" 7 "Starting highlighting from type info."
@@ -1112,13 +1154,13 @@ createInterface mname file isMain msrc = do
     -- stMetaStore `setTCLens` savedMetaStore
 
     -- Profiling: Print statistics.
-    printStatistics 30 (Just mname) =<< getStatistics
+    printStatistics (Just mname) =<< getStatistics
 
     -- Get the statistics of the current module
     -- and add it to the accumulated statistics.
     localStatistics <- getStatistics
     lensAccumStatistics `modifyTCLens` Map.unionWith (+) localStatistics
-    verboseS "profile" 1 $ reportSLn "import.iface" 5 "Accumulated statistics."
+    reportSLn "import.iface" 5 "Accumulated statistics."
 
     isPrimitiveModule <- Lens.isPrimitiveModule (filePath srcPath)
 
@@ -1158,7 +1200,7 @@ buildInterface
   -> TCM Interface
 buildInterface src topLevel = do
     reportSLn "import.iface" 5 "Building interface..."
-    let mname = topLevelModuleName topLevel
+    let mname = CToA.topLevelModuleName topLevel
         source   = srcText src
         fileType = srcFileType src
         defPragmas = srcDefaultPragmas src
@@ -1174,20 +1216,23 @@ buildInterface src topLevel = do
     -- faster and interface file sizes a bit smaller, at least for the
     -- standard library).
     builtin     <- useTC stLocalBuiltins
-    ms          <- getImports
-    mhs         <- mapM (\ m -> (m,) <$> moduleHash m) $ Set.toList ms
+    mhs         <- mapM (\top -> (top,) <$> moduleHash top) .
+                   HSet.toList =<<
+                   useR stImportedModules
     foreignCode <- useTC stForeignCode
     -- Ulf, 2016-04-12:
     -- Non-closed display forms are not applicable outside the module anyway,
     -- and should be dead-code eliminated (#1928).
     origDisplayForms <- HMap.filter (not . null) . HMap.map (filter isClosed) <$> useTC stImportsDisplayForms
     -- TODO: Kill some ranges?
-    (display, sig) <- eliminateDeadCode origDisplayForms =<< getSignature
-    userwarns      <- useTC stLocalUserWarnings
-    importwarn     <- useTC stWarningOnImport
-    syntaxInfo     <- useTC stSyntaxInfo
-    optionsUsed    <- useTC stPragmaOptions
-    partialDefs    <- useTC stLocalPartialDefs
+    (display, sig, solvedMetas) <-
+      eliminateDeadCode builtin origDisplayForms ==<<
+        (getSignature, useR stSolvedMetaStore)
+    userwarns   <- useTC stLocalUserWarnings
+    importwarn  <- useTC stWarningOnImport
+    syntaxInfo  <- useTC stSyntaxInfo
+    optionsUsed <- useTC stPragmaOptions
+    partialDefs <- useTC stLocalPartialDefs
 
     -- Andreas, 2015-02-09 kill ranges in pattern synonyms before
     -- serialization to avoid error locations pointing to external files
@@ -1195,31 +1240,39 @@ buildInterface src topLevel = do
     patsyns <- killRange <$> getPatternSyns
     let builtin' = Map.mapWithKey (\ x b -> (x,) . primFunName <$> b) builtin
     warnings <- getAllWarnings AllWarnings
-    reportSLn "import.iface" 7 "  instantiating all meta variables"
-    -- Note that the meta-variables in the definitions in "sig" have
-    -- already been instantiated (by eliminateDeadCode).
-    i <- instantiateFullExceptForDefinitions Interface
-      { iSourceHash      = hashText source
-      , iSource          = source
-      , iFileType        = fileType
-      , iImportedModules = mhs
-      , iModuleName      = mname
-      , iScope           = empty -- publicModules scope
-      , iInsideScope     = topLevelScope topLevel
-      , iSignature       = sig
-      , iDisplayForms    = display
-      , iUserWarnings    = userwarns
-      , iImportWarning   = importwarn
-      , iBuiltin         = builtin'
-      , iForeignCode     = foreignCode
-      , iHighlighting    = syntaxInfo
-      , iDefaultPragmaOptions = defPragmas
-      , iFilePragmaOptions    = filePragmas
-      , iOptionsUsed     = optionsUsed
-      , iPatternSyns     = patsyns
-      , iWarnings        = warnings
-      , iPartialDefs     = partialDefs
-      }
+    let i = Interface
+          { iSourceHash      = hashText source
+          , iSource          = source
+          , iFileType        = fileType
+          , iImportedModules = mhs
+          , iModuleName      = mname
+          , iTopLevelModuleName = srcModuleName src
+          , iScope           = empty -- publicModules scope
+          , iInsideScope     = topLevelScope topLevel
+          , iSignature       = sig
+          , iMetaBindings    = solvedMetas
+          , iDisplayForms    = display
+          , iUserWarnings    = userwarns
+          , iImportWarning   = importwarn
+          , iBuiltin         = builtin'
+          , iForeignCode     = foreignCode
+          , iHighlighting    = syntaxInfo
+          , iDefaultPragmaOptions = defPragmas
+          , iFilePragmaOptions    = filePragmas
+          , iOptionsUsed     = optionsUsed
+          , iPatternSyns     = patsyns
+          , iWarnings        = warnings
+          , iPartialDefs     = partialDefs
+          }
+    i <-
+      ifM (collapseDefault . optSaveMetas <$> pragmaOptions)
+        (return i)
+        (do reportSLn "import.iface" 7
+              "  instantiating all meta variables"
+            -- Note that the meta-variables in the definitions in
+            -- "sig" have already been instantiated (by
+            -- eliminateDeadCode).
+            instantiateFullExceptForDefinitions i)
     reportSLn "import.iface" 7 "  interface complete"
     return i
 
@@ -1234,5 +1287,5 @@ getInterfaceFileHashes fp = do
   maybe 0 (uncurry (+)) hs `seq` close
   return hs
 
-moduleHash :: ModuleName -> TCM Hash
-moduleHash m = iFullHash <$> getNonMainInterface (toTopLevelModuleName m) Nothing
+moduleHash :: TopLevelModuleName -> TCM Hash
+moduleHash m = iFullHash <$> getNonMainInterface m Nothing
