@@ -25,7 +25,7 @@ import Control.Monad.STM
 import Control.Monad.Trans          ( lift )
 
 import qualified Data.Char as Char
-import Data.Function
+import Data.Function (on)
 import qualified Data.List as List
 import qualified Data.Map as Map
 import Data.Maybe
@@ -39,7 +39,7 @@ import qualified Agda.TypeChecking.Monad as TCM
 import qualified Agda.TypeChecking.Pretty as TCP
 import Agda.TypeChecking.Rules.Term (checkExpr, isType_)
 import Agda.TypeChecking.Errors
-import Agda.TypeChecking.Warnings (runPM)
+import Agda.TypeChecking.Warnings (runPM, warning)
 
 import Agda.Syntax.Fixity
 import Agda.Syntax.Position
@@ -53,8 +53,10 @@ import Agda.Syntax.Info (mkDefInfo)
 import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Translation.AbstractToConcrete hiding (withScope)
 import Agda.Syntax.Scope.Base
+import Agda.Syntax.TopLevelModuleName
 
 import Agda.Interaction.Base
+import Agda.Interaction.ExitCode
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import Agda.Interaction.Options.Lenses as Lenses
@@ -81,6 +83,7 @@ import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Pretty hiding (Mode)
+import qualified Agda.Utils.ProfileOptions as Profile
 import Agda.Utils.Singleton
 import Agda.Utils.String
 import Agda.Utils.Time
@@ -233,7 +236,7 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
       let handle e =
             Right <$>
               toIO (handleErr (Just Direct) $
-                        Exception noRange $ text $ show e)
+                        Exception noRange $ text $ E.displayException e)
 
           asyncHandler e@AsyncCancelled = return (Left e)
 
@@ -263,7 +266,8 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
 
         showImpl <- lift $ optShowImplicit <$> useTC stPragmaOptions
         showIrr <- lift $ optShowIrrelevant <$> useTC stPragmaOptions
-        unless noError $ mapM_ putResponse $
+        unless noError $ do
+          mapM_ putResponse $
             [ Resp_DisplayInfo $ Info_Error $ Info_GenericError e ] ++
             tellEmacsToJumpToError (getRange e) ++
             [ Resp_HighlightingInfo info KeepHighlighting
@@ -272,6 +276,8 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
                                    , sShowImplicitArguments = showImpl
                                    , sShowIrrelevantArguments = showIrr
                                    } ]
+          whenM (optExitOnError <$> commandLineOptions) $
+            liftIO $ exitAgdaWith TCMError
 
 -- | Run an 'IOTCM' value, catch the exceptions, emit output
 --
@@ -280,15 +286,17 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
 --   loaded interfaces for example).
 
 runInteraction :: IOTCM -> CommandM ()
-runInteraction (IOTCM current highlighting highlightingMethod cmd) =
+runInteraction iotcm =
   handleCommand inEmacs onFail $ do
     currentAbs <- liftIO $ absolute current
-    -- Raises an error if the given file is not the one currently
-    -- loaded.
-    cf <- gets theCurrentFile
-    when (not (independent cmd) && Just currentAbs /= (currentFilePath <$> cf)) $ do
+    cf  <- gets theCurrentFile
+    cmd <- if independent cmd then return cmd else do
+      when (Just currentAbs /= (currentFilePath <$> cf)) $ do
         let mode = TypeCheck
         cmd_load' current [] True mode $ \_ -> return ()
+      cf <- fromMaybe __IMPOSSIBLE__ <$> gets theCurrentFile
+      return $ case iotcm (Just (currentFileModule cf)) of
+        IOTCM _ _ _ cmd -> cmd
 
     withCurrentFile $ interpret cmd
 
@@ -299,6 +307,10 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd) =
         putResponse . Resp_InteractionPoints =<< gets theInteractionPoints
 
   where
+    -- The ranges in cmd might be incorrect because of the use of
+    -- Nothing here. That is taken care of above.
+    IOTCM current highlighting highlightingMethod cmd = iotcm Nothing
+
     inEmacs :: forall a. CommandM a -> CommandM a
     inEmacs = liftCommandMT $ withEnv $ initEnv
             { envHighlightingLevel  = highlighting
@@ -337,13 +349,14 @@ maybeAbort m = do
       tcState <- getTC
       tcEnv   <- askTC
       result  <- liftIO $ race
-                   (runTCM tcEnv tcState $ runStateT (m c) commandState)
+                   (runTCM tcEnv tcState $
+                    runStateT (m c) commandState)
                    (waitForAbort n q)
       case result of
         Left ((x, commandState'), tcState') -> do
           putTC tcState'
           put commandState'
-          case c of
+          case c Nothing of
             IOTCM _ _ _ Cmd_exit -> do
               putResponse Resp_DoneExiting
               return Done
@@ -420,7 +433,7 @@ initialiseCommandQueue next = do
       readCommands n = do
         c <- next
         case c of
-          Command (IOTCM _ _ _ Cmd_abort) -> do
+          Command c | IOTCM _ _ _ Cmd_abort <- c Nothing -> do
             atomically $ writeTVar abort (Just n)
             readCommands n
           _ -> do
@@ -455,6 +468,7 @@ updateInteractionPointsAfter Cmd_load{}                          = True
 updateInteractionPointsAfter Cmd_compile{}                       = True
 updateInteractionPointsAfter Cmd_constraints{}                   = False
 updateInteractionPointsAfter Cmd_metas{}                         = False
+updateInteractionPointsAfter Cmd_no_metas{}                      = False
 updateInteractionPointsAfter Cmd_show_module_contents_toplevel{} = False
 updateInteractionPointsAfter Cmd_search_about_toplevel{}         = False
 updateInteractionPointsAfter Cmd_solveAll{}                      = True
@@ -524,6 +538,11 @@ interpret (Cmd_metas norm) = do
   ms <- lift $ B.getGoals' norm (max Simplified norm)
   display_info . Info_AllGoalsWarnings ms =<< lift B.getWarningsAndNonFatalErrors
 
+interpret Cmd_no_metas = do
+  metas <- getOpenMetas
+  unless (null metas) $
+    typeError $ GenericError "Unsolved meta-variables"
+
 interpret (Cmd_show_module_contents_toplevel norm s) =
   atTopLevel $ showModuleContents norm noRange s
 
@@ -846,7 +865,7 @@ solveInstantiatedGoals norm mii = do
   putResponse $ Resp_SolveAll out
   where
       prt (i, m, e) = do
-        mi <- getMetaInfo <$> lookupMeta m
+        mi <- getMetaInfo <$> lookupLocalMeta m
         e' <- withMetaInfo mi $ abstractToConcreteCtx TopCtx e
         return (i, e')
 
@@ -907,9 +926,10 @@ cmd_load' file argv unsolvedOK mode cmd = do
     -- choice of whether or not to display implicit arguments.
     opts0 <- gets optionsOnReload
     backends <- useTC stBackends
-    z <- runOptM $ parseBackendOptions backends argv opts0
+    let (z, warns) = runOptM $ parseBackendOptions backends argv opts0
+    mapM_ (lift . warning . OptionWarning) warns
     case z of
-      Left err   -> lift $ typeError $ GenericError err
+      Left err -> lift $ typeError $ GenericError err
       Right (_, opts) -> do
         opts <- lift $ addTrustedExecutables opts
         let update o = o { optAllowUnsolved = unsolvedOK && optAllowUnsolved o}
@@ -928,7 +948,12 @@ cmd_load' file argv unsolvedOK mode cmd = do
     when (t == t') $ do
       is <- lift $ sortInteractionPoints =<< getInteractionPoints
       modify $ \st -> st { theInteractionPoints = is
-                         , theCurrentFile       = Just $ CurrentFile fp argv t
+                         , theCurrentFile       = Just $ CurrentFile
+                             { currentFilePath   = fp
+                             , currentFileModule = Imp.srcModuleName src
+                             , currentFileArgs   = argv
+                             , currentFileStamp  = t
+                             }
                          }
 
     cmd ok
@@ -1091,10 +1116,10 @@ searchAbout norm rg names = do
 
 whyInScope :: String -> CommandM ()
 whyInScope s = do
-  Just (CurrentFile file _ _) <- gets theCurrentFile
-  let cwd = takeDirectory (filePath file)
-  (v, xs, ms) <- liftLocalState (B.whyInScope s)
-  display_info $ Info_WhyInScope s cwd v xs ms
+  Just file <- gets theCurrentFile
+  let cwd = takeDirectory (filePath $ currentFilePath file)
+  why <- liftLocalState $ B.whyInScope cwd s
+  display_info $ Info_WhyInScope why
 
 -- | Sets the command line options and updates the status information.
 
@@ -1118,16 +1143,13 @@ status = do
   -- changed since. Note: This code does not check if any dependencies
   -- have changed, and uses a time stamp to check for changes.
   checked  <- lift $ case cf of
-    Nothing     -> return False
-    Just (CurrentFile f _ t) -> do
-      t' <- liftIO $ getModificationTime $ filePath f
-      if t == t'
+    Nothing -> return False
+    Just f  -> do
+      t <- liftIO $ getModificationTime $ filePath (currentFilePath f)
+      if currentFileStamp f == t
         then
-          do
-            mm <- lookupModuleFromSource f
-            case mm of
-              Nothing -> return False -- work-around for Issue1007
-              Just m  -> maybe False (null . miWarnings) <$> getVisitedModule m
+          maybe False (null . miWarnings) <$>
+          getVisitedModule (currentFileModule f)
         else
             return False
 
@@ -1164,13 +1186,14 @@ parseAndDoAtToplevel
   -> CommandM (Maybe CPUTime, a)
 parseAndDoAtToplevel cmd s = do
   localStateCommandM $ do
-    e <- lift $ runPM $ parse exprParser s
+    (e, coh) <- lift $ runPM $ parse exprParser s
+    lift $ checkCohesionAttributes coh
     maybeTimed $ atTopLevel $ lift $
       cmd =<< concreteToAbstract_ e
 
 maybeTimed :: CommandM a -> CommandM (Maybe CPUTime, a)
 maybeTimed work = do
-  doTime <- lift $ hasVerbosity "profile.interactive" 10
+  doTime <- lift $ hasProfileOption Profile.Interactive
   if not doTime
     then (Nothing,) <$> work
     else do
@@ -1194,4 +1217,4 @@ tellEmacsToJumpToError r =
     Nothing                                           -> []
     Just (Pn { srcFile = Strict.Nothing })            -> []
     Just (Pn { srcFile = Strict.Just f, posPos = p }) ->
-       [ Resp_JumpToError (filePath f) p ]
+       [ Resp_JumpToError (filePath (rangeFilePath f)) p ]
