@@ -30,6 +30,7 @@ import Control.Monad ( guard, when )
 import Text.Read ( readMaybe )
 import Data.Set (Set)
 import qualified Data.Set as Set
+import qualified Data.HashMap.Strict as HMap
 import Data.List ( stripPrefix, intercalate )
 
 import GHC.Generics (Generic)
@@ -166,9 +167,10 @@ usualWarnings = allWarnings Set.\\ Set.fromList
 -- to existing warnings in the codebase.
 
 data WarningName
-  =
+  -- Option Warnings
+  = OptionRenamed_
   -- Parser Warnings
-    OverlappingTokensWarning_
+  | OverlappingTokensWarning_
   | UnsupportedAttribute_
   | MultipleAttributes_
   -- Library Warnings
@@ -186,6 +188,7 @@ data WarningName
   | EmptyPrivate_
   | EmptyRewritePragma_
   | EmptyWhere_
+  | HiddenGeneralize_
   | InvalidCatchallPragma_
   | InvalidConstructor_
   | InvalidConstructorBlock_
@@ -231,6 +234,7 @@ data WarningName
   | NoGuardednessFlag_
   | NotInScope_
   | NotStrictlyPositive_
+  | UnsupportedIndexedMatch_
   | OldBuiltin_
   | PragmaCompileErased_
   | RewriteMaybeNonConfluent_
@@ -273,9 +277,9 @@ instance NFData WarningName
 -- | The flag corresponding to a warning is precisely the name of the constructor
 -- minus the trailing underscore.
 
--- sorry
 string2WarningName :: String -> Maybe WarningName
-string2WarningName = readMaybe . (++ "_")
+string2WarningName = (`HMap.lookup` warnings) where
+  warnings = HMap.fromList $ map (\x -> (warningName2String x, x)) [minBound..maxBound]
 
 warningName2String :: WarningName -> String
 warningName2String = initWithDefault __IMPOSSIBLE__ . show
@@ -321,6 +325,8 @@ usageWarning = intercalate "\n"
 
 warningNameDescription :: WarningName -> String
 warningNameDescription = \case
+  -- Option Warnings
+  OptionRenamed_                   -> "Renamed options."
   -- Parser Warnings
   OverlappingTokensWarning_        -> "Multi-line comments spanning one or more literate text blocks."
   UnsupportedAttribute_            -> "Unsupported attributes."
@@ -340,6 +346,7 @@ warningNameDescription = \case
   EmptyPrivate_                    -> "Empty `private' blocks."
   EmptyRewritePragma_              -> "Empty `REWRITE' pragmas."
   EmptyWhere_                      -> "Empty `where' blocks."
+  HiddenGeneralize_                -> "Hidden identifiers in variable blocks."
   InvalidCatchallPragma_           -> "`CATCHALL' pragmas before a non-function clause."
   InvalidConstructor_              -> "`constructor' blocks may only contain type signatures for constructors."
   InvalidConstructorBlock_         -> "No `constructor' blocks outside of `interleaved mutual' blocks."
@@ -389,6 +396,7 @@ warningNameDescription = \case
   FixityInRenamingModule_          -> "Found fixity annotation in renaming directive for module."
   NotInScope_                      -> "Out of scope name."
   NotStrictlyPositive_             -> "Failed strict positivity checks."
+  UnsupportedIndexedMatch_         -> "Failed to compute full equivalence when splitting on indexed family."
   OldBuiltin_                      -> "Deprecated `BUILTIN' pragmas."
   PragmaCompileErased_             -> "`COMPILE' pragma targeting an erased symbol."
   RewriteMaybeNonConfluent_        -> "Failed local confluence check while computing overlap."
@@ -413,7 +421,7 @@ warningNameDescription = \case
   UnsolvedMetaVariables_           -> "Unsolved meta variables."
   UserWarning_                     -> "User-defined warning added using one of the 'WARNING_ON_*' pragmas."
   WithoutKFlagPrimEraseEquality_   -> "`primEraseEquality' usages with the without-K flags."
-  WrongInstanceDeclaration_        -> "Terms marked as eligible for instance search should end with a name."
+  WrongInstanceDeclaration_        -> "Instances that do not adhere to the required format."
   -- Checking consistency of options
   CoInfectiveImport_               -> "Importing a file not using e.g. `--safe'  from one which does."
   InfectiveImport_                 -> "Importing a file using e.g. `--cubical' into one which doesn't."
