@@ -5,6 +5,7 @@ module Agda.Main where
 
 import Prelude hiding (null)
 
+import qualified Control.Exception as E
 import Control.Monad          ( void )
 import Control.Monad.Except   ( MonadError(..), ExceptT(..), runExceptT )
 import Control.Monad.IO.Class ( MonadIO(..) )
@@ -13,7 +14,9 @@ import qualified Data.List as List
 import Data.Maybe
 
 import System.Environment
+import System.Exit
 import System.Console.GetOpt
+import qualified System.IO as IO
 
 import Paths_Agda            ( getDataDir )
 
@@ -51,25 +54,44 @@ runAgda backends = runAgda' $ builtinBackends ++ backends
 
 -- | The main function without importing built-in backends
 runAgda' :: [Backend] -> IO ()
-runAgda' backends = do
-  progName <- getProgName
-  argv     <- getArgs
-  conf     <- runExceptT $ do
-    (bs, opts) <- ExceptT $ runOptM $ parseBackendOptions backends argv defaultOptions
+runAgda' backends = runTCMPrettyErrors $ do
+  progName <- liftIO getProgName
+  argv     <- liftIO getArgs
+  let (z, warns) = runOptM $ parseBackendOptions backends argv defaultOptions
+  mapM_ (warning . OptionWarning) warns
+  conf     <- liftIO $ runExceptT $ do
+    (bs, opts) <- ExceptT $ pure z
     -- The absolute path of the input file, if provided
     inputFile <- liftIO $ mapM absolute $ optInputFile opts
     mode      <- getMainMode bs inputFile opts
     return (bs, opts, mode)
 
   case conf of
-    Left err -> optionError err
-    Right (bs, opts, mode) -> case mode of
-      MainModePrintHelp hp   -> printUsage bs hp
-      MainModePrintVersion   -> printVersion bs
-      MainModePrintAgdaDir   -> printAgdaDir
-      MainModeRun interactor -> runTCMPrettyErrors $ do
-        setTCLens stBackends bs
-        runAgdaWithOptions interactor progName opts
+    Left err -> liftIO $ optionError err
+    Right (bs, opts, mode) -> do
+
+      when (optTransliterate opts) $ liftIO $ do
+        -- When --interaction or --interaction-json is used, then we
+        -- use UTF-8 when writing to stdout (and when reading from
+        -- stdin).
+        if optGHCiInteraction opts || optJSONInteraction opts
+        then optionError $
+               "The option --transliterate must not be combined with " ++
+               "--interaction or --interaction-json"
+        else do
+          -- Transliterate unsupported code points.
+          enc <- IO.mkTextEncoding
+                   (show IO.localeEncoding ++ "//TRANSLIT")
+          IO.hSetEncoding IO.stdout enc
+          IO.hSetEncoding IO.stderr enc
+
+      case mode of
+        MainModePrintHelp hp   -> liftIO $ printUsage bs hp
+        MainModePrintVersion   -> liftIO $ printVersion bs
+        MainModePrintAgdaDir   -> liftIO $ printAgdaDir
+        MainModeRun interactor -> do
+          setTCLens stBackends bs
+          runAgdaWithOptions interactor progName opts
 
 -- | Main execution mode
 data MainMode
@@ -182,7 +204,7 @@ runAgdaWithOptions interactor progName opts = do
             Bench.print
 
             -- Print accumulated statistics.
-            printStatistics 1 Nothing =<< useTC lensAccumStatistics
+            printStatistics Nothing =<< useTC lensAccumStatistics
   where
     -- Options are fleshed out here so that (most) errors like
     -- "bad library path" are validated within the interactor,
@@ -253,17 +275,84 @@ optionError err = do
   exitAgdaWith OptionError
 
 -- | Run a TCM action in IO; catch and pretty print errors.
+
+-- If some error message cannot be printed due to locale issues, then
+-- one may get the "Error when handling error" error message. There is
+-- currently no test case for this error, but on some systems one can
+-- (at the time of writing) trigger it by running @LC_CTYPE=C agda
+-- --no-libraries Bug.agda@, where @Bug.agda@ contains the following
+-- code (if there is some other file in the same directory, for
+-- instance @Bug.lagda@, then the error message may be different):
+--
+-- @
+-- _ : Set
+-- _ = Set
+-- @
+
 runTCMPrettyErrors :: TCM () -> IO ()
 runTCMPrettyErrors tcm = do
-    r <- runTCMTop $ tcm `catchError` \err -> do
-      s2s <- prettyTCWarnings' =<< getAllWarningsOfTCErr err
-      s1  <- prettyError err
-      let ss = filter (not . null) $ s2s ++ [s1]
-      unless (null s1) (liftIO $ putStr $ unlines ss)
-      throwError err
-    case r of
-      Right _ -> exitSuccess
-      Left _  -> exitAgdaWith TCMError
-  `catchImpossible` \e -> do
-    putStr $ show e
-    exitAgdaWith ImpossibleError
+  r <- runTCMTop
+    ( ( (Nothing <$ tcm)
+          `catchError` \err -> do
+            s2s <- prettyTCWarnings' =<< getAllWarningsOfTCErr err
+            s1  <- prettyError err
+            let ss = filter (not . null) $ s2s ++ [s1]
+            unless (null s1) (liftIO $ putStr $ unlines ss)
+            liftIO $ helpForLocaleError err
+            return (Just TCMError)
+      ) `catchImpossible` \e -> do
+          liftIO $ putStr $ E.displayException e
+          return (Just ImpossibleError)
+    ) `E.catches`
+        -- Catch all exceptions except for those of type ExitCode
+        -- (which are thrown by exitWith) and asynchronous exceptions
+        -- (which are for instance raised when Ctrl-C is used, or if
+        -- the program runs out of heap or stack space).
+        [ E.Handler $ \(e :: ExitCode)         -> E.throw e
+        , E.Handler $ \(e :: E.AsyncException) -> E.throw e
+        , E.Handler $ \(e :: E.SomeException)  -> do
+            liftIO $ putStr $ E.displayException e
+            return $ Right (Just UnknownError)
+        ]
+  case r of
+    Right Nothing       -> exitSuccess
+    Right (Just reason) -> exitAgdaWith reason
+    Left err            -> do
+      liftIO $ do
+        putStrLn "\n\nError when handling error:"
+        putStrLn $ tcErrString err
+        helpForLocaleError err
+      exitAgdaWith UnknownError
+
+-- | If the error is an IO error, and the error message suggests that
+-- the problem is related to locales or code pages, print out some
+-- extra information.
+
+helpForLocaleError :: TCErr -> IO ()
+helpForLocaleError e = case e of
+  (IOException _ _ e)
+    | "invalid argument" `List.isInfixOf` show e -> msg
+  _                                              -> return ()
+  where
+  msg = putStr $ unlines
+    [ ""
+    , "This error may be due to the use of a locale or code page that does not"
+    , "support some character used in the program being type-checked."
+    , ""
+    , "If it is, then one option is to use the option --transliterate, in which"
+    , "case unsupported characters are (hopefully) replaced with something else,"
+    , "perhaps question marks. However, that could make the output harder to"
+    , "read."
+    , ""
+    , "If you want to fix the problem \"properly\", then you could try one of the"
+    , "following suggestions:"
+    , ""
+    , "* If you are using Windows, try switching to a different code page (for"
+    , "  instance by running the command 'CHCP 65001')."
+    , ""
+    , "* If you are using a Unix-like system, try using a different locale. The"
+    , "  installed locales are perhaps printed by the command 'locale -a'. If"
+    , "  you have a UTF-8 locale installed (for instance sv_SE.UTF-8), then you"
+    , "  can perhaps make Agda use this locale by running something like"
+    , "  'LC_ALL=sv_SE.UTF-8 agda <...>'."
+    ]
