@@ -61,7 +61,6 @@ module Agda.Syntax.Concrete
   , Module(..)
   , ThingWithFixity(..)
   , HoleContent, HoleContent'(..)
-  , topLevelModuleName
   , spanAllowedBeforeModule
   )
   where
@@ -70,7 +69,7 @@ import Prelude hiding (null)
 
 import Control.DeepSeq
 
-import Data.Data        ( Data )
+import qualified Data.DList as DL
 import Data.Functor.Identity
 import Data.Set         ( Set  )
 import Data.Text        ( Text )
@@ -103,14 +102,14 @@ data OpApp e
     -- ^ An abstraction inside a special syntax declaration
     --   (see Issue 358 why we introduce this).
   | Ordinary e
-  deriving (Data, Functor, Foldable, Traversable, Eq)
+  deriving (Functor, Foldable, Traversable, Eq)
 
 fromOrdinary :: e -> OpApp e -> e
 fromOrdinary d (Ordinary e) = e
 fromOrdinary d _            = d
 
 data FieldAssignment' a = FieldAssignment { _nameFieldA :: Name, _exprFieldA :: a }
-  deriving (Data, Functor, Foldable, Traversable, Show, Eq)
+  deriving (Functor, Foldable, Traversable, Show, Eq)
 
 type FieldAssignment = FieldAssignment' Expr
 
@@ -119,7 +118,7 @@ data ModuleAssignment  = ModuleAssignment
                            , _exprModA      :: [Expr]
                            , _importDirModA :: ImportDirective
                            }
-  deriving (Data, Eq)
+  deriving Eq
 
 type RecordAssignment  = Either FieldAssignment ModuleAssignment
 type RecordAssignments = [RecordAssignment]
@@ -170,7 +169,6 @@ data Expr
   | As Range Name Expr                         -- ^ ex: @x\@p@, only in patterns
   | Dot Range Expr                             -- ^ ex: @.p@, only in patterns
   | DoubleDot Range Expr                       -- ^ ex: @..A@, used for parsing @..A -> B@
-  | ETel Telescope                             -- ^ only used for printing telescopes
   | Quote Range                                -- ^ ex: @quote@, should be applied to a name
   | QuoteTerm Range                            -- ^ ex: @quoteTerm@, should be applied to a term
   | Tactic Range Expr                          -- ^ ex: @\@(tactic t)@, used to declare tactic arguments
@@ -179,7 +177,7 @@ data Expr
   | Equal Range Expr Expr                      -- ^ ex: @a = b@, used internally in the parser
   | Ellipsis Range                             -- ^ @...@, used internally to parse patterns.
   | Generalized Expr
-  deriving (Data, Eq)
+  deriving Eq
 
 type OpAppArgs = OpAppArgs' Expr
 type OpAppArgs' e = [NamedArg (MaybePlaceholder (OpApp e))]
@@ -210,19 +208,19 @@ data Pattern
                                            --   Second arg is @Nothing@ before expansion, and
                                            --   @Just p@ after expanding ellipsis to @p@.
   | WithP Range Pattern                    -- ^ @| p@, for with-patterns.
-  deriving (Data, Eq)
+  deriving Eq
 
 data DoStmt
   = DoBind Range Pattern Expr [LamClause]   -- ^ @p ← e where cs@
   | DoThen Expr
   | DoLet Range (List1 Declaration)
-  deriving (Data, Eq)
+  deriving Eq
 
 -- | A Binder @x\@p@, the pattern is optional
 data Binder' a = Binder
   { binderPattern :: Maybe Pattern
   , binderName    :: a
-  } deriving (Data, Eq, Functor, Foldable, Traversable)
+  } deriving (Eq, Functor, Foldable, Traversable)
 
 type Binder = Binder' BoundName
 
@@ -240,7 +238,7 @@ data LamBinding' a
     -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@ or @x\@p@ or @(p)@
   | DomainFull a
     -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Data, Functor, Foldable, Traversable, Eq)
+  deriving (Functor, Foldable, Traversable, Eq)
 
 -- | Drop type annotations and lets from bindings.
 dropTypeAndModality :: LamBinding -> [LamBinding]
@@ -250,11 +248,12 @@ dropTypeAndModality (DomainFull TLet{}) = []
 dropTypeAndModality (DomainFree x) = [DomainFree $ setModality defaultModality x]
 
 data BoundName = BName
-  { boundName   :: Name
-  , bnameFixity :: Fixity'
-  , bnameTactic :: TacticAttribute -- From @tactic attribute
+  { boundName       :: Name
+  , bnameFixity     :: Fixity'
+  , bnameTactic     :: TacticAttribute -- From @tactic attribute
+  , bnameIsFinite   :: Bool
   }
-  deriving (Data, Eq)
+  deriving Eq
 
 type TacticAttribute = Maybe Expr
 
@@ -262,7 +261,7 @@ mkBoundName_ :: Name -> BoundName
 mkBoundName_ x = mkBoundName x noFixity'
 
 mkBoundName :: Name -> Fixity' -> BoundName
-mkBoundName x f = BName x f Nothing
+mkBoundName x f = BName x f Nothing False
 
 -- | A typed binding.
 
@@ -273,7 +272,7 @@ data TypedBinding' e
     -- ^ Binding @(x1\@p1 ... xn\@pn : A)@.
   | TLet  Range (List1 Declaration)
     -- ^ Let binding @(let Ds)@ or @(open M args)@.
-  deriving (Data, Functor, Foldable, Traversable, Eq)
+  deriving (Functor, Foldable, Traversable, Eq)
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
 --   in later types.
@@ -327,7 +326,7 @@ data LHS = LHS  -- ^ Original pattern (including with-patterns), rewrite equatio
   , lhsWithExpr        :: [WithExpr]
     -- ^ @with e1 in eq | {e2} | ...@ (many)
   }
-  deriving (Data, Eq)
+  deriving Eq
 
 type RewriteEqn = RewriteEqn' () Name Pattern Expr
 type WithExpr   = Named Name (Arg Expr)
@@ -351,13 +350,13 @@ data LHSCore
              { lhsEllipsisRange :: Range
              , lhsEllipsisPat   :: LHSCore           -- ^ Pattern that was expanded from an ellipsis @...@.
              }
-  deriving (Data, Eq)
+  deriving Eq
 
 type RHS = RHS' Expr
 data RHS' e
   = AbsurdRHS -- ^ No right hand side because of absurd match.
   | RHS e
-  deriving (Data, Functor, Foldable, Traversable, Eq)
+  deriving (Functor, Foldable, Traversable, Eq)
 
 -- | @where@ block following a clause.
 type WhereClause = WhereClause' [Declaration]
@@ -375,14 +374,14 @@ data WhereClause' decls
       --   The 'Access' flag applies to the 'Name' (not the module contents!)
       --   and is propagated from the parent function.
       --   List of declarations can be empty.
-  deriving (Data, Eq, Functor, Foldable, Traversable)
+  deriving (Eq, Functor, Foldable, Traversable)
 
 data LamClause = LamClause
   { lamLHS      :: [Pattern]   -- ^ Possibly empty sequence.
   , lamRHS      :: RHS
   , lamCatchAll :: Bool
   }
-  deriving (Data, Eq)
+  deriving Eq
 
 -- | An expression followed by a where clause.
 --   Currently only used to give better a better error message in interaction.
@@ -406,7 +405,7 @@ data AsName' a = AsName
   , asRange :: Range
     -- ^ The range of the \"as\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Data, Show, Functor, Foldable, Traversable, Eq)
+  deriving (Show, Functor, Foldable, Traversable, Eq)
 
 -- | From the parser, we get an expression for the @as@-'Name', which
 --   we have to parse into a 'Name'.
@@ -434,7 +433,7 @@ data RecordDirective
        -- ^ Range of @[no-]eta-equality@ keyword.
    | PatternOrCopattern Range
        -- ^ If declaration @pattern@ is present, give its range.
-   deriving (Data,Eq,Show)
+   deriving (Eq, Show)
 
 type RecordDirectives = RecordDirectives' (Name, IsInstance)
 
@@ -479,9 +478,13 @@ data Declaration
   | ModuleMacro Range  Name ModuleApplication !OpenShortHand ImportDirective
   | Module      Range QName Telescope [Declaration]
   | UnquoteDecl Range [Name] Expr
+      -- ^ @unquoteDecl xs = e@
   | UnquoteDef  Range [Name] Expr
+      -- ^ @unquoteDef xs = e@
+  | UnquoteData Range Name [Name] Expr
+      -- ^ @unquoteDecl data d constructor xs = e@
   | Pragma      Pragma
-  deriving (Data, Eq)
+  deriving Eq
 
 -- | Extract a record directive
 isRecordDirective :: Declaration -> Maybe RecordDirective
@@ -494,10 +497,10 @@ data ModuleApplication
     -- ^ @tel. M args@
   | RecordModuleInstance Range QName
     -- ^ @M {{...}}@
-  deriving (Data, Eq)
+  deriving Eq
 
 data OpenShortHand = DoOpen | DontOpen
-  deriving (Data, Eq, Show, Generic)
+  deriving (Eq, Show, Generic)
 
 -- Pragmas ----------------------------------------------------------------
 
@@ -539,7 +542,9 @@ data Pragma
   | PolarityPragma            Range Name [Occurrence]
   | NoUniverseCheckPragma     Range
     -- ^ Applies to the following data/record type.
-  deriving (Data, Eq)
+  | NotProjectionLikePragma   Range QName
+    -- ^ Applies to the stated function
+  deriving Eq
 
 ---------------------------------------------------------------------------
 
@@ -550,19 +555,6 @@ data Module = Mod
   , modDecls   :: [Declaration]
   }
 
--- | Computes the top-level module name.
---
--- Precondition: The 'Module' has to be well-formed.
--- This means that there are only allowed declarations before the
--- first module declaration, typically import declarations.
--- See 'spanAllowedBeforeModule'.
-
-topLevelModuleName :: Module -> TopLevelModuleName
-topLevelModuleName (Mod _ []) = __IMPOSSIBLE__
-topLevelModuleName (Mod _ ds) = case spanAllowedBeforeModule ds of
-  (_, Module _ n _ _ : _) -> toTopLevelModuleName n
-  _ -> __IMPOSSIBLE__
-
 -- | Splits off allowed (= import) declarations before the first
 --   non-allowed declaration.
 --   After successful parsing, the first non-allowed declaration
@@ -610,13 +602,17 @@ rawAppP (p :| []) = p
 data AppView = AppView Expr [NamedArg Expr]
 
 appView :: Expr -> AppView
-appView = \case
-    App r e1 e2      -> vApp (appView e1) e2
-    RawApp _ (List2 e1 e2 es)
-                     -> AppView e1 $ map arg (e2:es)
-    e                -> AppView e []
+appView e = f (DL.toList ess)
   where
-    vApp (AppView e es) arg = AppView e (es ++ [arg])
+    (f, ess) = appView' e
+
+    appView' = \case
+      App r e1 e2      -> vApp (appView' e1) e2
+      RawApp _ (List2 e1 e2 es)
+                       -> (AppView e1, DL.fromList (map arg (e2 : es)))
+      e                -> (AppView e, mempty)
+
+    vApp (f, es) arg = (f, es `DL.snoc` arg)
 
     arg (HiddenArg   _ e) = hide         $ defaultArg e
     arg (InstanceArg _ e) = makeInstance $ defaultArg e
@@ -823,7 +819,6 @@ instance HasRange Expr where
       InstanceArg r _    -> r
       Rec r _            -> r
       RecUpdate r _ _    -> r
-      ETel tel           -> getRange tel
       Quote r            -> r
       QuoteTerm r        -> r
       Unquote r          -> r
@@ -903,6 +898,7 @@ instance HasRange Declaration where
   getRange (PatternSyn r _ _ _)    = r
   getRange (UnquoteDecl r _ _)     = r
   getRange (UnquoteDef r _ _)      = r
+  getRange (UnquoteData r _ _ _)   = r
   getRange (Pragma p)              = getRange p
 
 instance HasRange LHS where
@@ -946,6 +942,7 @@ instance HasRange Pragma where
   getRange (NoPositivityCheckPragma r)       = r
   getRange (PolarityPragma r _ _)            = r
   getRange (NoUniverseCheckPragma r)         = r
+  getRange (NotProjectionLikePragma r _)     = r
 
 instance HasRange AsName where
   getRange a = getRange (asRange a, asName a)
@@ -1011,7 +1008,7 @@ instance KillRange Binder where
   killRange (Binder a b) = killRange2 Binder a b
 
 instance KillRange BoundName where
-  killRange (BName n f t) = killRange3 BName n f t
+  killRange (BName n f t b) = killRange4 BName n f t b
 
 instance KillRange RecordDirective where
   killRange (Induction a)          = killRange1 Induction a
@@ -1050,6 +1047,7 @@ instance KillRange Declaration where
   killRange (Module _ q t d)        = killRange3 (Module noRange) q t d
   killRange (UnquoteDecl _ x t)     = killRange2 (UnquoteDecl noRange) x t
   killRange (UnquoteDef _ x t)      = killRange2 (UnquoteDef noRange) x t
+  killRange (UnquoteData _ xs cs t) = killRange3 (UnquoteData noRange) xs cs t
   killRange (Pragma p)              = killRange1 Pragma p
 
 instance KillRange Expr where
@@ -1078,7 +1076,6 @@ instance KillRange Expr where
   killRange (As _ n e)            = killRange2 (As noRange) n e
   killRange (Dot _ e)             = killRange1 (Dot noRange) e
   killRange (DoubleDot _ e)       = killRange1 (DoubleDot noRange) e
-  killRange (ETel t)              = killRange1 ETel t
   killRange (Quote _)             = Quote noRange
   killRange (QuoteTerm _)         = QuoteTerm noRange
   killRange (Unquote _)           = Unquote noRange
@@ -1150,6 +1147,7 @@ instance KillRange Pragma where
   killRange (NoPositivityCheckPragma _)       = NoPositivityCheckPragma noRange
   killRange (PolarityPragma _ q occs)         = killRange1 (\q -> PolarityPragma noRange q occs) q
   killRange (NoUniverseCheckPragma _)         = NoUniverseCheckPragma noRange
+  killRange (NotProjectionLikePragma _ q)     = NotProjectionLikePragma noRange q
 
 instance KillRange RHS where
   killRange AbsurdRHS = AbsurdRHS
@@ -1195,7 +1193,6 @@ instance NFData Expr where
   rnf (As _ a b)          = rnf a `seq` rnf b
   rnf (Dot _ a)           = rnf a
   rnf (DoubleDot _ a)     = rnf a
-  rnf (ETel a)            = rnf a
   rnf (Quote _)           = ()
   rnf (QuoteTerm _)       = ()
   rnf (Tactic _ a)        = rnf a
@@ -1265,6 +1262,7 @@ instance NFData Declaration where
   rnf (Module _ a b c)        = rnf a `seq` rnf b `seq` rnf c
   rnf (UnquoteDecl _ a b)     = rnf a `seq` rnf b
   rnf (UnquoteDef _ a b)      = rnf a `seq` rnf b
+  rnf (UnquoteData _ a b c)   = rnf a `seq` rnf b `seq` rnf c
   rnf (Pragma a)              = rnf a
 
 instance NFData OpenShortHand
@@ -1291,6 +1289,7 @@ instance NFData Pragma where
   rnf (NoPositivityCheckPragma _)       = ()
   rnf (PolarityPragma _ a b)            = rnf a `seq` rnf b
   rnf (NoUniverseCheckPragma _)         = ()
+  rnf (NotProjectionLikePragma _ q)     = rnf q
 
 -- | Ranges are not forced.
 
@@ -1342,7 +1341,7 @@ instance NFData Binder where
   rnf (Binder a b) = rnf a `seq` rnf b
 
 instance NFData BoundName where
-  rnf (BName a b c) = rnf a `seq` rnf b `seq` rnf c
+  rnf (BName a b c d) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
 
 instance NFData a => NFData (RHS' a) where
   rnf AbsurdRHS = ()
