@@ -2,8 +2,6 @@ module Agda.Syntax.Concrete.Definitions.Errors where
 
 import Control.DeepSeq
 
-import Data.Data
-
 import GHC.Generics (Generic)
 
 import Agda.Syntax.Position
@@ -47,7 +45,7 @@ data DeclarationException'
       --   Range is of mutual block.
   | UnquoteDefRequiresSignature (List1 Name)
   | BadMacroDef NiceDeclaration
-    deriving (Data, Show)
+    deriving Show
 
 ------------------------------------------------------------------------
 -- Warnings
@@ -70,6 +68,10 @@ data DeclarationWarning'
   | EmptyPostulate Range   -- ^ Empty @postulate@ block.
   | EmptyPrivate Range     -- ^ Empty @private@   block.
   | EmptyPrimitive Range   -- ^ Empty @primitive@ block.
+  | HiddenGeneralize Range
+      -- ^ A 'Hidden' identifier in a @variable@ declaration.
+      --   Hiding has no effect there as generalized variables are always hidden
+      --   (or instance variables).
   | InvalidCatchallPragma Range
       -- ^ A {-\# CATCHALL \#-} pragma
       --   that does not precede a function clause.
@@ -115,7 +117,7 @@ data DeclarationWarning'
       -- ^ @instance@ block with nothing that can (newly) become an instance.
   | UselessPrivate Range
       -- ^ @private@ block with nothing that can (newly) be made private.
-  deriving (Data, Show, Generic)
+  deriving (Show, Generic)
 
 declarationWarningName :: DeclarationWarning -> WarningName
 declarationWarningName = declarationWarningName' . dwWarning
@@ -133,6 +135,7 @@ declarationWarningName' = \case
   EmptyPrivate{}                    -> EmptyPrivate_
   EmptyPostulate{}                  -> EmptyPostulate_
   EmptyPrimitive{}                  -> EmptyPrimitive_
+  HiddenGeneralize{}                -> HiddenGeneralize_
   InvalidCatchallPragma{}           -> InvalidCatchallPragma_
   InvalidConstructor{}              -> InvalidConstructor_
   InvalidConstructorBlock{}         -> InvalidConstructorBlock_
@@ -174,6 +177,7 @@ unsafeDeclarationWarning' = \case
   EmptyPrivate{}                    -> False
   EmptyPostulate{}                  -> False
   EmptyPrimitive{}                  -> False
+  HiddenGeneralize{}                -> False
   InvalidCatchallPragma{}           -> False
   InvalidConstructor{}              -> False
   InvalidConstructorBlock{}         -> False
@@ -243,6 +247,7 @@ instance HasRange DeclarationWarning' where
   getRange (EmptyGeneralize r)                  = r
   getRange (EmptyPrimitive r)                   = r
   getRange (EmptyField r)                       = r
+  getRange (HiddenGeneralize r)                 = r
   getRange (InvalidTerminationCheckPragma r)    = r
   getRange (InvalidCoverageCheckPragma r)       = r
   getRange (InvalidNoPositivityCheckPragma r)   = r
@@ -340,6 +345,7 @@ instance Pretty DeclarationWarning' where
   pretty (EmptyGeneralize _) = fsep $ pwords "Empty variable block."
   pretty (EmptyPrimitive _)  = fsep $ pwords "Empty primitive block."
   pretty (EmptyField _)      = fsep $ pwords "Empty field block."
+  pretty (HiddenGeneralize _) = fsep $ pwords "Declaring a variable as hidden has no effect in a variable block. Generalization never introduces visible arguments."
   pretty InvalidRecordDirective{} = fsep $
     pwords "Record directives can only be used inside record definitions and before field declarations."
   pretty (InvalidTerminationCheckPragma _) = fsep $
