@@ -3,30 +3,26 @@
 -}
 module Agda.Syntax.Concrete.Name where
 
+import Prelude hiding ((!!))
+
 import Control.DeepSeq
 
 import Data.ByteString.Char8 (ByteString)
-import Data.Function
+import Data.Function (on)
 import qualified Data.Foldable as Fold
-import qualified Data.List as List
-import Data.Data (Data)
 
 import GHC.Generics (Generic)
 
-import System.FilePath
-
 import Agda.Syntax.Common
 import Agda.Syntax.Concrete.Glyph
 import Agda.Syntax.Position
 
-import Agda.Utils.FileName
 import Agda.Utils.Lens
-import Agda.Utils.List  (last1)
+import Agda.Utils.List  ((!!), last1)
 import Agda.Utils.List1 (List1, pattern (:|), (<|))
 import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Pretty
 import Agda.Utils.Singleton
-import Agda.Utils.Size
 import Agda.Utils.Suffix
 
 import Agda.Utils.Impossible
@@ -49,7 +45,6 @@ data Name
     { nameRange     :: Range
     , nameId        :: NameId
     }
-  deriving Data
 
 type NameParts = List1 NamePart
 
@@ -72,7 +67,7 @@ instance Underscore Name where
 data NamePart
   = Hole       -- ^ @_@ part.
   | Id RawName  -- ^ Identifier part.
-  deriving (Data, Generic)
+  deriving Generic
 
 -- | Define equality on @Name@ to ignore range so same names in different
 --   locations are equal.
@@ -115,27 +110,13 @@ instance Ord NamePart where
 data QName
   = Qual  Name QName -- ^ @A.rest@.
   | QName Name       -- ^ @x@.
-  deriving (Data, Eq, Ord)
+  deriving (Eq, Ord)
 
 instance Underscore QName where
   underscore = QName underscore
   isUnderscore (QName x) = isUnderscore x
   isUnderscore Qual{}    = False
 
--- | Top-level module names.  Used in connection with the file system.
---
---   Invariant: The list must not be empty.
-
-data TopLevelModuleName = TopLevelModuleName
-  { moduleNameRange :: Range
-  , moduleNameParts :: List1 String
-  }
-  deriving (Show, Data, Generic)
-
-instance Eq    TopLevelModuleName where (==)    = (==)    `on` moduleNameParts
-instance Ord   TopLevelModuleName where compare = compare `on` moduleNameParts
-instance Sized TopLevelModuleName where size    = size     .   moduleNameParts
-
 ------------------------------------------------------------------------
 -- * Constructing simple 'Name's.
 ------------------------------------------------------------------------
@@ -179,7 +160,7 @@ nameStringParts n = [ s | Id s <- List1.toList $ nameParts n ]
 stringNameParts :: String -> NameParts
 stringNameParts ""  = singleton $ Id "_"  -- NoName
 stringNameParts "_" = singleton $ Id "_"  -- NoName
-stringNameParts s = List1.fromList $ loop s
+stringNameParts s = List1.fromListSafe __IMPOSSIBLE__ $ loop s
   where
   loop ""                              = []
   loop ('_':s)                         = Hole : loop s
@@ -223,7 +204,7 @@ isNonfix  x = not (isHole (List1.head xs)) && not (isHole (List1.last xs)) where
 ------------------------------------------------------------------------
 
 data NameInScope = InScope | NotInScope
-  deriving (Eq, Show, Data)
+  deriving (Eq, Show)
 
 class LensInScope a where
   lensInScope :: Lens' NameInScope a
@@ -376,37 +357,6 @@ isUnqualified :: QName -> Maybe Name
 isUnqualified Qual{}    = Nothing
 isUnqualified (QName n) = Just n
 
-------------------------------------------------------------------------
--- * Operations on 'TopLevelModuleName'
-------------------------------------------------------------------------
-
--- | Turns a qualified name into a 'TopLevelModuleName'. The qualified
--- name is assumed to represent a top-level module name.
-
-toTopLevelModuleName :: QName -> TopLevelModuleName
-toTopLevelModuleName q = TopLevelModuleName (getRange q) $
-  fmap nameToRawName $ qnameParts q
-
--- | Turns a top-level module name into a file name with the given
--- suffix.
-
-moduleNameToFileName :: TopLevelModuleName -> String -> FilePath
-moduleNameToFileName (TopLevelModuleName _ ms) ext =
-  joinPath (List1.init ms) </> List1.last ms <.> ext
-
--- | Finds the current project's \"root\" directory, given a project
--- file and the corresponding top-level module name.
---
--- Example: If the module \"A.B.C\" is located in the file
--- \"/foo/A/B/C.agda\", then the root is \"/foo/\".
---
--- Precondition: The module name must be well-formed.
-
-projectRoot :: AbsolutePath -> TopLevelModuleName -> AbsolutePath
-projectRoot file (TopLevelModuleName _ m) =
-  mkAbsolute $
-    iterate takeDirectory (filePath file) !! length m
-
 ------------------------------------------------------------------------
 -- * No name stuff
 ------------------------------------------------------------------------
@@ -445,8 +395,6 @@ instance IsNoName QName where
 instance IsNoName a => IsNoName (Ranged a) where
 instance IsNoName a => IsNoName (WithOrigin a) where
 
--- no instance for TopLevelModuleName
-
 ------------------------------------------------------------------------
 -- * Showing names
 ------------------------------------------------------------------------
@@ -473,9 +421,6 @@ instance Pretty QName where
     | otherwise      = pretty m <> "." <> pretty x
   pretty (QName x)  = pretty x
 
-instance Pretty TopLevelModuleName where
-  pretty (TopLevelModuleName _ ms) = text $ List.intercalate "." $ List1.toList ms
-
 ------------------------------------------------------------------------
 -- * Range instances
 ------------------------------------------------------------------------
@@ -488,9 +433,6 @@ instance HasRange QName where
     getRange (QName  x) = getRange x
     getRange (Qual n x) = fuseRange n x
 
-instance HasRange TopLevelModuleName where
-  getRange = moduleNameRange
-
 instance SetRange Name where
   setRange r (Name _ nis ps) = Name r nis ps
   setRange r (NoName _ i)  = NoName r i
@@ -499,9 +441,6 @@ instance SetRange QName where
   setRange r (QName x)  = QName (setRange r x)
   setRange r (Qual n x) = Qual (setRange r n) (setRange r x)
 
-instance SetRange TopLevelModuleName where
-  setRange r (TopLevelModuleName _ x) = TopLevelModuleName r x
-
 instance KillRange QName where
   killRange (QName x) = QName $ killRange x
   killRange (Qual n x) = killRange n `Qual` killRange x
@@ -510,9 +449,6 @@ instance KillRange Name where
   killRange (Name r nis ps)  = Name (killRange r) nis ps
   killRange (NoName r i)     = NoName (killRange r) i
 
-instance KillRange TopLevelModuleName where
-  killRange (TopLevelModuleName _ x) = TopLevelModuleName noRange x
-
 ------------------------------------------------------------------------
 -- * NFData instances
 ------------------------------------------------------------------------
@@ -534,5 +470,3 @@ instance NFData NamePart where
 instance NFData QName where
   rnf (Qual a b) = rnf a `seq` rnf b
   rnf (QName a)  = rnf a
-
-instance NFData TopLevelModuleName
