@@ -19,12 +19,12 @@ module Agda.Syntax.Concrete.Operators
     ) where
 
 import Control.Applicative ( Alternative((<|>)))
-import Control.Arrow (second)
 import Control.Monad.Except (throwError)
 
 import Data.Either (partitionEithers)
 import qualified Data.Foldable as Fold
-import Data.Function
+import Data.Function (on)
+import qualified Data.Function
 import qualified Data.List as List
 import Data.Maybe
 import Data.Map (Map)
@@ -42,6 +42,7 @@ import qualified Agda.Syntax.Abstract.Name as A
 import Agda.Syntax.Position
 import Agda.Syntax.Notation
 import Agda.Syntax.Scope.Base
+import Agda.Syntax.Scope.Flat
 import Agda.Syntax.Scope.Monad
 
 import Agda.TypeChecking.Monad.Base (typeError, TypeError(..), LHSOrPatSyn(..))
@@ -81,45 +82,6 @@ billToParser k = Bench.billTo
 -- * Building the parser
 ---------------------------------------------------------------------------
 
-type FlatScope = Map QName [AbstractName]
-
--- | Compute all defined names in scope and their fixities/notations.
--- Note that overloaded names (constructors) can have several
--- fixities/notations. Then we 'mergeNotations'. (See issue 1194.)
-getDefinedNames :: KindsOfNames -> FlatScope -> [[NewNotation]]
-getDefinedNames kinds names =
-  [ mergeNotations $ map (namesToNotation x . A.qnameName . anameName) ds
-  | (x, ds) <- Map.toList names
-  , any ((`elemKindsOfNames` kinds) . anameKind) ds
-  , not (null ds)
-  ]
-  -- Andreas, 2013-03-21 see Issue 822
-  -- Names can have different kinds, i.e., 'defined' and 'constructor'.
-  -- We need to consider all names that have *any* matching kind,
-  -- not only those whose first appearing kind is matching.
-
--- | Compute all names (first component) and operators/notations
--- (second component) in scope.
-localNames :: FlatScope -> ScopeM ([QName], [NewNotation])
-localNames flat = do
-  let defs = getDefinedNames allKindsOfNames flat
-  locals <- nubOn fst . notShadowedLocals <$> getLocalVars
-  -- Note: Debug printout aligned with the one in buildParsers.
-  reportS "scope.operators" 50
-    [ "flat  = " ++ prettyShow flat
-    , "defs  = " ++ prettyShow defs
-    , "locals= " ++ prettyShow locals
-    ]
-  let localNots  = map localOp locals
-      notLocal   = not . hasElem (map notaName localNots) . notaName
-      otherNots  = concatMap (filter notLocal) defs
-  return $ second (map useDefaultFixity) $ split $ localNots ++ otherNots
-  where
-    localOp (x, y) = namesToNotation (QName x) y
-    split          = partitionEithers . concatMap opOrNot
-    opOrNot n      = Left (notaName n) :
-                     [Right n | not (null (notation n))]
-
 -- | A data structure used internally by 'buildParsers'.
 data InternalParsers e = InternalParsers
   { pTop    :: Parser e e
@@ -181,7 +143,9 @@ buildParsers
 buildParsers kind exprNames = do
     flat         <- flattenScope (qualifierModules exprNames) <$>
                       getScope
-    (names, ops) <- localNames flat
+    (names, ops0) <- localNames flat
+    let ops | kind == IsPattern = filter (not . isLambdaNotation) ops0
+            | otherwise         = ops0
 
     let -- All names.
         namesInExpr :: Set QName
@@ -224,9 +188,9 @@ buildParsers kind exprNames = do
           Trie.member (addHole withHole p) partListsInExpr
           where
           p = case n of
-            NormalHole{} : IdPart p : _ -> rangedThing p
-            IdPart p : _                -> rangedThing p
-            _                           -> __IMPOSSIBLE__
+            HolePart{} : IdPart p : _ -> rangedThing p
+            IdPart p : _              -> rangedThing p
+            _                         -> __IMPOSSIBLE__
 
         -- Is the last identifier part present in n present in the
         -- expression, without any succeeding name parts, except for a
@@ -235,9 +199,9 @@ buildParsers kind exprNames = do
           Trie.member (addHole withHole p) reversedPartListsInExpr
           where
           p = case reverse n of
-            NormalHole{} : IdPart p : _ -> rangedThing p
-            IdPart p : _                -> rangedThing p
-            _                           -> __IMPOSSIBLE__
+            HolePart{} : IdPart p : _ -> rangedThing p
+            IdPart p : _              -> rangedThing p
+            _                         -> __IMPOSSIBLE__
 
         -- Are the initial and final identifier parts present with
         -- the right mix of leading and trailing underscores?
