@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE PatternSynonyms            #-}
 
 module Agda.Syntax.Internal
@@ -13,7 +14,7 @@ import Prelude hiding (null)
 import Control.Monad.Identity
 import Control.DeepSeq
 
-import Data.Function
+import Data.Function (on)
 import qualified Data.List as List
 import Data.Maybe
 import Data.Semigroup ( Semigroup, (<>), Sum(..) )
@@ -21,7 +22,6 @@ import qualified Data.Set as Set
 import Data.Set (Set)
 
 import Data.Traversable
-import Data.Data (Data)
 
 import GHC.Generics (Generic)
 
@@ -65,34 +65,35 @@ import Agda.Utils.Impossible
 --   'Arg' is used for actual arguments ('Var', 'Con', 'Def' etc.)
 --   and in 'Abstract' syntax and other situations.
 --
---   [ cubical ] When @domFinite = True@ for the domain of a 'Pi'
---   type, the elements should be compared by tabulating the domain type.
---   Only supported in case the domain type is primIsOne, to obtain
---   the correct equality for partial elements.
+--   [ cubical ] When @annFinite (argInfoAnnotation domInfo) = True@ for
+--   the domain of a 'Pi' type, the elements should be compared by
+--   tabulating the domain type.  Only supported in case the domain type
+--   is primIsOne, to obtain the correct equality for partial elements.
 --
 data Dom' t e = Dom
   { domInfo   :: ArgInfo
-  , domFinite :: !Bool
   , domName   :: Maybe NamedName  -- ^ e.g. @x@ in @{x = y : A} -> B@.
+  , domIsFinite :: Bool
+    -- ^ Is this a Π-type (False), or a partial type (True)?
   , domTactic :: Maybe t        -- ^ "@tactic e".
   , unDom     :: e
-  } deriving (Data, Show, Functor, Foldable, Traversable)
+  } deriving (Show, Functor, Foldable, Traversable)
 
 type Dom = Dom' Term
 
 instance Decoration (Dom' t) where
-  traverseF f (Dom ai b x t a) = Dom ai b x t <$> f a
+  traverseF f (Dom ai x t b a) = Dom ai x t b <$> f a
 
 instance HasRange a => HasRange (Dom' t a) where
   getRange = getRange . unDom
 
 instance (KillRange t, KillRange a) => KillRange (Dom' t a) where
-  killRange (Dom info b x t a) = killRange5 Dom info b x t a
+  killRange (Dom info x t b a) = killRange4 Dom info x t b a
 
 -- | Ignores 'Origin' and 'FreeVariables' and tactic.
 instance Eq a => Eq (Dom' t a) where
-  Dom (ArgInfo h1 m1 _ _ a1) b1 s1 _ x1 == Dom (ArgInfo h2 m2 _ _ a2) b2 s2 _ x2 =
-    (h1, m1, a1, b1, s1, x1) == (h2, m2, a2, b2, s2, x2)
+  Dom (ArgInfo h1 m1 _ _ a1) s1 f1 _ x1 == Dom (ArgInfo h2 m2 _ _ a2) s2 f2 _ x2 =
+    (h1, m1, a1, s1, f1, x1) == (h2, m2, a2, s2, f2, x2)
 
 instance LensNamed (Dom' t e) where
   type NameOf (Dom' t e) = NamedName
@@ -129,10 +130,10 @@ namedArgFromDom Dom{domInfo = i, domName = s, unDom = a} = Arg i $ Named s a
 -- often for class AddContext.
 
 domFromArg :: Arg a -> Dom a
-domFromArg (Arg i a) = Dom i False Nothing Nothing a
+domFromArg (Arg i a) = Dom i Nothing False Nothing a
 
 domFromNamedArg :: NamedArg a -> Dom a
-domFromNamedArg (Arg i a) = Dom i False (nameOf a) Nothing (namedThing a)
+domFromNamedArg (Arg i a) = Dom i (nameOf a) False Nothing (namedThing a)
 
 defaultDom :: a -> Dom a
 defaultDom = defaultArgDom defaultArgInfo
@@ -151,7 +152,7 @@ type NamedArgs  = [NamedArg Term]
 data DataOrRecord
   = IsData
   | IsRecord PatternOrCopattern
-  deriving (Data, Show, Eq, Generic)
+  deriving (Show, Eq, Generic)
 
 -- | Store the names of the record fields in the constructor.
 --   This allows reduction of projection redexes outside of TCM.
@@ -163,7 +164,7 @@ data ConHead = ConHead
   , conFields     :: [Arg QName]   -- ^ The name of the record fields.
       --   'Arg' is stored since the info in the constructor args
       --   might not be accurate because of subtyping (issue #2170).
-  } deriving (Data, Show, Generic)
+  } deriving (Show, Generic)
 
 instance Eq ConHead where
   (==) = (==) `on` conName
@@ -225,7 +226,7 @@ data Term = Var {-# UNPACK #-} !Int Elims -- ^ @x es@ neutral
             --   where they can affect type checking, so syntactic checks are free to ignore the
             --   eliminators, which are only there to ease debugging when a dummy term incorrectly
             --   leaks into a relevant position.
-  deriving (Data, Show)
+  deriving Show
 
 type ConInfo = ConOrigin
 
@@ -242,7 +243,7 @@ data Abs a = Abs   { absName :: ArgName, unAbs :: a }
                -- ^ The body has (at least) one free variable.
                --   Danger: 'unAbs' doesn't shift variables properly
            | NoAbs { absName :: ArgName, unAbs :: a }
-  deriving (Data, Functor, Foldable, Traversable, Generic)
+  deriving (Functor, Foldable, Traversable, Generic)
 
 instance Decoration Abs where
   traverseF f (Abs   x a) = Abs   x <$> f a
@@ -251,7 +252,7 @@ instance Decoration Abs where
 -- | Types are terms with a sort annotation.
 --
 data Type'' t a = El { _getSort :: Sort' t, unEl :: a }
-  deriving (Data, Show, Functor, Foldable, Traversable)
+  deriving (Show, Functor, Foldable, Traversable)
 
 type Type' a = Type'' Term a
 
@@ -284,22 +285,23 @@ instance LensSort a => LensSort (Arg a) where
 --   and so on.
 data Tele a = EmptyTel
             | ExtendTel a (Abs (Tele a))  -- ^ 'Abs' is never 'NoAbs'.
-  deriving (Data, Show, Functor, Foldable, Traversable, Generic)
+  deriving (Show, Functor, Foldable, Traversable, Generic)
 
 type Telescope = Tele (Dom Type)
 
 data IsFibrant = IsFibrant | IsStrict
-  deriving (Data, Show, Eq, Ord, Generic)
+  deriving (Show, Eq, Ord, Generic)
 
 -- | Sorts.
 --
 data Sort' t
   = Type (Level' t)  -- ^ @Set ℓ@.
   | Prop (Level' t)  -- ^ @Prop ℓ@.
-  | Inf IsFibrant Integer      -- ^ @Setωᵢ@.
+  | Inf IsFibrant !Integer      -- ^ @Setωᵢ@.
   | SSet (Level' t)  -- ^ @SSet ℓ@.
   | SizeUniv    -- ^ @SizeUniv@, a sort inhabited by type @Size@.
   | LockUniv    -- ^ @LockUniv@, a sort for locks.
+  | IntervalUniv -- ^ @IntervalUniv@, a sort inhabited by the cubical interval.
   | PiSort (Dom' t t) (Sort' t) (Abs (Sort' t)) -- ^ Sort of the pi type.
   | FunSort (Sort' t) (Sort' t) -- ^ Sort of a (non-dependent) function type.
   | UnivSort (Sort' t) -- ^ Sort of another sort.
@@ -310,19 +312,19 @@ data Sort' t
     --   Replaces the abuse of @Prop@ for a dummy sort.
     --   The @String@ typically describes the location where we create this dummy,
     --   but can contain other information as well.
-  deriving (Data, Show)
+  deriving Show
 
 type Sort = Sort' Term
 
 -- | A level is a maximum expression of a closed level and 0..n
 --   'PlusLevel' expressions each of which is an atom plus a number.
-data Level' t = Max Integer [PlusLevel' t]
-  deriving (Show, Data, Functor, Foldable, Traversable)
+data Level' t = Max !Integer [PlusLevel' t]
+  deriving (Show, Functor, Foldable, Traversable)
 
 type Level = Level' Term
 
-data PlusLevel' t = Plus Integer t
-  deriving (Show, Data, Functor, Foldable, Traversable)
+data PlusLevel' t = Plus !Integer t
+  deriving (Show, Functor, Foldable, Traversable)
 
 type PlusLevel = PlusLevel' Term
 type LevelAtom = Term
@@ -333,7 +335,7 @@ type LevelAtom = Term
 
 -- | Newtypes for terms that produce a dummy, rather than crash, when
 --   applied to incompatible eliminations.
-newtype BraveTerm = BraveTerm { unBrave :: Term } deriving (Data, Show)
+newtype BraveTerm = BraveTerm { unBrave :: Term } deriving Show
 
 ---------------------------------------------------------------------------
 -- * Blocked Terms
@@ -366,21 +368,21 @@ type NAPs = [NamedArg DeBruijnPattern]
 --  For the purpose of the permutation and the body dot patterns count
 --  as variables. TODO: Change this!
 data Clause = Clause
-    { clauseLHSRange  :: Range
-    , clauseFullRange :: Range
-    , clauseTel       :: Telescope
+    { clauseLHSRange    :: Range
+    , clauseFullRange   :: Range
+    , clauseTel         :: Telescope
       -- ^ @Δ@: The types of the pattern variables in dependency order.
-    , namedClausePats :: NAPs
+    , namedClausePats   :: NAPs
       -- ^ @Δ ⊢ ps@.  The de Bruijn indices refer to @Δ@.
-    , clauseBody      :: Maybe Term
+    , clauseBody        :: Maybe Term
       -- ^ @Just v@ with @Δ ⊢ v@ for a regular clause, or @Nothing@ for an
       --   absurd one.
-    , clauseType      :: Maybe (Arg Type)
+    , clauseType        :: Maybe (Arg Type)
       -- ^ @Δ ⊢ t@.  The type of the rhs under @clauseTel@.
       --   Used, e.g., by @TermCheck@.
       --   Can be 'Irrelevant' if we encountered an irrelevant projection
       --   pattern on the lhs.
-    , clauseCatchall  :: Bool
+    , clauseCatchall    :: Bool
       -- ^ Clause has been labelled as CATCHALL.
     , clauseExact       :: Maybe Bool
       -- ^ Pattern matching of this clause is exact, no catch-all case.
@@ -400,10 +402,12 @@ data Clause = Clause
       --   @Nothing@ means coverage checker has not run yet (clause may be unreachable).
       --   @Just False@ means clause is not unreachable.
       --   @Just True@ means clause is unreachable.
-    , clauseEllipsis  :: ExpandedEllipsis
+    , clauseEllipsis    :: ExpandedEllipsis
       -- ^ Was this clause created by expansion of an ellipsis?
+    , clauseWhereModule :: Maybe ModuleName
+      -- ^ Keeps track of the module name associate with the clause's where clause.
     }
-  deriving (Data, Show, Generic)
+  deriving (Show, Generic)
 
 clausePats :: Clause -> [Arg DeBruijnPattern]
 clausePats = map (fmap namedThing) . namedClausePats
@@ -423,7 +427,7 @@ nameToPatVarName = nameToArgName
 data PatternInfo = PatternInfo
   { patOrigin :: PatOrigin
   , patAsNames :: [Name]
-  } deriving (Data, Show, Eq, Generic)
+  } deriving (Show, Eq, Generic)
 
 defaultPatternInfo :: PatternInfo
 defaultPatternInfo = PatternInfo PatOSystem []
@@ -439,7 +443,7 @@ data PatOrigin
   | PatORec            -- ^ User wrote a record pattern
   | PatOLit            -- ^ User wrote a literal pattern
   | PatOAbsurd         -- ^ User wrote an absurd pattern
-  deriving (Data, Show, Eq, Generic)
+  deriving (Show, Eq, Generic)
 
 -- | Patterns are variables, constructors, or wildcards.
 --   @QName@ is used in @ConP@ rather than @Name@ since
@@ -463,7 +467,7 @@ data Pattern' x
     -- ^ Path elimination pattern, like @VarP@ but keeps track of endpoints.
   | DefP PatternInfo QName [NamedArg (Pattern' x)]
     -- ^ Used for HITs, the QName should be the one from primHComp.
-  deriving (Data, Show, Functor, Foldable, Traversable, Generic)
+  deriving (Show, Functor, Foldable, Traversable, Generic)
 
 type Pattern = Pattern' PatVarName
     -- ^ The @PatVarName@ is a name suggestion.
@@ -480,8 +484,8 @@ litP = LitP defaultPatternInfo
 -- | Type used when numbering pattern variables.
 data DBPatVar = DBPatVar
   { dbPatVarName  :: PatVarName
-  , dbPatVarIndex :: Int
-  } deriving (Data, Show, Eq, Generic)
+  , dbPatVarIndex :: !Int
+  } deriving (Show, Eq, Generic)
 
 type DeBruijnPattern = Pattern' DBPatVar
 
@@ -528,7 +532,7 @@ data ConPatternInfo = ConPatternInfo
     --   variables they bind are unused. The GHC backend compiles lazy matches
     --   to lazy patterns in Haskell (TODO: not yet).
   }
-  deriving (Data, Show, Generic)
+  deriving (Show, Generic)
 
 noConPatternInfo :: ConPatternInfo
 noConPatternInfo = ConPatternInfo defaultPatternInfo False False Nothing False
@@ -648,14 +652,16 @@ data Substitution' a
     --     Γ ⊢ u :# ρ : Δ, A
     --   @
 
-  | Strengthen Impossible (Substitution' a)
+  | Strengthen Impossible !Int (Substitution' a)
     -- ^ Strengthening substitution.  First argument is @__IMPOSSIBLE__@.
-    --   Apply this to a term which does not contain variable 0
-    --   to lower all de Bruijn indices by one.
-    --   @
-    --             Γ ⊢ ρ : Δ
+    --   In @'Strengthen err n ρ@ the number @n@ must be non-negative.
+    --   This substitution should only be applied to values @t@ for
+    --   which none of the variables @0@ up to @n - 1@ are free in
+    --   @t[ρ]@, and in that case @n@ is subtracted from all free de
+    --   Bruijn indices in @t[ρ]@.
+    --        Γ ⊢ ρ : Δ    |Θ| = n
     --     ---------------------------
-    --     Γ ⊢ Strengthen ρ : Δ, A
+    --     Γ ⊢ Strengthen n ρ : Δ, Θ
     --   @
 
   | Wk !Int (Substitution' a)
@@ -701,17 +707,35 @@ instance Null (Substitution' a) where
 -- | View type as equality type.
 
 data EqualityView
-  = EqualityType
-    { eqtSort  :: Sort     -- ^ Sort of this type.
-    , eqtName  :: QName    -- ^ Builtin EQUALITY.
-    , eqtParams :: [Arg Term] -- ^ Hidden.  Empty or @Level@.
-    , eqtType  :: Arg Term -- ^ Hidden
-    , eqtLhs   :: Arg Term -- ^ NotHidden
-    , eqtRhs   :: Arg Term -- ^ NotHidden
-    }
+  = EqualityViewType EqualityTypeData
   | OtherType Type -- ^ reduced
   | IdiomType Type -- ^ reduced
 
+data EqualityTypeData = EqualityTypeData
+    { _eqtSort   :: Sort        -- ^ Sort of this type.
+    , _eqtName   :: QName       -- ^ Builtin EQUALITY.
+    , _eqtParams :: Args        -- ^ Hidden.  Empty or @Level@.
+    , _eqtType   :: Arg Term    -- ^ Hidden.
+    , _eqtLhs    :: Arg Term    -- ^ NotHidden.
+    , _eqtRhs    :: Arg Term    -- ^ NotHidden.
+    }
+
+pattern EqualityType
+  :: Sort
+  -> QName
+  -> Args
+  -> Arg Term
+  -> Arg Term
+  -> Arg Term
+  -> EqualityView
+pattern EqualityType{ eqtSort, eqtName, eqtParams, eqtType, eqtLhs, eqtRhs } =
+  EqualityViewType (EqualityTypeData eqtSort eqtName eqtParams eqtType eqtLhs eqtRhs)
+
+-- The COMPLETE pragma is new in GHC 8.2
+#if __GLASGOW_HASKELL__ >= 802
+{-# COMPLETE EqualityType, OtherType, IdiomType #-}
+#endif
+
 isEqualityType :: EqualityView -> Bool
 isEqualityType EqualityType{} = True
 isEqualityType OtherType{}    = False
@@ -1022,10 +1046,6 @@ hasElims v =
     MetaV x es -> Just (MetaV x, es)
     Con{}      -> Nothing
     Lit{}      -> Nothing
-    -- Andreas, 2016-04-13, Issue 1932: We convert λ x → x .f  into f
-    Lam h (Abs _ v) | visible h -> case v of
-      Var 0 [Proj _o f] -> Just (Def f, [])
-      _ -> Nothing
     Lam{}      -> Nothing
     Pi{}       -> Nothing
     Sort{}     -> Nothing
@@ -1045,11 +1065,12 @@ instance Null (Tele a) where
 -- | A 'null' clause is one with no patterns and no rhs.
 --   Should not exist in practice.
 instance Null Clause where
-  empty = Clause empty empty empty empty empty empty False Nothing Nothing Nothing empty
-  null (Clause _ _ tel pats body _ _ _ _ _ _)
+  empty = Clause empty empty empty empty empty empty False Nothing Nothing Nothing empty empty
+  null (Clause _ _ tel pats body _ _ _ _ _ _ wm)
     =  null tel
     && null pats
     && null body
+    && null wm
 
 
 ---------------------------------------------------------------------------
@@ -1120,6 +1141,7 @@ instance TermSize Sort where
     SSet l    -> 1 + tsize l
     SizeUniv  -> 1
     LockUniv  -> 1
+    IntervalUniv -> 1
     PiSort a s1 s2 -> 1 + tsize a + tsize s1 + tsize s2
     FunSort s1 s2 -> 1 + tsize s1 + tsize s2
     UnivSort s -> 1 + tsize s
@@ -1134,12 +1156,12 @@ instance TermSize PlusLevel where
   tsize (Plus _ a)      = tsize a
 
 instance TermSize a => TermSize (Substitution' a) where
-  tsize IdS                = 1
-  tsize (EmptyS _)         = 1
-  tsize (Wk _ rho)         = 1 + tsize rho
-  tsize (t :# rho)         = 1 + tsize t + tsize rho
-  tsize (Strengthen _ rho) = 1 + tsize rho
-  tsize (Lift _ rho)       = 1 + tsize rho
+  tsize IdS                  = 1
+  tsize (EmptyS _)           = 1
+  tsize (Wk _ rho)           = 1 + tsize rho
+  tsize (t :# rho)           = 1 + tsize t + tsize rho
+  tsize (Strengthen _ _ rho) = 1 + tsize rho
+  tsize (Lift _ rho)         = 1 + tsize rho
 
 ---------------------------------------------------------------------------
 -- * KillRange instances.
@@ -1179,6 +1201,7 @@ instance KillRange Sort where
     Inf f n    -> Inf f n
     SizeUniv   -> SizeUniv
     LockUniv   -> LockUniv
+    IntervalUniv -> IntervalUniv
     Type a     -> killRange1 Type a
     Prop a     -> killRange1 Prop a
     SSet a     -> killRange1 SSet a
@@ -1190,12 +1213,12 @@ instance KillRange Sort where
     s@DummyS{} -> s
 
 instance KillRange Substitution where
-  killRange IdS                  = IdS
-  killRange (EmptyS err)         = EmptyS err
-  killRange (Wk n rho)           = killRange1 (Wk n) rho
-  killRange (t :# rho)           = killRange2 (:#) t rho
-  killRange (Strengthen err rho) = killRange1 (Strengthen err) rho
-  killRange (Lift n rho)         = killRange1 (Lift n) rho
+  killRange IdS                    = IdS
+  killRange (EmptyS err)           = EmptyS err
+  killRange (Wk n rho)             = killRange1 (Wk n) rho
+  killRange (t :# rho)             = killRange2 (:#) t rho
+  killRange (Strengthen err n rho) = killRange1 (Strengthen err n) rho
+  killRange (Lift n rho)           = killRange1 (Lift n) rho
 
 instance KillRange PatOrigin where
   killRange = id
@@ -1221,8 +1244,8 @@ instance KillRange a => KillRange (Pattern' a) where
       DefP o q ps      -> killRange2 (DefP o) q ps
 
 instance KillRange Clause where
-  killRange (Clause rl rf tel ps body t catchall exact recursive unreachable ell) =
-    killRange10 Clause rl rf tel ps body t catchall exact recursive unreachable ell
+  killRange (Clause rl rf tel ps body t catchall exact recursive unreachable ell wm) =
+    killRange11 Clause rl rf tel ps body t catchall exact recursive unreachable ell wm
 
 instance KillRange a => KillRange (Tele a) where
   killRange = fmap killRange
@@ -1241,12 +1264,16 @@ instance Pretty a => Pretty (Substitution' a) where
   prettyPrec = pr
     where
     pr p rho = case rho of
-      IdS              -> "idS"
-      EmptyS err       -> "emptyS"
-      t :# rho         -> mparens (p > 2) $ sep [ pr 2 rho <> ",", prettyPrec 3 t ]
-      Strengthen _ rho -> mparens (p > 9) $ "strS" <+> pr 10 rho
-      Wk n rho         -> mparens (p > 9) $ text ("wkS " ++ show n) <+> pr 10 rho
-      Lift n rho       -> mparens (p > 9) $ text ("liftS " ++ show n) <+> pr 10 rho
+      IdS                -> "idS"
+      EmptyS err         -> "emptyS"
+      t :# rho           -> mparens (p > 2) $
+                            sep [ pr 2 rho <> ",", prettyPrec 3 t ]
+      Strengthen _ n rho -> mparens (p > 9) $
+                            text ("strS " ++ show n) <+> pr 10 rho
+      Wk n rho           -> mparens (p > 9) $
+                            text ("wkS " ++ show n) <+> pr 10 rho
+      Lift n rho         -> mparens (p > 9) $
+                            text ("liftS " ++ show n) <+> pr 10 rho
 
 instance Pretty Term where
   prettyPrec p v =
@@ -1338,10 +1365,10 @@ instance Pretty Sort where
       SSet l -> mparens (p > 9) $ "SSet" <+> prettyPrec 10 l
       SizeUniv -> "SizeUniv"
       LockUniv -> "LockUniv"
+      IntervalUniv -> "IntervalUniv"
       PiSort a s1 s2 -> mparens (p > 9) $
-        "piSort" <+> pDom (domInfo a) (text (absName s2) <+> ":" <+> pretty (unDom a))
-                      <+> parens (sep [ text ("λ " ++ absName s2 ++ " ->")
-                                      , nest 2 $ pretty (unAbs s2) ])
+        "piSort" <+> pDom (domInfo a) (text (absName s2) <+> ":" <+> pretty (unDom a) <+> ":" <+> pretty s1)
+                      <+> parens (pretty (unAbs s2))
       FunSort a b -> mparens (p > 9) $
         "funSort" <+> prettyPrec 10 a <+> prettyPrec 10 b
       UnivSort s -> mparens (p > 9) $ "univSort" <+> prettyPrec 10 s
@@ -1380,6 +1407,10 @@ instance Pretty a => Pretty (Pattern' a) where
   prettyPrec n (IApplyP _o _ _ x) = prettyPrec n x
 --  prettyPrec n (IApplyP _o u0 u1 x) = text "@[" <> prettyPrec 0 u0 <> text ", " <> prettyPrec 0 u1 <> text "]" <> prettyPrec n x
 
+instance Pretty a => Pretty (Blocked a) where
+  pretty (Blocked x a) = ("[" <+> pretty a <+> "]") <> pretty x
+  pretty (NotBlocked _ x) = pretty x
+
 -----------------------------------------------------------------------------
 -- * NFData instances
 -----------------------------------------------------------------------------
@@ -1411,6 +1442,7 @@ instance NFData Sort where
     SSet l   -> rnf l
     SizeUniv -> ()
     LockUniv -> ()
+    IntervalUniv -> ()
     PiSort a b c -> rnf (a, b, unAbs c)
     FunSort a b -> rnf (a, b)
     UnivSort a -> rnf a
@@ -1425,7 +1457,7 @@ instance NFData PlusLevel where
   rnf (Plus n l) = rnf (n, l)
 
 instance NFData e => NFData (Dom e) where
-  rnf (Dom a b c d e) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d `seq` rnf e
+  rnf (Dom a c d e f) = rnf a `seq` rnf c `seq` rnf d `seq` rnf e `seq` rnf f
 
 instance NFData DataOrRecord
 instance NFData ConHead
