@@ -96,7 +96,7 @@ instance TermLike Term where
     t@Lit{}     -> f t
     Sort s      -> f =<< Sort <$> traverseTermM f s
     DontCare mv -> f =<< DontCare <$> traverseTermM f mv
-    t@Dummy{}   -> f t
+    Dummy s xs  -> f =<< Dummy s <$> traverseTermM f xs
 
   foldTerm f t = f t `mappend` case t of
     Var i xs    -> foldTerm f xs
@@ -109,7 +109,7 @@ instance TermLike Term where
     Lit _       -> mempty
     Sort s      -> foldTerm f s
     DontCare mv -> foldTerm f mv
-    Dummy{}     -> mempty
+    Dummy _ xs  -> foldTerm f xs
 
 instance TermLike Level where
   traverseTermM f (Max n as) = Max n <$> traverseTermM f as
@@ -131,12 +131,13 @@ instance TermLike Sort where
     SSet l     -> SSet <$> traverseTermM f l
     s@SizeUniv -> pure s
     s@LockUniv -> pure s
+    s@IntervalUniv -> pure s
     PiSort a b c -> PiSort   <$> traverseTermM f a <*> traverseTermM f b <*> traverseTermM f c
     FunSort a b -> FunSort   <$> traverseTermM f a <*> traverseTermM f b
     UnivSort a -> UnivSort <$> traverseTermM f a
     MetaS x es -> MetaS x  <$> traverseTermM f es
     DefS q es  -> DefS q   <$> traverseTermM f es
-    s@DummyS{} -> pure s
+    s@(DummyS _) -> pure s
 
   foldTerm f = \case
     Type l     -> foldTerm f l
@@ -145,12 +146,13 @@ instance TermLike Sort where
     SSet l     -> foldTerm f l
     SizeUniv   -> mempty
     LockUniv   -> mempty
+    IntervalUniv -> mempty
     PiSort a b c -> foldTerm f a <> foldTerm f b <> foldTerm f c
     FunSort a b -> foldTerm f a <> foldTerm f b
     UnivSort a -> foldTerm f a
     MetaS _ es -> foldTerm f es
     DefS _ es  -> foldTerm f es
-    DummyS{}   -> mempty
+    DummyS _   -> mempty
 
 instance TermLike EqualityView where
 
