@@ -37,11 +37,11 @@ import Control.Monad.Except
 import Control.Monad.State
 
 import Data.Int
-import Data.Data  ( Data )
 import Data.Maybe ( listToMaybe )
 
 import Agda.Interaction.Options.Warnings
 
+import Agda.Syntax.Concrete.Attribute
 import Agda.Syntax.Position
 import Agda.Syntax.Parser.Tokens ( Keyword( KwMutual ) )
 
@@ -76,6 +76,8 @@ data ParseState = PState
                                              --   (states can be nested so we need a stack)
     , parseFlags    :: ParseFlags            -- ^ parametrization of the parser
     , parseWarnings :: ![ParseWarning]       -- ^ In reverse order.
+    , parseCohesion :: !CohesionAttributes
+      -- ^ Every encountered occurrence of a cohesion attribute.
     }
     deriving Show
 
@@ -151,12 +153,12 @@ data ParseError
 
   -- | Parse errors that concern a whole file.
   | InvalidExtensionError
-    { errPath      :: !AbsolutePath
+    { errPath      :: !RangeFile
                       -- ^ The file which the error concerns.
     , errValidExts :: [String]
     }
   | ReadFileError
-    { errPath      :: !AbsolutePath
+    { errPath      :: !RangeFile
     , errIOError   :: IOError
     }
   deriving Show
@@ -172,7 +174,7 @@ data ParseWarning
     -- ^ Unsupported attribute.
   | MultipleAttributes Range !(Maybe String)
     -- ^ Multiple attributes.
-  deriving (Data, Show)
+  deriving Show
 
 instance NFData ParseWarning where
   rnf (OverlappingTokensWarning _) = ()
@@ -293,6 +295,7 @@ initStatePos pos flags inp st =
                                                 -- Just for better errors on stray @constructor@ decls.
                 , parseFlags        = flags
                 , parseWarnings     = []
+                , parseCohesion     = []
                 }
   where
   pos' = pos { srcFile = () }
@@ -300,8 +303,8 @@ initStatePos pos flags inp st =
 -- | Constructs the initial state of the parser. The string argument
 --   is the input string, the file path is only there because it's part
 --   of a position.
-initState :: Maybe AbsolutePath -> ParseFlags -> String -> [LexState]
-          -> ParseState
+initState ::
+  Maybe RangeFile -> ParseFlags -> String -> [LexState] -> ParseState
 initState file = initStatePos (startPos file)
 
 -- | The default flags.
