@@ -36,15 +36,15 @@ import qualified Control.Monad.Fail as Fail
 
 import qualified Data.Map as Map
 import Data.Maybe
-import Data.Monoid
+import Data.Monoid hiding ((<>))
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Map (Map)
 import qualified Data.Foldable as Fold
 import Data.Void
 import Data.List (sortBy)
-import Data.List.NonEmpty (NonEmpty(..))
-import qualified Data.List.NonEmpty as NonEmpty
+import Data.Semigroup (Semigroup, (<>))
+import Data.String
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
@@ -64,8 +64,13 @@ import Agda.Syntax.Scope.Monad ( tryResolveName )
 
 import Agda.TypeChecking.Monad.State (getScope, getAllPatternSyns)
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Monad.MetaVars
+import Agda.TypeChecking.Monad.Pure
+import Agda.TypeChecking.Monad.Signature
+import {-# SOURCE #-} Agda.TypeChecking.Pretty (prettyTCM)
 import Agda.Interaction.Options
 
 import qualified Agda.Utils.AssocList as AssocList
@@ -79,7 +84,7 @@ import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty hiding ((<>))
 import Agda.Utils.Singleton
 import Agda.Utils.Suffix
 
@@ -91,7 +96,7 @@ data Env = Env { takenVarNames :: Set A.Name
                   -- ^ Abstract names currently in scope. Unlike the
                   --   ScopeInfo, this includes names for hidden
                   --   arguments inserted by the system.
-               , takenDefNames :: Set C.Name
+               , takenDefNames :: Set C.NameParts
                   -- ^ Concrete names of all definitions in scope
                , currentScope :: ScopeInfo
                , builtins     :: Map String A.QName
@@ -132,14 +137,20 @@ makeEnv scope = do
         , foldPatternSynonyms = foldPatSyns
         }
   where
+    defs = Set.map nameParts . Map.keysSet $
+        Map.filterWithKey usefulDef $
+        nsNames $ everythingInScope scope
+
     -- Jesper, 2018-12-10: It's fine to shadow generalizable names as
     -- they will never show up directly in printed terms.
     notGeneralizeName AbsName{ anameKind = k }  =
       not (k == GeneralizeName || k == DisallowedGeneralizeName)
 
-    defs = Map.keysSet $
-           Map.filter (all notGeneralizeName) $
-           nsNames $ everythingInScope scope
+    usefulDef C.NoName{} _ = False
+    usefulDef C.Name{} names = all notGeneralizeName names
+
+    nameParts (C.NoName {}) = __IMPOSSIBLE__
+    nameParts (C.Name { nameNameParts }) = nameNameParts
 
 currentPrecedence :: AbsToCon PrecedenceStack
 currentPrecedence = asks $ (^. scopePrecedence) . currentScope
@@ -179,7 +190,7 @@ isBuiltinFun = asks $ is . builtins
   where is m q b = Just q == Map.lookup b m
 
 -- | Resolve a concrete name. If illegally ambiguous fail with the ambiguous names.
-resolveName :: KindsOfNames -> Maybe (Set A.Name) -> C.QName -> AbsToCon (Either (NonEmpty A.QName) ResolvedName)
+resolveName :: KindsOfNames -> Maybe (Set A.Name) -> C.QName -> AbsToCon (Either AmbiguousNameReason ResolvedName)
 resolveName kinds candidates q = runExceptT $ tryResolveName kinds candidates q
 
 -- | Treat illegally ambiguous names as UnknownNames.
@@ -188,20 +199,17 @@ resolveName_ q cands = fromRight (const UnknownName) <$> resolveName allKindsOfN
 
 -- The Monad --------------------------------------------------------------
 
--- | We need:
---   - Read access to the AbsToCon environment
---   - Read access to the TC environment
---   - Read access to the TC state
---   - Read and write access to the stConcreteNames part of the TC state
---   - Read access to the options
---   - Permission to print debug messages
+-- | The function 'runAbsToCon' can target any monad that satisfies
+-- the constraints of 'MonadAbsToCon'.
 type MonadAbsToCon m =
-  ( MonadTCEnv m
-  , ReadTCState m
+  ( MonadFresh NameId m
+  , MonadInteractionPoints m
   , MonadStConcreteNames m
   , HasOptions m
-  , HasBuiltins m
-  , MonadDebug m
+  , PureTCM m
+  , IsString (m Doc)
+  , Null (m Doc)
+  , Semigroup (m Doc)
   )
 
 newtype AbsToCon a = AbsToCon
@@ -261,9 +269,47 @@ instance MonadDebug AbsToCon where
   traceDebugMessage  k n s cont = AbsToCon $ traceDebugMessage  k n s $ unAbsToCon cont  -- can't eta-reduce!
   verboseBracket     k n s cont = AbsToCon $ verboseBracket     k n s $ unAbsToCon cont  -- because of GHC-9.0
 
-  getVerbosity     = defaultGetVerbosity
-  isDebugPrinting  = defaultIsDebugPrinting
-  nowDebugPrinting = defaultNowDebugPrinting
+  getVerbosity      = defaultGetVerbosity
+  getProfileOptions = defaultGetProfileOptions
+  isDebugPrinting   = defaultIsDebugPrinting
+  nowDebugPrinting  = defaultNowDebugPrinting
+
+instance HasConstInfo AbsToCon where
+  getConstInfo' a      = AbsToCon (getConstInfo' a)
+  getRewriteRulesFor a = AbsToCon (getRewriteRulesFor a)
+
+instance MonadAddContext AbsToCon where
+  addCtx a b c = AbsToCon (addCtx a b (unAbsToCon c))
+
+  addLetBinding' a b c d =
+    AbsToCon (addLetBinding' a b c (unAbsToCon d))
+
+  updateContext a b c = AbsToCon (updateContext a b (unAbsToCon c))
+
+  withFreshName a b c =
+    AbsToCon (withFreshName a b (\x -> unAbsToCon (c x)))
+
+instance MonadReduce AbsToCon where
+  liftReduce a = AbsToCon (liftReduce a)
+
+instance PureTCM AbsToCon where
+
+instance MonadFresh NameId AbsToCon where
+  fresh = AbsToCon fresh
+
+instance MonadInteractionPoints AbsToCon where
+  freshInteractionId        = AbsToCon freshInteractionId
+  modifyInteractionPoints a = AbsToCon (modifyInteractionPoints a)
+
+instance IsString (AbsToCon Doc) where
+  fromString a = AbsToCon (fromString a)
+
+instance Null (AbsToCon Doc) where
+  empty = AbsToCon empty
+  null  = __IMPOSSIBLE__
+
+instance Semigroup (AbsToCon Doc) where
+  a <> b = AbsToCon (unAbsToCon a <> unAbsToCon b)
 
 runAbsToCon :: MonadAbsToCon m => AbsToCon c -> m c
 runAbsToCon m = do
@@ -362,7 +408,9 @@ pickConcreteName x y = modifyConcreteNames $ flip Map.alter x $ \case
 -- | For the given abstract name, return the names that could shadow it.
 shadowingNames :: (ReadTCState m, MonadStConcreteNames m)
                => A.Name -> m (Set RawName)
-shadowingNames x = Set.fromList . Map.findWithDefault [] x <$> useR stShadowingNames
+shadowingNames x =
+  Set.fromList . Fold.toList . Map.findWithDefault mempty x <$>
+    useR stShadowingNames
 
 toConcreteName :: A.Name -> AbsToCon C.Name
 toConcreteName x | y <- nameConcrete x , isNoName y = return y
@@ -399,13 +447,21 @@ chooseName x = lookupNameInScope (nameConcrete x) >>= \case
     return $ nameConcrete x
   -- Otherwise we pick a name that does not shadow other names
   _ -> do
+    takenDefs <- asks takenDefNames
     taken   <- takenNames
     toAvoid <- shadowingNames x
     glyphMode <- optUseUnicode <$> pragmaOptions
     let freshNameMode = case glyphMode of
           UnicodeOk -> A.UnicodeSubscript
           AsciiOnly -> A.AsciiCounter
-    let shouldAvoid = (`Set.member` (taken `Set.union` toAvoid)) . C.nameToRawName
+
+        shouldAvoid C.NoName {} = False
+        shouldAvoid name@C.Name { nameNameParts } =
+          let raw = C.nameToRawName name in
+          nameNameParts `Set.member` takenDefs ||
+          raw `Set.member` taken ||
+          raw `Set.member` toAvoid
+
         y = firstNonTakenName freshNameMode shouldAvoid $ nameConcrete x
     reportSLn "toConcrete.bindName" 80 $ render $ vcat
       [ "picking concrete name for:" <+> text (C.nameToRawName $ nameConcrete x)
@@ -418,11 +474,10 @@ chooseName x = lookupNameInScope (nameConcrete x) >>= \case
   where
     takenNames :: AbsToCon (Set RawName)
     takenNames = do
-      xs <- asks takenDefNames
       ys0 <- asks takenVarNames
       reportSLn "toConcrete.bindName" 90 $ render $ "abstract names of local vars: " <+> prettyList_ (map (C.nameToRawName . nameConcrete) $ Set.toList ys0)
       ys <- Set.fromList . concat <$> mapM hasConcreteNames (Set.toList ys0)
-      return $ Set.map C.nameToRawName $ xs `Set.union` ys
+      return $ Set.map C.nameToRawName ys
 
 
 -- | Add a abstract name to the scope and produce an available concrete version of it.
@@ -680,9 +735,9 @@ instance ToConcrete ResolvedName where
   toConcrete = \case
     VarName x _          -> C.QName <$> toConcrete x
     DefinedName _ x s    -> addSuffixConcrete s $ toConcrete x
-    FieldName xs         -> toConcrete (NonEmpty.head xs)
-    ConstructorName _ xs -> toConcrete (NonEmpty.head xs)
-    PatternSynResName xs -> toConcrete (NonEmpty.head xs)
+    FieldName xs         -> toConcrete (List1.head xs)
+    ConstructorName _ xs -> toConcrete (List1.head xs)
+    PatternSynResName xs -> toConcrete (List1.head xs)
     UnknownName          -> __IMPOSSIBLE__
 
 addSuffixConcrete :: HasOptions m => A.Suffix -> m C.QName -> m C.QName
@@ -727,9 +782,10 @@ instance ToConcrete A.Expr where
       return $ C.QuestionMark (getRange i) $
                  interactionId ii <$ guard (preserve || isJust (metaNumber i))
 
-    toConcrete (A.Underscore i)     = return $
-      C.Underscore (getRange i) $
-        prettyShow . NamedMeta (metaNameSuggestion i) . MetaId . metaId <$> metaNumber i
+    toConcrete (A.Underscore i) =
+      C.Underscore (getRange i) <$>
+      traverse (render <.> prettyTCM)
+        (NamedMeta (metaNameSuggestion i) <$> metaNumber i)
 
     toConcrete (A.Dot i e) =
       C.Dot (getRange i) <$> toConcrete e
@@ -826,7 +882,7 @@ instance ToConcrete A.Expr where
                 reportSLn "extendedlambda" 50 $ "abstractToConcrete extended lambda patterns ps = " ++ prettyShow ps
                 return $ LamClause ps rhs ca
               decl2clause _ = __IMPOSSIBLE__
-          C.ExtendedLam (getRange i) erased . List1.fromList <$>
+          C.ExtendedLam (getRange i) erased . List1.fromListSafe __IMPOSSIBLE__ <$>
             mapM decl2clause decls
             -- TODO List1: can we demonstrate non-emptiness?
 
@@ -875,8 +931,6 @@ instance ToConcrete A.Expr where
       bracket appBrackets $ do
         C.RecUpdate (getRange i) <$> toConcrete e <*> toConcreteTop fs
 
-    toConcrete (A.ETel tel) = C.ETel . catMaybes <$> toConcrete tel
-
     toConcrete (A.ScopedExpr _ e) = toConcrete e
     toConcrete (A.Quote i) = return $ C.Quote (getRange i)
     toConcrete (A.QuoteTerm i) = return $ C.QuoteTerm (getRange i)
@@ -892,7 +946,7 @@ makeDomainFree :: A.LamBinding -> A.LamBinding
 makeDomainFree b@(A.DomainFull (A.TBind _ tac (x :| []) t)) =
   case unScope t of
     A.Underscore A.MetaInfo{metaNumber = Nothing} ->
-      A.DomainFree tac x
+      A.DomainFree (tbTacticAttr tac) x
     _ -> b
 makeDomainFree b = b
 
@@ -946,10 +1000,10 @@ instance ToConcrete A.TypedBinding where
     type ConOfAbs A.TypedBinding = Maybe C.TypedBinding
 
     bindToConcrete (A.TBind r t xs e) ret = do
-        t <- traverse toConcrete t
+        tac <- traverse toConcrete (tbTacticAttr t)
         bindToConcrete (fmap forceNameIfHidden xs) $ \ xs -> do
           e <- toConcreteTop e
-          let setTac x = x { bnameTactic = t }
+          let setTac x = x { bnameTactic = tac , C.bnameIsFinite = tbFinite t }
           ret $ Just $ C.TBind r (fmap (updateNamedArg (fmap setTac)) xs) e
     bindToConcrete (A.TLet r lbs) ret =
         bindToConcrete lbs $ \ ds -> do
@@ -992,15 +1046,15 @@ instance ToConcrete A.LetBinding where
 instance ToConcrete A.WhereDeclarations where
   type ConOfAbs A.WhereDeclarations = WhereClause
 
-  bindToConcrete (A.WhereDecls _ Nothing) ret = ret C.NoWhere
-  bindToConcrete (A.WhereDecls (Just am) (Just (A.Section _ _ _ ds))) ret = do
+  bindToConcrete (A.WhereDecls _ _ Nothing) ret = ret C.NoWhere
+  bindToConcrete (A.WhereDecls (Just am) False (Just (A.Section _ _ _ ds))) ret = do
     ds' <- declsToConcrete ds
     cm  <- unqualify <$> lookupModule am
     -- Andreas, 2016-07-08 I put PublicAccess in the following SomeWhere
     -- Should not really matter for printing...
     let wh' = (if isNoName cm then AnyWhere noRange else SomeWhere noRange cm PublicAccess) $ ds'
     local (openModule' am defaultImportDir id) $ ret wh'
-  bindToConcrete (A.WhereDecls _ (Just d)) ret =
+  bindToConcrete (A.WhereDecls _ _ (Just d)) ret =
     ret . AnyWhere noRange =<< toConcrete d
 
 mergeSigAndDef :: [C.Declaration] -> [C.Declaration]
@@ -1223,6 +1277,8 @@ instance ToConcrete A.Declaration where
     xs <- mapM (unqual <=< toConcrete) xs
     (:[]) . C.UnquoteDef (getRange i) xs <$> toConcrete e
 
+  toConcrete (A.UnquoteData i xs uc j cs e) = __IMPOSSIBLE__
+
 
 data RangeAndPragma = RangeAndPragma Range A.Pragma
 
@@ -1240,6 +1296,7 @@ instance ToConcrete RangeAndPragma where
     A.StaticPragma x -> C.StaticPragma r <$> toConcrete x
     A.InjectivePragma x -> C.InjectivePragma r <$> toConcrete x
     A.InlinePragma b x -> C.InlinePragma r b <$> toConcrete x
+    A.NotProjectionLikePragma q -> C.NotProjectionLikePragma r <$> toConcrete q
     A.EtaPragma x    -> C.EtaPragma    r <$> toConcrete x
     A.DisplayPragma f ps rhs ->
       C.DisplayPragma r <$> toConcrete (A.DefP (PatRange noRange) (unambiguous f) ps) <*> toConcrete rhs
@@ -1776,5 +1833,5 @@ instance ToConcrete InteractionId where
 
 instance ToConcrete NamedMeta where
     type ConOfAbs NamedMeta = C.Expr
-    toConcrete i = do
-      return $ C.Underscore noRange (Just $ prettyShow i)
+    toConcrete i =
+      C.Underscore noRange . Just . render <$> prettyTCM i
