@@ -177,9 +177,9 @@ instance Reify MetaId where
     type ReifiesTo MetaId = Expr
 
     reifyWhen = reifyWhenE
-    reify x@(MetaId n) = do
+    reify x = do
       b <- asksTC envPrintMetasBare
-      mi  <- mvInfo <$> lookupMeta x
+      mi  <- mvInfo <$> lookupLocalMeta x
       let mi' = Info.MetaInfo
                  { metaRange          = getRange $ miClosRange mi
                  , metaScope          = clScope $ miClosRange mi
@@ -435,9 +435,11 @@ reifyTerm :: MonadReify m => Bool -> Term -> m Expr
 reifyTerm expandAnonDefs0 v0 = do
   -- Jesper 2018-11-02: If 'PrintMetasBare', drop all meta eliminations.
   metasBare <- asksTC envPrintMetasBare
+  reportSDoc "reify.term" 80 $ pure $ "reifyTerm v0 = " <+> pretty v0
   v <- instantiate v0 >>= \case
     I.MetaV x _ | metasBare -> return $ I.MetaV x []
     v -> return v
+  reportSDoc "reify.term" 80 $ pure $ "reifyTerm v = " <+> pretty v
   -- Ulf 2014-07-10: Don't expand anonymous when display forms are disabled
   -- (i.e. when we don't care about nice printing)
   expandAnonDefs <- return expandAnonDefs0 `and2M` displayFormsEnabled
@@ -445,6 +447,7 @@ reifyTerm expandAnonDefs0 v0 = do
   -- then we print system-generated projections as postfix, else prefix.
   havePfp <- optPostfixProjections <$> pragmaOptions
   let pred = if havePfp then (== ProjPrefix) else (/= ProjPostfix)
+  reportSDoc "reify.term" 80 $ pure $ "reifyTerm (unSpine v) = " <+> pretty (unSpine' pred v)
   case unSpine' pred v of
     -- Hack to print generalized field projections with nicer names. Should
     -- only show up in errors. Check the spined form!
@@ -528,11 +531,15 @@ reifyTerm expandAnonDefs0 v0 = do
     I.Level l      -> reify l
     I.Pi a b       -> case b of
         NoAbs _ b'
-          | visible a   -> uncurry (A.Fun $ noExprInfo) <$> reify (a, b')
+          | visible a, not (domIsFinite a) -> uncurry (A.Fun $ noExprInfo) <$> reify (a, b')
             -- Andreas, 2013-11-11 Hidden/Instance I.Pi must be A.Pi
             -- since (a) the syntax {A} -> B or {{A}} -> B is not legal
             -- and (b) the name of the binder might matter.
             -- See issue 951 (a) and 952 (b).
+            --
+            -- Amy, 2022-09-05: Can't be finite either, since otherwise
+            -- we say ".(IsOne φ) → A ≠ .(IsOne φ) → A" with no
+            -- indication of which is finite and which isn't
           | otherwise   -> mkPi b =<< reify a
         b               -> mkPi b =<< do
           ifM (domainFree a (absBody b))
@@ -543,7 +550,9 @@ reifyTerm expandAnonDefs0 v0 = do
           tac <- traverse reify $ domTactic a
           (x, b) <- reify b
           let xs = singleton $ Arg info $ Named (domName a) $ mkBinder_ x
-          return $ A.Pi noExprInfo (singleton $ TBind noRange tac xs a') b
+          return $ A.Pi noExprInfo
+            (singleton $ TBind noRange (TypedBindingInfo tac (domIsFinite a)) xs a')
+            b
         -- We can omit the domain type if it doesn't have any free variables
         -- and it's mentioned in the target type.
         domainFree a b = do
@@ -556,7 +565,7 @@ reifyTerm expandAnonDefs0 v0 = do
 
           es' <- reify es
 
-          mv <- lookupMeta x
+          mv <- lookupLocalMeta x
           (msub1,meta_tel,msub2) <- do
             local_chkpt <- viewTC eCurrentCheckpoint
             (chkpt, tel, msub2) <- enterClosure mv $ \ _ ->
@@ -689,7 +698,7 @@ reifyTerm expandAnonDefs0 v0 = do
         alreadyPrinting <- viewTC ePrintingPatternLambdas
 
         extLam <- case def of
-          Function{ funExtLam = Just{}, funProjection = Just{} } -> __IMPOSSIBLE__
+          Function{ funExtLam = Just{}, funProjection = Right{} } -> __IMPOSSIBLE__
           Function{ funExtLam = Just (ExtLamInfo m b sys) } ->
             Just . (,Strict.toLazy sys) . size <$> lookupSection m
           _ -> return Nothing
@@ -703,7 +712,7 @@ reifyTerm expandAnonDefs0 v0 = do
           _ -> do
            (pad, nes :: [Elim' (Named_ Term)]) <- case def of
 
-            Function{ funProjection = Just Projection{ projIndex = np } } | np > 0 -> do
+            Function{ funProjection = Right Projection{ projIndex = np } } | np > 0 -> do
               reportSLn "reify.def" 70 $ "  def. is a projection with projIndex = " ++ show np
 
               -- This is tricky:
@@ -795,7 +804,7 @@ reifyTerm expandAnonDefs0 v0 = do
       :: MonadReify m
       => QName -> ArgInfo -> Int -> Maybe System -> [I.Clause]
       -> I.Elims -> m Expr
-    reifyExtLam x i npars msys cls es = do
+    reifyExtLam x ai npars msys cls es = do
       reportSLn "reify.def" 10 $ "reifying extended lambda " ++ prettyShow x
       reportSLn "reify.def" 50 $ render $ nest 2 $ vcat
         [ "npars =" <+> pretty npars
@@ -815,13 +824,14 @@ reifyTerm expandAnonDefs0 v0 = do
       let cx     = nameConcrete $ qnameName x
           dInfo  = mkDefInfo cx noFixity' PublicAccess ConcreteDef
                      (getRange x)
-          erased = case getQuantity i of
+          erased = case getQuantity ai of
             Quantity0 o -> Erased o
             Quantityω o -> NotErased o
             Quantity1 o -> __IMPOSSIBLE__
-      elims (A.ExtendedLam exprNoRange dInfo erased x $
-             List1.fromList cls)
-        =<< reify rest
+          lam = case cls of
+            []       -> A.AbsurdLam exprNoRange NotHidden
+            (cl:cls) -> A.ExtendedLam exprNoRange dInfo erased x (cl :| cls)
+      elims lam =<< reify rest
 
 -- | @nameFirstIfHidden (x:a) ({e} es) = {x = e} es@
 nameFirstIfHidden :: Dom (ArgName, t) -> [Elim' a] -> [Elim' (Named_ a)]
@@ -1047,7 +1057,6 @@ instance BlankVars A.Expr where
     A.Let _ _ _              -> __IMPOSSIBLE__
     A.Rec i es               -> A.Rec i $ blank bound es
     A.RecUpdate i e es       -> uncurry (A.RecUpdate i) $ blank bound (e, es)
-    A.ETel _                 -> __IMPOSSIBLE__
     A.Quote {}               -> __IMPOSSIBLE__
     A.QuoteTerm {}           -> __IMPOSSIBLE__
     A.Unquote {}             -> __IMPOSSIBLE__
@@ -1347,8 +1356,8 @@ instance Reify (QNamed System) where
         result = A.Clause (spineToLhs lhs) [] rhs A.noWhereDecls False
       return result
 
-instance Reify Type where
-    type ReifiesTo Type = Expr
+instance Reify I.Type where
+    type ReifiesTo I.Type = A.Type
 
     reifyWhen = reifyWhenE
     reify (I.El _ t) = reify t
@@ -1359,7 +1368,7 @@ instance Reify Sort where
     reifyWhen = reifyWhenE
     reify s = do
       s <- instantiateFull s
-      SortKit{..} <- sortKit
+      SortKit{..} <- infallibleSortKit
       case s of
         I.Type (I.ClosedLevel 0) -> return $ A.Def' nameOfSet A.NoSuffix
         I.Type (I.ClosedLevel n) -> return $ A.Def' nameOfSet (A.Suffix n)
@@ -1383,6 +1392,9 @@ instance Reify Sort where
         I.LockUniv  -> do
           lockU <- fromMaybe __IMPOSSIBLE__ <$> getName' builtinLockUniv
           return $ A.Def lockU
+        I.IntervalUniv -> do
+          intervalU <- fromMaybe __IMPOSSIBLE__ <$> getName' builtinIntervalUniv
+          return $ A.Def intervalU
         I.PiSort a s1 s2 -> do
           pis <- freshName_ ("piSort" :: String) -- TODO: hack
           (e1,e2) <- reify (s1, I.Lam defaultArgInfo $ fmap Sort s2)
@@ -1438,7 +1450,7 @@ instance Reify I.Telescope where
         name = domName arg
     tac <- traverse reify $ domTactic arg
     let xs = singleton $ Arg info $ Named name $ A.mkBinder_ x
-    return $ TBind r tac xs e : bs
+    return $ TBind r (TypedBindingInfo tac (domIsFinite arg)) xs e : bs
 
 instance Reify i => Reify (Dom i) where
     type ReifiesTo (Dom i) = Arg (ReifiesTo i)
