@@ -18,7 +18,10 @@ import Control.Monad.IO.Class ( MonadIO(..) )
 import Control.Monad.Except
 import Control.Monad.Reader
 
+import Data.DList (DList)
+import qualified Data.DList as DL
 import Data.Semigroup ( Semigroup(..) )
+import Data.Set (Set)
 import qualified Data.Set as Set
 
 import Agda.Interaction.Options (optTerminationDepth)
@@ -31,7 +34,7 @@ import Agda.Syntax.Position (noRange)
 
 import Agda.Termination.CutOff
 import Agda.Termination.Order (Order,le,unknown)
-import Agda.Termination.RecCheck (anyDefs)
+import Agda.Termination.RecCheck (MutualNames, anyDefs)
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Benchmark
@@ -56,16 +59,16 @@ import qualified Agda.Utils.VarSet as VarSet
 
 import Agda.Utils.Impossible
 
--- | The mutual block we are checking.
---
---   The functions are numbered according to their order of appearance
---   in this list.
-
-type MutualNames = [QName]
-
 -- | The target of the function we are checking.
 
-type Target = QName
+data Target
+  = TargetDef QName
+      -- ^ The target of recursion is a @record@, @data@, or unreducible @Def@.
+  | TargetRecord
+      -- ^ We are termination-checking a record.
+  | TargetOther
+      -- ^ None of the above two or unknown.
+  deriving (Eq, Show)
 
 -- | The current guardedness level.
 
@@ -94,13 +97,13 @@ data TerEnv = TerEnv
     -- ^ The names of the functions in the mutual block we are checking.
     --   This includes the internally generated functions
     --   (with, extendedlambda, coinduction).
-  , terUserNames :: [QName]
+  , terUserNames :: Set QName
     -- ^ The list of name actually appearing in the file (abstract syntax).
     --   Excludes the internally generated functions.
   , terHaveInlinedWith :: Bool
     -- ^ Does the actual clause result from with-inlining?
     --   (If yes, it may be ill-typed.)
-  , terTarget  :: Maybe Target
+  , terTarget  :: Target
     -- ^ Target type of the function we are currently termination checking.
     --   Only the constructors of 'Target' are considered guarding.
   , terDelayed :: Delayed
@@ -153,7 +156,7 @@ defaultTerEnv = TerEnv
   , terMutual                   = __IMPOSSIBLE__ -- needs to be set!
   , terCurrent                  = __IMPOSSIBLE__ -- needs to be set!
   , terHaveInlinedWith          = False
-  , terTarget                   = Nothing
+  , terTarget                   = TargetOther
   , terDelayed                  = NotDelayed
   , terMaskArgs                 = repeat False   -- use all arguments (mask none)
   , terMaskResult               = False          -- use result (do not mask)
@@ -279,13 +282,13 @@ terGetCutOff = terAsks terCutOff
 terGetMutual :: TerM MutualNames
 terGetMutual = terAsks terMutual
 
-terGetUserNames :: TerM [QName]
+terGetUserNames :: TerM (Set QName)
 terGetUserNames = terAsks terUserNames
 
-terGetTarget :: TerM (Maybe Target)
+terGetTarget :: TerM Target
 terGetTarget = terAsks terTarget
 
-terSetTarget :: Maybe Target -> TerM a -> TerM a
+terSetTarget :: Target -> TerM a -> TerM a
 terSetTarget t = terLocal $ \ e -> e { terTarget = t }
 
 terGetHaveInlinedWith :: TerM Bool
@@ -575,7 +578,11 @@ instance PrettyTCM a => PrettyTCM (Masked a) where
 
 -- * Call pathes
 
--- | The call information is stored as free monoid
+-- | Call paths.
+
+-- An old comment:
+--
+--   The call information is stored as free monoid
 --   over 'CallInfo'.  As long as we never look at it,
 --   only accumulate it, it does not matter whether we use
 --   'Set', (nub) list, or 'Tree'.
@@ -584,16 +591,24 @@ instance PrettyTCM a => PrettyTCM (Masked a) where
 --   Since we define no order on 'CallInfo' (expensive),
 --   we cannot use a 'Set' or nub list.
 --   Performance-wise, I could not see a difference between Set and list.
+--
+-- If the binary tree is balanced "incorrectly", then forcing it could
+-- be expensive, so a switch was made to difference lists.
 
-newtype CallPath = CallPath { callInfos :: [CallInfo] }
+newtype CallPath = CallPath (DList CallInfo)
   deriving (Show, Semigroup, Monoid)
 
+-- | The calls making up the call path.
+
+callInfos :: CallPath -> [CallInfo]
+callInfos (CallPath cs) = DL.toList cs
+
 -- | Only show intermediate nodes.  (Drop last 'CallInfo').
 instance Pretty CallPath where
-  pretty (CallPath cis0) = if null cis then empty else
+  pretty cis0 = if null cis then empty else
     P.hsep (map (\ ci -> arrow P.<+> P.pretty ci) cis) P.<+> arrow
     where
-      cis   = init cis0
+      cis   = init (callInfos cis0)
       arrow = "-->"
 
 -- * Size depth estimation
@@ -601,16 +616,23 @@ instance Pretty CallPath where
 -- | A very crude way of estimating the @SIZELT@ chains
 --   @i > j > k@ in context.  Returns 3 in this case.
 --   Overapproximates.
+class TerSetSizeDepth b where
+  terSetSizeDepth :: b -> TerM a -> TerM a
+
+instance TerSetSizeDepth Telescope where
+  terSetSizeDepth = terSetSizeDepth . telToList
 
 -- TODO: more precise analysis, constructing a tree
 -- of relations between size variables.
-terSetSizeDepth :: Telescope -> TerM a -> TerM a
-terSetSizeDepth tel cont = do
-  n <- liftTCM $ sum <$> do
-    forM (telToList tel) $ \ dom -> do
-      a <- reduce $ snd $ unDom dom
-      ifM (isJust <$> isSizeType a) (return 1) {- else -} $
-        case unEl a of
-          MetaV{} -> return 1
-          _       -> return 0
-  terLocal (set terSizeDepth n) cont
+instance TerSetSizeDepth ListTel where
+  terSetSizeDepth doms cont = do
+    n <- liftTCM $ sum <$> do
+      forM doms $ \ dom -> do
+        -- Andreas, 2022-03-12, TODO:
+        -- use ifBlocked?  Shouldn't blocked types be treated like metas?
+        a <- reduce $ snd $ unDom dom
+        ifM (isJust <$> isSizeType a) (return 1) {- else -} $
+          case unEl a of
+            MetaV{} -> return 1
+            _       -> return 0
+    terLocal (set terSizeDepth n) cont
