@@ -13,16 +13,19 @@
 
 
 module Agda.Termination.RecCheck
-    ( recursive
+    ( MutualNames
+    , recursive
     , anyDefs
     )
  where
 
 import Control.Monad (forM, forM_)
+import Data.Foldable
 import Data.Graph
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import qualified Data.Map as Map
+import qualified Data.Map.Strict as MapS
 import Data.Maybe
 import Data.Set (Set)
 import qualified Data.Set as Set
@@ -37,6 +40,13 @@ import Agda.Utils.Pretty  (prettyShow)
 
 import Agda.Utils.Impossible
 
+-- | The mutual block we are checking.
+--
+--   The functions are numbered according to their order of appearance
+--   in this set.
+
+type MutualNames = Set QName
+
 -- | We compute for each clause the set of potentially recursive names.
 type NamesPerClause = IntMap (Set QName)
 
@@ -46,16 +56,21 @@ type NamesPerClause = IntMap (Set QName)
 --
 --   As a side effect, update the 'clauseRecursive' field in the
 --   clauses belonging to the given functions.
-recursive :: [QName] -> TCM [[QName]]
+recursive :: Set QName -> TCM [MutualNames]
 recursive names = do
+  let names' = toList names
   -- For each function, get names per clause and total.
-  (perClauses, nss) <- unzip <$> mapM (recDef (names `hasElem`)) names
+  (perClauses, nss) <- unzip <$> mapM (recDef (`Set.member` names)) names'
   -- Create graph suitable for stronglyConnComp.
   -- Nodes are identical to node keys.
-  let graph  = zipWith (\ x ns -> (x, x, Set.toList ns)) names nss
+  let graph  = zipWith (\ x ns -> (x, x, Set.toList ns)) names' nss
   let sccs   = stronglyConnComp graph
-  let nonRec = mapMaybe (\case{ AcyclicSCC x -> Just x ; _ -> Nothing}) sccs
-  let recs   = mapMaybe (\case{ CyclicSCC xs -> Just xs; _ -> Nothing}) sccs
+  let nonRec = mapMaybe (\case AcyclicSCC x -> Just x
+                               _            -> Nothing)
+                 sccs
+  let recs   = mapMaybe (\case CyclicSCC xs -> Just (Set.fromList xs)
+                               _            -> Nothing)
+                 sccs
 
   reportSLn "rec.graph" 60 $ show graph
 
@@ -65,7 +80,7 @@ recursive names = do
   -- Mark individual clauses of recursive functions:
   --------------------------------------------------
   -- Map names to clause numbers to sets of mentioned names.
-  let clMap = Map.fromListWith __IMPOSSIBLE__ $ zip names perClauses
+  let clMap = Map.fromListWith __IMPOSSIBLE__ $ zip names' perClauses
   -- Walk through SCCs.
   forM_ recs $ \ scc -> do
     -- Does a set of names have an overlap with the current scc?
@@ -88,6 +103,9 @@ markNonRecursive q = modifySignature $ updateDefinition q $ updateTheDef $ \case
    { funTerminates = Just True
    , funClauses    = map (\ cl -> cl { clauseRecursive = Just False }) $ funClauses def
    }
+  def@Record{} -> def
+   { recTerminates = Just True
+   }
   def -> def
 
 -- | Mark all clauses of a function as recursive or non-recursive.
@@ -112,11 +130,17 @@ recDef include name = do
 
   -- Get names in body
   (perClause, ns2) <- case theDef def of
+
     Function{ funClauses = cls } -> do
       perClause <- do
         forM (zip [0..] cls) $ \ (i, cl) ->
           (i,) <$> anyDefs include cl
       return (IntMap.fromList perClause, mconcat $ map snd perClause)
+
+    Record{ recTel } -> do
+      ns <- anyDefs include recTel
+      return (IntMap.singleton 0 ns, ns)
+
     _ -> return (mempty, mempty)
 
   reportS "rec.graph" 20
@@ -131,14 +155,16 @@ recDef include name = do
 anyDefs :: GetDefs a => (QName -> Bool) -> a -> TCM (Set QName)
 anyDefs include a = do
   -- Prepare function to lookup metas outside of TCM
-  st <- getMetaStore
-  let lookup (MetaId x) = (mvInstantiation <$> IntMap.lookup x st) >>= \case
-        InstV _ v                      -> Just v    -- TODO: ignoring the lambdas might be bad?
-        Open                           -> Nothing
-        OpenInstance                   -> Nothing
-        BlockedConst{}                 -> Nothing
-        PostponedTypeCheckingProblem{} -> Nothing
+  st <- useR stSolvedMetaStore
+  let lookup x = inst . mvInstantiation <$> MapS.lookup x st
       -- we collect only those used definitions that are in @names@
       emb d = if include d then Set.singleton d else Set.empty
   -- get all the Defs that are in names
   return $ getDefs' lookup emb a
+  where
+  -- TODO: Is it bad to ignore the lambdas?
+  inst (InstV i)                      = instBody i
+  inst Open                           = __IMPOSSIBLE__
+  inst OpenInstance                   = __IMPOSSIBLE__
+  inst BlockedConst{}                 = __IMPOSSIBLE__
+  inst PostponedTypeCheckingProblem{} = __IMPOSSIBLE__
