@@ -24,6 +24,7 @@ import Control.Monad        ( (<=<), filterM, forM, forM_, zipWithM )
 import Data.Foldable (toList)
 import qualified Data.List as List
 import Data.Monoid hiding ((<>))
+import Data.Set (Set)
 import qualified Data.Set as Set
 
 import qualified Agda.Syntax.Abstract as A
@@ -115,13 +116,17 @@ termDecl' = \case
     A.ScopedDecl scope ds -> {- withScope_ scope $ -} termDecls ds
         -- scope is irrelevant as we are termination checking Syntax.Internal
     A.RecSig{}            -> return mempty
-    A.RecDef _ r _ _ _ _ ds -> termDecls ds
-    -- These should all be wrapped in mutual blocks
+    A.RecDef _ x _ _ _ _ ds -> termMutual [x] <> termDecls ds
+        -- Andreas, 2022-10-23, issue #5823
+        -- Also check record types for termination.
+        -- They are unfolded during construction of unique inhabitants of eta-records.
+    -- These should all be wrapped in mutual blocks:
     A.FunDef{}      -> __IMPOSSIBLE__
     A.DataSig{}     -> __IMPOSSIBLE__
     A.DataDef{}     -> __IMPOSSIBLE__
     A.UnquoteDecl{} -> __IMPOSSIBLE__
     A.UnquoteDef{}  -> __IMPOSSIBLE__
+    A.UnquoteData{} -> __IMPOSSIBLE__
   where
     termDecls ds = concat <$> mapM termDecl' ds
 
@@ -129,7 +134,7 @@ termDecl' = \case
     -- for symbols that need to be termination-checked.
     getNames = concatMap getName
     getName (A.FunDef i x delayed cs)   = [x]
-    getName (A.RecDef _ _ _ _ _ _ ds)   = getNames ds
+    getName (A.RecDef _ x _ _ _ _ ds)   = x : getNames ds
     getName (A.Mutual _ ds)             = getNames ds
     getName (A.Section _ _ _ ds)        = getNames ds
     getName (A.ScopedDecl _ ds)         = getNames ds
@@ -153,8 +158,9 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
   -- during type-checking.
   mid <- fromMaybe __IMPOSSIBLE__ <$> asksTC envMutualBlock
   mutualBlock <- lookupMutualBlock mid
-  let allNames = filter (not . isAbsurdLambdaName) $ Set.elems $ mutualNames mutualBlock
-      names    = if null names0 then allNames else names0
+  let allNames = Set.filter (not . isAbsurdLambdaName) $
+                 mutualNames mutualBlock
+      names    = if null names0 then allNames else Set.fromList names0
       i        = mutualInfo mutualBlock
 
   -- We set the range to avoid panics when printing error messages.
@@ -196,7 +202,7 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
      forM sccs $ \ allNames -> do
 
      -- Set the mutual names in the termination environment.
-     let namesSCC = filter (allNames `hasElem`) names
+     let namesSCC = Set.filter (`Set.member` allNames) names
      let setNames e = e
            { terMutual    = allNames
            , terUserNames = namesSCC
@@ -207,23 +213,15 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
              " with cutoff=" ++ show cutoff ++ "..."
            terLocal setNames cont
 
-     -- New check currently only makes a difference for copatterns.
+     -- New check currently only makes a difference for copatterns and record types.
      -- Since it is slow, only invoke it if
-     -- any of the definitions uses copatterns.
-     res <- ifM (orM $ map usesCopatterns allNames)
+     -- any of the definitions uses copatterns or is a record type.
+     ifM (anyM allNames $ \ q -> usesCopatterns q `or2M` (isJust <$> isRecord q))
          -- Then: New check, one after another.
          (runTerm $ forM' allNames $ termFunction)
          -- Else: Old check, all at once.
          (runTerm $ termMutual')
 
-     -- Record result of termination check in signature.
-     -- If there are some termination errors, we collect them in
-     -- the state and mark the definition as non-terminating so
-     -- that it does not get unfolded
-     let terminates = null res
-     forM_ allNames $ \ q -> setTerminates q terminates
-     return res
-
 -- | @termMutual'@ checks all names of the current mutual block,
 --   henceforth called @allNames@, for termination.
 --
@@ -268,19 +266,25 @@ termMutual' = do
   -- the names the user has declared.  This is for error reporting.
   names <- terGetUserNames
   case r of
-    Left calls -> return $ singleton $ terminationError names $ callInfos calls
+
+    Left calls -> do
+      mapM_ (`setTerminates` False) allNames
+      return $ singleton $ terminationError names calls
+
     Right{} -> do
       liftTCM $ reportSLn "term.warn.yes" 2 $
         prettyShow (names) ++ " does termination check"
+      mapM_ (`setTerminates` True) allNames
       return mempty
 
 -- | Smart constructor for 'TerminationError'.
 --   Removes 'termErrFunctions' that are not mentioned in 'termErrCalls'.
-terminationError :: [QName] -> [CallInfo] -> TerminationError
-terminationError names calls = TerminationError names' calls
+terminationError :: Set QName -> CallPath -> TerminationError
+terminationError names calls = TerminationError names' calls'
   where
-  names'    = filter (hasElem mentioned) names
-  mentioned = map callInfoTarget calls
+  calls'    = callInfos calls
+  mentioned = map callInfoTarget calls'
+  names'    = filter (hasElem mentioned) $ toList names
 
 billToTerGraph :: a -> TerM a
 billToTerGraph a = liftTCM $ billPureTo [Benchmark.Termination, Benchmark.Graph] a
@@ -343,20 +347,27 @@ reportCalls no calls = do
     return ()
 
 -- | @termFunction name@ checks @name@ for termination.
+-- If it passes the termination check it is marked as "terminates" in the signature.
 
 termFunction :: QName -> TerM Result
-termFunction name = do
+termFunction name = inConcreteOrAbstractMode name $ \ def -> do
 
   -- Function @name@ is henceforth referred to by its @index@
   -- in the list of @allNames@ of the mutual block.
 
   allNames <- terGetMutual
-  let index = fromMaybe __IMPOSSIBLE__ $ List.elemIndex name allNames
+  let index = fromMaybe __IMPOSSIBLE__ $ Set.lookupIndex name allNames
 
   -- Retrieve the target type of the function to check.
   -- #4256: Don't use typeOfConst (which instantiates type with module params), since termination
   -- checking is running in the empty context, but with the current module unchanged.
-  target <- liftTCM $ do typeEndsInDef . defType =<< getConstInfo name
+  target <- case theDef def of
+    -- We are termination-checking a record (calls to record will not be guarding):
+    Record{} -> return TargetRecord
+    -- We are termination-checking a definition:
+    _ -> typeEndsInDef (defType def) <&> \case
+           Just d  -> TargetDef d
+           Nothing -> TargetOther
   reportTarget target
   terSetTarget target $ do
 
@@ -368,7 +379,10 @@ termFunction name = do
           if null todo then return $ Left calls else do
             -- Extract calls originating from indices in @todo@.
             new <- forM' todo $ \ i ->
-              termDef $ fromMaybe __IMPOSSIBLE__ $ allNames !!! i
+              termDef $
+              if i < 0 || i >= Set.size allNames
+              then __IMPOSSIBLE__
+              else Set.elemAt i allNames
             -- Mark those functions as processed and add the calls to the result.
             let done'  = done `mappend` todo
                 calls' = new  `mappend` calls
@@ -385,26 +399,61 @@ termFunction name = do
      cutoff <- terGetCutOff
      let ?cutoff = cutoff
      r <- billToTerGraph $ Term.terminatesFilter (== index) calls1
-     case r of
+
+     -- Andrea: 22/04/2020.
+     -- With cubical we will always have a clause where the dot
+     -- patterns are instead replaced with a variable, so they
+     -- cannot be relied on for termination.
+     -- See issue #4606 for a counterexample involving HITs.
+     --
+     -- Without the presence of HITs I conjecture that dot patterns
+     -- could be turned into actual splits, because no-confusion
+     -- would make the other cases impossible, so I do not disable
+     -- this for --without-K entirely.
+     --
+     -- Andreas, 2022-03-21: The check for --cubical was missing here.
+     ifM (isJust . optCubical <$> pragmaOptions) (return r) {- else -} $ case r of
        Right () -> return $ Right ()
-       Left{}    -> do
+       Left{}   -> do
          -- Try again, but include the dot patterns this time.
          calls2 <- terSetUseDotPatterns True $ collect
          reportCalls "" calls2
-         billToTerGraph $ mapLeft callInfos $ Term.terminatesFilter (== index) calls2
+         billToTerGraph $ Term.terminatesFilter (== index) calls2
 
     names <- terGetUserNames
-    case r of
-      Left calls -> return $ singleton $ terminationError ([name] `List.intersect` names) calls
+    case mapLeft callInfos r of
+
+      Left calls -> do
+        -- Mark as non-terminating.
+        setTerminates name False
+
+        -- Functions must be terminating, records types need not...
+        case theDef def of
+
+          -- Records need not terminate, so we just put the error on the debug log.
+          Record{} -> do
+            reportSDoc "term.warn.no" 10 $ vcat $
+              hsep [ "Record type", prettyTCM name, "does not termination check.", "Problematic calls:" ] :
+              (map (nest 2 . prettyTCM) $ List.sortOn getRange calls)
+            mempty
+
+          -- Functions must terminate, so we report the error.
+          _ -> do
+            let err = TerminationError [name | name `elem` names] calls
+            return $ singleton err
+
       Right () -> do
-        liftTCM $ reportSLn "term.warn.yes" 2 $
+        reportSLn "term.warn.yes" 2 $
           prettyShow name ++ " does termination check"
+        setTerminates name True
         return mempty
    where
-     reportTarget r = liftTCM $
-       reportSLn "term.target" 20 $ "  target type " ++
-         caseMaybe r "not recognized" (\ q ->
-           "ends in " ++ prettyShow q)
+     reportTarget :: MonadDebug m => Target -> m ()
+     reportTarget tgt = reportSLn "term.target" 20 $ ("  " ++) $
+       case tgt of
+         TargetRecord -> "termination checking a record type"
+         TargetDef q  -> unwords [ "target type ends in", prettyShow q ]
+         TargetOther  -> "target type not recognized"
 
 -- | To process the target type.
 typeEndsInDef :: MonadTCM tcm => Type -> tcm (Maybe QName)
@@ -427,6 +476,13 @@ typeEndsInDef t = liftTCM $ do
 termDef :: QName -> TerM Calls
 termDef name = terSetCurrent name $ inConcreteOrAbstractMode name $ \ def -> do
 
+ -- Skip calls to record types unless we are checking a record type in the first place.
+ let isRecord_ = case theDef def of { Record{} -> True; _ -> False }
+ let notTargetRecord = terGetTarget <&> \case
+       TargetRecord -> False
+       _ -> True
+ ifM (pure isRecord_ `and2M` notTargetRecord) mempty {-else-} $ do
+
   -- Retrieve definition
   let t = defType def
 
@@ -459,14 +515,28 @@ termDef name = terSetCurrent name $ inConcreteOrAbstractMode name $ \ def -> do
               then return empty
               else termClause cl
 
-        _ -> return empty
-  where
-    hasDefP :: [NamedArg DeBruijnPattern] -> Bool
-    hasDefP ps = getAny $ flip foldPattern ps $ \ (x :: DeBruijnPattern) ->
-                  case x of
-                    DefP{} -> Any True
-                    _      -> Any False
+        -- @record R pars : Set where field tel@
+        -- is treated like function @R pars = tel@.
+        Record{ recPars, recTel } -> termRecTel recPars recTel
 
+        _ -> return empty
+
+-- | Extract "calls" to the field types from a record constructor telescope.
+-- Does not extract from the parameters, but treats these as the "pattern variables"
+-- (the lhs of the "function").
+termRecTel :: Nat -> Telescope -> TerM Calls
+termRecTel npars tel = do
+  -- Set up the record parameters like function parameters.
+  let (pars, fields) = splitAt npars $ telToList tel
+  addContext pars $ do
+    ps <- mkPats npars
+    terSetPatterns ps $ terSetSizeDepth pars $ do
+      -- Treat the record fields like the body of a function.
+      extract $ telFromList fields
+  where
+  -- create n variable patterns
+  mkPats n  = zipWith mkPat (downFrom n) <$> getContextNames
+  mkPat i x = notMasked $ VarP defaultPatternInfo $ DBPatVar (prettyShow x) i
 
 -- | Collect calls in type signature @f : (x1:A1)...(xn:An) -> B@.
 --   It is treated as if there were the additional function clauses.
@@ -505,50 +575,43 @@ setMasks t cont = do
   (ds, d) <- liftTCM $ do
     TelV tel core <- telViewPath t
     -- Check argument types
-    ds <- forM (telToList tel) $ \ t -> do
-      TelV _ t <- telViewPath $ snd $ unDom t
-      d <- (isNothing <$> isDataOrRecord (unEl t)) `or2M` (isJust <$> isSizeType t)
-      when d $
-        reportSDoc "term.mask" 20 $ do
-          "argument type "
-            <+> prettyTCM t
-            <+> " is not data or record type, ignoring structural descent for --without-K"
-      return d
+    ds <- checkArgumentTypes tel
     -- Check result types
-    d  <- isNothing <.> isDataOrRecord . unEl $ core
+    d  <- addContext tel $ isNothing <.> isDataOrRecord . unEl $ core
     when d $
       reportSLn "term.mask" 20 $ "result type is not data or record type, ignoring guardedness for --without-K"
     return (ds, d)
   terSetMaskArgs (ds ++ repeat True) $ terSetMaskResult d $ cont
 
+  where
+    checkArgumentTypes :: Telescope -> TCM [Bool]
+    checkArgumentTypes EmptyTel = return []
+    checkArgumentTypes (ExtendTel dom atel) = do
+      TelV tel2 t <- telViewPath $ unDom dom
+      d <- addContext tel2 $
+        (isNothing <$> isDataOrRecord (unEl t)) `or2M` (isJust <$> isSizeType t)
+      when d $
+        reportSDoc "term.mask" 20 $ do
+          "argument type "
+            <+> prettyTCM t
+            <+> " is not data or record type, ignoring structural descent for --without-K"
+      underAbstraction dom atel $ \tel -> (d:) <$> checkArgumentTypes tel
 
 -- | Is the current target type among the given ones?
 
-targetElem :: [Target] -> TerM Bool
-targetElem ds = maybe False (`elem` ds) <$> terGetTarget
+targetElem :: [QName] -> TerM Bool
+targetElem ds = terGetTarget <&> \case
+  TargetDef d  -> d `elem` ds
+  TargetRecord -> False
+  TargetOther  -> False
 
-{-
--- | The target type of the considered recursive definition.
-data Target
-  = Set        -- ^ Constructing a Set (only meaningful with 'guardingTypeConstructors').
-  | Data QName -- ^ Constructing a coinductive or mixed type (could be data or record).
-  deriving (Eq, Show)
-
--- | Check wether a 'Target" corresponds to the current one.
-matchingTarget :: DBPConf -> Target -> TCM Bool
-matchingTarget conf t = maybe (return True) (match t) (currentTarget conf)
-  where
-    match Set      Set       = return True
-    match (Data d) (Data d') = mutuallyRecursive d d'
-    match _ _                = return False
--}
 
 -- | Convert a term (from a dot pattern) to a DeBruijn pattern.
 --
 --   The term is first normalized and stripped of all non-coinductive projections.
 
 termToDBP :: Term -> TerM DeBruijnPattern
-termToDBP t = ifNotM terGetUseDotPatterns (return unusedVar) $ {- else -} do
+termToDBP t =
   termToPattern =<< do liftTCM $ stripAllProjections =<< normalise t
 
 -- | Convert a term (from a dot pattern) to a pattern for the purposes of the termination checker.
@@ -572,18 +635,25 @@ instance TermToPattern a b => TermToPattern (Named c a) (Named c b) where
 instance TermToPattern Term DeBruijnPattern where
   termToPattern t = liftTCM (constructorForm t) >>= \case
     -- Constructors.
-    Con c _ args -> ConP c noConPatternInfo . map (fmap unnamed) <$> termToPattern (fromMaybe __IMPOSSIBLE__ $ allApplyElims args)
-    Def s [Apply arg] -> do
+    Con c _ args -> ifDotPatsOrRecord c $
+      ConP c noConPatternInfo . map (fmap unnamed) <$> termToPattern (fromMaybe __IMPOSSIBLE__ $ allApplyElims args)
+    Def s [Apply arg] -> ifDotPats $ do
       suc <- terGetSizeSuc
       if Just s == suc then ConP (ConHead s IsData Inductive []) noConPatternInfo . map (fmap unnamed) <$> termToPattern [arg]
-       else return $ dotP t
+       else fallback
     DontCare t  -> termToPattern t -- OR: __IMPOSSIBLE__  -- removed by stripAllProjections
     -- Leaves.
     Var i []    -> varP . (`DBPatVar` i) . prettyShow <$> nameOfBV i
     Lit l       -> return $ litP l
     Dummy s _   -> __IMPOSSIBLE_VERBOSE__ s
-    t           -> return $ dotP t
-
+    t           -> fallback
+    where
+    -- Andreas, 2022-06-14, issues #5953 and #4725
+    -- Recognize variable and record patterns in dot patterns regardless
+    -- of whether dot-pattern termination is on.
+    ifDotPats           = ifNotM terGetUseDotPatterns fallback
+    ifDotPatsOrRecord c = ifM (pure (IsData == conDataRecord c) `and2M` do not <$> terGetUseDotPatterns) fallback
+    fallback            = return $ dotP t
 
 -- | Masks all non-data/record type patterns if --without-K.
 --   See issue #1023.
@@ -686,6 +756,7 @@ instance ExtractCalls Sort where
       Inf f n    -> return empty
       SizeUniv   -> return empty
       LockUniv   -> return empty
+      IntervalUniv -> return empty
       Type t     -> terUnguarded $ extract t  -- no guarded levels
       Prop t     -> terUnguarded $ extract t
       SSet t     -> terUnguarded $ extract t
@@ -701,6 +772,11 @@ instance ExtractCalls Sort where
 instance ExtractCalls Type where
   extract (El s t) = extract (s, t)
 
+instance ExtractCalls a => ExtractCalls (Tele a) where
+  extract = \case
+    EmptyTel        -> mempty
+    ExtendTel a tel -> extract a <> extract tel
+
 -- | Extract recursive calls from a constructor application.
 
 constructor
@@ -756,13 +832,16 @@ function g es0 = do
           ]
 
     -- insert this call into the call list
-    case List.elemIndex g names of
+    case Set.lookupIndex g names of
 
        -- call leads outside the mutual block and can be ignored
        Nothing   -> return calls
 
-       -- call is to one of the mutally recursive functions
+       -- call is to one of the mutally recursive functions/record
        Just gInd -> do
+         cutoff <- terGetCutOff
+         let ?cutoff = cutoff
+
          delayed <- terGetDelayed
          -- Andreas, 2017-02-14, issue #2458:
          -- If we have inlined with-functions, we could be illtyped,
@@ -798,15 +877,30 @@ function g es0 = do
          -- spent on call matrix generation
          (nrows, ncols, matrix) <- billTo [Benchmark.Termination, Benchmark.Compare] $
            compareArgs es
-         -- only a delayed definition can be guarded
-         let ifDelayed o | Order.decreasing o && delayed == NotDelayed = Order.le
-                         | otherwise                                  = o
+
+         -- Andreas, 2022-03-21, #5823:
+         -- If we are "calling" a record type we are guarded unless the origin
+         -- of the termination analysis is itself a record.
+         -- This is because we usually do not "unfold" record types into their
+         -- field telescope.  We only do so when trying to construct the
+         -- unique inhabitant of record type (singleton analysis).
+         -- In the latter case, a call to a record type is not guarding.
+         guarded' <- isRecord g >>= \case
+           Just{} -> terGetTarget >>= \case
+             TargetRecord
+               -> return guarded
+             _ -> return (guarded .*. Order.lt)
+                    -- guarding when we call a record and not termination checking a record
+           Nothing
+             -- only a delayed definition can be guarded
+             | Order.decreasing guarded && delayed == NotDelayed
+               -> return Order.le
+             | otherwise
+               -> return guarded
          liftTCM $ reportSLn "term.guardedness" 20 $
            "composing with guardedness " ++ prettyShow guarded ++
-           " counting as " ++ prettyShow (ifDelayed guarded)
-         cutoff <- terGetCutOff
-         let ?cutoff = cutoff
-         let matrix' = composeGuardedness (ifDelayed guarded) matrix
+           " counting as " ++ prettyShow guarded'
+         let matrix' = composeGuardedness guarded' matrix
 
          -- Andreas, 2013-04-26 FORBIDDINGLY expensive!
          -- This PrettyTCM QName cost 50% of the termination time for std-lib!!
@@ -830,14 +924,14 @@ function g es0 = do
            -- Andreas, 2017-01-05, issue #2376
            -- Remove arguments inserted by etaExpandClause.
 
-         let src  = fromMaybe __IMPOSSIBLE__ $ List.elemIndex f names
+         let src  = fromMaybe __IMPOSSIBLE__ $ Set.lookupIndex f names
              tgt  = gInd
              cm   = makeCM ncols nrows matrix'
-             info = CallPath [CallInfo
+             info = CallPath $ singleton $
+                    CallInfo
                       { callInfoTarget = g
-                      , callInfoRange  = getRange g
                       , callInfoCall   = doc
-                      }]
+                      }
          verboseS "term.kept.call" 5 $ do
            pats <- terGetPatterns
            reportSDoc "term.kept.call" 5 $ vcat
@@ -887,7 +981,7 @@ tryReduceNonRecursiveClause g es continue fallback = do
 
   -- Finally, try to reduce with the non-recursive clauses (and no rewrite rules).
   r <- liftTCM $ modifyAllowedReductions (SmallSet.delete UnconfirmedReductions) $
-    runReduceM $ appDefE' v0 cls [] (map notReduced es)
+    runReduceM $ appDefE' g v0 cls [] (map notReduced es)
   case r of
     NoReduction{}    -> fallback
     YesReduction _ v -> do
@@ -909,29 +1003,36 @@ getNonRecursiveClauses q =
 
 instance ExtractCalls Term where
   extract t = do
-    liftTCM $ reportSDoc "term.check.term" 50 $ do
+    reportSDoc "term.check.term" 50 $ do
       "looking for calls in" <+> prettyTCM t
 
     -- Instantiate top-level MetaVar.
-    t <- liftTCM $ instantiate t
-    case t of
+    instantiate t >>= \case
 
       -- Constructed value.
       Con ConHead{conName = c, conDataRecord = dataOrRec} _ es -> do
         let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
         -- A constructor preserves the guardedness of all its arguments.
-        let argsg = zip args $ repeat True
+        -- Andreas, 2022-09-19, issue #6108:
+        -- A higher constructor does not.  So check if there is an @IApply@ amoung @es@.
+        let argsg = zip args $ repeat $ all isProperApplyElim es
 
         -- If we encounter a coinductive record constructor
         -- in a type mutual with the current target
         -- then we count it as guarding.
-        let inductive   = return Inductive
-            coinductive = return CoInductive
+        let inductive   = return Inductive    -- not guarding, but preserving
+            coinductive = return CoInductive  -- guarding
+        -- ♯ is guarding
         ind <- ifM ((Just c ==) <$> terGetSharp) coinductive $ {-else-} do
+          -- data constructors are not guarding
           if dataOrRec == IsData then inductive else do
-          caseMaybeM (liftTCM $ isRecordConstructor c) inductive $ \ (q, def) -> do
+          -- abstract constructors are not guarding
+          caseMaybeM (isRecordConstructor c) inductive $ \ (q, def) -> do
             reportSLn "term.check.term" 50 $ "constructor " ++ prettyShow c ++ " has record type " ++ prettyShow q
+            -- inductive record constructors are not guarding
             if recInduction def /= Just CoInductive then inductive else do
+            -- coinductive constructors unrelated to the mutually
+            -- constructed inhabitants of coinductive types are not guarding
             ifM (targetElem . fromMaybe __IMPOSSIBLE__ $ recMutual def)
                {-then-} coinductive
                {-else-} inductive
@@ -946,17 +1047,16 @@ instance ExtractCalls Term where
       -- Neutral term. Destroys guardedness.
       Var i es -> terUnguarded $ extract es
 
-      -- Dependent function space. Destroys guardedness.
+      -- Dependent function space.
       Pi a (Abs x b) ->
-        terUnguarded $
         CallGraph.union <$>
         extract a <*> do
           a <- maskSizeLt a  -- OR: just do not add a to the context!
           addContext (x, a) $ terRaise $ extract b
 
-      -- Non-dependent function space. Destroys guardedness.
+      -- Non-dependent function space.
       Pi a (NoAbs _ b) ->
-        terUnguarded $ CallGraph.union <$> extract a <*> extract b
+        CallGraph.union <$> extract a <*> extract b
 
       -- Literal.
       Lit l -> return empty
@@ -1035,7 +1135,7 @@ compareArgs es = do
   let guardedness =
         if useGuardedness
         then decr True $ projsCaller - projsCallee
-        else Order.Unknown
+        else Order.le
   liftTCM $ reportSDoc "term.guardedness" 30 $ sep
     [ "compareArgs:"
     , nest 2 $ text $ "projsCaller = " ++ prettyShow projsCaller
@@ -1328,14 +1428,17 @@ compareConArgs :: Args -> [NamedArg DeBruijnPattern] -> TerM Order
 compareConArgs ts ps = do
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
-  -- we may assume |ps| >= |ts|, otherwise c ps would be of functional type
-  -- which is impossible
-  case (length ts, length ps) of
-    (0,0) -> return Order.le        -- c <= c
-    (0,1) -> return Order.unknown   -- c not<= c x
-    (1,0) -> __IMPOSSIBLE__
-    (1,1) -> compareTerm' (unArg (head ts)) (notMasked $ namedArg $ head ps)
-    (_,_) -> foldl (Order..*.) Order.le <$>
+  case compare (length ts) (length ps) of
+
+    -- We may assume |ps| >= |ts|, otherwise c ps would be of functional type
+    -- which is impossible.
+    GT -> __IMPOSSIBLE__
+
+    -- Andreas, 2022-08-31, issue #6059: doing anything smarter than
+    -- @unknown@ here can lead to non-termination.
+    LT -> return Order.unknown
+
+    EQ -> List.foldl' (Order..*.) Order.le <$>
                zipWithM compareTerm' (map unArg ts) (map (notMasked . namedArg) ps)
        -- corresponds to taking the size, not the height
        -- allows examples like (x, y) < (Succ x, y)
