@@ -41,6 +41,7 @@ import Agda.TypeChecking.Telescope
 
 
 import Agda.Utils.Functor (($>))
+import Agda.Utils.Pretty  (prettyShow)
 import Agda.Utils.Size
 
 import Agda.Utils.Impossible
@@ -162,12 +163,12 @@ checkInternal' action v cmp t = verboseBracket "tc.check.internal" 20 "" $ do
     [ "checking internal "
     , nest 2 $ sep [ prettyTCM v <+> ":"
                    , nest 2 $ prettyTCM t ] ]
-  reportSDoc "tc.check.internal" 30 $ sep
+  reportSDoc "tc.check.internal" 60 $ sep
     [ "checking internal with DB indices"
     , nest 2 $ sep [ pretty v <+> ":"
                    , nest 2 $ pretty t ] ]
   ctx <- getContextTelescope
-  reportSDoc "tc.check.internal" 30 $ sep
+  unless (null ctx) $ reportSDoc "tc.check.internal" 30 $ sep
     [ "In context"
     , nest 2 $ sep [ prettyTCM ctx ] ]
   -- Bring projection-like funs in post-fix form,
@@ -215,7 +216,7 @@ checkInternal' action v cmp t = verboseBracket "tc.check.internal" 20 "" $ do
                                NoAbs{} -> id
       a <- mkDom <$> checkInternal' action (unEl $ unDom a) CmpLeq (sort sa)
       v' <- goInside $ Pi a . mkRng <$> checkInternal' action (unEl $ unAbs b) CmpLeq (sort sb)
-      s' <- sortOf v'
+      s' <- sortOf v -- Issue #6205: do not use v' since it might not be valid syntax
       compareSort cmp s' s
       return v'
     Sort s     -> do
@@ -227,7 +228,7 @@ checkInternal' action v cmp t = verboseBracket "tc.check.internal" 20 "" $ do
       return $ Sort s
     Level l    -> do
       l <- checkLevel action l
-      lt <- levelType
+      lt <- levelType'
       compareType cmp lt t
       return $ Level l
     DontCare v -> DontCare <$> checkInternal' action v cmp t
@@ -322,8 +323,7 @@ checkModality action mod mod' = do
 
 -- | Infer type of a neutral term.
 infer :: (MonadCheckInternal m) => Term -> m Type
-infer v = do
-  case v of
+infer = \case
     Var i es   -> do
       a <- typeOfBV i
       snd <$> inferSpine a (Var i   []) es
@@ -332,7 +332,10 @@ infer v = do
     MetaV x es -> do -- we assume meta instantiations to be well-typed
       a <- metaType x
       snd <$> inferSpine a (MetaV x []) es
-    _ -> __IMPOSSIBLE__
+    v -> __IMPOSSIBLE_VERBOSE__ $ unlines
+      [ "CheckInternal.infer: non-inferable term:"
+      , "  " ++ prettyShow v
+      ]
 
 -- | Infer ordinary function application.
 inferDef :: (MonadCheckInternal m) => QName -> Elims -> m Type
@@ -370,7 +373,7 @@ inferSpine' action t self self' (e : es) = do
     , "type t = " <+> pretty t
     , "self  = " <+> pretty self
     , "self' = " <+> pretty self'
-    , "eliminated by e = " <+> text (show e)
+    , "eliminated by e = " <+> pretty e
     ]
   case e of
     IApply x y r -> do
@@ -426,6 +429,7 @@ checkSort action s =
     SSet l   -> SSet <$> checkLevel action l
     SizeUniv -> return SizeUniv
     LockUniv -> return LockUniv
+    IntervalUniv -> return IntervalUniv
     PiSort dom s1 s2 -> do
       let a = unDom dom
       s1' <- checkSort action s1
@@ -465,7 +469,7 @@ checkLevel action (Max n ls) = Max n <$> mapM checkPlusLevel ls
     checkPlusLevel (Plus k l)      = Plus k <$> checkLevelAtom l
 
     checkLevelAtom l = do
-      lvl <- levelType
+      lvl <- levelType'
       checkInternal' action l CmpLeq lvl
 
 -- | Universe subsumption and type equality (subtyping for sizes, resp.).
