@@ -9,8 +9,6 @@ import Control.Monad.Trans.Identity
 
 import Data.Maybe
 import qualified Data.Map as Map
-import Data.List (nubBy)
-import Data.Function
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -18,7 +16,6 @@ import Agda.Syntax.Internal.Pattern
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Coverage.SplitTree
-import Agda.TypeChecking.Forcing
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Substitute
@@ -155,14 +152,14 @@ compileWithSplitTree t cs = case t of
     compiles _ _ Branches{etaBranch = Just{}} = __IMPOSSIBLE__  -- we haven't inserted eta matches yet
 
 compile :: Cls -> CompiledClauses
-compile [] = Fail
+compile [] = Fail []
 compile cs = case nextSplit cs of
-  Just (isRecP, n) -> Case n $ fmap compile $ splitOn isRecP (unArg n) cs
+  Just (isRecP, n) -> Case n $ compile <$> splitOn isRecP (unArg n) cs
   Nothing -> case clBody c of
     -- It's possible to get more than one clause here due to
     -- catch-all expansion.
     Just t  -> Done (map (fmap name) $ clPats c) t
-    Nothing -> Fail
+    Nothing -> Fail (map (fmap name) $ clPats c)
   where
     -- If there are more than one clauses, take the first one.
     c = headWithDefault __IMPOSSIBLE__ cs
@@ -399,7 +396,7 @@ ensureNPatterns n ais0 cl@(Cl ps b)
   ps'  = for ais $ \ ai -> Arg ai $ varP "_"
   args = zipWith (\ i ai -> Arg ai $ var i) (downFrom m) ais
 
-substBody :: (Subst t a) => Int -> Int -> t -> a -> a
+substBody :: Subst a => Int -> Int -> SubstArg a -> a -> a
 substBody n m v = applySubst $ liftS n $ v :# raiseS m
 
 instance PrecomputeFreeVars a => PrecomputeFreeVars (CompiledClauses' a) where
