@@ -55,8 +55,8 @@ instance DropArgs FunctionInverse where
 -- | Use for dropping initial lambdas in clause bodies.
 --   NOTE: does not reduce term, need lambdas to be present.
 instance DropArgs Term where
-  dropArgs 0 v = v
-  dropArgs n v = case v of
+  dropArgs 0 = id
+  dropArgs n = \case
     Lam h b -> dropArgs (n - 1) (absBody b)
     _       -> __IMPOSSIBLE__
 
@@ -71,9 +71,9 @@ instance DropArgs CompiledClauses where
               | otherwise     -> Case (i <&> \ j -> j - n) $ fmap (dropArgs n) br
     Done xs t | length xs < n -> __IMPOSSIBLE__
               | otherwise     -> Done (drop n xs) t
-    Fail                      -> Fail
+    Fail xs   | length xs < n -> __IMPOSSIBLE__
+              | otherwise     -> Fail (drop n xs)
 
 instance DropArgs SplitTree where
   dropArgs n (SplittingDone m) = SplittingDone (m - n)
   dropArgs n (SplitAt i lz ts) = SplitAt (subtract n <$> i) lz $ map (second $ dropArgs n) ts
-
