@@ -21,6 +21,7 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Coverage.Match ( fromSplitPatterns )
 import Agda.TypeChecking.Records
+import Agda.TypeChecking.Reduce ( instantiateFull )
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 
@@ -36,7 +37,7 @@ data ErrorNonEmpty
   | DontKnow Blocker   -- ^ Emptyness check blocked
 
 instance Semigroup ErrorNonEmpty where
-  DontKnow u1     <> DontKnow u2  = DontKnow $ u1 <> u2  -- Both must unblock for this to proceed
+  DontKnow u1     <> DontKnow u2  = DontKnow $ unblockOnBoth u1 u2  -- Both must unblock for this to proceed
   e@DontKnow{}    <> _            = e
   _               <> e@DontKnow{} = e
   FailBecause err <> _            = FailBecause err
@@ -88,7 +89,9 @@ checkEmptyType range t = do
       dontAssignMetas $ do
         r <- splitLast Inductive tel ps
         case r of
-          Left UnificationStuck{} -> return $ Left $ DontKnow $ unblockOnAnyMetaIn tel
+          Left UnificationStuck{} -> do
+            blocker <- unblockOnAnyMetaIn <$> instantiateFull tel -- TODO Jesper: get proper blocking information from unification
+            return $ Left $ DontKnow blocker
           Left _                  -> return $ Left Fail
           Right cov -> do
             let ps = map (namedArg . lastWithDefault __IMPOSSIBLE__ . fromSplitPatterns . scPats) $ splitClauses cov
