@@ -17,17 +17,19 @@ module Agda.TypeChecking.Errors
   , dropTopLevelModule
   , topLevelModuleDropper
   , stringTCErr
+  , explainWhyInScope
+  , Verbalize(verbalize)
   ) where
 
 import Prelude hiding ( null, foldl )
 
+import qualified Control.Exception as E
 import Control.Monad.Except
 
 import qualified Data.CaseInsensitive as CaseInsens
 import Data.Foldable (foldl)
-import Data.Function
+import Data.Function (on)
 import Data.List (sortBy, dropWhileEnd, intercalate)
-import qualified Data.List.NonEmpty as NonEmpty
 import Data.Maybe
 import qualified Data.Set as Set
 import qualified Text.PrettyPrint.Boxes as Boxes
@@ -48,6 +50,7 @@ import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Debug
+import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.SizedTypes ( sizeType )
 import Agda.TypeChecking.Monad.State
@@ -107,7 +110,7 @@ tcErrString :: TCErr -> String
 tcErrString err = prettyShow (getRange err) ++ " " ++ case err of
   TypeError _ _ cl  -> errorString $ clValue cl
   Exception r s     -> prettyShow r ++ " " ++ show s
-  IOException _ r e -> prettyShow r ++ " " ++ show e
+  IOException _ r e -> prettyShow r ++ " " ++ E.displayException e
   PatternErr{}      -> "PatternErr"
 
 stringTCErr :: String -> TCErr
@@ -232,6 +235,7 @@ errorString err = case err of
   UnequalRelevance{}                       -> "UnequalRelevance"
   UnequalQuantity{}                        -> "UnequalQuantity"
   UnequalCohesion{}                        -> "UnequalCohesion"
+  UnequalFiniteness{}                      -> "UnequalFiniteness"
   UnequalHiding{}                          -> "UnequalHiding"
   UnequalLevel{}                           -> "UnequalLevel"
   UnequalSorts{}                           -> "UnequalSorts"
@@ -291,14 +295,10 @@ dropTopLevelModule q = ($ q) <$> topLevelModuleDropper
 
 -- | Produces a function which drops the filename component of the qualified name.
 topLevelModuleDropper :: (MonadDebug m, MonadTCEnv m, ReadTCState m) => m (QName -> QName)
-topLevelModuleDropper = do
-  caseMaybeM (asksTC envCurrentPath) (return id) $ \ f -> do
-  reportSDoc "err.dropTopLevel" 60 $ vcat
-    [ "current path =" <+> (text . filePath) f
-    , "moduleToSource =" <+> do text . show =<< useR stModuleToSource
-    ]
-  m <- fromMaybe __IMPOSSIBLE__ <$> lookupModuleFromSource f
-  return $ dropTopLevelModule' $ size m
+topLevelModuleDropper =
+  caseMaybeM currentTopLevelModule
+    (return id)
+    (return . dropTopLevelModule' . size)
 
 instance PrettyTCM TypeError where
   prettyTCM err = case err of
@@ -597,6 +597,11 @@ instance PrettyTCM TypeError where
       pwords "because one is a non-flat function type and the other is a flat function type"
       -- FUTURE Cohesion: update message if/when introducing sharp.
 
+    UnequalFiniteness cmp a b -> fsep $
+      [prettyTCM a, notCmp cmp, prettyTCM b] ++
+      pwords "because one is a type of partial elements and the other is a function type"
+      -- FUTURE Cohesion: update message if/when introducing sharp.
+
     UnequalHiding a b -> fsep $
       [prettyTCM a, "!=", prettyTCM b] ++
       pwords "because one is an implicit function type and the other is an explicit function type"
@@ -804,11 +809,11 @@ instance PrettyTCM TypeError where
 
     NoSuchModule x -> fsep $ pwords "No module" ++ [pretty x] ++ pwords "in scope"
 
-    AmbiguousName x ys -> vcat
+    AmbiguousName x reason -> vcat
       [ fsep $ pwords "Ambiguous name" ++ [pretty x <> "."] ++
                pwords "It could refer to any one of"
-      , nest 2 $ vcat $ fmap nameWithBinding ys
-      , fwords "(hint: Use C-c C-w (in Emacs) if you want to know why)"
+      , nest 2 $ vcat $ fmap nameWithBinding $ ambiguousNamesInReason reason
+      , explainWhyInScope $ whyInScopeDataFromAmbiguousNameReason x reason
       ]
 
     AmbiguousModule x ys -> vcat
@@ -937,7 +942,7 @@ instance PrettyTCM TypeError where
       , fsep $ pwords "(hint: overloaded pattern synonyms must be equal up to variable and constructor names)"
       ]
       where
-        (x, _) = NonEmpty.head defs
+        (x, _) = List1.head defs
         prDef (x, (xs, p)) = prettyA (A.PatternSynDef x (map (fmap BindName) xs) p) <?> ("at" <+> pretty r)
           where r = nameBindingSite $ qnameName x
 
@@ -1005,14 +1010,14 @@ instance PrettyTCM TypeError where
                sortBy (compare `on` prettyShow . notaName . sectNotation) $
                filter (not . closedWithoutHoles) sects))
       where
-      trimLeft  = dropWhile isNormalHole
-      trimRight = dropWhileEnd isNormalHole
+      trimLeft  = dropWhile isAHole
+      trimRight = dropWhileEnd isAHole
 
       closedWithoutHoles sect =
         sectKind sect == NonfixNotation
           &&
-        null [ () | NormalHole {} <- trimLeft $ trimRight $
-                                       notation (sectNotation sect) ]
+        null [ () | HolePart{} <- trimLeft $ trimRight $
+                                    notation (sectNotation sect) ]
 
       prettyName n = Boxes.text $
         P.render (P.pretty n) ++
@@ -1261,7 +1266,7 @@ instance PrettyUnequal Term where
     (d1, d2, d) <- prettyInEqual t1 t2
     fsep $ return d1 : ncmp : return d2 : return d : []
 
-instance PrettyUnequal Type where
+instance PrettyUnequal I.Type where
   prettyUnequal t1 ncmp t2 = prettyUnequal (unEl t1) ncmp (unEl t2)
 
 instance PrettyTCM SplitError where
@@ -1328,7 +1333,7 @@ instance PrettyTCM SplitError where
       pwords "Case to handle:") $$ nest 2 (vcat $ [display cl])
                                 $$ ((pure msg <+> enterClosure t displayAbs) <> ".")
         where
-        displayAbs :: Abs Type -> m Doc
+        displayAbs :: Abs I.Type -> m Doc
         displayAbs (Abs x t) = addContext x $ prettyTCM t
         displayAbs (NoAbs x t) = prettyTCM t
         display (tel, ps) = prettyTCM $ NamedClause f True $
@@ -1379,6 +1384,84 @@ instance PrettyTCM UnificationFailure where
       pwords "modality"
 
 
+
+explainWhyInScope :: forall m. MonadPretty m => WhyInScopeData -> m Doc
+explainWhyInScope (WhyInScopeData y _ Nothing [] []) = text (prettyShow  y ++ " is not in scope.")
+explainWhyInScope (WhyInScopeData y _ v xs ms) = vcat
+  [ text (prettyShow y ++ " is in scope as")
+  , nest 2 $ vcat [variable v xs, modules ms]
+  ]
+  where
+    -- variable :: Maybe _ -> [_] -> m Doc
+    variable Nothing vs = names vs
+    variable (Just x) vs
+      | null vs   = asVar
+      | otherwise = vcat
+         [ sep [ asVar, nest 2 $ shadowing x]
+         , nest 2 $ names vs
+         ]
+      where
+        asVar :: m Doc
+        asVar = do
+          "* a variable bound at" <+> prettyTCM (nameBindingSite $ localVar x)
+        shadowing :: LocalVar -> m Doc
+        shadowing (LocalVar _ _ [])    = "shadowing"
+        shadowing _ = "in conflict with"
+    names   = vcat . map pName
+    modules = vcat . map pMod
+
+    pKind = \case
+      ConName                  -> "constructor"
+      CoConName                -> "coinductive constructor"
+      FldName                  -> "record field"
+      PatternSynName           -> "pattern synonym"
+      GeneralizeName           -> "generalizable variable"
+      DisallowedGeneralizeName -> "generalizable variable from let open"
+      MacroName                -> "macro name"
+      QuotableName             -> "quotable name"
+      -- previously DefName:
+      DataName                 -> "data type"
+      RecName                  -> "record type"
+      AxiomName                -> "postulate"
+      PrimName                 -> "primitive function"
+      FunName                  -> "defined name"
+      OtherDefName             -> "defined name"
+
+    pName :: AbstractName -> m Doc
+    pName a = sep
+      [ "* a"
+        <+> pKind (anameKind a)
+        <+> text (prettyShow $ anameName a)
+      , nest 2 $ "brought into scope by"
+      ] $$
+      nest 2 (pWhy (nameBindingSite $ qnameName $ anameName a) (anameLineage a))
+    pMod :: AbstractModule -> m Doc
+    pMod  a = sep
+      [ "* a module" <+> text (prettyShow $ amodName a)
+      , nest 2 $ "brought into scope by"
+      ] $$
+      nest 2 (pWhy (nameBindingSite $ qnameName $ mnameToQName $ amodName a) (amodLineage a))
+
+    pWhy :: Range -> WhyInScope -> m Doc
+    pWhy r Defined = "- its definition at" <+> prettyTCM r
+    pWhy r (Opened (C.QName x) w) | isNoName x = pWhy r w
+    pWhy r (Opened m w) =
+      "- the opening of"
+      <+> prettyTCM m
+      <+> "at"
+      <+> prettyTCM (getRange m)
+      $$
+      pWhy r w
+    pWhy r (Applied m w) =
+      "- the application of"
+      <+> prettyTCM m
+      <+> "at"
+      <+> prettyTCM (getRange m)
+      $$
+      pWhy r w
+
+
+
 ---------------------------------------------------------------------------
 -- * Natural language
 ---------------------------------------------------------------------------
@@ -1413,6 +1496,13 @@ instance Verbalize Cohesion where
       Continuous -> "continuous"
       Squash     -> "squashed"
 
+instance Verbalize Modality where
+  verbalize mod | mod == defaultModality = "default"
+  verbalize (Modality rel qnt coh) = intercalate "," $
+    [ verbalize rel | rel /= defaultRelevance ] ++
+    [ verbalize qnt | qnt /= defaultQuantity ] ++
+    [ verbalize coh | coh /= defaultCohesion ]
+
 -- | Indefinite article.
 data Indefinite a = Indefinite a
 
