@@ -65,10 +65,11 @@ import Control.Monad.Reader ( MonadReader(..), asks, ReaderT, Reader, runReader
 
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
-import Data.IntSet (IntSet)
-import qualified Data.IntSet as IntSet
+import Data.HashSet (HashSet)
+import qualified Data.HashSet as HashSet
 import Data.Semigroup ( Semigroup, (<>) )
-
+import qualified Data.Set as Set
+import Data.Set (Set)
 
 
 import Agda.Syntax.Common
@@ -87,17 +88,20 @@ import Agda.Utils.Size
 
 -- | A set of meta variables.  Forms a monoid under union.
 
-newtype MetaSet = MetaSet { theMetaSet :: IntSet }
+newtype MetaSet = MetaSet { theMetaSet :: HashSet MetaId }
   deriving (Eq, Show, Null, Semigroup, Monoid)
 
 instance Singleton MetaId MetaSet where
-  singleton = MetaSet . singleton . metaId
+  singleton = MetaSet . singleton
 
 insertMetaSet :: MetaId -> MetaSet -> MetaSet
-insertMetaSet (MetaId m) (MetaSet ms) = MetaSet $ IntSet.insert m ms
+insertMetaSet m (MetaSet ms) = MetaSet $ HashSet.insert m ms
 
 foldrMetaSet :: (MetaId -> a -> a) -> a -> MetaSet -> a
-foldrMetaSet f e ms = IntSet.foldr (f . MetaId) e $ theMetaSet ms
+foldrMetaSet f e ms = HashSet.foldr f e $ theMetaSet ms
+
+metaSetToBlocker :: MetaSet -> Blocker
+metaSetToBlocker ms = unblockOnAny $ foldrMetaSet (Set.insert . unblockOnMeta) Set.empty ms
 
 ---------------------------------------------------------------------------
 -- * Flexible and rigid occurrences (semigroup)
@@ -548,6 +552,7 @@ instance Free Sort where
       SSet a     -> freeVars' a
       SizeUniv   -> mempty
       LockUniv   -> mempty
+      IntervalUniv -> mempty
       PiSort a s1 s2 -> underFlexRig (Flexible mempty) (freeVars' $ unDom a) `mappend`
                         underFlexRig WeaklyRigid (freeVars' (s1, s2))
       FunSort s1 s2 -> freeVars' s1 `mappend` freeVars' s2
