@@ -10,12 +10,12 @@ import Control.Arrow (first)
 import Control.Monad
 import Control.Monad.Except
 
-import qualified Data.IntMap as IntMap
 import qualified Data.IntSet as IntSet
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Map (Map)
 import qualified Data.Map as Map
+import qualified Data.Map.Strict as MapS
 import Data.List (partition, sortBy)
 import Data.Monoid
 import Data.Function (on)
@@ -46,14 +46,15 @@ import Agda.Utils.Benchmark
 import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.Functor
 import Agda.Utils.Impossible
-import Agda.Utils.List   (hasElem)
+import Agda.Utils.Lens
+import Agda.Utils.List (downFrom, hasElem)
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Size
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty (prettyShow)
-
+import Agda.Utils.Tuple
 
 -- | Generalize a telescope over a set of generalizable variables.
 generalizeTelescope :: Map QName Name -> (forall a. (Telescope -> TCM a) -> TCM a) -> ([Maybe Name] -> Telescope -> TCM a) -> TCM a
@@ -128,7 +129,7 @@ generalizeType' s typecheckAction = billTo [Typing, Generalize] $ withGenRecVar
 
 -- | Create metas for the generalizable variables and run the type check action.
 createMetasAndTypeCheck ::
-  Set QName -> TCM a -> TCM (a, Map MetaId QName, MetaStore)
+  Set QName -> TCM a -> TCM (a, Map MetaId QName, LocalMetaStores)
 createMetasAndTypeCheck s typecheckAction = do
   ((namedMetas, x), allmetas) <- metasCreatedBy $ do
     (metamap, genvals) <- createGenValues s
@@ -147,26 +148,42 @@ withGenRecVar ret = do
   genRecMeta <- newTypeMeta (mkType 0)
   addContext (defaultDom ("genTel" :: String, genRecMeta)) $ ret genRecMeta
 
--- | Compute the generalized telescope from metas created when checking the type/telescope to be
+-- | Compute the generalized telescope from metas created when checking the *thing* (type or telescope) to be
 --   generalized. Called in the context extended with the telescope record variable (whose type is
 --   the first argument). Returns the telescope of generalized variables and a substitution from
 --   this telescope to the current context.
-computeGeneralization ::
-  Type -> Map MetaId name -> MetaStore ->
-  TCM (Telescope, [Maybe name], Substitution)
+computeGeneralization
+  :: Type
+       -- ^ The metavariable to be instantiated with record type containing
+       --   as fields the variables generalized in the *thing*.
+  -> Map MetaId name
+       -- ^ Metas created from an occurrence of a @variable@. (The original free variables.)
+       --   E.g. if you have
+       --   @
+       --     variable  l : Level; A : Set l
+       --     postulate f : A → A
+       --   @
+       --   then @A@ would be in this @Map@, but not @l@.
+  -> LocalMetaStores
+       -- ^ The metas created when type-checking the *thing*.
+  -> TCM (Telescope, [Maybe name], Substitution)
+       -- ^ The telescope together with binder name (left-to-right order),
+       --   and substitution from this telescope to the current context.
 computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints $ do
 
   reportSDoc "tc.generalize" 10 $ "computing generalization for type" <+> prettyTCM genRecMeta
 
   -- Pair metas with their metaInfo
-  mvs <- mapM ((\ x -> (x,) <$> lookupMeta x) . MetaId) $ IntMap.keys allmetas
+  let mvs = MapS.assocs (openMetas allmetas) ++
+            MapS.assocs (solvedMetas allmetas)
 
   -- Issue 4727: filter out metavariables that were created before the
   -- current checkpoint, since they are too old to be generalized.
   -- TODO: make metasCreatedBy smarter so it doesn't see pruned
   -- versions of old metas as new metas.
   cp <- viewTC eCurrentCheckpoint
-  let isFreshMeta (x,mv) = enterClosure mv $ \ _ -> isJust <$> checkpointSubstitution' cp
+  let isFreshMeta :: MonadReduce m => (MetaId, MetaVariable) -> m Bool
+      isFreshMeta (x,mv) = enterClosure mv $ \ _ -> isJust <$> checkpointSubstitution' cp
   mvs <- filterM isFreshMeta mvs
   cs <- (++) <$> useTC stAwakeConstraints
              <*> useTC stSleepingConstraints
@@ -208,7 +225,7 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
   cp <- viewTC eCurrentCheckpoint
   let canGeneralize x | isConstrained x = return False
       canGeneralize x = do
-          mv   <- lookupMeta x
+          mv   <- lookupLocalMeta x
           msub <- enterClosure mv $ \ _ ->
                     checkpointSubstitution' cp
           let sameContext =
@@ -223,10 +240,10 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
                   (Just IdS, Perm m xs)        -> xs == [0 .. m - 1]
                   (Just (Wk n IdS), Perm m xs) -> xs == [0 .. m - n - 1]
                   _                            -> False
-          unless sameContext $ do
+          unless sameContext $ reportSDoc "tc.generalize" 20 $ do
             ty <- getMetaType x
             let Perm m xs = mvPermutation mv
-            reportSDoc "tc.generalize" 20 $ vcat
+            vcat
               [ text "Don't know how to generalize over"
               , nest 2 $ prettyTCM x <+> text ":" <+> prettyTCM ty
               , text "in context"
@@ -236,15 +253,18 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
           return sameContext
   inherited <- fmap Set.unions $ forM generalizableClosed $ \ (x, mv) ->
     case mvInstantiation mv of
-      InstV _ v -> do
+      InstV inst -> do
         parentName <- getMetaNameSuggestion x
-        metas <- filterM canGeneralize . Set.toList . allMetas Set.singleton =<< instantiateFull v
+        metas <- filterM canGeneralize . Set.toList .
+                 allMetas Set.singleton =<<
+                 instantiateFull (instBody inst)
         let suggestNames i [] = return ()
             suggestNames i (m : ms) = do
               -- #4291: Override existing meta name suggestion. If we solved the parent with a new
               --        meta use the parent name for that, otherwise suffix with a number.
-              let suf | null ms && i == 1, MetaV{} <- v = ""
-                      | otherwise                       = "." ++ show i
+              let suf | null ms && i == 1,
+                        MetaV{} <- instBody inst = ""
+                      | otherwise                = "." ++ show i
               setMetaNameSuggestion m (parentName ++ suf)
               suggestNames (i + 1) ms
         unless (null metas) $
@@ -304,7 +324,7 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
   teleTypes <- do
     args <- getContextArgs
     fmap concat $ forM sortedMetas $ \ m -> do
-      mv   <- lookupMeta m
+      mv <- lookupLocalMeta m
       let info =
             hideOrKeepInstance $
             getArgInfo $ miGeneralizable $ mvInfo mv
@@ -320,7 +340,9 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
   -- Now we need to prune the unsolved metas to make sure they respect the new
   -- dependencies (#3672). Also update interaction points to point to pruned metas.
   let inscope (ii, InteractionPoint{ipMeta = Just x})
-        | IntMap.member (metaId x) allmetas = Just (x, ii)
+        | MapS.member x (openMetas allmetas) ||
+          MapS.member x (solvedMetas allmetas) =
+          Just (x, ii)
       inscope _ = Nothing
   ips <- Map.fromDistinctAscList . mapMaybe inscope . fst . BiMap.toDistinctAscendingLists <$> useTC stInteractionPoints
   pruneUnsolvedMetas genRecName genRecCon genTel genRecFields ips shouldGeneralize allSortedMetas
@@ -332,6 +354,15 @@ computeGeneralization genRecMeta nameMap allmetas = postponeInstanceConstraints
   -- value packing up the generalized variables for the genTel variable.
   let sub = unpackSub genRecCon (map (argInfo . fst) teleTypes) (length teleTypes)
 
+  -- Instantiate all fresh meta-variables to get rid of
+  -- __DUMMY_TERM__.
+  genTel <- flip instantiateWhen genTel $ \m -> do
+    mv <- lookupMeta m
+    case mv of
+      Nothing         -> __IMPOSSIBLE__
+      Just Left{}     -> return False
+      Just (Right mv) -> isFreshMeta (m, mv)
+
   return (genTel, telNames, sub)
 
 -- | Prune unsolved metas (#3672). The input includes also the generalized metas and is sorted in
@@ -380,7 +411,7 @@ pruneUnsolvedMetas genRecName genRecCon genTel genRecFields interactionPoints is
     -- on any variables introduced after the genRec. See test/Fail/Issue3672b.agda for a test case.
     prePrune x = do
       cp <- viewTC eCurrentCheckpoint
-      mv <- lookupMeta x
+      mv <- lookupLocalMeta x
       (i, _A) <- enterClosure mv $ \ _ -> do
         δ <- checkpointSubstitution cp
         _A <- case mvJudgement mv of
@@ -432,7 +463,7 @@ pruneUnsolvedMetas genRecName genRecCon genTel genRecFields interactionPoints is
 
     pruneMeta _Θ x = do
       cp <- viewTC eCurrentCheckpoint
-      mv <- lookupMeta x
+      mv <- lookupLocalMeta x
       -- The reason we are doing all this inside the closure of x is so that if x is an interaction
       -- meta we get the right context for the pruned interaction meta.
       enterClosure mv $ \ _ ->
@@ -555,10 +586,15 @@ pruneUnsolvedMetas genRecName genRecCon genTel genRecFields interactionPoints is
     findGenRec :: MetaVariable -> TCM (Maybe Int)
     findGenRec mv = do
       cxt <- instantiateFull =<< getContext
-      let notPruned = permute (takeP (length cxt) $ mvPermutation mv) $
-               reverse $ zipWith const [0..] cxt
-      case [ i | (i, Dom{unDom = (_, El _ (Def q _))}) <- zip [0..] cxt,
-                 q == genRecName, i `elem` notPruned ] of
+      let n         = length cxt
+          notPruned = IntSet.fromList $
+                      permute (takeP n $ mvPermutation mv) $
+                      downFrom n
+      case [ i
+           | (i, Dom{unDom = (_, El _ (Def q _))}) <- zip [0..] cxt
+           , q == genRecName
+           , i `IntSet.member` notPruned
+           ] of
         []    -> return Nothing
         _:_:_ -> __IMPOSSIBLE__
         [i]   -> return (Just i)
@@ -782,12 +818,13 @@ createGenRecordType genRecMeta@(El genRecSort _) sortedMetas = do
                , funMutual       = Just []
                , funAbstr        = ConcreteDef
                , funDelayed      = NotDelayed
-               , funProjection   = Just proj
+               , funProjection   = Right proj
                , funFlags        = Set.empty
                , funTerminates   = Just True
                , funExtLam       = Nothing
                , funWith         = Nothing
                , funCovering     = []
+               , funIsKanOp      = Nothing
                }
   addConstant' (conName genRecCon) defaultArgInfo (conName genRecCon) __DUMMY_TYPE__ $ -- Filled in later
     Constructor { conPars   = 0
@@ -814,6 +851,7 @@ createGenRecordType genRecMeta@(El genRecSort _) sortedMetas = do
            , recEtaEquality' = Inferred YesEta
            , recPatternMatching = CopatternMatching
            , recInduction    = Nothing
+           , recTerminates   = Just True    -- not recursive
            , recAbstr        = ConcreteDef
            , recComp         = emptyCompKit
            }
@@ -849,7 +887,6 @@ fillInGenRecordDetails name con fields recTy fieldTel = do
   reportSDoc "tc.generalize" 40 $ text "Final genRecCon type:" <+> inTopContext (prettyTCM conType)
   setType (conName con) conType
   -- Record telescope: Includes both parameters and fields.
-  modifyGlobalDefinition name $ \ r ->
-    r { theDef = (theDef r) { recTel = fullTel } }
+  modifyGlobalDefinition name $ set (lensTheDef . lensRecord . lensRecTel) fullTel
   where
     setType q ty = modifyGlobalDefinition q $ \ d -> d { defType = ty }
