@@ -3,7 +3,6 @@ module Agda.TypeChecking.Level where
 
 import Data.Maybe
 import qualified Data.List as List
-import Data.List.NonEmpty (NonEmpty(..))
 import Data.Traversable (Traversable)
 
 import Agda.Syntax.Common
@@ -14,6 +13,7 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Reduce
 
+import Agda.Utils.List1 ( List1, pattern (:|) )
 import Agda.Utils.Maybe ( caseMaybeM, allJustM )
 import Agda.Utils.Monad ( tryMaybe )
 import Agda.Utils.Singleton
@@ -31,9 +31,20 @@ data LevelKit = LevelKit
   , zeroName :: QName
   }
 
--- | Get the 'primLevel' as a 'Type'.
-levelType :: (HasBuiltins m) => m Type
-levelType = El (mkType 0) . fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
+{-# SPECIALIZE levelType :: TCM Type #-}
+-- | Get the 'primLevel' as a 'Type'.  Aborts if any of the level BUILTINs is undefined.
+levelType :: (HasBuiltins m, MonadTCError m) => m Type
+levelType = El (mkType 0) . lvlType <$> requireLevels
+  -- Andreas, 2022-10-11, issue #6168
+  -- It seems superfluous to require all level builtins here,
+  -- but since we are in MonadTCError here, this is our chance to make sure
+  -- that all level builtins are defined.
+  -- Otherwise, we might run into an __IMPOSSIBLE__ later,
+  -- e.g. if only BUILTIN LEVEL was defined by reallyUnLevelView requires all builtins.
+
+-- | Get the 'primLevel' as a 'Type'.  Unsafe, crashes if the BUILTIN LEVEL is undefined.
+levelType' :: (HasBuiltins m) => m Type
+levelType' = El (mkType 0) . fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevel
 
 isLevelType :: PureTCM m => Type -> m Bool
 isLevelType a = reduce (unEl a) >>= \case
@@ -42,9 +53,6 @@ isLevelType a = reduce (unEl a) >>= \case
     return $ f == lvl
   _ -> return False
 
-levelSucFunction :: TCM (Term -> Term)
-levelSucFunction = apply1 <$> primLevelSuc
-
 {-# SPECIALIZE builtinLevelKit :: TCM LevelKit #-}
 {-# SPECIALIZE builtinLevelKit :: ReduceM LevelKit #-}
 builtinLevelKit :: (HasBuiltins m) => m LevelKit
@@ -64,9 +72,24 @@ builtinLevelKit = do
       , zeroName = z
       }
 
+{-# SPECIALIZE requireLevels :: TCM LevelKit #-}
 -- | Raises an error if no level kit is available.
-requireLevels :: HasBuiltins m => m LevelKit
-requireLevels = builtinLevelKit
+requireLevels :: (HasBuiltins m, MonadTCError m) => m LevelKit
+requireLevels = do
+    level@(Def l []) <- getBuiltin builtinLevel
+    zero@(Def z [])  <- getBuiltin builtinLevelZero
+    suc@(Def s [])   <- getBuiltin builtinLevelSuc
+    max@(Def m [])   <- getBuiltin builtinLevelMax
+    return $ LevelKit
+      { lvlType  = level
+      , lvlSuc   = \ a -> suc `apply1` a
+      , lvlMax   = \ a b -> max `applys` [a, b]
+      , lvlZero  = zero
+      , typeName = l
+      , sucName  = s
+      , maxName  = m
+      , zeroName = z
+      }
 
 -- | Checks whether level kit is fully available.
 haveLevels :: HasBuiltins m => m Bool
@@ -90,13 +113,7 @@ unLevel v = return v
 {-# SPECIALIZE reallyUnLevelView :: Level -> TCM Term #-}
 {-# SPECIALIZE reallyUnLevelView :: Level -> ReduceM Term #-}
 reallyUnLevelView :: (HasBuiltins m) => Level -> m Term
-reallyUnLevelView nv = do
-  suc <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelSuc
-  zer <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinLevelZero
-  case nv of
-    Max n []       -> return $ unConstV zer (apply1 suc) n
-    Max 0 [a]      -> return $ unPlusV (apply1 suc) a
-    _              -> (`unlevelWithKit` nv) <$> builtinLevelKit
+reallyUnLevelView nv = (`unlevelWithKit` nv) <$> builtinLevelKit
 
 unlevelWithKit :: LevelKit -> Level -> Term
 unlevelWithKit LevelKit{ lvlZero = zer, lvlSuc = suc, lvlMax = max } = \case
@@ -223,7 +240,7 @@ unSingleLevels ls = levelMax n as
     n = maximum $ 0 : [m | SingleClosed m <- ls]
     as = [a | SinglePlus a <- ls]
 
-levelMaxView :: Level' t -> NonEmpty (SingleLevel' t)
+levelMaxView :: Level' t -> List1 (SingleLevel' t)
 levelMaxView (Max n [])     = singleton $ SingleClosed n
 levelMaxView (Max 0 (a:as)) = SinglePlus a :| map SinglePlus as
 levelMaxView (Max n as)     = SingleClosed n :| map SinglePlus as
