@@ -9,10 +9,10 @@ import Control.Monad        ( foldM, forM, forM_, liftM2, void )
 import Control.Monad.Except ( MonadError(..), ExceptT, runExceptT )
 import Control.Monad.Trans  ( lift )
 
-import Data.Function
-import qualified Data.IntMap as IntMap
+import Data.Function (on)
 import qualified Data.IntSet as IntSet
 import qualified Data.List as List
+import qualified Data.Map.Strict as MapS
 import qualified Data.Set as Set
 import qualified Data.Foldable as Fold
 import qualified Data.Traversable as Trav
@@ -58,7 +58,8 @@ import Agda.Utils.Monad
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import Agda.Utils.Permutation
-import Agda.Utils.Pretty ( Pretty, prettyShow )
+import Agda.Utils.Pretty (Pretty, prettyShow, render)
+import qualified Agda.Utils.ProfileOptions as Profile
 import Agda.Utils.Singleton
 import qualified Agda.Utils.Graph.TopSort as Graph
 import Agda.Utils.VarSet (VarSet)
@@ -90,16 +91,18 @@ instance MonadMetaSolver TCM where
 findIdx :: Eq a => [a] -> a -> Maybe Int
 findIdx vs v = List.elemIndex v (reverse vs)
 
+-- | Does the given local meta-variable have a twin meta-variable?
+
 hasTwinMeta :: MetaId -> TCM Bool
 hasTwinMeta x = do
-    m <- lookupMeta x
+    m <- lookupLocalMeta x
     return $ isJust $ mvTwin m
 
 -- | Check whether a meta variable is a place holder for a blocked term.
 isBlockedTerm :: MetaId -> TCM Bool
 isBlockedTerm x = do
     reportSLn "tc.meta.blocked" 12 $ "is " ++ prettyShow x ++ " a blocked term? "
-    i <- mvInstantiation <$> lookupMeta x
+    i <- lookupMetaInstantiation x
     let r = case i of
             BlockedConst{}                 -> True
             PostponedTypeCheckingProblem{} -> True
@@ -107,12 +110,12 @@ isBlockedTerm x = do
             Open{}                         -> False
             OpenInstance{}                 -> False
     reportSLn "tc.meta.blocked" 12 $
-      if r then "  yes, because " ++ show i else "  no"
+      if r then "  yes, because " ++ prettyShow i else "  no"
     return r
 
 isEtaExpandable :: [MetaKind] -> MetaId -> TCM Bool
 isEtaExpandable kinds x = do
-    i <- mvInstantiation <$> lookupMeta x
+    i <- lookupMetaInstantiation x
     return $ case i of
       Open{}                         -> True
       OpenInstance{}                 -> Records `notElem` kinds
@@ -143,13 +146,21 @@ assignTermTCM' x tel v = do
      -- verify (new) invariants
     whenM (not <$> asksTC envAssignMetas) __IMPOSSIBLE__
 
-    verboseS "profile.metas" 10 $ liftTCM $ return () {-tickMax "max-open-metas" . (fromIntegral . size) =<< getOpenMetas-}
-    modifyMetaStore $ ins x $ InstV tel $ killRange v
+    whenProfile Profile.Metas $ liftTCM $ return () {-tickMax "max-open-metas" . (fromIntegral . size) =<< getOpenMetas-}
+    updateMetaVarTCM x $ \ mv ->
+      mv { mvInstantiation = InstV $ Instantiation
+             { instTel = tel
+             , instBody = v
+             -- Andreas, 2022-04-28, issue #5875:
+             -- Can't killRange the meta-solution, since this will destroy
+             -- ranges of termination errors (and potentially other passes
+             -- that run on internal syntax)!
+             -- , instBody = killRange v
+             }
+         }
     etaExpandListeners x
     wakeupConstraints x
     reportSLn "tc.meta.assign" 20 $ "completed assignment of " ++ prettyShow x
-  where
-    ins x i = IntMap.adjust (\ mv -> mv { mvInstantiation = i }) $ metaId x
 
 -- * Creating meta variables.
 
@@ -394,7 +405,7 @@ newQuestionMark' new ii cmp t = lookupInteractionMeta ii >>= \case
     MetaVar
       { mvInfo = MetaInfo{ miClosRange = Closure{ clEnv = TCEnv{ envContext = gamma }}}
       , mvPermutation = p
-      } <- fromMaybe __IMPOSSIBLE__ <$> lookupMeta' x
+      } <- fromMaybe __IMPOSSIBLE__ <$> lookupLocalMeta' x
     -- Get the current context Δ.
     delta <- getContext
     -- A bit hazardous:
@@ -617,28 +628,8 @@ problemType (CheckProjAppToKnownPrincipalArg _ _ _ _ _ t _ _ _ _) = t -- The tar
 problemType (CheckLambda _ _ _ t     ) = t
 problemType (DoQuoteTerm _ _ t)        = t
 
--- | Eta expand metavariables listening on the current meta.
-etaExpandListeners :: MetaId -> TCM ()
-etaExpandListeners m = do
-  ls <- getMetaListeners m
-  clearMetaListeners m  -- we don't really have to do this
-  mapM_ wakeupListener ls
-
--- | Wake up a meta listener and let it do its thing
-wakeupListener :: Listener -> TCM ()
-  -- Andreas 2010-10-15: do not expand record mvars, lazyness needed for irrelevance
-wakeupListener (EtaExpand x)         = etaExpandMetaSafe x
-wakeupListener (CheckConstraint _ c) = do
-  reportSDoc "tc.meta.blocked" 20 $ "waking boxed constraint" <+> prettyTCM c
-  modifyAwakeConstraints (c:)
-  solveAwakeConstraints
-
--- | Do safe eta-expansions for meta (@SingletonRecords,Levels@).
-etaExpandMetaSafe :: (MonadMetaSolver m) => MetaId -> m ()
-etaExpandMetaSafe = etaExpandMeta [SingletonRecords,Levels]
-
--- | Eta expand a metavariable, if it is of the specified kind.
---   Don't do anything if the metavariable is a blocked term.
+-- | Eta-expand a local meta-variable, if it is of the specified kind.
+--   Don't do anything if the meta-variable is a blocked term.
 etaExpandMetaTCM :: [MetaKind] -> MetaId -> TCM ()
 etaExpandMetaTCM kinds m = whenM ((not <$> isFrozen m) `and2M` asksTC envAssignMetas `and2M` isEtaExpandable kinds m) $ do
   verboseBracket "tc.meta.eta" 20 ("etaExpandMeta " ++ prettyShow m) $ do
@@ -652,7 +643,7 @@ etaExpandMetaTCM kinds m = whenM ((not <$> isFrozen m) `and2M` asksTC envAssignM
           reportSDoc "tc.meta.eta" 20 $ do
             "we do not expand meta variable" <+> prettyTCM m <+>
               text ("(requested was expansion of " ++ show kinds ++ ")")
-    meta <- lookupMeta m
+    meta <- lookupLocalMeta m
     case mvJudgement meta of
       IsSort{} -> dontExpand
       HasType _ cmp a -> do
@@ -676,7 +667,7 @@ etaExpandMetaTCM kinds m = whenM ((not <$> isFrozen m) `and2M` asksTC envAssignM
         -- they go unsolved: conversion will compare them at the
         -- different cases for the domain, so it will not find the
         -- solution for the whole meta.
-        if any domFinite (flattenTel tel) then dontExpand else do
+        if any domIsFinite (flattenTel tel) then dontExpand else do
 
         -- Issue #3774: continue with the right context for b
         addContext tel $ do
@@ -742,7 +733,7 @@ assignWrapper :: (MonadMetaSolver m, MonadConstraint m, MonadError TCErr m, Mona
 assignWrapper dir x es v doAssign = do
   ifNotM (asksTC envAssignMetas) dontAssign $ {- else -} do
     reportSDoc "tc.meta.assign" 10 $ do
-      "term" <+> prettyTCM (MetaV x es) <+> text (":" ++ show dir) <+> prettyTCM v
+      "term" <+> prettyTCM (MetaV x es) <+> text (":" ++ prettyShow dir) <+> prettyTCM v
     nowSolvingConstraints doAssign `finally` solveAwakeConstraints
 
   where dontAssign = do
@@ -768,7 +759,7 @@ assignWrapper dir x es v doAssign = do
 assign :: CompareDirection -> MetaId -> Args -> Term -> CompareAs -> TCM ()
 assign dir x args v target = addOrUnblocker (unblockOnMeta x) $ do
 
-  mvar <- lookupMeta x  -- information associated with meta x
+  mvar <- lookupLocalMeta x  -- information associated with meta x
   let t = jMetaType $ mvJudgement mvar
 
   -- Andreas, 2011-05-20 TODO!
@@ -1057,7 +1048,7 @@ assign dir x args v target = addOrUnblocker (unblockOnMeta x) $ do
           --   @
           -- we need to check that @A <: A'@ (due to contravariance).
           let sigma = parallelS $ reverse $ map unArg args
-          hasSubtyping <- collapseDefault . optSubtyping <$> pragmaOptions
+          hasSubtyping <- optCumulativity <$> pragmaOptions
           when hasSubtyping $ forM_ ids $ \(i , u) -> do
             -- @u@ is a (projected) variable, so we can infer its type
             a  <- applySubst sigma <$> addContext tel' (infer u)
@@ -1083,10 +1074,12 @@ assign dir x args v target = addOrUnblocker (unblockOnMeta x) $ do
       let success = killResult `elem` [PrunedSomething,PrunedEverything]
       reportSDoc "tc.meta.assign" 10 $
         "pruning" <+> prettyTCM x <+> do text $ if success then "succeeded" else "failed"
-      patternViolation (if success then alwaysUnblock  -- If pruning succeeded we want to retry right away
-                                   else unblockOnAnyMetaIn $ MetaV x $ map Apply args)
-                                        -- TODO: could be more precise: only unblock on metas
-                                        --       applied to offending variables
+      blocker <- if
+        | success   -> return alwaysUnblock  -- If pruning succeeded we want to retry right away
+        | otherwise -> unblockOnAnyMetaIn . MetaV x . map Apply <$> instantiateFull args
+             -- TODO: could be more precise: only unblock on metas
+             --       applied to offending variables
+      patternViolation blocker
 
 {- UNUSED
 -- | When faced with @_X us == D vs@ for an inert D we can solve this by
@@ -1254,7 +1247,7 @@ assignMeta' m x t n ids v = do
 
     -- Andreas, 2013-10-25 double check solution before assigning
     whenM (optDoubleCheck  <$> pragmaOptions) $ do
-      m <- lookupMeta x
+      m <- lookupLocalMeta x
       reportSDoc "tc.meta.check" 30 $ "double checking solution"
       catchConstraint (CheckMetaInst x) $
         addContext tel' $ checkSolutionForMeta x m v' a
@@ -1281,18 +1274,19 @@ assignMeta' m x t n ids v = do
 --   instantiated, add a constraint to check the instantiation later.
 checkMetaInst :: MetaId -> TCM ()
 checkMetaInst x = do
-  m <- lookupMeta x
+  m <- lookupLocalMeta x
   let postpone = addConstraint (unblockOnMeta x) $ CheckMetaInst x
   case mvInstantiation m of
     BlockedConst{} -> postpone
     PostponedTypeCheckingProblem{} -> postpone
     Open{} -> postpone
     OpenInstance{} -> postpone
-    InstV xs v -> do
-      let n = size xs
+    InstV inst -> do
+      let n = size (instTel inst)
           t = jMetaType $ mvJudgement m
       (telv@(TelV tel a),bs) <- telViewUpToPathBoundary n t
-      catchConstraint (CheckMetaInst x) $ addContext tel $ checkSolutionForMeta x m v a
+      catchConstraint (CheckMetaInst x) $ addContext tel $
+        checkSolutionForMeta x m (instBody inst) a
 
 -- | Check that the instantiation of the metavariable with the given
 --   term is well-typed.
@@ -1321,8 +1315,7 @@ checkSubtypeIsEqual a b = do
   reportSDoc "tc.meta.subtype" 30 $
     "checking that subtype" <+> prettyTCM a <+>
     "of" <+> prettyTCM b <+> "is actually equal."
-  ((a, b), equal) <- SynEq.checkSyntacticEquality a b
-  unless equal $ do
+  SynEq.checkSyntacticEquality a b (\_ _ -> return ()) $ \a b -> do
     cumulativity <- optCumulativity <$> pragmaOptions
     abortIfBlocked (unEl b) >>= \case
       Sort sb -> abortIfBlocked (unEl a) >>= \case
@@ -1353,8 +1346,8 @@ checkSubtypeIsEqual a b = do
 -- @_Y args <= u@.
 subtypingForSizeLt
   :: CompareDirection -- ^ @dir@
-  -> MetaId           -- ^ The meta variable @x@.
-  -> MetaVariable     -- ^ Its associated information @mvar <- lookupMeta x@.
+  -> MetaId           -- ^ The local meta-variable @x@.
+  -> MetaVariable     -- ^ Its associated information @mvar <- lookupLocalMeta x@.
   -> Type             -- ^ Its type @t = jMetaType $ mvJudgement mvar@
   -> Args             -- ^ Its arguments.
   -> Term             -- ^ Its to-be-assigned value @v@, such that @x args `dir` v@.
@@ -1571,86 +1564,88 @@ data InvertExcept
 inverseSubst :: Args -> ExceptT InvertExcept TCM SubstCand
 inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
   where
-    loop  = foldM isVarOrIrrelevant []
-    terms = map var (downFrom (size args))
-    failure = do
-      lift $ reportSDoc "tc.meta.assign" 15 $ vcat
-        [ "not all arguments are variables: " <+> prettyTCM args
-        , "  aborting assignment" ]
-      throwError CantInvert
-    neutralArg = throwError NeutralArg
+  loop  = foldM isVarOrIrrelevant []
+  terms = map var (downFrom (size args))
+  failure = do
+    lift $ reportSDoc "tc.meta.assign" 15 $ vcat
+      [ "not all arguments are variables: " <+> prettyTCM args
+      , "  aborting assignment" ]
+    throwError CantInvert
+  neutralArg = throwError NeutralArg
 
-    isVarOrIrrelevant :: Res -> (Arg Term, Term) -> ExceptT InvertExcept TCM Res
-    isVarOrIrrelevant vars (Arg info v, t) = do
-      let irr | isIrrelevant info = True
-              | DontCare{} <- v   = True
-              | otherwise         = False
-      case stripDontCare v of
-        -- i := x
-        Var i [] -> return $ (Arg info i, t) `cons` vars
+  isVarOrIrrelevant :: Res -> (Arg Term, Term) -> ExceptT InvertExcept TCM Res
+  isVarOrIrrelevant vars (Arg info v, t) = do
+    let irr | isIrrelevant info = True
+            | DontCare{} <- v   = True
+            | otherwise         = False
+    case stripDontCare v of
+      -- i := x
+      Var i [] -> return $ (Arg info i, t) `cons` vars
 
-        -- π i := x  try to eta-expand projection π away!
-        Var i es | Just qs <- mapM isProjElim es ->
-          throwError $ ProjVar $ ProjectedVar i qs
+      -- π i := x  try to eta-expand projection π away!
+      Var i es | Just qs <- mapM isProjElim es ->
+        throwError $ ProjVar $ ProjectedVar i qs
 
-        -- (i, j) := x  becomes  [i := fst x, j := snd x]
-        -- Andreas, 2013-09-17 but only if constructor is fully applied
-        Con c ci es -> do
-          let fallback
-               | isIrrelevant info = return vars
-               | otherwise                              = failure
-          irrProj <- optIrrelevantProjections <$> pragmaOptions
-          lift (isRecordConstructor $ conName c) >>= \case
-            Just (_, r@Record{ recFields = fs })
-              | YesEta <- recEtaEquality r  -- Andreas, 2019-11-10, issue #4185: only for eta-records
-              , length fs == length es
-              , hasQuantity0 info || all usableQuantity fs     -- Andreas, 2019-11-12/17, issue #4168b
-              , irrProj || all isRelevant fs -> do
-                let aux (Arg _ v) Dom{domInfo = info', unDom = f} = (Arg ai v,) $ t `applyE` [Proj ProjSystem f] where
-                     ai = ArgInfo
-                       { argInfoHiding   = min (getHiding info) (getHiding info')
-                       , argInfoModality = Modality
-                         { modRelevance  = max (getRelevance info) (getRelevance info')
-                         , modQuantity   = max (getQuantity  info) (getQuantity  info')
-                         , modCohesion   = max (getCohesion  info) (getCohesion  info')
-                         }
-                       , argInfoOrigin   = min (getOrigin info) (getOrigin info')
-                       , argInfoFreeVariables = unknownFreeVariables
-                       , argInfoAnnotation    = argInfoAnnotation info'
-                       }
-                    vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-                res <- loop $ zipWith aux vs fs
-                return $ res `append` vars
-              | otherwise -> fallback
-            _ -> fallback
+      -- (i, j) := x  becomes  [i := fst x, j := snd x]
+      -- Andreas, 2013-09-17 but only if constructor is fully applied
+      Con c ci es -> do
+        let fallback
+             | isIrrelevant info = return vars
+             | otherwise                              = failure
+        irrProj <- optIrrelevantProjections <$> pragmaOptions
+        lift (isRecordConstructor $ conName c) >>= \case
+          Just (_, r@Record{ recFields = fs })
+            | YesEta <- recEtaEquality r  -- Andreas, 2019-11-10, issue #4185: only for eta-records
+            , length fs == length es
+            , hasQuantity0 info || all usableQuantity fs     -- Andreas, 2019-11-12/17, issue #4168b
+            , irrProj || all isRelevant fs -> do
+              let aux (Arg _ v) Dom{domInfo = info', unDom = f} =
+                    (Arg ai v,) $ t `applyE` [Proj ProjSystem f]
+                    where
+                    ai = ArgInfo
+                      { argInfoHiding   = min (getHiding info) (getHiding info')
+                      , argInfoModality = Modality
+                        { modRelevance  = max (getRelevance info) (getRelevance info')
+                        , modQuantity   = max (getQuantity  info) (getQuantity  info')
+                        , modCohesion   = max (getCohesion  info) (getCohesion  info')
+                        }
+                      , argInfoOrigin   = min (getOrigin info) (getOrigin info')
+                      , argInfoFreeVariables = unknownFreeVariables
+                      , argInfoAnnotation    = argInfoAnnotation info'
+                      }
+                  vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+              res <- loop $ zipWith aux vs fs
+              return $ res `append` vars
+            | otherwise -> fallback
+          _ -> fallback
 
-        -- An irrelevant argument which is not an irrefutable pattern is dropped
-        _ | irr -> return vars
+      -- An irrelevant argument which is not an irrefutable pattern is dropped
+      _ | irr -> return vars
 
-        -- Distinguish args that can be eliminated (Con,Lit,Lam,unsure) ==> failure
-        -- from those that can only put somewhere as a whole ==> neutralArg
-        Var{}      -> neutralArg
-        Def{}      -> neutralArg  -- Note that this Def{} is in normal form and might be prunable.
-        Lam{}      -> failure
-        Lit{}      -> failure
-        MetaV{}    -> failure
-        Pi{}       -> neutralArg
-        Sort{}     -> neutralArg
-        Level{}    -> neutralArg
-        DontCare{} -> __IMPOSSIBLE__ -- Ruled out by stripDontCare
-        Dummy s _  -> __IMPOSSIBLE_VERBOSE__ s
+      -- Distinguish args that can be eliminated (Con,Lit,Lam,unsure) ==> failure
+      -- from those that can only put somewhere as a whole ==> neutralArg
+      Var{}      -> neutralArg
+      Def{}      -> neutralArg  -- Note that this Def{} is in normal form and might be prunable.
+      Lam{}      -> failure
+      Lit{}      -> failure
+      MetaV{}    -> failure
+      Pi{}       -> neutralArg
+      Sort{}     -> neutralArg
+      Level{}    -> neutralArg
+      DontCare{} -> __IMPOSSIBLE__ -- Ruled out by stripDontCare
+      Dummy s _  -> __IMPOSSIBLE_VERBOSE__ s
 
-    -- managing an assoc list where duplicate indizes cannot be irrelevant vars
-    append :: Res -> Res -> Res
-    append res vars = foldr cons vars res
+  -- managing an assoc list where duplicate indizes cannot be irrelevant vars
+  append :: Res -> Res -> Res
+  append res vars = foldr cons vars res
 
-    -- adding an irrelevant entry only if not present
-    cons :: (Arg Nat, Term) -> Res -> Res
-    cons a@(Arg ai i, t) vars
-      | isIrrelevant ai = applyUnless (any ((i==) . unArg . fst) vars) (a :) vars
-      | otherwise       = a :  -- adding a relevant entry
-          -- filter out duplicate irrelevants
-          filter (not . (\ a@(Arg info j, t) -> isIrrelevant info && i == j)) vars
+  -- adding an irrelevant entry only if not present
+  cons :: (Arg Nat, Term) -> Res -> Res
+  cons a@(Arg ai i, t) vars
+    | isIrrelevant ai = applyUnless (any ((i==) . unArg . fst) vars) (a :) vars
+    | otherwise       = a :  -- adding a relevant entry
+        -- filter out duplicate irrelevants
+        filter (not . (\ a@(Arg info j, t) -> isIrrelevant info && i == j)) vars
 
 
 -- | Turn open metas into postulates.
@@ -1666,35 +1661,35 @@ openMetasToPostulates = do
   m <- asksTC envCurrentModule
 
   -- Go through all open metas.
-  ms <- IntMap.assocs <$> useTC stMetaStore
+  ms <- MapS.assocs <$> useTC stOpenMetaStore
   forM_ ms $ \ (x, mv) -> do
-    when (isOpenMeta $ mvInstantiation mv) $ do
-      let t = dummyTypeToOmega $ jMetaType $ mvJudgement mv
+    let t = dummyTypeToOmega $ jMetaType $ mvJudgement mv
 
-      -- Create a name for the new postulate.
-      let r = clValue $ miClosRange $ mvInfo mv
-      -- s <- render <$> prettyTCM x -- Using _ is a bad idea, as it prints as prefix op
-      let s = "unsolved#meta." ++ prettyShow x
-      n <- freshName r s
-      let q = A.QName m n
+    -- Create a name for the new postulate.
+    let r = clValue $ miClosRange $ mvInfo mv
+    s' <- render <$> prettyTCM x -- Using _ is a bad idea, as it prints as prefix op
+    let s = "unsolved#meta." ++ filter (/= '_') s'
+    n <- freshName r s
+    let q = A.QName m n
 
-      -- Debug.
-      reportSDoc "meta.postulate" 20 $ vcat
-        [ text ("Turning " ++ if isSortMeta_ mv then "sort" else "value" ++ " meta ")
-            <+> prettyTCM (MetaId x) <+> " into postulate."
-        , nest 2 $ vcat
-          [ "Name: " <+> prettyTCM q
-          , "Type: " <+> prettyTCM t
-          ]
+    -- Debug.
+    reportSDoc "meta.postulate" 20 $ vcat
+      [ text ("Turning " ++ if isSortMeta_ mv then "sort" else "value" ++ " meta ")
+          <+> prettyTCM x <+> " into postulate."
+      , nest 2 $ vcat
+        [ "Name: " <+> prettyTCM q
+        , "Type: " <+> prettyTCM t
         ]
+      ]
 
-      -- Add the new postulate to the signature.
-      addConstant' q defaultArgInfo q t defaultAxiom
+    -- Add the new postulate to the signature.
+    addConstant' q defaultArgInfo q t defaultAxiom
 
-      -- Solve the meta.
-      let inst = InstV [] $ Def q []
-      updateMetaVar (MetaId x) $ \ mv0 -> mv0 { mvInstantiation = inst }
-      return ()
+    -- Solve the meta.
+    let inst = InstV $ Instantiation
+                 { instTel = [], instBody = Def q [] }
+    updateMetaVar x $ \ mv0 -> mv0 { mvInstantiation = inst }
+    return ()
   where
     -- Unsolved sort metas can have a type ending in a Dummy if they are allowed to be instantiated
     -- to Setω. This will crash the serializer (issue #3730). To avoid this we replace dummy type
@@ -1709,15 +1704,20 @@ dependencySortMetas :: [MetaId] -> TCM (Maybe [MetaId])
 dependencySortMetas metas = do
   metaGraph <- concat <$> do
     forM metas $ \ m -> do
-      deps <- allMetas (\ m' -> if m' `elem` metas then singleton m' else mempty) <$> getType m
+      deps <- allMetas (\m' -> if m' `Set.member` metas'
+                               then singleton m'
+                               else mempty) <$>
+                getType m
       return [ (m, m') | m' <- Set.toList deps ]
 
-  return $ Graph.topSort metas metaGraph
+  return $ Graph.topSort metas' metaGraph
 
   where
+    metas' = Set.fromList metas
+
     -- Sort metas don't have types, but we still want to sort them.
     getType m = do
-      mv <- lookupMeta m
-      case mvJudgement mv of
+      j <- lookupMetaJudgement m
+      case j of
         IsSort{}                 -> return Nothing
         HasType{ jMetaType = t } -> Just <$> instantiateFull t
