@@ -43,6 +43,7 @@ instance MentionsMeta Blocker where
   mentionsMetas xs (UnblockOnAny bs)  = mentionsMetas xs $ Set.toList bs
   mentionsMetas xs (UnblockOnMeta x)  = HashSet.member x xs
   mentionsMetas xs UnblockOnProblem{} = False
+  mentionsMetas xs UnblockOnDef{}     = False
 
 instance MentionsMeta Type where
     mentionsMetas xs (El s t) = mentionsMetas xs (s, t)
@@ -55,6 +56,7 @@ instance MentionsMeta Sort where
     SSet l     -> mentionsMetas xs l
     SizeUniv   -> False
     LockUniv   -> False
+    IntervalUniv -> False
     PiSort a s1 s2 -> mentionsMetas xs (a, s1, s2)
     FunSort s1 s2 -> mentionsMetas xs (s1, s2)
     UnivSort s -> mentionsMetas xs s
@@ -117,10 +119,11 @@ instance MentionsMeta Constraint where
     HasBiggerSort a     -> mm a
     HasPTSRule a b      -> mm (a, b)
     UnquoteTactic tac hole goal -> False
+    CheckDataSort q s   -> mm s
     CheckMetaInst m     -> True   -- TODO
     CheckType t         -> mm t
     CheckLockedVars a b c d -> mm ((a, b), (c, d))
-    UsableAtModality mod t -> mm t
+    UsableAtModality _ ms mod t -> mm (ms, t)
     where
       mm :: forall t. MentionsMeta t => t -> Bool
       mm = mentionsMetas xs
