@@ -32,7 +32,7 @@ import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.MetaVars
 
-import Agda.TypeChecking.Constraints () -- instances
+import Agda.TypeChecking.Constraints ( wakeupConstraints )
 import Agda.TypeChecking.Monad
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 import Agda.TypeChecking.Reduce
@@ -190,7 +190,7 @@ definitionCheck d = do
         [ "occursCheck: definition"
         , prettyTCM d
         , "has relevance"
-        , prettyTCM (getRelevance dmod)
+        , text . show $ getRelevance dmod
         ]
       abort neverUnblock $ MetaIrrelevantSolution m $ Def d []
     unless (er || usableQuantity dmod) $ do
@@ -198,7 +198,7 @@ definitionCheck d = do
         [ "occursCheck: definition"
         , prettyTCM d
         , "has quantity"
-        , prettyTCM (getQuantity dmod)
+        , text . show $ getQuantity dmod
         ]
       abort neverUnblock $ MetaErasedSolution m $ Def d []
 
@@ -225,17 +225,17 @@ metaCheck m = do
   --   abort ctx $ MetaOccursInItself m'
   when (m == m0) $ patternViolation' neverUnblock 50 $ "occursCheck failed: Found " ++ prettyShow m
 
-  mv <- lookupMeta m
-  let mmod = getMetaModality mv
+  mv <- lookupLocalMeta m
+  let mmod = getModality mv
       mmod' = setRelevance rel $ setQuantity qnt $ mmod
   if (mmod `moreUsableModality` mmod') then return m else do
     reportSDoc "tc.meta.occurs" 35 $ hsep
       [ "occursCheck: meta variable"
       , prettyTCM m
       , "has relevance"
-      , prettyTCM (getRelevance mmod)
+      , text . show $ getRelevance mmod
       , "and quantity"
-      , prettyTCM (getQuantity mmod)
+      , text . show $ getQuantity mmod
       ]
     allowAssign <- asksTC envAssignMetas
     -- Jesper, 2020-11-10: if we encounter a metavariable that is
@@ -260,15 +260,10 @@ metaCheck m = do
     when (isUnguarded cxt)                   $ fail "occurrence is unguarded"
 
     reportSDoc "tc.meta.occurs" 20 $ "Promoting meta" <+> prettyTCM m <+> "to modality" <+> prettyTCM mmod'
-    let info' = setModality mmod' $ mvInfo mv
-    m' <- liftTCM $ newMeta Instantiable info' (mvPriority mv) (mvPermutation mv) (mvJudgement mv)
-    reportSDoc "tc.meta.occurs.qnt" 20 $ hsep
-       [ "occursCheck: new meta variable"
-       , prettyTCM m'
-       ]
-    liftTCM $ assignTerm m [] $ MetaV m' []
-    reportSDoc "tc.meta.occurs" 35 $ "New name for" <+> prettyTCM m <+> "is" <+> prettyTCM m'
-    return m'
+    updateMetaVar m $ \ mv -> mv { mvInfo = setModality mmod' $ mvInfo mv }
+    etaExpandListeners m
+    wakeupConstraints m
+    return m
 
 -- | Construct a test whether a de Bruijn index is allowed
 --   or needs to be pruned.
@@ -276,9 +271,9 @@ allowedVars :: OccursM (Nat -> Bool)
 allowedVars = do
   -- @n@ is the number of binders we have stepped under.
   n  <- liftM2 (-) getContextSize (asks (occCxtSize . feExtra))
-  xs <- asks ((IntMap.keysSet . theVarMap) . (occVars . feExtra))
+  xs <- asks (theVarMap . occVars . feExtra)
   -- Bound variables are allowed, and those mentioned in occVars.
-  return $ \ i -> i < n || (i - n) `IntSet.member` xs
+  return $ \ i -> i < n || (i - n) `IntMap.member` xs
 
 -- ** Unfolding during occurs check.
 
@@ -367,7 +362,7 @@ occursCheck
   :: (Occurs a, InstantiateFull a, PrettyTCM a)
   => MetaId -> VarMap -> a -> TCM a
 occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
-  mv <- lookupMeta m
+  mv <- lookupLocalMeta m
   n  <- getContextSize
   reportSLn "tc.meta.occurs" 35 $ "occursCheck " ++ show m ++ " " ++ show xs
   let initEnv unf = FreeEnv
@@ -378,7 +373,7 @@ occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
           , occCxtSize = n
           }
         , feFlexRig   = StronglyRigid -- ? Unguarded
-        , feModality  = getMetaModality mv
+        , feModality  = getModality mv
         , feSingleton = variableCheck xs
         }
   initOccursCheck mv
@@ -389,7 +384,7 @@ occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
       -- (unless metavariable is irrelevant, in which case the
       -- constraint will anyway be dropped)
       case err of
-        PatternErr{} | not (isIrrelevant $ getMetaModality mv) -> do
+        PatternErr{} | not (isIrrelevant $ getModality mv) -> do
           initOccursCheck mv
           occurs v `runReaderT` initEnv YesUnfold
         _ -> throwError err
@@ -401,19 +396,25 @@ occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
       TypeError _ _ cl -> case clValue cl of
         MetaOccursInItself{} ->
           typeError . GenericDocError =<<
-            fsep [ text ("Refuse to construct infinite term by instantiating " ++ prettyShow m ++ " to")
+            fsep [ text "Refuse to construct infinite term by instantiating"
+                 , prettyTCM m
+                 , "to"
                  , prettyTCM =<< instantiateFull v
                  ]
         MetaCannotDependOn _ i ->
           ifM (isSortMeta m `and2M` (not <$> hasUniversePolymorphism))
           ( typeError . GenericDocError =<<
-            fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to")
+            fsep [ text "Cannot instantiate the metavariable"
+                 , prettyTCM m
+                 , "to"
                  , prettyTCM v
                  , "since universe polymorphism is disabled"
                  ]
           ) {- else -}
           ( typeError . GenericDocError =<<
-              fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to solution")
+              fsep [ text "Cannot instantiate the metavariable"
+                   , prettyTCM m
+                   , "to solution"
                    , prettyTCM v
                    , "since it contains the variable"
                    , enterClosure cl $ \_ -> prettyTCM (Var i [])
@@ -422,13 +423,17 @@ occursCheck m xs v = Bench.billTo [ Bench.Typing, Bench.OccursCheck ] $ do
             )
         MetaIrrelevantSolution _ _ ->
           typeError . GenericDocError =<<
-            fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to solution")
+            fsep [ text "Cannot instantiate the metavariable"
+                 , prettyTCM m
+                 , "to solution"
                  , prettyTCM v
                  , "since (part of) the solution was created in an irrelevant context"
                  ]
         MetaErasedSolution _ _ ->
           typeError . GenericDocError =<<
-            fsep [ text ("Cannot instantiate the metavariable " ++ prettyShow m ++ " to solution")
+            fsep [ text "Cannot instantiate the metavariable"
+                 , prettyTCM m
+                 , "to solution"
                  , prettyTCM v
                  , "since (part of) the solution was created in an erased context"
                  ]
@@ -613,6 +618,7 @@ instance Occurs Sort where
       SSet a     -> SSet <$> occurs a
       s@SizeUniv -> return s
       s@LockUniv -> return s
+      s@IntervalUniv -> return s
       UnivSort s -> UnivSort <$> do flexibly $ occurs s
       MetaS x es -> do
         MetaV x es <- occurs (MetaV x es)
@@ -633,6 +639,7 @@ instance Occurs Sort where
       SSet a     -> metaOccurs m a
       SizeUniv   -> return ()
       LockUniv   -> return ()
+      IntervalUniv -> return ()
       UnivSort s -> metaOccurs m s
       MetaS x es -> metaOccurs m $ MetaV x es
       DefS d es  -> metaOccurs m $ Def d es
@@ -822,7 +829,7 @@ instance AnyRigid Term where
         -- If the definition is incomplete, arguments might disappear
         -- by reductions that come with more clauses, thus, these
         -- arguments are not rigid.
-        NotBlocked MissingClauses _ -> return False
+        NotBlocked (MissingClauses _) _ -> return False
         -- _        -> mempty -- breaks: ImproveInertRHS, Issue442, PruneRecord, PruningNonMillerPattern
         _        -> anyRigid f es
       Con _ _ ts -> anyRigid f ts
@@ -845,6 +852,7 @@ instance AnyRigid Sort where
       SSet l     -> anyRigid f l
       SizeUniv   -> return False
       LockUniv   -> return False
+      IntervalUniv -> return False
       PiSort a s1 s2 -> return False
       FunSort s1 s2 -> return False
       UnivSort s -> anyRigid f s
@@ -898,7 +906,7 @@ killArgs :: (MonadMetaSolver m) => [Bool] -> MetaId -> m PruneResult
 killArgs kills _
   | not (or kills) = return NothingToPrune  -- nothing to kill
 killArgs kills m = do
-  mv <- lookupMeta m
+  mv <- lookupLocalMeta m
   allowAssign <- asksTC envAssignMetas
   if mvFrozen mv == Frozen || not allowAssign then return PrunedNothing else do
       -- Andreas 2011-04-26, we allow pruning in MetaV and MetaS
@@ -940,9 +948,12 @@ killArgs kills m = do
 killedType :: (MonadReduce m) => [(Dom (ArgName, Type), Bool)] -> Type -> m ([Arg Bool], Type)
 killedType args b = do
 
+  let n = length args
+  let iargs = zip (downFrom n) args
+
   -- Turn list of bools into an IntSet containing the variables we want to kill
   -- (indices relative to b).
-  let tokill = IntSet.fromList [ i | ((_, True), i) <- zip (reverse args) [0..] ]
+  let tokill = IntSet.fromList [ i | (i, (_, True)) <- iargs ]
 
   -- First, check the free variables of b to see if they prevent any kills.
   (tokill, b) <- reallyNotFreeIn tokill b
@@ -952,7 +963,7 @@ killedType args b = do
 
   -- Turn the IntSet of killed variables into the list of Arg Bool's to return.
   let kills = [ Arg (getArgInfo dom) (IntSet.member i killed)
-              | (i, (dom, _)) <- reverse $ zip [0..] $ reverse args ]
+              | (i, (dom, _)) <- iargs ]
   return (kills, b)
   where
     down = IntSet.map pred
@@ -990,7 +1001,7 @@ killedType args b = do
           (ys, a) <- reallyNotFreeIn xs' a
           -- Recurse on Δ, ys, and (x : A') → B, where A reduces to A' and ys ⊆ xs'
           -- not free in A'. We already know ys not free in B.
-          (zs, b) <- go args ys (mkPi ((name, a) <$ arg) b)
+          (zs, b) <- go args ys $ mkPi ((name, a) <$ arg) b
           -- Shift back up to make it relative to Δ (x : A) again.
           return (up zs, b)
 
@@ -1024,7 +1035,7 @@ performKill
   -> Type          -- ^ The pruned type of the new meta var.
   -> m ()
 performKill kills m a = do
-  mv <- lookupMeta m
+  mv <- lookupLocalMeta m
   when (mvFrozen mv == Frozen) __IMPOSSIBLE__
   -- Arity of the old meta.
   let n = size kills
