@@ -15,21 +15,25 @@ import Control.Monad.Trans.Identity (IdentityT)
 import Control.Monad.Trans.Maybe
 import Control.Monad.Writer
 
-import qualified Data.Map as Map
 import Data.Function ( on )
+import qualified Data.Map as Map
+import Data.Set (Set)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Agda.Syntax.Literal
 import Agda.Syntax.Builtin
 import Agda.Syntax.Internal as I
+import Agda.Interaction.Options.Base (PragmaOptions(..))
 import Agda.TypeChecking.Monad.Base
 -- import Agda.TypeChecking.Functions  -- LEADS TO IMPORT CYCLE
 import Agda.TypeChecking.Substitute
 
+import Agda.Utils.Functor
 import Agda.Utils.ListT
 import Agda.Utils.Monad
 import Agda.Utils.Maybe
+import Agda.Utils.Singleton
 import Agda.Utils.Tuple
 
 import Agda.Utils.Impossible
@@ -54,8 +58,10 @@ instance (HasBuiltins m, Monoid w) => HasBuiltins (WriterT w m)
 deriving instance HasBuiltins m => HasBuiltins (BlockT m)
 
 instance MonadIO m => HasBuiltins (TCMT m) where
-  getBuiltinThing b = liftM2 mplus (Map.lookup b <$> useTC stLocalBuiltins)
-                      (Map.lookup b <$> useTC stImportedBuiltins)
+  getBuiltinThing b =
+    liftM2 (unionMaybeWith unionBuiltin)
+      (Map.lookup b <$> useTC stLocalBuiltins)
+      (Map.lookup b <$> useTC stImportedBuiltins)
 
 -- If Agda is changed so that the type of a literal can belong to an
 -- inductive family (with at least one index), then the implementation
@@ -87,6 +93,7 @@ bindBuiltinName b x = do
   case builtin of
     Just (Builtin y) -> typeError $ DuplicateBuiltinBinding b y x
     Just (Prim _)    -> typeError $ NoSuchBuiltinName b
+    Just BuiltinRewriteRelations{} -> __IMPOSSIBLE__
     Nothing          -> stLocalBuiltins `modifyTCLens` Map.insert b (Builtin x)
 
 bindPrimitive :: String -> PrimFun -> TCM ()
@@ -95,9 +102,25 @@ bindPrimitive b pf = do
   case builtin of
     Just (Builtin _) -> typeError $ NoSuchPrimitiveFunction b
     Just (Prim x)    -> typeError $ (DuplicatePrimitiveBinding b `on` primFunName) x pf
+    Just BuiltinRewriteRelations{} -> __IMPOSSIBLE__
     Nothing          -> stLocalBuiltins `modifyTCLens` Map.insert b (Prim pf)
 
-getBuiltin :: (HasBuiltins m, MonadError TCErr m, MonadTCEnv m, ReadTCState m)
+-- | Add one (more) relation symbol to the rewrite relations.
+bindBuiltinRewriteRelation :: QName -> TCM ()
+bindBuiltinRewriteRelation x =
+  stLocalBuiltins `modifyTCLens`
+    Map.insertWith unionBuiltin builtinRewrite (BuiltinRewriteRelations $ singleton x)
+
+-- | Get the currently registered rewrite relation symbols.
+getBuiltinRewriteRelations :: HasBuiltins m => m (Maybe (Set QName))
+getBuiltinRewriteRelations = fmap rels <$> getBuiltinThing builtinRewrite
+  where
+  rels = \case
+    BuiltinRewriteRelations xs -> xs
+    Prim{}    -> __IMPOSSIBLE__
+    Builtin{} -> __IMPOSSIBLE__
+
+getBuiltin :: (HasBuiltins m, MonadTCError m)
            => String -> m Term
 getBuiltin x =
   fromMaybeM (typeError $ NoBindingForBuiltin x) $ getBuiltin' x
@@ -106,6 +129,7 @@ getBuiltin' :: HasBuiltins m => String -> m (Maybe Term)
 getBuiltin' x = do
     builtin <- getBuiltinThing x
     case builtin of
+        Just BuiltinRewriteRelations{} -> __IMPOSSIBLE__
         Just (Builtin t) -> return $ Just $ killRange t
         _                -> return Nothing
 
@@ -113,6 +137,7 @@ getPrimitive' :: HasBuiltins m => String -> m (Maybe PrimFun)
 getPrimitive' x = (getPrim =<<) <$> getBuiltinThing x
   where
     getPrim (Prim pf) = return pf
+    getPrim BuiltinRewriteRelations{} = __IMPOSSIBLE__
     getPrim _         = Nothing
 
 getPrimitive :: (HasBuiltins m, MonadError TCErr m, MonadTCEnv m, ReadTCState m)
@@ -166,7 +191,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primFloat, primChar, primString, primUnit, primUnitUnit, primBool, primTrue, primFalse,
     primSigma,
     primList, primNil, primCons, primIO, primNat, primSuc, primZero, primMaybe, primNothing, primJust,
-    primPath, primPathP, primInterval, primIZero, primIOne, primPartial, primPartialP,
+    primPath, primPathP, primIntervalUniv, primInterval, primIZero, primIOne, primPartial, primPartialP,
     primIMin, primIMax, primINeg,
     primIsOne, primItIsOne, primIsOne1, primIsOne2, primIsOneEmpty,
     primSub, primSubIn, primSubOut,
@@ -184,7 +209,6 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primSizeUniv, primSize, primSizeLt, primSizeSuc, primSizeInf, primSizeMax,
     primInf, primSharp, primFlat,
     primEquality, primRefl,
-    primRewrite, -- Name of rewrite relation
     primLevel, primLevelZero, primLevelSuc, primLevelMax,
     primLockUniv,
     primSet, primProp, primSetOmega, primStrictSet, primSSetOmega,
@@ -193,7 +217,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primQName, primArgInfo, primArgArgInfo, primArg, primArgArg, primAbs, primAbsAbs, primAgdaTerm, primAgdaTermVar,
     primAgdaTermLam, primAgdaTermExtLam, primAgdaTermDef, primAgdaTermCon, primAgdaTermPi,
     primAgdaTermSort, primAgdaTermLit, primAgdaTermUnsupported, primAgdaTermMeta,
-    primAgdaErrorPart, primAgdaErrorPartString, primAgdaErrorPartTerm, primAgdaErrorPartName,
+    primAgdaErrorPart, primAgdaErrorPartString, primAgdaErrorPartTerm, primAgdaErrorPartPatt, primAgdaErrorPartName,
     primHiding, primHidden, primInstance, primVisible,
     primRelevance, primRelevant, primIrrelevant,
     primQuantity, primQuantity0, primQuantityω,
@@ -214,15 +238,17 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaTCMTypeError, primAgdaTCMInferType, primAgdaTCMCheckType,
     primAgdaTCMNormalise, primAgdaTCMReduce,
     primAgdaTCMCatchError, primAgdaTCMGetContext, primAgdaTCMExtendContext, primAgdaTCMInContext,
-    primAgdaTCMFreshName, primAgdaTCMDeclareDef, primAgdaTCMDeclarePostulate, primAgdaTCMDefineFun,
+    primAgdaTCMFreshName, primAgdaTCMDeclareDef, primAgdaTCMDeclarePostulate, primAgdaTCMDeclareData, primAgdaTCMDefineData, primAgdaTCMDefineFun,
     primAgdaTCMGetType, primAgdaTCMGetDefinition,
     primAgdaTCMQuoteTerm, primAgdaTCMUnquoteTerm, primAgdaTCMQuoteOmegaTerm,
     primAgdaTCMBlockOnMeta, primAgdaTCMCommit, primAgdaTCMIsMacro,
-    primAgdaTCMWithNormalisation, primAgdaTCMDebugPrint, primAgdaTCMWithReconsParams,
+    primAgdaTCMFormatErrorParts, primAgdaTCMDebugPrint,
+    primAgdaTCMWithNormalisation, primAgdaTCMWithReconsParams,
     primAgdaTCMOnlyReduceDefs, primAgdaTCMDontReduceDefs,
     primAgdaTCMNoConstraints,
     primAgdaTCMRunSpeculative,
-    primAgdaTCMExec
+    primAgdaTCMExec,
+    primAgdaTCMGetInstances
     :: (HasBuiltins m, MonadError TCErr m, MonadTCEnv m, ReadTCState m) => m Term
 
 primInteger                           = getBuiltin builtinInteger
@@ -245,10 +271,11 @@ primNothing                           = getBuiltin builtinNothing
 primJust                              = getBuiltin builtinJust
 primIO                                = getBuiltin builtinIO
 primId                                = getBuiltin builtinId
-primConId                             = getBuiltin builtinConId
+primConId                             = getPrimitiveTerm builtinConId
 primIdElim                            = getPrimitiveTerm builtinIdElim
 primPath                              = getBuiltin builtinPath
 primPathP                             = getBuiltin builtinPathP
+primIntervalUniv                      = getBuiltin builtinIntervalUniv
 primInterval                          = getBuiltin builtinInterval
 primIZero                             = getBuiltin builtinIZero
 primIOne                              = getBuiltin builtinIOne
@@ -299,7 +326,6 @@ primSharp                             = getBuiltin builtinSharp
 primFlat                              = getBuiltin builtinFlat
 primEquality                          = getBuiltin builtinEquality
 primRefl                              = getBuiltin builtinRefl
-primRewrite                           = getBuiltin builtinRewrite
 primLevel                             = getBuiltin builtinLevel
 primLevelZero                         = getBuiltin builtinLevelZero
 primLevelSuc                          = getBuiltin builtinLevelSuc
@@ -362,6 +388,7 @@ primAgdaTermMeta                      = getBuiltin builtinAgdaTermMeta
 primAgdaErrorPart                     = getBuiltin builtinAgdaErrorPart
 primAgdaErrorPartString               = getBuiltin builtinAgdaErrorPartString
 primAgdaErrorPartTerm                 = getBuiltin builtinAgdaErrorPartTerm
+primAgdaErrorPartPatt                 = getBuiltin builtinAgdaErrorPartPatt
 primAgdaErrorPartName                 = getBuiltin builtinAgdaErrorPartName
 primAgdaLiteral                       = getBuiltin builtinAgdaLiteral
 primAgdaLitNat                        = getBuiltin builtinAgdaLitNat
@@ -405,6 +432,8 @@ primAgdaTCMInContext                  = getBuiltin builtinAgdaTCMInContext
 primAgdaTCMFreshName                  = getBuiltin builtinAgdaTCMFreshName
 primAgdaTCMDeclareDef                 = getBuiltin builtinAgdaTCMDeclareDef
 primAgdaTCMDeclarePostulate           = getBuiltin builtinAgdaTCMDeclarePostulate
+primAgdaTCMDeclareData                = getBuiltin builtinAgdaTCMDeclareData
+primAgdaTCMDefineData                 = getBuiltin builtinAgdaTCMDefineData
 primAgdaTCMDefineFun                  = getBuiltin builtinAgdaTCMDefineFun
 primAgdaTCMGetType                    = getBuiltin builtinAgdaTCMGetType
 primAgdaTCMGetDefinition              = getBuiltin builtinAgdaTCMGetDefinition
@@ -416,12 +445,14 @@ primAgdaTCMCommit                     = getBuiltin builtinAgdaTCMCommit
 primAgdaTCMIsMacro                    = getBuiltin builtinAgdaTCMIsMacro
 primAgdaTCMWithNormalisation          = getBuiltin builtinAgdaTCMWithNormalisation
 primAgdaTCMWithReconsParams           = getBuiltin builtinAgdaTCMWithReconsParams
+primAgdaTCMFormatErrorParts           = getBuiltin builtinAgdaTCMFormatErrorParts
 primAgdaTCMDebugPrint                 = getBuiltin builtinAgdaTCMDebugPrint
 primAgdaTCMOnlyReduceDefs             = getBuiltin builtinAgdaTCMOnlyReduceDefs
 primAgdaTCMDontReduceDefs             = getBuiltin builtinAgdaTCMDontReduceDefs
 primAgdaTCMNoConstraints              = getBuiltin builtinAgdaTCMNoConstraints
 primAgdaTCMRunSpeculative             = getBuiltin builtinAgdaTCMRunSpeculative
 primAgdaTCMExec                       = getBuiltin builtinAgdaTCMExec
+primAgdaTCMGetInstances               = getBuiltin builtinAgdaTCMGetInstances
 
 -- | The coinductive primitives.
 
@@ -456,13 +487,18 @@ data SortKit = SortKit
   , nameOfSetOmega :: IsFibrant -> QName
   }
 
-sortKit :: HasBuiltins m => m SortKit
+-- | Compute a 'SortKit' in an environment that supports failures.
+--
+-- When 'optLoadPrimitives' is set to 'False', 'sortKit' is a fallible operation,
+-- so for the uses of 'sortKit' in fallible contexts (e.g. 'TCM'),
+-- we report a type error rather than exploding.
+sortKit :: (HasBuiltins m, MonadTCError m, HasOptions m) => m SortKit
 sortKit = do
-  Def set      _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSet
-  Def prop     _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinProp
-  Def setomega _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSetOmega
-  Def sset     _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinStrictSet
-  Def ssetomega _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSSetOmega
+  Def set      _  <- getBuiltin builtinSet
+  Def prop     _  <- getBuiltin builtinProp
+  Def setomega _  <- getBuiltin builtinSetOmega
+  Def sset     _  <- getBuiltin builtinStrictSet
+  Def ssetomega _ <- getBuiltin builtinSSetOmega
   return $ SortKit
     { nameOfSet      = set
     , nameOfProp     = prop
@@ -472,6 +508,25 @@ sortKit = do
         IsStrict  -> ssetomega
     }
 
+-- | Compute a 'SortKit' in contexts that do not support failure (e.g.
+-- 'Reify'). This should only be used when we are sure that the
+-- primitive sorts have been bound, i.e. because it is "after" type
+-- checking.
+infallibleSortKit :: HasBuiltins m => m SortKit
+infallibleSortKit = do
+  Def set      _  <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSet
+  Def prop     _  <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinProp
+  Def setomega _  <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSetOmega
+  Def sset     _  <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinStrictSet
+  Def ssetomega _ <- fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSSetOmega
+  return $ SortKit
+    { nameOfSet      = set
+    , nameOfProp     = prop
+    , nameOfSSet     = sset
+    , nameOfSetOmega = \case
+        IsFibrant -> setomega
+        IsStrict  -> ssetomega
+    }
 
 ------------------------------------------------------------------------
 -- * Path equality
@@ -495,6 +550,9 @@ getPrimitiveName' n = fmap primFunName <$> getPrimitive' n
 isPrimitive :: HasBuiltins m => String -> QName -> m Bool
 isPrimitive n q = (Just q ==) <$> getPrimitiveName' n
 
+intervalSort :: Sort
+intervalSort = IntervalUniv
+
 intervalView' :: HasBuiltins m => m (Term -> IntervalView)
 intervalView' = do
   iz <- getBuiltinName' builtinIZero
@@ -594,6 +652,23 @@ pathUnview (OType t) = t
 pathUnview (PathType s path l t lhs rhs) =
   El s $ Def path $ map Apply [l, t, lhs, rhs]
 
+------------------------------------------------------------------------
+-- * Swan's Id Equality
+------------------------------------------------------------------------
+
+-- f x (< phi , p > : Id A x _) = Just (phi,p)
+conidView' :: HasBuiltins m => m (Term -> Term -> Maybe (Arg Term,Arg Term))
+conidView' = do
+  mn <- sequence <$> mapM getName' [builtinReflId, builtinConId]
+  mio <- getTerm' builtinIOne
+  let fallback = return $ \ _ _ -> Nothing
+  caseMaybe mn fallback $ \ [refl,conid] ->
+   caseMaybe mio fallback $ \ io -> return $ \ x t ->
+    case t of
+      Con h _ [] | conName h == refl -> Just (defaultArg io,defaultArg (Lam defaultArgInfo $ NoAbs "_" $ x))
+      Def d es | Just [l,a,x,y,phi,p] <- allApplyElims es, d == conid -> Just (phi, p)
+      _ -> Nothing
+
 ------------------------------------------------------------------------
 -- * Builtin equality
 ------------------------------------------------------------------------
@@ -638,11 +713,18 @@ equalityView t0@(El s t) = do
 --
 --   Postcondition: type is reduced.
 
-equalityUnview :: EqualityView -> Type
-equalityUnview (OtherType t) = t
-equalityUnview (IdiomType t) = t
-equalityUnview (EqualityType s equality l t lhs rhs) =
-  El s $ Def equality $ map Apply (l ++ [t, lhs, rhs])
+class EqualityUnview a where
+  equalityUnview :: a -> Type
+
+instance EqualityUnview EqualityView where
+  equalityUnview = \case
+    OtherType t -> t
+    IdiomType t -> t
+    EqualityViewType eqt -> equalityUnview eqt
+
+instance EqualityUnview EqualityTypeData where
+  equalityUnview (EqualityTypeData s equality l t lhs rhs) =
+    El s $ Def equality $ map Apply (l ++ [t, lhs, rhs])
 
 -- | Primitives with typechecking constrants.
 constrainedPrims :: [String]
