@@ -1,6 +1,9 @@
 
 module Agda.TypeChecking.Monad.Context where
 
+import Data.Text (Text)
+import qualified Data.Text as T
+
 import Control.Monad                ( (<=<), forM, when )
 import Control.Monad.Except
 import Control.Monad.Reader
@@ -8,10 +11,12 @@ import Control.Monad.State
 import Control.Monad.Trans.Control  ( MonadTransControl(..), liftThrough )
 import Control.Monad.Trans.Identity ( IdentityT )
 import Control.Monad.Trans.Maybe
-import Control.Monad.Writer
+import Control.Monad.Writer         ( WriterT )
 -- Control.Monad.Fail import is redundant since GHC 8.8.1
 import Control.Monad.Fail (MonadFail)
 
+import qualified Data.DList as DL
+import Data.Foldable
 import qualified Data.List as List
 import qualified Data.Map as Map
 
@@ -122,7 +127,7 @@ checkpoint sub k = do
   -- created under refined parent parameters, but as long as those modules
   -- aren't named we shouldn't look at the checkpoint. The right thing to do
   -- would be to not store these modules in the checkpoint map, but todo..
-  stModuleCheckpoints `setTCLens` Map.union oldMods (old <$ Map.difference newMods oldMods)
+  stModuleCheckpoints `setTCLens` Map.union oldMods (old <$ newMods)
   unlessDebugPrinting $ reportSLn "tc.cxt.checkpoint" 105 "}"
   return x
 
@@ -228,19 +233,22 @@ withShadowingNameTCM x f = do
         setTCLens stUsedNames Map.empty
         result <- f
         newUsedNames <- useTC stUsedNames
-        setTCLens stUsedNames $ Map.unionWith (++) origUsedNames newUsedNames
+        setTCLens stUsedNames $ Map.unionWith (<>) origUsedNames newUsedNames
         return (result , newUsedNames)
 
       tellUsedName x = do
         let concreteX = nameConcrete x
             rawX      = nameToRawName concreteX
             rootX     = nameRoot concreteX
-        modifyTCLens (stUsedNames . key rootX) $ Just . (rawX:) . concat
+        modifyTCLens (stUsedNames . key rootX) $
+          Just . (rawX `DL.cons`) . fold
 
       tellShadowing x useds = case Map.lookup (nameRoot $ nameConcrete x) useds of
         Just shadows -> do
-          reportSDoc "tc.cxt.shadowing" 80 $ pure $ "names shadowing" <+> pretty x <+> ": " <+> prettyList_ (map pretty shadows)
-          modifyTCLens stShadowingNames $ Map.insertWith (++) x shadows
+          reportSDoc "tc.cxt.shadowing" 80 $ pure $
+            "names shadowing" <+> pretty x <+> ": " <+>
+            prettyList_ (map pretty $ toList shadows)
+          modifyTCLens stShadowingNames $ Map.insertWith (<>) x shadows
         Nothing      -> return ()
 
 instance MonadAddContext TCM where
@@ -331,6 +339,10 @@ instance AddContext (String, Dom Type) where
     withFreshName noRange s $ \x -> addCtx (setNotInScope x) dom ret
   contextSize _ = 1
 
+instance AddContext (Text, Dom Type) where
+  addContext (s, dom) ret = addContext (T.unpack s, dom) ret
+  contextSize _ = 1
+
 instance AddContext (KeepNames String, Dom Type) where
   addContext (KeepNames s, dom) ret =
     withFreshName noRange s $ \ x -> addCtx x dom ret
@@ -414,8 +426,8 @@ addLetBinding info x v t0 ret = addLetBinding' x v (defaultArgDom info t0) ret
 -- * Querying the context
 
 -- | Get the current context.
-{-# SPECIALIZE getContext :: TCM [Dom (Name, Type)] #-}
-getContext :: MonadTCEnv m => m [Dom (Name, Type)]
+{-# SPECIALIZE getContext :: TCM Context #-}
+getContext :: MonadTCEnv m => m Context
 getContext = asksTC envContext
 
 -- | Get the size of the current context.
@@ -446,8 +458,8 @@ getContextNames = map (fst . unDom) <$> getContext
 
 -- | get type of bound variable (i.e. deBruijn index)
 --
-{-# SPECIALIZE lookupBV' :: Nat -> TCM (Maybe (Dom (Name, Type))) #-}
-lookupBV' :: MonadTCEnv m => Nat -> m (Maybe (Dom (Name, Type)))
+{-# SPECIALIZE lookupBV' :: Nat -> TCM (Maybe ContextEntry) #-}
+lookupBV' :: MonadTCEnv m => Nat -> m (Maybe ContextEntry)
 lookupBV' n = do
   ctx <- getContext
   return $ raise (n + 1) <$> ctx !!! n
