@@ -1,6 +1,8 @@
 
 module Agda.TypeChecking.Monad.Options where
 
+import Prelude hiding (null)
+
 import Control.Arrow          ( (&&&) )
 import Control.Monad          ( unless, when )
 import Control.Monad.IO.Class ( MonadIO(..) )
@@ -19,24 +21,28 @@ import System.Directory
 import System.FilePath
 
 import Agda.Syntax.Common
-import Agda.Syntax.Concrete (TopLevelModuleName)
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.TopLevelModuleName
+
 import Agda.TypeChecking.Monad.Debug (reportSDoc)
 import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Imports
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Benchmark
+
 import Agda.Interaction.FindFile
 import Agda.Interaction.Options
 import qualified Agda.Interaction.Options.Lenses as Lens
 import Agda.Interaction.Library
+
 import Agda.Utils.FileName
 import Agda.Utils.Functor
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as G
 import Agda.Utils.List
 import Agda.Utils.Maybe
+import Agda.Utils.Null
 import Agda.Utils.Pretty
 import Agda.Utils.Tuple
 import Agda.Utils.WithDefault
@@ -44,26 +50,23 @@ import Agda.Utils.WithDefault
 import Agda.Utils.Impossible
 
 -- | Sets the pragma options.
-
+--   Checks for unsafe combinations.
 setPragmaOptions :: PragmaOptions -> TCM ()
 setPragmaOptions opts = do
-  stPragmaOptions `modifyTCLens` Lens.mapSafeMode (Lens.getSafeMode opts ||)
-  clo <- commandLineOptions
-  let unsafe = unsafePragmaOptions clo opts
-  when (Lens.getSafeMode opts && not (null unsafe)) $ warning $ SafeFlagPragma unsafe
-  runOptM (checkOpts clo{ optPragmaOptions = opts }) >>= \case
-    Left err   -> __IMPOSSIBLE__
-    Right opts -> do
-      stPragmaOptions `setTCLens` optPragmaOptions opts
-      updateBenchmarkingStatus
+  -- Check for unsafe pragma options if @--safe@ is on.
+  when (Lens.getSafeMode opts) $
+    unlessNull (unsafePragmaOptions opts) $ \ unsafe ->
+      warning $ SafeFlagPragma unsafe
+
+  stPragmaOptions `setTCLens` opts
+  updateBenchmarkingStatus
 
 -- | Sets the command line options (both persistent and pragma options
 -- are updated).
 --
 -- Relative include directories are made absolute with respect to the
 -- current working directory. If the include directories have changed
--- (thus, they are 'Left' now, and were previously @'Right' something@),
--- then the state is reset (completely, see setIncludeDirs) .
+-- then the state is reset (partly, see 'setIncludeDirs').
 --
 -- An empty list of relative include directories (@'Left' []@) is
 -- interpreted as @["."]@.
@@ -78,9 +81,7 @@ setCommandLineOptions'
   -> CommandLineOptions
   -> TCM ()
 setCommandLineOptions' root opts = do
-  runOptM (checkOpts opts) >>= \case
-    Left err   -> __IMPOSSIBLE__
-    Right opts -> do
+  -- Andreas, 2022-11-19: removed a call to checkOpts which did nothing.
       incs <- case optAbsoluteIncludePaths opts of
         [] -> do
           opts' <- setLibraryPaths root opts
@@ -108,14 +109,26 @@ libToTCM m = do
     Left s  -> typeError $ GenericDocError s
     Right x -> return x
 
-getAgdaLibFiles :: FilePath -> TCM [AgdaLibFile]
-getAgdaLibFiles root = do
+-- | Returns the library files for a given file.
+
+getAgdaLibFiles
+  :: AbsolutePath        -- ^ The file name.
+  -> TopLevelModuleName  -- ^ The top-level module name.
+  -> TCM [AgdaLibFile]
+getAgdaLibFiles f m = do
   useLibs <- optUseLibs <$> commandLineOptions
   if | useLibs   -> libToTCM $ mkLibM [] $ getAgdaLibFiles' root
      | otherwise -> return []
+  where
+  root = filePath (projectRoot f m)
 
-getLibraryOptions :: FilePath -> TCM [OptionsPragma]
-getLibraryOptions root = map _libPragmas <$> getAgdaLibFiles root
+-- | Returns the library options for a given file.
+
+getLibraryOptions
+  :: AbsolutePath        -- ^ The file name.
+  -> TopLevelModuleName  -- ^ The top-level module name.
+  -> TCM [OptionsPragma]
+getLibraryOptions f m = map _libPragmas <$> getAgdaLibFiles f m
 
 setLibraryPaths
   :: AbsolutePath
@@ -161,7 +174,9 @@ addTrustedExecutables o = do
 setOptionsFromPragma :: OptionsPragma -> TCM ()
 setOptionsFromPragma ps = do
     opts <- commandLineOptions
-    runOptM (parsePragmaOptions ps opts) >>= \case
+    let (z, warns) = runOptM (parsePragmaOptions ps opts)
+    mapM_ (warning . OptionWarning) warns
+    case z of
       Left err    -> typeError $ GenericError err
       Right opts' -> setPragmaOptions opts'
 
@@ -278,16 +293,16 @@ setIncludeDirs incs root = do
     -- A graph with one node per module in old, and an edge from m to
     -- n if the module corresponding to m imports the module
     -- corresponding to n.
-    dependencyGraph :: G.Graph A.ModuleName ()
+    dependencyGraph :: G.Graph TopLevelModuleName ()
     dependencyGraph =
       G.fromNodes
-        [ iModuleName $ miInterface m
+        [ iTopLevelModuleName $ miInterface m
         | m <- Map.elems old
         ]
         `G.union`
       G.fromEdges
         [ G.Edge
-            { source = iModuleName $ miInterface m
+            { source = iTopLevelModuleName $ miInterface m
             , target = d
             , label = ()
             }
@@ -304,17 +319,16 @@ setIncludeDirs incs root = do
                 -- Agda does not allow cycles in the dependency graph.
                 __IMPOSSIBLE__
               Graph.AcyclicSCC m ->
-                case Map.lookup (A.toTopLevelModuleName m) old of
+                case Map.lookup m old of
                   Just m  -> m
                   Nothing -> __IMPOSSIBLE__) $
       G.sccs' dependencyGraph
 
     process ::
-      Map A.ModuleName ModuleInfo -> ModuleToSource -> [ModuleInfo] ->
-      TCM (DecodedModules, ModuleToSource)
+      Map TopLevelModuleName ModuleInfo -> ModuleToSource ->
+      [ModuleInfo] -> TCM (DecodedModules, ModuleToSource)
     process !keep !modFile [] = return
       ( Map.fromList $
-        map (mapFst A.toTopLevelModuleName) $
         Map.toList keep
       , modFile
       )
@@ -323,13 +337,12 @@ setIncludeDirs incs root = do
           depsKept = all (`Map.member` keep) deps
       (keep, modFile) <-
         if not depsKept then return (keep, modFile) else do
-        let n = iModuleName $ miInterface m
-            t = A.toTopLevelModuleName n
+        let t = iTopLevelModuleName $ miInterface m
         oldF            <- findFile' t
         (newF, modFile) <- liftIO $ findFile'' incs t modFile
         return $ case (oldF, newF) of
           (Right f1, Right f2) | f1 == f2 ->
-            (Map.insert n m keep, modFile)
+            (Map.insert t m keep, modFile)
           _ -> (keep, modFile)
       process keep modFile ms
 
