@@ -38,6 +38,7 @@ import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Context
+import Agda.TypeChecking.Monad.Constraints
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
@@ -82,7 +83,7 @@ addConstant q d = do
   tel <- getContextTelescope
   let tel' = replaceEmptyName "r" $ killRange $ case theDef d of
               Constructor{} -> fmap hideOrKeepInstance tel
-              Function{ funProjection = Just Projection{ projProper = Just{}, projIndex = n } } ->
+              Function{ funProjection = Right Projection{ projProper = Just{}, projIndex = n } } ->
                 let fallback = fmap hideOrKeepInstance tel in
                 if n > 0 then fallback else
                 -- if the record value is part of the telescope, its hiding should left unchanged
@@ -109,9 +110,10 @@ addConstant' q info x t def = do
   addConstant q $ defaultDefn info x t lang def
 
 -- | Set termination info of a defined function symbol.
-setTerminates :: QName -> Bool -> TCM ()
+setTerminates :: MonadTCState m => QName -> Bool -> m ()
 setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ \case
     def@Function{} -> def { funTerminates = Just b }
+    def@Record{}   -> def { recTerminates = Just b }
     def -> def
 
 -- | Set CompiledClauses of a defined function symbol.
@@ -135,13 +137,17 @@ modifyFunClauses q f =
 
 -- | Lifts clauses to the top-level and adds them to definition.
 --   Also adjusts the 'funCopatternLHS' field if necessary.
-addClauses :: QName -> [Clause] -> TCM ()
+addClauses :: (MonadConstraint m, MonadTCState m) => QName -> [Clause] -> m ()
 addClauses q cls = do
   tel <- getContextTelescope
   modifySignature $ updateDefinition q $
     updateTheDef (updateFunClauses (++ abstract tel cls))
     . updateDefCopatternLHS (|| isCopatternLHS cls)
 
+  -- Jesper, 2022-10-13: unblock any constraints that were
+  -- waiting for more clauses of this function
+  wakeConstraints' $ wakeIfBlockedOnDef q . constraintUnblocker
+
 mkPragma :: String -> TCM CompilerPragma
 mkPragma s = CompilerPragma <$> getCurrentRange <*> pure s
 
@@ -180,7 +186,7 @@ getUniqueCompilerPragma backend q = do
                        vcat [ "-" <+> pretty (getRange p) | p <- ps ]
 
 setFunctionFlag :: FunctionFlag -> Bool -> QName -> TCM ()
-setFunctionFlag flag val q = modifyGlobalDefinition q $ set (theDefLens . funFlag flag) val
+setFunctionFlag flag val q = modifyGlobalDefinition q $ set (lensTheDef . funFlag flag) val
 
 markStatic :: QName -> TCM ()
 markStatic = setFunctionFlag FunStatic True
@@ -507,14 +513,20 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
             -- This is because we may abstract the record argument later again.
             -- See succeed/ProjectionNotNormalized.agda
             isVar0 t = case unArg t of Var 0 [] -> True; _ -> False
+            proj :: Either ProjectionLikenessMissing Projection
             proj   = case oldDef of
-              Function{funProjection = Just p@Projection{projIndex = n}}
+              Function{funProjection = Right p@Projection{projIndex = n}}
                 | size ts' < n || (size ts' == n && maybe True isVar0 (lastMaybe ts'))
-                -> Just $ p { projIndex = n - size ts'
-                            , projLams  = projLams p `apply` ts'
-                            , projProper= copyName <$> projProper p
-                            }
-              _ -> Nothing
+                -> Right p { projIndex = n - size ts'
+                           , projLams  = projLams p `apply` ts'
+                           , projProper= copyName <$> projProper p
+                           }
+              -- Preserve no-projection-likeness flag if it exists, and
+              -- it's set to @Left _@. For future reference: The match
+              -- on left can't be simplified or it accidentally
+              -- circumvents the guard above.
+              Function{funProjection = Left projl} -> Left projl
+              _ -> Left MaybeProjection
             def =
               case oldDef of
                 Constructor{ conPars = np, conData = d } -> return $
@@ -538,37 +550,38 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         set funMacro  (oldDef ^. funMacro) $
                         set funStatic (oldDef ^. funStatic) $
                         set funInline True $
-                        emptyFunction
-                        { funClauses        = [cl]
-                        , funCompiled       = Just cc
-                        , funSplitTree      = mst
-                        , funMutual         = mutual
-                        , funProjection     = proj
-                        , funTerminates     = Just True
-                        , funExtLam         = extlam
-                        , funWith           = with
+                        FunctionDefn emptyFunctionData
+                        { _funClauses        = [cl]
+                        , _funCompiled       = Just cc
+                        , _funSplitTree      = mst
+                        , _funMutual         = mutual
+                        , _funProjection     = proj
+                        , _funTerminates     = Just True
+                        , _funExtLam         = extlam
+                        , _funWith           = with
                         }
                   reportSDoc "tc.mod.apply" 80 $ ("new def for" <+> pretty x) <?> pretty newDef
                   return newDef
 
-            cl = Clause { clauseLHSRange  = getRange $ defClauses d
-                        , clauseFullRange = getRange $ defClauses d
-                        , clauseTel       = EmptyTel
-                        , namedClausePats = []
-                        , clauseBody      = Just $ dropArgs pars $ case oldDef of
-                            Function{funProjection = Just p} -> projDropParsApply p ProjSystem rel ts'
+            cl = Clause { clauseLHSRange    = getRange $ defClauses d
+                        , clauseFullRange   = getRange $ defClauses d
+                        , clauseTel         = EmptyTel
+                        , namedClausePats   = []
+                        , clauseBody        = Just $ dropArgs pars $ case oldDef of
+                            Function{funProjection = Right p} -> projDropParsApply p ProjSystem rel ts'
                             _ -> Def x $ map Apply ts'
-                        , clauseType      = Just $ defaultArg t
-                        , clauseCatchall  = False
+                        , clauseType        = Just $ defaultArg t
+                        , clauseCatchall    = False
                         , clauseExact       = Just True
                         , clauseRecursive   = Just False -- definitely not recursive
                         , clauseUnreachable = Just False -- definitely not unreachable
-                        , clauseEllipsis  = NoEllipsis
+                        , clauseEllipsis    = NoEllipsis
+                        , clauseWhereModule = Nothing
                         }
               where
                 -- The number of remaining parameters. We need to drop the
                 -- lambdas corresponding to these from the clause body above.
-                pars = max 0 $ maybe 0 (pred . projIndex) proj
+                pars = max 0 $ either (const 0) (pred . projIndex) proj
                 rel  = getRelevance $ defArgInfo d
 
     {- Example
@@ -1241,8 +1254,8 @@ isProjection qn = isProjection_ . theDef <$> getConstInfo qn
 isProjection_ :: Defn -> Maybe Projection
 isProjection_ def =
   case def of
-    Function { funProjection = result } -> result
-    _                                   -> Nothing
+    Function { funProjection = Right result } -> Just result
+    _                                         -> Nothing
 
 -- | Is it the name of a non-irrelevant record projection?
 {-# SPECIALIZE isProjection :: QName -> TCM (Maybe Projection) #-}
