@@ -5,14 +5,13 @@ module Agda.TypeChecking.Monad.State where
 
 import qualified Control.Exception as E
 
-import Control.Monad       (void)
+import Control.Monad       (void, when)
 import Control.Monad.Trans (MonadIO, liftIO)
 
 import Data.Maybe
 
 import qualified Data.Map as Map
 
-import qualified Data.List.NonEmpty as NonEmpty
 import Data.Set (Set)
 import qualified Data.Set as Set
 import qualified Data.HashMap.Strict as HMap
@@ -29,6 +28,8 @@ import Agda.Syntax.Abstract (PatternSynDefn, PatternSynDefns)
 import Agda.Syntax.Abstract.PatternSynonyms
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Internal
+import Agda.Syntax.Position
+import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Warnings
@@ -37,8 +38,10 @@ import Agda.TypeChecking.Monad.Debug (reportSDoc, reportSLn, verboseS)
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.CompiledClause
 
+import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.Hash
 import Agda.Utils.Lens
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.Monad (bracket_)
 import Agda.Utils.Pretty
 import Agda.Utils.Tuple
@@ -326,23 +329,70 @@ updateDefBlocked f def@Defn{ defBlocked = b } = def { defBlocked = f b }
 -- * Top level module
 ---------------------------------------------------------------------------
 
+-- | Tries to convert a raw top-level module name to a top-level
+-- module name.
+
+topLevelModuleName :: RawTopLevelModuleName -> TCM TopLevelModuleName
+topLevelModuleName raw = do
+  hash <- BiMap.lookup raw <$> useR stTopLevelModuleNames
+  case hash of
+    Just hash -> return (unsafeTopLevelModuleName raw hash)
+    Nothing   -> do
+      let hash = hashRawTopLevelModuleName raw
+      when (hash == noModuleNameHash) $ typeError $ GenericError $
+        "The module name " ++ prettyShow raw ++ " has a reserved " ++
+        "hash (you may want to consider renaming the module with " ++
+        "this name)"
+      raw' <- BiMap.invLookup hash <$> useR stTopLevelModuleNames
+      case raw' of
+        Just raw' -> typeError $ GenericError $
+          "Module name hash collision for " ++ prettyShow raw ++
+          " and " ++ prettyShow raw' ++ " (you may want to consider " ++
+          "renaming one of these modules)"
+        Nothing -> do
+          stTopLevelModuleNames `modifyTCLens'`
+            BiMap.insert (killRange raw) hash
+          return (unsafeTopLevelModuleName raw hash)
+
 -- | Set the top-level module. This affects the global module id of freshly
 --   generated names.
 
--- TODO: Is the hash-function collision-free? If not, then the
--- implementation of 'setTopLevelModule' should be changed.
+setTopLevelModule :: TopLevelModuleName -> TCM ()
+setTopLevelModule top = do
+  let hash = moduleNameId top
+  stFreshNameId `setTCLens'` NameId 0 hash
+  stFreshMetaId `setTCLens'`
+    MetaId { metaId     = 0
+           , metaModule = hash
+           }
 
-setTopLevelModule :: C.QName -> TCM ()
-setTopLevelModule x = stFreshNameId `setTCLens` NameId 0 (ModuleNameHash $ hashString $ prettyShow x)
+-- | The name of the current top-level module, if any.
+{-# SPECIALIZE
+    currentTopLevelModule :: TCM (Maybe TopLevelModuleName) #-}
+{-# SPECIALIZE
+    currentTopLevelModule :: ReduceM (Maybe TopLevelModuleName) #-}
+currentTopLevelModule ::
+  (MonadTCEnv m, ReadTCState m) => m (Maybe TopLevelModuleName)
+currentTopLevelModule = do
+  m <- useR stCurrentModule
+  case m of
+    Just (_, top) -> return (Just top)
+    Nothing       -> do
+      p <- asksTC envImportPath
+      return $ case p of
+        top : _ -> Just top
+        []      -> Nothing
 
 -- | Use a different top-level module for a computation. Used when generating
 --   names for imported modules.
-withTopLevelModule :: C.QName -> TCM a -> TCM a
+withTopLevelModule :: TopLevelModuleName -> TCM a -> TCM a
 withTopLevelModule x m = do
-  next <- useTC stFreshNameId
+  nextN <- useTC stFreshNameId
+  nextM <- useTC stFreshMetaId
   setTopLevelModule x
   y <- m
-  stFreshNameId `setTCLens` next
+  stFreshMetaId `setTCLens` nextM
+  stFreshNameId `setTCLens` nextN
   return y
 
 currentModuleNameHash :: ReadTCState m => m ModuleNameHash
@@ -401,7 +451,7 @@ lookupPatternSyn (AmbQ xs) = do
   defs <- traverse lookupSinglePatternSyn xs
   case mergePatternSynDefs defs of
     Just def   -> return def
-    Nothing    -> typeError $ CannotResolveAmbiguousPatternSynonym (NonEmpty.zip xs defs)
+    Nothing    -> typeError $ CannotResolveAmbiguousPatternSynonym $ List1.zip xs defs
 
 lookupSinglePatternSyn :: QName -> TCM PatternSynDefn
 lookupSinglePatternSyn x = do
