@@ -1,10 +1,11 @@
-{-# LANGUAGE TypeFamilies #-}
 
 -- | Tools to manipulate patterns in abstract syntax
 --   in the TCM (type checking monad).
 
 module Agda.TypeChecking.Patterns.Abstract where
 
+import Control.Monad.Except
+
 import qualified Data.List as List
 import Data.Void
 
@@ -19,27 +20,28 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Warnings (raiseWarningsOnUsage)
 
 import Agda.Utils.Impossible
 
 -- | Expand literal integer pattern into suc/zero constructor patterns.
 --
-expandLitPattern :: A.Pattern -> TCM A.Pattern
-expandLitPattern p = case asView p of
-  (xs, A.LitP (LitNat r n))
+expandLitPattern
+  :: (MonadError TCErr m, MonadTCEnv m, ReadTCState m, HasBuiltins m)
+  => A.Pattern -> m A.Pattern
+expandLitPattern = \case
+  A.LitP info (LitNat n)
     | n < 0     -> negLit -- Andreas, issue #2365, negative literals not yet supported.
     | n > 20    -> tooBig
     | otherwise -> do
       Con z _ _ <- primZero
       Con s _ _ <- primSuc
+      let r     = getRange info
       let zero  = A.ConP cinfo (unambiguous $ setRange r $ conName z) []
           suc p = A.ConP cinfo (unambiguous $ setRange r $ conName s) [defaultNamedArg p]
-          info  = A.PatRange r
           cinfo = A.ConPatInfo ConOCon info ConPatEager
-          p'    = foldr ($) zero $ List.genericReplicate n suc
-      return $ foldr (A.AsP info) p' (map A.mkBindName xs)
-  _ -> return p
+      return $ foldr ($) zero $ List.genericReplicate n suc
+  p -> return p
 
   where
     tooBig = typeError $ GenericError $
@@ -74,6 +76,15 @@ expandPatternSynonyms' = postTraverseAPatternM $ \case
   A.PatternSynP i x as -> setCurrentRange i $ do
     (ns, p) <- killRange <$> lookupPatternSyn x
 
+    -- Andreas, 2020-02-11, issue #3734
+    -- If lookup of ambiguous pattern synonym was successful,
+    -- we are justified to complain if one of the definitions
+    -- involved in the resolution is tagged with a warning.
+    -- This is less than optimal, since we do not rule out
+    -- the invalid alternatives by typing, but we cannot do
+    -- better here.
+    mapM_ raiseWarningsOnUsage $ A.unAmbQ x
+
     -- Must expand arguments before instantiating otherwise pattern
     -- synonyms could get into dot patterns (which is __IMPOSSIBLE__).
     p <- expandPatternSynonyms' (vacuous p :: A.Pattern' e)
@@ -94,11 +105,11 @@ class ExpandPatternSynonyms a where
     :: (Traversable f, ExpandPatternSynonyms b, f b ~ a) => a -> TCM a
   expandPatternSynonyms = traverse expandPatternSynonyms
 
-instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Maybe a)            where
-instance ExpandPatternSynonyms a => ExpandPatternSynonyms [a]                  where
-instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Arg a)              where
-instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Named n a)          where
-instance ExpandPatternSynonyms a => ExpandPatternSynonyms (FieldAssignment' a) where
+instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Maybe a)
+instance ExpandPatternSynonyms a => ExpandPatternSynonyms [a]
+instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Arg a)
+instance ExpandPatternSynonyms a => ExpandPatternSynonyms (Named n a)
+instance ExpandPatternSynonyms a => ExpandPatternSynonyms (FieldAssignment' a)
 
 instance ExpandPatternSynonyms (A.Pattern' e) where
   expandPatternSynonyms = expandPatternSynonyms'
