@@ -19,6 +19,7 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad hiding (constructorForm)
+import Agda.TypeChecking.Monad.Builtin (getName',builtinHComp, builtinConId)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
 
@@ -56,7 +57,7 @@ matchedArgs' n vs = map get [0 .. n - 1]
 buildSubstitution :: (DeBruijn a)
                   => Impossible -> Int -> IntMap (Arg a) -> Substitution' a
 buildSubstitution err n vs = foldr cons idS $ matchedArgs' n vs
-  where cons Nothing  = Strengthen err
+  where cons Nothing  = strengthenS' err 1
         cons (Just v) = consS (unArg v)
 
 
@@ -262,12 +263,16 @@ matchPattern p u = case (p, u) of
   -- can be matched on.
   isMatchable' :: HasBuiltins m => m (Blocked Term -> Maybe Term)
   isMatchable' = do
-    mhcomp <- getName' builtinHComp
+    [mhcomp,mconid] <- mapM getName' [builtinHComp, builtinConId]
     return $ \ r ->
       case ignoreBlocking r of
         t@Con{} -> Just t
         t@(Def q [l,a,phi,u,u0]) | Just q == mhcomp
                 -> Just t
+        t@(Def q [l,a,x,y,phi,p]) | Just q == mconid
+                -> Just t
+        -- TODO this covers the transpIx functions, but it's a hack.
+        t@(Def q _) | NotBlocked{blockingStatus = MissingClauses _} <- r -> Just t
         _       -> Nothing
 
   -- DefP hcomp and ConP matching.
