@@ -39,12 +39,14 @@ import qualified Agda.Syntax.Concrete.Pretty as CP
 import qualified Agda.Syntax.Info as A
 import Agda.Syntax.Scope.Base  (AbstractName(..))
 import Agda.Syntax.Scope.Monad (withContextPrecedence)
+import Agda.Syntax.TopLevelModuleName
 
 import Agda.TypeChecking.Coverage.SplitTree
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
 
+import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.Graph.AdjacencyMap.Unidirectional (Graph)
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as Graph
 import Agda.Utils.List1 ( List1, pattern (:|) )
@@ -144,13 +146,7 @@ punctuate d ts
 -- * The PrettyTCM class
 ---------------------------------------------------------------------------
 
-type MonadPretty m =
-  ( MonadReify m
-  , MonadAbsToCon m
-  , IsString (m Doc)
-  , Null (m Doc)
-  , Semigroup (m Doc)
-  )
+type MonadPretty m = MonadAbsToCon m
 
 -- This instance is to satify the constraints of superclass MonadPretty:
 -- | This instance is more specific than a generic instance
@@ -169,6 +165,8 @@ instance {-# OVERLAPPING #-} PrettyTCM String where prettyTCM = text
 instance PrettyTCM Bool        where prettyTCM = pretty
 instance PrettyTCM C.Name      where prettyTCM = pretty
 instance PrettyTCM C.QName     where prettyTCM = pretty
+instance PrettyTCM TopLevelModuleName
+                               where prettyTCM = pretty
 instance PrettyTCM Comparison  where prettyTCM = pretty
 instance PrettyTCM Literal     where prettyTCM = pretty
 instance PrettyTCM Nat         where prettyTCM = pretty
@@ -209,7 +207,7 @@ instance PrettyTCM (Arg A.Expr)       where prettyTCM = prettyA <=< reify
 instance PrettyTCM (NamedArg A.Expr)  where prettyTCM = prettyA <=< reify
 instance PrettyTCM (NamedArg Term)    where prettyTCM = prettyA <=< reify
 instance PrettyTCM (Dom Type)         where prettyTCM = prettyA <=< reify
-instance PrettyTCM (Dom (Name, Type)) where prettyTCM = prettyA <=< reify
+instance PrettyTCM ContextEntry       where prettyTCM = prettyA <=< reify
 
 instance PrettyTCM Permutation where prettyTCM = text . show
 instance PrettyTCM Polarity    where prettyTCM = text . show
@@ -241,7 +239,25 @@ instance PrettyTCM a => PrettyTCM (Judgement a) where
 instance PrettyTCM MetaId where
   prettyTCM x = do
     mn <- getMetaNameSuggestion x
-    pretty $ NamedMeta mn x
+    prettyTCM $ NamedMeta mn x
+
+instance PrettyTCM NamedMeta where
+  prettyTCM (NamedMeta s m) = do
+    current <- currentModuleNameHash
+    prefix  <-
+      if metaModule m == current
+      then return empty
+      else do
+        modName <- BiMap.invLookup (metaModule m) <$>
+                   useR stTopLevelModuleNames
+        return $ case modName of
+          Nothing      -> __IMPOSSIBLE__
+          Just modName -> pretty modName <> text "."
+    let inBetween = case s of
+          ""  -> text "_"
+          "_" -> text "_"
+          s   -> text $ "_" ++ s ++ "_"
+    prefix <> inBetween <> text (show (metaId m))
 
 instance PrettyTCM a => PrettyTCM (Blocked a) where
   prettyTCM (Blocked x a) = ("[" <+> prettyTCM a <+> "]") <> text (P.prettyShow x)
@@ -272,6 +288,9 @@ instance PrettyTCM A.Expr where
 instance PrettyTCM A.TypedBinding where
   prettyTCM = prettyA
 
+instance PrettyTCM A.Pattern where
+  prettyTCM = prettyA
+
 instance PrettyTCM Relevance where
   prettyTCM = pretty
 
@@ -289,6 +308,7 @@ instance PrettyTCM Blocker where
   prettyTCM (UnblockOnAny us) = "any" <> parens (fsep $ punctuate "," $ map prettyTCM $ Set.toList us)
   prettyTCM (UnblockOnMeta m) = prettyTCM m
   prettyTCM (UnblockOnProblem p) = "problem" <+> pretty p
+  prettyTCM (UnblockOnDef q) = "definition" <+> pretty q
 
 instance PrettyTCM CompareAs where
   prettyTCM (AsTermsOf a) = ":" <+> prettyTCMCtx TopCtx a
@@ -412,9 +432,11 @@ instance PrettyTCM NLPSort where
   prettyTCM = \case
     PType l   -> parens $ "Set" <+> prettyTCM l
     PProp l   -> parens $ "Prop" <+> prettyTCM l
+    PSSet l   -> parens $ "SSet" <+> prettyTCM l
     PInf f n  -> prettyTCM (Inf f n :: Sort)
     PSizeUniv -> prettyTCM (SizeUniv :: Sort)
     PLockUniv -> prettyTCM (LockUniv :: Sort)
+    PIntervalUniv -> prettyTCM (IntervalUniv :: Sort)
 
 instance PrettyTCM (Elim' NLPat) where
   prettyTCM (IApply x y v) = prettyTCM v
