@@ -15,6 +15,7 @@ import Agda.Syntax.Translation.AbstractToConcrete
 
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Context
+import Agda.TypeChecking.Monad.Closure
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Pretty
 
@@ -24,6 +25,8 @@ import qualified Agda.Utils.Pretty as P
 
 import Agda.Utils.Impossible
 
+import Agda.Version (docsUrl)
+
 sayWhere :: MonadPretty m => HasRange a => a -> m Doc -> m Doc
 sayWhere x d = applyUnless (null r) (prettyTCM r $$) d
   where r = getRange x
@@ -33,9 +36,9 @@ sayWhen r Nothing   m = sayWhere r m
 sayWhen r (Just cl) m = sayWhere r (m $$ prettyTCM cl)
 
 instance PrettyTCM CallInfo where
-  prettyTCM c = do
-    let call = prettyTCM $ callInfoCall c
-        r    = callInfoRange c
+  prettyTCM (CallInfo callInfoTarget callInfoCall) = do
+    let call = prettyTCM callInfoCall
+        r    = getRange callInfoTarget
     if null $ P.pretty r
       then call
       else call $$ nest 2 ("(at" <+> prettyTCM r) <> ")"
@@ -142,7 +145,8 @@ instance PrettyTCM Call where
 
     CheckWithFunctionType a -> fsep $
       pwords "when checking that the type" ++
-      [prettyTCM a] ++ pwords "of the generated with function is well-formed"
+      [prettyTCM a] ++ pwords "of the generated with function is well-formed" ++
+      [parens $ text $ docsUrl "language/with-abstraction.html#ill-typed-with-abstractions"]
 
     CheckDotPattern e v -> fsep $
       pwords "when checking that the given dot pattern" ++ [prettyA e] ++
@@ -193,6 +197,17 @@ instance PrettyTCM Call where
 
     ModuleContents -> fsep $ pwords "when retrieving the contents of a module"
 
+    CheckIApplyConfluence _ qn fn l r t -> do
+      vcat
+        [ fsep (pwords "when checking that a clause of" ++ [prettyTCM qn] ++ pwords "has the correct boundary.")
+        , ""
+        , "Specifically, the terms"
+        , nest 2 (prettyTCM l)
+        , "and"
+        , nest 2 (prettyTCM r)
+        , fsep (pwords "must be equal, since" ++ [prettyTCM fn] ++ pwords "could reduce to either.")
+        ]
+
     where
     hPretty :: MonadPretty m => Arg (Named_ Expr) -> m Doc
     hPretty a = do
