@@ -9,7 +9,7 @@ import Control.Monad ( guard )
 import Control.Monad.Fail ( MonadFail )
 
 import Data.Char ( toLower )
-import Data.Function
+import Data.Function (on)
 import Data.Maybe
 
 import qualified Data.Set as Set
@@ -82,7 +82,7 @@ prettyWarning = \case
                concatMap termErrFunctions because)
         $$ fwords "Problematic calls:"
         $$ nest 2 (fmap (P.vcat . List.nub) $
-              mapM prettyTCM $ List.sortBy (compare `on` callInfoRange) $
+              mapM prettyTCM $ List.sortOn getRange $
               concatMap termErrCalls because)
 
     UnreachableClauses f pss -> fsep $
@@ -108,6 +108,15 @@ prettyWarning = \case
       [prettyTCM d] ++ pwords "is not strictly positive, because it occurs"
       ++ [prettyTCM ocs]
 
+    UnsupportedIndexedMatch doc -> vcat
+      [ fsep (pwords "This clause uses pattern-matching features that are not yet supported by Cubical Agda,"
+           ++ pwords "the function to which it belongs will not compute when applied to transports."
+             )
+      , ""
+      , "Reason:" <+> pure doc
+      , ""
+      ]
+
     CantGeneralizeOverSorts ms -> vcat
             [ text "Cannot generalize over unsolved sort metas:"
             , nest 2 $ vcat [ prettyTCM x <+> text "at" <+> (pretty =<< getMetaRange x) | x <- ms ]
@@ -137,6 +146,7 @@ prettyWarning = \case
 
     UselessPatternDeclarationForRecord s -> fwords $ unwords
       [ "`pattern' attribute ignored for", s, "record" ]
+      -- the @s@ is a qualifier like "eta" or "coinductive"
 
     UselessPublic -> fwords $ "Keyword `public' is ignored here"
 
@@ -149,14 +159,17 @@ prettyWarning = \case
       pwords "It is pointless for INLINE'd function" ++ [prettyTCM q] ++
       pwords "to have a separate Haskell definition"
 
-    WrongInstanceDeclaration -> fwords "Terms marked as eligible for instance search should end with a name, so `instance' is ignored here."
+    WrongInstanceDeclaration -> fwords $
+      "Instances should be of type {Γ} → C, where C evaluates to a postulated name or the name of " ++
+      "a data or record type, so `instance' is ignored here."
 
     InstanceWithExplicitArg q -> fsep $
       pwords "Instance declarations with explicit arguments are never considered by instance search," ++
       pwords "so making" ++ [prettyTCM q] ++ pwords "into an instance has no effect."
 
     InstanceNoOutputTypeName b -> fsep $
-      pwords "Instance arguments whose type does not end in a named or variable type are never considered by instance search," ++
+      pwords "Instance arguments whose type is not {Γ} → C, where C evaluates to a postulated type, " ++
+      pwords "a parameter type or the name of a data or record type, are never considered by instance search," ++
       pwords "so having an instance argument" ++ [return b] ++ pwords "has no effect."
 
     InstanceArgWithExplicitArg b -> fsep $
@@ -220,6 +233,8 @@ prettyWarning = \case
     SafeFlagNoCoverageCheck -> fsep $
       pwords "Cannot use NON_COVERING pragma with safe flag."
 
+    OptionWarning ow -> pretty ow
+
     ParseWarning pw -> pretty pw
 
     DeprecationWarning old new version -> fsep $
@@ -248,13 +263,9 @@ prettyWarning = \case
 
     LibraryWarning lw -> pretty lw
 
-    InfectiveImport o m -> fsep $
-      pwords "Importing module" ++ [pretty m] ++ pwords "using the" ++
-      [pretty o] ++ pwords "flag from a module which does not."
+    InfectiveImport msg -> return msg
 
-    CoInfectiveImport o m -> fsep $
-      pwords "Importing module" ++ [pretty m] ++ pwords "not using the" ++
-      [pretty o] ++ pwords "flag from a module which does."
+    CoInfectiveImport msg -> return msg
 
     RewriteNonConfluent lhs rhs1 rhs2 err -> fsep
       [ "Local confluence check failed:"
