@@ -2,8 +2,10 @@
 
 module Agda.TypeChecking.Primitive.Base where
 
--- Control.Monad.Fail import is redundant since GHC 8.8.1
-import Control.Monad.Fail (MonadFail)
+import Control.Monad             ( mzero )
+import Control.Monad.Fail        ( MonadFail )
+  -- Control.Monad.Fail import is redundant since GHC 8.8.1
+import Control.Monad.Trans.Maybe ( MaybeT(..), runMaybeT )
 
 import qualified Data.Map as Map
 
@@ -62,19 +64,24 @@ nPi = gpi defaultArgInfo
 
 hPi', nPi' :: (MonadFail m, MonadAddContext m, MonadDebug m)
            => String -> NamesT m Type -> (NamesT m Term -> NamesT m Type) -> NamesT m Type
-hPi' s a b = hPi s a (bind' s b)
-nPi' s a b = nPi s a (bind' s b)
+hPi' s a b = hPi s a (bind' s (\ x -> b x))
+nPi' s a b = nPi s a (bind' s (\ x -> b x))
 
 pPi' :: (MonadAddContext m, HasBuiltins m, MonadDebug m)
      => String -> NamesT m Term -> (NamesT m Term -> NamesT m Type) -> NamesT m Type
 pPi' n phi b = toFinitePi <$> nPi' n (elSSet $ cl isOne <@> phi) b
  where
-   toFinitePi :: Type -> Type
-   toFinitePi (El s (Pi d b)) = El s $ Pi (setRelevance Irrelevant $ d { domFinite = True }) b
-   toFinitePi _               = __IMPOSSIBLE__
-
    isOne = fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinIsOne
 
+-- | Turn a 'Pi' type into one whose domain is annotated finite, i.e.,
+-- one that represents a @Partial@ element rather than an actual
+-- function.
+toFinitePi :: Type -> Type
+toFinitePi (El s (Pi d b)) = El s $ Pi
+  (setRelevance Irrelevant $ d { domIsFinite = True })
+  b
+toFinitePi _ = __IMPOSSIBLE__
+
 el' :: Applicative m => m Term -> m Term -> m Type
 el' l a = El <$> (tmSort <$> l) <*> a
 
@@ -132,22 +139,16 @@ path t = primPath <@> t
 el :: Functor m => m Term -> m Type
 el t = El (mkType 0) <$> t
 
+-- | The universe @Set0@ as a type.
 tset :: Applicative m => m Type
 tset = pure $ sort (mkType 0)
 
+-- | @SizeUniv@ as a sort.
 sSizeUniv :: Sort
--- sSizeUniv = mkType 0
--- Andreas, 2016-04-14 switching off SizeUniv, unfixing issue #1428
 sSizeUniv = SizeUniv
 
+-- | @SizeUniv@ as a type.
 tSizeUniv :: Applicative m => m Type
--- tSizeUniv = tset
--- Andreas, 2016-04-14 switching off SizeUniv, unfixing issue #1428
--- tSizeUniv = return $ El sSizeUniv $ Sort sSizeUniv
--- Andreas, 2015-03-16 Since equality checking for types
--- includes equality checking for sorts, we cannot put
--- SizeUniv in Setω.  (SizeUniv : Setω) == (_0 : suc _0)
--- will first instantiate _0 := Setω, which is wrong.
 tSizeUniv = pure $ sort sSizeUniv
 
 -- | Abbreviation: @argN = 'Arg' 'defaultArgInfo'@.
@@ -182,16 +183,18 @@ lookupPrimitiveFunctionQ q = do
   return (s, PrimImpl t $ pf { primFunName = q })
 
 getBuiltinName :: (HasBuiltins m, MonadReduce m) => String -> m (Maybe QName)
-getBuiltinName b = do
-  caseMaybeM (getBuiltin' b) (return Nothing) (Just <.> getName)
-  where
-  getName v = do
+getBuiltinName b = runMaybeT $ getQNameFromTerm =<< MaybeT (getBuiltin' b)
+
+-- | Convert a name in 'Term' form back to 'QName'.
+--
+getQNameFromTerm :: MonadReduce m => Term -> MaybeT m QName
+getQNameFromTerm v = do
     v <- reduce v
-    case unSpine $ v of
+    case unSpine v of
       Def x _   -> return x
       Con x _ _ -> return $ conName x
-      Lam _ b   -> getName $ unAbs b
-      _ -> __IMPOSSIBLE__
+      Lam _ b   -> getQNameFromTerm $ unAbs b
+      _ -> mzero
 
 isBuiltin :: (HasBuiltins m, MonadReduce m) => QName -> String -> m Bool
 isBuiltin q b = (Just q ==) <$> getBuiltinName b
