@@ -2,14 +2,18 @@
 
 module Agda.TypeChecking.Records where
 
+import Prelude hiding (null)
+
 import Control.Monad
 import Control.Monad.Except
 import Control.Monad.Trans.Maybe
 import Control.Monad.Writer
+import Control.Applicative
 
 import Data.Bifunctor
 import qualified Data.List as List
 import Data.Maybe
+import Data.Set (Set)
 import qualified Data.Set as Set
 import qualified Data.HashMap.Strict as HMap
 
@@ -17,6 +21,7 @@ import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Concrete (FieldAssignment'(..))
 import Agda.Syntax.Abstract.Name
+import Agda.Syntax.Internal.MetaVars (unblockOnAnyMetaIn)
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Position
 import Agda.Syntax.Scope.Base (isNameInScope)
@@ -29,11 +34,14 @@ import Agda.TypeChecking.Reduce.Monad () --instance only
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Warnings
+import {-# SOURCE #-} Agda.TypeChecking.Primitive.Cubical.Base (isCubicalSubtype)
 
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike (eligibleForProjectionLike)
 
 import Agda.Utils.Either
-import Agda.Utils.Functor (for, ($>))
+import Agda.Utils.Function (applyWhen)
+import Agda.Utils.Functor (for, ($>), (<&>))
+import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
@@ -395,10 +403,23 @@ typeElims a self (e : es) = do
     IApply{} -> __IMPOSSIBLE__
 
 -- | Check if a name refers to an eta expandable record.
+--
+-- The answer is no for a record type with an erased constructor
+-- unless the current quantity is \"erased\".
 {-# SPECIALIZE isEtaRecord :: QName -> TCM Bool #-}
 {-# SPECIALIZE isEtaRecord :: QName -> ReduceM Bool #-}
 isEtaRecord :: HasConstInfo m => QName -> m Bool
-isEtaRecord r = maybe False ((YesEta ==) . recEtaEquality) <$> isRecord r
+isEtaRecord r = do
+  isRec <- isRecord r
+  case isRec of
+    Nothing -> return False
+    Just r
+      | recEtaEquality r /= YesEta -> return False
+      | otherwise                  -> do
+        constructorQ <- getQuantity <$>
+                          getConstInfo (conName (recConHead r))
+        currentQ     <- viewTC eQuantity
+        return $ constructorQ `moreQuantity` currentQ
 
 isEtaCon :: HasConstInfo m => QName -> m Bool
 isEtaCon c = getConstInfo' c >>= \case
@@ -457,30 +478,31 @@ unguardedRecord q pat = modifySignature $ updateDefinition q $ updateTheDef $ \c
   r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) $ NoEta pat }
   _ -> __IMPOSSIBLE__
 
+-- | Turn on eta for non-recursive and inductive guarded recursive records,
+--   unless user declared otherwise.
+--   Projections do not preserve guardedness.
+updateEtaForRecord :: QName -> TCM ()
+updateEtaForRecord q = whenM etaEnabled $ do
+
+  -- Do we need to switch on eta for record q?
+  switchEta <- getConstInfo q <&> theDef <&> \case
+    Record{ recInduction = ind, recEtaEquality' = eta }
+      | Inferred NoEta{} <- eta, ind /= Just CoInductive -> True
+      | otherwise -> False
+    _ -> __IMPOSSIBLE__
+
+  when switchEta $ do
+    modifySignature $ updateDefinition q $ over (lensTheDef . lensRecord) $ \ d ->
+      d{ _recEtaEquality' = Inferred YesEta }
+
 -- | Turn on eta for inductive guarded recursive records.
 --   Projections do not preserve guardedness.
 recursiveRecord :: QName -> TCM ()
-recursiveRecord q = do
-  ok <- etaEnabled
-  modifySignature $ updateDefinition q $ updateTheDef $ \case
-    r@Record{ recInduction = ind, recEtaEquality' = eta } ->
-      r { recEtaEquality' = eta' }
-      where
-      eta' | ok, Inferred NoEta{} <- eta, ind /= Just CoInductive = Inferred YesEta
-           | otherwise = eta
-    _ -> __IMPOSSIBLE__
+recursiveRecord = updateEtaForRecord
 
 -- | Turn on eta for non-recursive record, unless user declared otherwise.
 nonRecursiveRecord :: QName -> TCM ()
-nonRecursiveRecord q = whenM etaEnabled $ do
-  -- Do nothing if eta is disabled by option.
-  modifySignature $ updateDefinition q $ updateTheDef $ \case
-    r@Record{ recInduction = ind, recEtaEquality' = Inferred (NoEta _) }
-      | ind /= Just CoInductive ->
-      r { recEtaEquality' = Inferred YesEta }
-    r@Record{} -> r
-    _          -> __IMPOSSIBLE__
-
+nonRecursiveRecord = updateEtaForRecord
 
 -- | Check whether record type is marked as recursive.
 --
@@ -508,7 +530,7 @@ etaExpandBoundVar i = fmap (\ (delta, sigma, tau, _) -> (delta, sigma, tau)) <$>
 --
 --   Postcondition: @Δ = Γ₁, Γ', Γ₂[c Γ']@ and @Γ ⊢ σ : Δ@ and @Δ ⊢ τ : Γ@.
 
-expandRecordVar :: Int -> Telescope -> TCM (Maybe (Telescope, Substitution, Substitution, Telescope))
+expandRecordVar :: PureTCM m => Int -> Telescope -> m (Maybe (Telescope, Substitution, Substitution, Telescope))
 expandRecordVar i gamma0 = do
   -- Get the context with last variable added last in list.
   let gamma = telToList gamma0
@@ -759,59 +781,117 @@ etaContractRecord r c ci args = if all (not . usableModality) args then fallBack
 -- Precondition: The name should refer to a record type, and the
 -- arguments should be the parameters to the type.
 isSingletonRecord :: (PureTCM m, MonadBlock m) => QName -> Args -> m Bool
-isSingletonRecord r ps = isJust <$> isSingletonRecord' False r ps
+isSingletonRecord r ps = isJust <$> isSingletonRecord' False r ps mempty
 
 isSingletonRecordModuloRelevance :: (PureTCM m, MonadBlock m)
                                  => QName -> Args -> m Bool
-isSingletonRecordModuloRelevance r ps = isJust <$> isSingletonRecord' True r ps
+isSingletonRecordModuloRelevance r ps = isJust <$> isSingletonRecord' True r ps mempty
 
 -- | Return the unique (closed) inhabitant if exists.
 --   In case of counting irrelevance in, the returned inhabitant
 --   contains dummy terms.
-isSingletonRecord' :: forall m. (PureTCM m, MonadBlock m)
-                   => Bool -> QName -> Args -> m (Maybe Term)
-isSingletonRecord' regardIrrelevance r ps = do
-  reportSDoc "tc.meta.eta" 30 $ "Is" <+> prettyTCM (Def r $ map Apply ps) <+> "a singleton record type?"
-  isRecord r >>= \case
-    Nothing  -> return Nothing
-    Just def -> do
-      fmap (mkCon (recConHead def) ConOSystem) <$> check (recTel def `apply` ps)
+isSingletonRecord'
+  :: forall m. (PureTCM m, MonadBlock m)
+  => Bool            -- ^ Should disregard irrelevant fields?
+  -> QName           -- ^ Name of record type to check.
+  -> Args            -- ^ Parameters given to the record type.
+  -> Set QName       -- ^ Non-terminating record types we already encountered.
+                     --   These are considered as non-singletons,
+                     --   otherwise we would construct an infinite inhabitant (in an infinite time...).
+  -> m (Maybe Term)  -- ^ The unique inhabitant, if any.  May contain dummy terms in irrelevant positions.
+isSingletonRecord' regardIrrelevance r ps rs = do
+  reportSDoc "tc.meta.eta" 30 $ vcat
+    [ "Is" <+> prettyTCM (Def r $ map Apply ps) <+> "a singleton record type?"
+    , "  already visited:" <+> hsep (map prettyTCM $ Set.toList rs)
+    ]
+  -- Andreas, 2022-03-10, issue #5823
+  -- We need to make sure we are not infinitely unfolding records, so we only expand each once,
+  -- and keep track of the recursive ones we have already seen.
+  if r `Set.member` rs then no else do
+    caseMaybeM (isRecord r) no $ \ def -> do
+      -- We might not know yet whether a record type is recursive because the positivity checker hasn't run yet.
+      -- In this case, we pessimistically consider the record type to be recursive (@True@).
+      let recursive = maybe True (not . null) $ recMutual def
+      -- Andreas, 2022-03-23, issue #5823
+      -- We may pass through terminating record types as often as we want.
+      -- If the termination checker has not run yet, we pessimistically consider the record type
+      -- to be non-terminating.
+      let nonTerminating = maybe True not $ recTerminates def
+      reportSDoc "tc.meta.eta" 30 $ vcat
+        [ hsep [ prettyTCM r, "is recursive      :", prettyTCM recursive      ]
+        , hsep [ prettyTCM r, "is non-terminating:", prettyTCM nonTerminating ]
+        ]
+      fmap (mkCon (recConHead def) ConOSystem) <$> do
+        check (applyWhen (recursive && nonTerminating) (Set.insert r) rs) $ recTel def `apply` ps
   where
-  check :: Telescope -> m (Maybe [Arg Term])
-  check tel = do
+  -- Check that all entries of the constructor telescope are singletons.
+  check :: Set QName -> Telescope -> m (Maybe [Arg Term])
+  check rs tel = do
     reportSDoc "tc.meta.eta" 30 $
       "isSingletonRecord' checking telescope " <+> prettyTCM tel
     case tel of
-      EmptyTel -> return $ Just []
+      EmptyTel -> yes
       ExtendTel dom tel -> ifM (return regardIrrelevance `and2M` isIrrelevantOrPropM dom)
         {-then-}
-          (underAbstraction dom tel $ fmap (fmap (Arg (domInfo dom) __DUMMY_TERM__ :)) . check)
+          (underAbstraction dom tel $ fmap (fmap (Arg (domInfo dom) __DUMMY_TERM__ :)) . check rs)
         {-else-} $ do
-          isSing <- isSingletonType' regardIrrelevance $ unDom dom
-          case isSing of
-            Nothing  -> return Nothing
-            (Just v) -> underAbstraction dom tel $ fmap (fmap (Arg (domInfo dom) v :)) . check
+          caseMaybeM (isSingletonType' regardIrrelevance (unDom dom) rs) no $ \ v -> do
+            underAbstraction dom tel $ fmap (fmap (Arg (domInfo dom) v :)) . check rs
+  no  = return Nothing
+  yes = return $ Just []
 
 -- | Check whether a type has a unique inhabitant and return it.
 --   Can be blocked by a metavar.
 isSingletonType :: (PureTCM m, MonadBlock m) => Type -> m (Maybe Term)
-isSingletonType = isSingletonType' False
+isSingletonType t = isSingletonType' False t mempty
 
 -- | Check whether a type has a unique inhabitant (irrelevant parts ignored).
 --   Can be blocked by a metavar.
 isSingletonTypeModuloRelevance :: (PureTCM m, MonadBlock m) => Type -> m Bool
-isSingletonTypeModuloRelevance t = isJust <$> isSingletonType' True t
+isSingletonTypeModuloRelevance t = isJust <$> isSingletonType' True t mempty
 
-isSingletonType' :: (PureTCM m, MonadBlock m) => Bool -> Type -> m (Maybe Term)
-isSingletonType' regardIrrelevance t = do
+isSingletonType'
+  :: forall m. (PureTCM m, MonadBlock m)
+  => Bool            -- ^ Should disregard irrelevant fields?
+  -> Type            -- ^ Type to check.
+  -> Set QName       -- ^ Non-terminating record typess we already encountered.
+                     --   These are considered as non-singletons,
+                     --   otherwise we would construct an infinite inhabitant (in an infinite time...).
+  -> m (Maybe Term)  -- ^ The unique inhabitant, if any.  May contain dummy terms in irrelevant positions.
+isSingletonType' regardIrrelevance t rs = do
     TelV tel t <- telView t
     t <- abortIfBlocked t
     addContext tel $ do
-      res <- isRecordType t
-      case res of
-        Just (r, ps, def) | YesEta <- recEtaEquality def -> do
-          fmap (abstract tel) <$> isSingletonRecord' regardIrrelevance r ps
-        _ -> return Nothing
+      let
+        -- Easy case: η for records.
+        record :: m (Maybe Term)
+        record = runMaybeT $ do
+          (r, ps, def) <- MaybeT $ isRecordType t
+          guard (YesEta == recEtaEquality def)
+          abstract tel <$> MaybeT (isSingletonRecord' regardIrrelevance r ps rs)
+
+        -- Slightly harder case: η for Sub {level} tA phi elt.
+        -- tA : Type level, phi : I, elt : Partial phi tA.
+        subtype :: m (Maybe Term)
+        subtype = runMaybeT $ do
+          (level, tA, phi, elt) <- MaybeT $ isCubicalSubtype t
+          subin <- MaybeT $ getBuiltinName' builtinSubIn
+          itIsOne <- MaybeT $ getBuiltinName' builtinIsOne
+          phiV <- intervalView phi
+          case phiV of
+            -- If phi = i1, then inS (elt 1=1) is the only inhabitant.
+            IOne -> do
+              let
+                argH = Arg $ setHiding Hidden defaultArgInfo
+                it = elt `apply` [defaultArg (Def itIsOne [])]
+              pure (Def subin [] `apply` [argH level, argH tA, argH phi, defaultArg it])
+            -- Otherwise we're blocked
+            OTerm phi' -> patternViolation (unblockOnAnyMetaIn phi')
+            -- This fails the MaybeT: we're not looking at a
+            -- definitional singleton.
+            _ -> fail ""
+
+      (<|>) <$> record <*> subtype
 
 -- | Checks whether the given term (of the given type) is beta-eta-equivalent
 --   to a variable. Returns just the de Bruijn-index of the variable if it is,
