@@ -1,6 +1,29 @@
 {-# LANGUAGE NondecreasingIndentation #-}
 
-module Agda.TypeChecking.Reduce where
+module Agda.TypeChecking.Reduce
+ -- Meta instantiation
+ ( Instantiate, instantiate', instantiate, instantiateWhen
+ -- Recursive meta instantiation
+ , InstantiateFull, instantiateFull', instantiateFull
+ , instantiateFullExceptForDefinitions
+ -- Check for meta (no reduction)
+ , IsMeta, isMeta
+ -- Reduction and blocking
+ , Reduce, reduce', reduceB', reduce, reduceB, reduceWithBlocker, reduceIApply'
+ , reduceDefCopy, reduceDefCopyTCM
+ , reduceHead
+ , slowReduceTerm
+ , unfoldCorecursion, unfoldCorecursionE
+ , unfoldDefinitionE, unfoldDefinitionStep
+ , unfoldInlined
+ , appDef', appDefE'
+ , abortIfBlocked, ifBlocked, isBlocked
+ -- Simplification
+ , Simplify, simplify, simplifyBlocked'
+ -- Normalization
+ , Normalise, normalise', normalise
+ , slowNormaliseArgs
+ ) where
 
 import Control.Monad ( (>=>), void )
 
@@ -56,6 +79,18 @@ instantiate = liftReduce . instantiate'
 instantiateFull :: (InstantiateFull a, MonadReduce m) => a -> m a
 instantiateFull = liftReduce . instantiateFull'
 
+-- | A variant of 'instantiateFull' that only instantiates those
+-- meta-variables that satisfy the predicate.
+
+instantiateWhen ::
+  (InstantiateFull a, MonadReduce m) =>
+  (MetaId -> ReduceM Bool) ->
+  a -> m a
+instantiateWhen p =
+  liftReduce .
+  localR (\env -> env { redPred = Just p }) .
+  instantiateFull'
+
 reduce :: (Reduce a, MonadReduce m) => a -> m a
 reduce = liftReduce . reduce'
 
@@ -80,10 +115,11 @@ withReduced a cont = ifBlocked a (\b a' -> addOrUnblocker b $ cont a') (\_ a' ->
 normalise :: (Normalise a, MonadReduce m) => a -> m a
 normalise = liftReduce . normalise'
 
--- | Normalise the given term but also preserve blocking tags
---   TODO: implement a more efficient version of this.
-normaliseB :: (MonadReduce m, Reduce t, Normalise t) => t -> m (Blocked t)
-normaliseB = normalise >=> reduceB
+-- UNUSED
+-- -- | Normalise the given term but also preserve blocking tags
+-- --   TODO: implement a more efficient version of this.
+-- normaliseB :: (MonadReduce m, Reduce t, Normalise t) => t -> m (Blocked t)
+-- normaliseB = normalise >=> reduceB
 
 simplify :: (Simplify a, MonadReduce m) => a -> m a
 simplify = liftReduce . simplify'
@@ -91,9 +127,9 @@ simplify = liftReduce . simplify'
 -- | Meaning no metas left in the instantiation.
 isFullyInstantiatedMeta :: MetaId -> TCM Bool
 isFullyInstantiatedMeta m = do
-  mv <- lookupMeta m
-  case mvInstantiation mv of
-    InstV _tel v -> noMetas <$> instantiateFull v
+  inst <- lookupMetaInstantiation m
+  case inst of
+    InstV inst -> noMetas <$> instantiateFull (instBody inst)
     _ -> return False
 
 -- | Blocking on all blockers.
@@ -139,34 +175,64 @@ instance (Instantiate a, Instantiate b) => Instantiate (a,b) where
 instance (Instantiate a, Instantiate b,Instantiate c) => Instantiate (a,b,c) where
     instantiate' (x,y,z) = (,,) <$> instantiate' x <*> instantiate' y <*> instantiate' z
 
+-- | Run the second computation if the 'redPred' predicate holds for
+-- the given meta-variable (or if the predicate is not defined), and
+-- otherwise the first computation.
+
+ifPredicateDoesNotHoldFor ::
+  MetaId -> ReduceM a -> ReduceM a -> ReduceM a
+ifPredicateDoesNotHoldFor m doesNotHold holds = do
+  pred <- redPred <$> askR
+  case pred of
+    Nothing -> holds
+    Just p  -> ifM (p m) holds doesNotHold
+
 instance Instantiate Term where
-  instantiate' t@(MetaV x es) = do
+  instantiate' t@(MetaV x es) = ifPredicateDoesNotHoldFor x (return t) $ do
     blocking <- view stInstantiateBlocking <$> getTCState
 
-    mv <- lookupMeta x
-    let mi = mvInstantiation mv
+    m <- lookupMeta x
+    case m of
+      Just (Left rmv) -> cont (rmvInstantiation rmv)
+
+      Just (Right mv) -> case mvInstantiation mv of
+         InstV inst -> cont inst
+
+         _ | Just m' <- mvTwin mv, blocking ->
+           instantiate' (MetaV m' es)
+
+         Open -> return t
+
+         OpenInstance -> return t
+
+         BlockedConst u
+           | blocking  -> instantiate' . unBrave $
+                          BraveTerm u `applyE` es
+           | otherwise -> return t
+
+         PostponedTypeCheckingProblem _ -> return t
+
+      Nothing -> __IMPOSSIBLE_VERBOSE__
+                   ("Meta-variable not found: " ++ prettyShow x)
+    where
+    cont i = instantiate' inst
+      where
+      -- A slight complication here is that the meta might be underapplied,
+      -- in which case we have to build the lambda abstraction before
+      -- applying the substitution, or overapplied in which case we need to
+      -- fall back to applyE.
+      (es1, es2) = splitAt (length (instTel i)) es
+      vs1 = reverse $ map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es1
+      rho = vs1 ++# wkS (length vs1) idS
+            -- really should be .. ++# emptyS but using wkS makes it reduce to idS
+            -- when applicable
+      -- specification:
+      -- inst == foldr mkLam (instBody i) (instTel i) `applyE` es
+      inst =
+        applySubst rho
+          (foldr mkLam (instBody i) $ drop (length es1) (instTel i))
+          `applyE` es2
 
-    case mi of
-      InstV tel v -> instantiate' inst
-        where
-          -- A slight complication here is that the meta might be underapplied,
-          -- in which case we have to build the lambda abstraction before
-          -- applying the substitution, or overapplied in which case we need to
-          -- fall back to applyE.
-          (es1, es2) = splitAt (length tel) es
-          vs1 = reverse $ map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es1
-          rho = vs1 ++# wkS (length vs1) idS
-                -- really should be .. ++# emptyS but using wkS makes it reduce to idS
-                -- when applicable
-          -- specification: inst == foldr mkLam v tel `applyE` es
-          inst = applySubst rho (foldr mkLam v $ drop (length es1) tel) `applyE` es2
-      _ | Just m' <- mvTwin mv, blocking -> do
-            instantiate' (MetaV m' es)
-      Open                             -> return t
-      OpenInstance                     -> return t
-      BlockedConst u | blocking  -> instantiate' . unBrave $ BraveTerm u `applyE` es
-                     | otherwise -> return t
-      PostponedTypeCheckingProblem _ -> return t
   instantiate' (Level l) = levelTm <$> instantiate' l
   instantiate' (Sort s) = Sort <$> instantiate' s
   instantiate' t = return t
@@ -192,6 +258,7 @@ instance Instantiate Blocker where
   instantiate' b@(UnblockOnMeta x) =
     ifM (isInstantiatedMeta x) (return alwaysUnblock) (return b)
   instantiate' b@UnblockOnProblem{} = return b
+  instantiate' b@UnblockOnDef{} = return b
 
 instance Instantiate Sort where
   instantiate' = \case
@@ -203,7 +270,7 @@ instance Instantiate Sort where
     s -> return s
 
 instance (Instantiate t, Instantiate e) => Instantiate (Dom' t e) where
-    instantiate' (Dom i fin n tac x) = Dom i fin n <$> instantiate' tac <*> instantiate' x
+    instantiate' (Dom i n b tac x) = Dom i n b <$> instantiate' tac <*> instantiate' x
 
 instance Instantiate a => Instantiate (Closure a) where
     instantiate' cl = do
@@ -231,9 +298,10 @@ instance Instantiate Constraint where
   instantiate' (CheckLockedVars a b c d) =
     CheckLockedVars <$> instantiate' a <*> instantiate' b <*> instantiate' c <*> instantiate' d
   instantiate' (UnquoteTactic t h g) = UnquoteTactic <$> instantiate' t <*> instantiate' h <*> instantiate' g
+  instantiate' (CheckDataSort q s)  = CheckDataSort q <$> instantiate' s
   instantiate' c@CheckMetaInst{}    = return c
   instantiate' (CheckType t)        = CheckType <$> instantiate' t
-  instantiate' (UsableAtModality mod t) = UsableAtModality mod <$> instantiate' t
+  instantiate' (UsableAtModality cc ms mod t) = flip (UsableAtModality cc) mod <$> instantiate' ms <*> instantiate' t
 
 instance Instantiate CompareAs where
   instantiate' (AsTermsOf a) = AsTermsOf <$> instantiate' a
@@ -332,27 +400,42 @@ instance Reduce Type where
     reduceB' (El s t) = workOnTypes $ fmap (El s) <$> reduceB' t
 
 instance Reduce Sort where
-    reduce' s = do
+    reduceB' s = do
       s <- instantiate' s
+      let done | MetaS x _ <- s = return $ blocked x s
+               | otherwise      = return $ notBlocked s
       case s of
-        PiSort a s1 s2 -> do
-          (s1' , s2') <- reduce' (s1 , s2)
-          maybe (return $ PiSort a s1' s2') reduce' $ piSort' a s1' s2'
-        FunSort s1 s2 -> do
-          (s1' , s2') <- reduce (s1 , s2)
-          maybe (return $ FunSort s1' s2') reduce' $ funSort' s1' s2'
-        UnivSort s' -> do
-          s' <- reduce' s'
-          caseMaybe (univSort' s') (return $ UnivSort s') reduce'
-        Prop s'    -> Prop <$> reduce' s'
-        Type s'    -> Type <$> reduce' s'
-        Inf f n    -> return $ Inf f n
-        SSet s'    -> SSet <$> reduce' s'
-        SizeUniv   -> return SizeUniv
-        LockUniv   -> return LockUniv
-        MetaS x es -> return s
-        DefS d es  -> return s -- postulated sorts do not reduce
-        DummyS{}   -> return s
+        PiSort a s1 s2 -> reduceB' (s1 , s2) >>= \case
+          Blocked b (s1',s2') -> return $ Blocked b $ PiSort a s1' s2'
+          NotBlocked _ (s1',s2') -> do
+            -- Jesper, 2022-10-12: do instantiateFull here because
+            -- `piSort'` does checking of free variables, and if we
+            -- don't instantiate we might end up blocking on a solved
+            -- metavariable.
+            s2' <- instantiateFull s2'
+            case piSort' a s1' s2' of
+              Left b -> return $ Blocked b $ PiSort a s1' s2'
+              Right s -> reduceB' s
+        FunSort s1 s2 -> reduceB' (s1 , s2) >>= \case
+          Blocked b (s1',s2') -> return $ Blocked b $ FunSort s1' s2'
+          NotBlocked _ (s1',s2') -> case funSort' s1' s2' of
+            Left b -> return $ Blocked b $ FunSort s1' s2'
+            Right s -> reduceB' s
+        UnivSort s1 -> reduceB' s1 >>= \case
+          Blocked b s1' -> return $ Blocked b $ UnivSort s1'
+          NotBlocked _ s1' -> case univSort' s1' of
+            Left b -> return $ Blocked b $ UnivSort s1'
+            Right s -> reduceB' s
+        Prop l     -> notBlocked . Prop <$> reduce l
+        Type l     -> notBlocked . Type <$> reduce l
+        Inf f n    -> done
+        SSet l     -> notBlocked . SSet <$> reduce l
+        SizeUniv   -> done
+        LockUniv   -> done
+        IntervalUniv -> done
+        MetaS x es -> done
+        DefS d es  -> done -- postulated sorts do not reduce
+        DummyS{}   -> done
 
 instance Reduce Elim where
   reduce' (Apply v) = Apply <$> reduce' v
@@ -367,13 +450,17 @@ instance Reduce PlusLevel where
   reduceB' (Plus n l) = fmap (Plus n) <$> reduceB' l
 
 instance (Subst a, Reduce a) => Reduce (Abs a) where
-  reduce' b@(Abs x _) = Abs x <$> underAbstraction_ b reduce'
-  reduce' (NoAbs x v) = NoAbs x <$> reduce' v
+  reduceB' b@(Abs x _) = fmap (Abs x) <$> underAbstraction_ b reduceB'
+  reduceB' (NoAbs x v) = fmap (NoAbs x) <$> reduceB' v
 
 -- Lists are never blocked
 instance Reduce t => Reduce [t] where
     reduce' = traverse reduce'
 
+-- Maybes are never blocked
+instance Reduce t => Reduce (Maybe t) where
+    reduce' = traverse reduce'
+
 instance Reduce t => Reduce (Arg t) where
     reduce' a = case getRelevance a of
       Irrelevant -> return a             -- Don't reduce' irr. args!?
@@ -448,7 +535,7 @@ maybeFastReduceTerm v = do
       MetaV x _ -> ifM (isOpen x) (return $ blocked x v) (maybeFast v)
       _         -> maybeFast v
   where
-    isOpen x = isOpenMeta . mvInstantiation <$> lookupMeta x
+    isOpen x = isOpenMeta <$> lookupMetaInstantiation x
     maybeFast v = ifM shouldTryFastReduce (fastReduce v) (slowReduceTerm v)
 
 slowReduceTerm :: Term -> ReduceM (Blocked Term)
@@ -470,7 +557,7 @@ slowReduceTerm v = do
           v <- flip reduceIApply es
                  $ unfoldDefinitionE False reduceB' (Con c ci []) (conName c) es
           traverse reduceNat v
-      Sort s   -> fmap Sort <$> reduceB' s
+      Sort s   -> done
       Level l  -> ifM (SmallSet.member LevelReductions <$> asksTC envAllowedReductions)
                     {- then -} (fmap levelTm <$> reduceB' l)
                     {- else -} done
@@ -575,7 +662,7 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
         then reducePrimitive x v0 f es pf dontUnfold
                              cls (defCompiled info) rewr
         else noReduction $ notBlocked v
-    PrimitiveSort{ primSort = s } -> yesReduction NoSimplification $ Sort s `applyE` es
+    PrimitiveSort{ primSortSort = s } -> yesReduction NoSimplification $ Sort s `applyE` es
 
     _  -> do
       if or
@@ -753,7 +840,7 @@ appDef_ f v0 cls mcc rewr args = appDefE_ f v0 cls mcc rewr $ map (fmap Apply) a
 appDefE_ :: QName -> Term -> [Clause] -> Maybe CompiledClauses -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
 appDefE_ f v0 cls mcc rewr args =
   localTC (\ e -> e { envAppDef = Just f }) $
-  maybe (appDefE' v0 cls rewr args)
+  maybe (appDefE'' v0 cls rewr args)
         (\cc -> appDefE v0 cc rewr args) mcc
 
 
@@ -771,11 +858,17 @@ appDefE v cc rewr es = do
     NoReduction es'      -> rewrite (void es') (applyE v) rewr (ignoreBlocking es')
 
 -- | Apply a defined function to it's arguments, using the original clauses.
-appDef' :: Term -> [Clause] -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
-appDef' v cls rewr args = appDefE' v cls rewr $ map (fmap Apply) args
+appDef' :: QName -> Term -> [Clause] -> RewriteRules -> MaybeReducedArgs -> ReduceM (Reduced (Blocked Term) Term)
+appDef' f v cls rewr args = appDefE' f v cls rewr $ map (fmap Apply) args
 
-appDefE' :: Term -> [Clause] -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
-appDefE' v cls rewr es = traceSDoc "tc.reduce" 90 ("appDefE' v = " <+> pretty v) $ do
+appDefE' :: QName -> Term -> [Clause] -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
+appDefE' f v cls rewr es =
+  localTC (\ e -> e { envAppDef = Just f }) $
+  appDefE'' v cls rewr es
+
+-- | Expects @'envAppDef' = Just f@ in 'TCEnv' to be able to report @'MissingClauses' f@.
+appDefE'' :: Term -> [Clause] -> RewriteRules -> MaybeReducedElims -> ReduceM (Reduced (Blocked Term) Term)
+appDefE'' v cls rewr es = traceSDoc "tc.reduce" 90 ("appDefE' v = " <+> pretty v) $ do
   goCls cls $ map ignoreReduced es
   where
     goCls :: [Clause] -> [Elim] -> ReduceM (Reduced (Blocked Term) Term)
@@ -787,7 +880,9 @@ appDefE' v cls rewr es = traceSDoc "tc.reduce" 90 ("appDefE' v = " <+> pretty v)
         -- the remaining clauses (see Issue 907).
         -- Andrea(s), 2014-12-05:  We return 'MissingClauses' here, since this
         -- is the most conservative reason.
-        [] -> rewrite (NotBlocked MissingClauses ()) (applyE v) rewr es
+        [] -> do
+          f <- fromMaybe __IMPOSSIBLE__ <$> asksTC envAppDef
+          rewrite (NotBlocked (MissingClauses f) ()) (applyE v) rewr es
         cl : cls -> do
           let pats = namedClausePats cl
               body = clauseBody cl
@@ -840,9 +935,10 @@ instance Reduce Constraint where
   reduce' (UnquoteTactic t h g) = UnquoteTactic <$> reduce' t <*> reduce' h <*> reduce' g
   reduce' (CheckLockedVars a b c d) =
     CheckLockedVars <$> reduce' a <*> reduce' b <*> reduce' c <*> reduce' d
+  reduce' (CheckDataSort q s)   = CheckDataSort q <$> reduce' s
   reduce' c@CheckMetaInst{}     = return c
   reduce' (CheckType t)         = CheckType <$> reduce' t
-  reduce' (UsableAtModality mod t) = UsableAtModality mod <$> reduce' t
+  reduce' (UsableAtModality cc ms mod t) = flip (UsableAtModality cc) mod <$> reduce' ms <*> reduce' t
 
 instance Reduce CompareAs where
   reduce' (AsTermsOf a) = AsTermsOf <$> reduce' a
@@ -919,9 +1015,9 @@ instance Simplify Term where
       Def f vs   -> iapp vs $ do
         let keepGoing simp v = return (simp, notBlocked v)
         (simpl, v) <- unfoldDefinition' False keepGoing (Def f []) f vs
-        traceSDoc "tc.simplify'" 90 (
-          text ("simplify': unfolding definition returns " ++ show simpl)
-            <+> pretty (ignoreBlocking v)) $ do
+        when (simpl == YesSimplification) $
+          reportSDoc "tc.simplify'" 90 $
+            pretty f <+> text ("simplify': unfolding definition returns " ++ show simpl) <+> pretty (ignoreBlocking v)
         case simpl of
           YesSimplification -> simplifyBlocked' v -- Dangerous, but if @simpl@ then @v /= Def f vs@
           NoSimplification  -> Def f <$> simplify' vs
@@ -955,6 +1051,7 @@ instance Simplify Sort where
         SSet s     -> SSet <$> simplify' s
         SizeUniv   -> return s
         LockUniv   -> return s
+        IntervalUniv -> return s
         MetaS x es -> MetaS x <$> simplify' es
         DefS d es  -> DefS d <$> simplify' es
         DummyS{}   -> return s
@@ -1005,9 +1102,10 @@ instance Simplify Constraint where
   simplify' (UnquoteTactic t h g) = UnquoteTactic <$> simplify' t <*> simplify' h <*> simplify' g
   simplify' (CheckLockedVars a b c d) =
     CheckLockedVars <$> simplify' a <*> simplify' b <*> simplify' c <*> simplify' d
+  simplify' (CheckDataSort q s)   = CheckDataSort q <$> simplify' s
   simplify' c@CheckMetaInst{}     = return c
   simplify' (CheckType t)         = CheckType <$> simplify' t
-  simplify' (UsableAtModality mod t) = UsableAtModality mod <$> simplify' t
+  simplify' (UsableAtModality cc ms mod t) = flip (UsableAtModality cc) mod <$> simplify' ms <*> simplify' t
 
 instance Simplify CompareAs where
   simplify' (AsTermsOf a) = AsTermsOf <$> simplify' a
@@ -1104,6 +1202,7 @@ instance Normalise Sort where
         SSet s     -> SSet <$> normalise' s
         SizeUniv   -> return SizeUniv
         LockUniv   -> return LockUniv
+        IntervalUniv -> return IntervalUniv
         MetaS x es -> return s
         DefS d es  -> return s
         DummyS{}   -> return s
@@ -1185,9 +1284,10 @@ instance Normalise Constraint where
   normalise' (UnquoteTactic t h g) = UnquoteTactic <$> normalise' t <*> normalise' h <*> normalise' g
   normalise' (CheckLockedVars a b c d) =
     CheckLockedVars <$> normalise' a <*> normalise' b <*> normalise' c <*> normalise' d
+  normalise' (CheckDataSort q s)   = CheckDataSort q <$> normalise' s
   normalise' c@CheckMetaInst{}     = return c
   normalise' (CheckType t)         = CheckType <$> normalise' t
-  normalise' (UsableAtModality mod t) = UsableAtModality mod <$> normalise' t
+  normalise' (UsableAtModality cc ms mod t) = flip (UsableAtModality cc) mod <$> normalise' ms <*> normalise' t
 
 instance Normalise CompareAs where
   normalise' (AsTermsOf a) = AsTermsOf <$> normalise' a
@@ -1311,6 +1411,7 @@ instance InstantiateFull Sort where
             Inf _ _    -> return s
             SizeUniv   -> return s
             LockUniv   -> return s
+            IntervalUniv -> return s
             MetaS x es -> MetaS x <$> instantiateFull' es
             DefS d es  -> DefS d <$> instantiateFull' es
             DummyS{}   -> return s
@@ -1346,13 +1447,13 @@ instance InstantiateFull PlusLevel where
 instance InstantiateFull Substitution where
   instantiateFull' sigma =
     case sigma of
-      IdS                  -> return IdS
-      EmptyS err           -> return $ EmptyS err
-      Wk   n sigma         -> Wk   n         <$> instantiateFull' sigma
-      Lift n sigma         -> Lift n         <$> instantiateFull' sigma
-      Strengthen bot sigma -> Strengthen bot <$> instantiateFull' sigma
-      t :# sigma           -> consS <$> instantiateFull' t
-                                    <*> instantiateFull' sigma
+      IdS                    -> return IdS
+      EmptyS err             -> return $ EmptyS err
+      Wk   n sigma           -> Wk   n           <$> instantiateFull' sigma
+      Lift n sigma           -> Lift n           <$> instantiateFull' sigma
+      Strengthen bot n sigma -> Strengthen bot n <$> instantiateFull' sigma
+      t :# sigma             -> consS <$> instantiateFull' t
+                                      <*> instantiateFull' sigma
 
 instance InstantiateFull ConPatternInfo where
     instantiateFull' i = instantiateFull' (conPType i) <&> \ t -> i { conPType = t }
@@ -1371,7 +1472,7 @@ instance (Subst a, InstantiateFull a) => InstantiateFull (Abs a) where
     instantiateFull' (NoAbs x a) = NoAbs x <$> instantiateFull' a
 
 instance (InstantiateFull t, InstantiateFull e) => InstantiateFull (Dom' t e) where
-    instantiateFull' (Dom i fin n tac x) = Dom i fin n <$> instantiateFull' tac <*> instantiateFull' x
+    instantiateFull' (Dom i n b tac x) = Dom i n b <$> instantiateFull' tac <*> instantiateFull' x
 
 -- Andreas, 2021-09-13, issue #5544, need to traverse @checkpoints@ map
 instance InstantiateFull t => InstantiateFull (Open t) where
@@ -1418,9 +1519,10 @@ instance InstantiateFull Constraint where
     UnquoteTactic t g h -> UnquoteTactic <$> instantiateFull' t <*> instantiateFull' g <*> instantiateFull' h
     CheckLockedVars a b c d ->
       CheckLockedVars <$> instantiateFull' a <*> instantiateFull' b <*> instantiateFull' c <*> instantiateFull' d
+    CheckDataSort q s   -> CheckDataSort q <$> instantiateFull' s
     c@CheckMetaInst{}   -> return c
     CheckType t         -> CheckType <$> instantiateFull' t
-    UsableAtModality mod t -> UsableAtModality mod <$> instantiateFull' t
+    UsableAtModality cc ms mod t -> flip (UsableAtModality cc) mod <$> instantiateFull' ms <*> instantiateFull' t
 
 instance InstantiateFull CompareAs where
   instantiateFull' (AsTermsOf a) = AsTermsOf <$> instantiateFull' a
@@ -1459,9 +1561,11 @@ instance InstantiateFull NLPType where
 instance InstantiateFull NLPSort where
   instantiateFull' (PType x) = PType <$> instantiateFull' x
   instantiateFull' (PProp x) = PProp <$> instantiateFull' x
+  instantiateFull' (PSSet x) = PSSet <$> instantiateFull' x
   instantiateFull' (PInf f n) = return $ PInf f n
   instantiateFull' PSizeUniv = return PSizeUniv
   instantiateFull' PLockUniv = return PLockUniv
+  instantiateFull' PIntervalUniv = return PIntervalUniv
 
 instance InstantiateFull RewriteRule where
   instantiateFull' (RewriteRule q gamma f ps rhs t c) =
@@ -1535,7 +1639,7 @@ instance InstantiateFull CompiledClauses where
   instantiateFull' (Case n bs) = Case n <$> instantiateFull' bs
 
 instance InstantiateFull Clause where
-    instantiateFull' (Clause rl rf tel ps b t catchall exact recursive unreachable ell) =
+    instantiateFull' (Clause rl rf tel ps b t catchall exact recursive unreachable ell wm) =
        Clause rl rf <$> instantiateFull' tel
        <*> instantiateFull' ps
        <*> instantiateFull' b
@@ -1545,6 +1649,23 @@ instance InstantiateFull Clause where
        <*> return recursive
        <*> return unreachable
        <*> return ell
+       <*> return wm
+
+instance InstantiateFull Instantiation where
+  instantiateFull' (Instantiation a b) =
+    Instantiation a <$> instantiateFull' b
+
+instance InstantiateFull (Judgement MetaId) where
+  instantiateFull' (HasType a b c) =
+    HasType a b <$> instantiateFull' c
+  instantiateFull' (IsSort a b) =
+    IsSort a <$> instantiateFull' b
+
+instance InstantiateFull RemoteMetaVariable where
+  instantiateFull' (RemoteMetaVariable a b c) = RemoteMetaVariable
+    <$> instantiateFull' a
+    <*> return b
+    <*> instantiateFull' c
 
 instance InstantiateFull Interface where
   instantiateFull' i = do
@@ -1556,16 +1677,17 @@ instance InstantiateFull Interface where
 
 instantiateFullExceptForDefinitions' :: Interface -> ReduceM Interface
 instantiateFullExceptForDefinitions'
-  (Interface h s ft ms mod scope inside sig display userwarn importwarn
-     b foreignCode highlighting libPragmas filePragmas usedOpts patsyns
-     warnings partialdefs) =
-  Interface h s ft ms mod scope inside
+  (Interface h s ft ms mod tlmod scope inside sig metas display userwarn
+     importwarn b foreignCode highlighting libPragmas filePragmas
+     usedOpts patsyns warnings partialdefs) =
+  Interface h s ft ms mod tlmod scope inside
     <$> ((\s r -> Sig { _sigSections     = s
                       , _sigDefinitions  = sig ^. sigDefinitions
                       , _sigRewriteRules = r
                       })
          <$> instantiateFull' (sig ^. sigSections)
          <*> instantiateFull' (sig ^. sigRewriteRules))
+    <*> instantiateFull' metas
     <*> instantiateFull' display
     <*> return userwarn
     <*> return importwarn
@@ -1589,6 +1711,7 @@ instantiateFullExceptForDefinitions =
 instance InstantiateFull a => InstantiateFull (Builtin a) where
     instantiateFull' (Builtin t) = Builtin <$> instantiateFull' t
     instantiateFull' (Prim x)   = Prim <$> instantiateFull' x
+    instantiateFull' b@(BuiltinRewriteRelations xs) = pure b
 
 instance InstantiateFull Candidate where
   instantiateFull' (Candidate q u t ov) =
