@@ -18,18 +18,20 @@ import Data.Maybe
 import System.IO.Unsafe
 
 import Agda.Syntax.Internal
-import Agda.TypeChecking.Monad hiding
-  ( enterClosure, isInstantiatedMeta, verboseS, typeOfConst, lookupMeta, lookupMeta', constructorForm )
+import Agda.TypeChecking.Monad hiding (enterClosure, constructorForm)
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Lens
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Pretty () --instance only
 
 
 instance HasBuiltins ReduceM where
-  getBuiltinThing b = liftM2 mplus (Map.lookup b <$> useR stLocalBuiltins)
-                                   (Map.lookup b <$> useR stImportedBuiltins)
+  getBuiltinThing b =
+    liftM2 (unionMaybeWith unionBuiltin)
+      (Map.lookup b <$> useR stLocalBuiltins)
+      (Map.lookup b <$> useR stImportedBuiltins)
 
 constructorForm :: HasBuiltins m => Term -> m Term
 constructorForm v = do
@@ -72,13 +74,13 @@ instance MonadAddContext ReduceM where
 instance MonadDebug ReduceM where
 
   traceDebugMessage k n s cont = do
-    ReduceEnv env st <- askR
+    ReduceEnv env st _ <- askR
     unsafePerformIO $ do
       _ <- runTCM env st $ displayDebugMessage k n s
       return $ cont
 
   formatDebugMessage k n d = do
-    ReduceEnv env st <- askR
+    ReduceEnv env st _ <- askR
     unsafePerformIO $ do
       (s , _) <- runTCM env st $ formatDebugMessage k n d
       return $ return s
@@ -86,14 +88,15 @@ instance MonadDebug ReduceM where
   verboseBracket k n s = applyWhenVerboseS k n $
     bracket_ (openVerboseBracket k n s) (const $ closeVerboseBracket k n)
 
-  getVerbosity     = defaultGetVerbosity
-  isDebugPrinting  = defaultIsDebugPrinting
-  nowDebugPrinting = defaultNowDebugPrinting
+  getVerbosity      = defaultGetVerbosity
+  getProfileOptions = defaultGetProfileOptions
+  isDebugPrinting   = defaultIsDebugPrinting
+  nowDebugPrinting  = defaultNowDebugPrinting
 
 instance HasConstInfo ReduceM where
   getRewriteRulesFor = defaultGetRewriteRulesFor
   getConstInfo' q = do
-    ReduceEnv env st <- askR
+    ReduceEnv env st _ <- askR
     defaultGetConstInfo st env q
 
 instance PureTCM ReduceM where
