@@ -48,9 +48,12 @@ import Prelude hiding (null)
 
 import Control.Monad
 
+import Data.Foldable (toList)
 import Data.IntSet (IntSet)
 import qualified Data.IntSet as IntSet
 import qualified Data.List as List
+import Data.Set (Set)
+import qualified Data.Set as Set
 
 import Agda.Interaction.Options
 
@@ -83,6 +86,7 @@ import Agda.Utils.Size
 import qualified Agda.Utils.SmallSet as SmallSet
 
 import Agda.Utils.Impossible
+import Agda.Utils.Either
 
 requireOptionRewriting :: TCM ()
 requireOptionRewriting =
@@ -152,6 +156,9 @@ addRewriteRules qs = do
   -- Run confluence check for the new rules
   -- (should be done after adding all rules, see #3795)
   whenJustM (optConfluenceCheck <$> pragmaOptions) $ \confChk -> do
+    -- Warn if --cubical is enabled
+    whenJustM (optCubical <$> pragmaOptions) $ \_ -> genericWarning
+      "Confluence checking for --cubical is not yet supported, confluence checking might be incomplete"
     -- Global confluence checker requires rules to be sorted
     -- according to the generality of their lhs
     when (confChk == GlobalConfluenceCheck) $
@@ -160,6 +167,19 @@ addRewriteRules qs = do
     reportSDoc "rewriting" 10 $
       "done checking confluence of rules" <+> prettyList_ (map (prettyTCM . rewName) rews)
 
+-- Auxiliary function for checkRewriteRule.
+-- | Get domain of rewrite relation.
+rewriteRelationDom :: QName -> TCM (ListTel, Dom Type)
+rewriteRelationDom rel = do
+  -- We know that the type of rel is that of a relation.
+  relV <- relView =<< do defType <$> getConstInfo rel
+  let RelView _tel delta a _a' _core = fromMaybe __IMPOSSIBLE__ relV
+  reportSDoc "rewriting" 30 $ do
+    "rewrite relation at type " <+> do
+      inTopContext $ prettyTCM (telFromList delta) <+> " |- " <+> do
+        addContext delta $ prettyTCM a
+  return (delta, a)
+
 -- | Check the validity of @q : Γ → rel us lhs rhs@ as rewrite rule
 --   @
 --       Γ ⊢ lhs ↦ rhs : B
@@ -173,7 +193,11 @@ checkRewriteRule q = do
   requireOptionRewriting
   let failNoBuiltin = typeError $ GenericError $
         "Cannot add rewrite rule without prior BUILTIN REWRITE"
-  rel <- fromMaybeM failNoBuiltin $ getBuiltinName builtinRewrite
+  rels <- fromMaybeM failNoBuiltin getBuiltinRewriteRelations
+  reportSDoc "rewriting.relations" 40 $ vcat
+    [ "Rewrite relations:"
+    , prettyList $ map prettyTCM $ toList rels
+    ]
   def <- instantiateDef =<< getConstInfo q
   -- Issue 1651: Check that we are not adding a rewrite rule
   -- for a type signature whose body has not been type-checked yet.
@@ -183,13 +207,6 @@ checkRewriteRule q = do
       , prettyTCM q
       , " cannot be added before the function definition"
       ]
-  -- We know that the type of rel is that of a relation.
-  relV <- relView =<< do defType <$> getConstInfo rel
-  let RelView _tel delta a _a' _core = fromMaybe __IMPOSSIBLE__ relV
-  reportSDoc "rewriting" 30 $ do
-    "rewrite relation at type " <+> do
-      inTopContext $ prettyTCM (telFromList delta) <+> " |- " <+> do
-        addContext delta $ prettyTCM a
   -- Get rewrite rule (type of q).
   TelV gamma1 core <- telView $ defType def
   reportSDoc "rewriting" 30 $ vcat
@@ -200,22 +217,38 @@ checkRewriteRule q = do
   let failureWrongTarget :: TCM a
       failureWrongTarget = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " does not target rewrite relation" ]
-  let failureMetas :: TCM a
-      failureMetas       = typeError . GenericDocError =<< hsep
-        [ prettyTCM q , " is not a legal rewrite rule, since it contains unsolved meta variables" ]
+  let failureBlocked :: Blocker -> TCM a
+      failureBlocked b
+        | not (null ms) = err $ "it contains the unsolved meta variable(s)" <+> prettyList_ (map prettyTCM $ Set.toList ms)
+        | not (null ps) = err $ "it is blocked on problem(s)" <+> prettyList_ (map prettyTCM $ Set.toList ps)
+        | not (null qs) = err $ "it requires the definition(s) of" <+> prettyList_ (map prettyTCM $ Set.toList qs)
+        | otherwise = __IMPOSSIBLE__
+        where
+          err reason = typeError . GenericDocError =<< hsep
+            [ prettyTCM q , " is not a legal rewrite rule, since" , reason ]
+          ms = allBlockingMetas b
+          ps = allBlockingProblems b
+          qs = allBlockingDefs b
   let failureNotDefOrCon :: TCM a
       failureNotDefOrCon = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule, since the left-hand side is neither a defined symbol nor a constructor" ]
   let failureFreeVars :: IntSet -> TCM a
       failureFreeVars xs = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule, since the following variables are not bound by the left hand side: " , prettyList_ (map (prettyTCM . var) $ IntSet.toList xs) ]
+  let failureNonLinearPars :: IntSet -> TCM a
+      failureNonLinearPars xs = typeError . GenericDocError =<< do
+        (prettyTCM q
+          <+> " is not a legal rewrite rule, since the following parameters are bound more than once on the left hand side: "
+          <+> hsep (List.intersperse "," $ map (prettyTCM . var) $ IntSet.toList xs))
+          <> ". Perhaps you can use a postulate instead of a constructor as the head symbol?"
   let failureIllegalRule :: TCM a -- TODO:: Defined but not used
       failureIllegalRule = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule" ]
 
   -- Check that type of q targets rel.
   case unEl core of
-    Def rel' es@(_:_:_) | rel == rel' -> do
+    Def rel es@(_:_:_) | rel `elem` rels -> do
+      (delta, a) <- rewriteRelationDom rel
       -- Because of the type of rel (Γ → sort), all es are applications.
       let vs = map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       -- The last two arguments are lhs and rhs.
@@ -230,10 +263,6 @@ checkRewriteRule q = do
       gamma1 <- instantiateFull gamma1
       let gamma = gamma0 `abstract` gamma1
 
-      unless (noMetas (telToList gamma1)) $ do
-        reportSDoc "rewriting" 30 $ "metas in gamma1: " <+> text (show $ allMetasList $ telToList gamma1)
-        failureMetas
-
       -- 2017-06-18, Jesper: Unfold inlined definitions on the LHS.
       -- This is necessary to replace copies created by imports by their
       -- original definition.
@@ -258,13 +287,9 @@ checkRewriteRule q = do
 
         checkNoLhsReduction f hd es
 
-        unless (noMetas (es, rhs, b)) $ do
-          reportSDoc "rewriting" 30 $ "metas in lhs: " <+> text (show $ allMetasList es)
-          reportSDoc "rewriting" 30 $ "metas in rhs: " <+> text (show $ allMetasList rhs)
-          reportSDoc "rewriting" 30 $ "metas in b  : " <+> text (show $ allMetasList b)
-          failureMetas
+        ps <- catchPatternErr failureBlocked $
+          patternFrom Relevant 0 (t , Def f []) es
 
-        ps <- patternFrom Relevant 0 (t , Def f []) es
         reportSDoc "rewriting" 30 $
           "Pattern generated from lhs: " <+> prettyTCM (PDef f ps)
 
@@ -295,6 +320,10 @@ checkRewriteRule q = do
         unlessNull (freeVars IntSet.\\ boundVars) failureFreeVars
         unlessNull (usedVars IntSet.\\ (boundVars `IntSet.union` IntSet.fromList pars)) failureFreeVars
 
+        reportSDoc "rewriting" 70 $
+          "variables bound in (erased) parameter position: " <+> text (show pars)
+        unlessNull (boundVars `IntSet.intersection` IntSet.fromList pars) failureNonLinearPars
+
         let rew = RewriteRule q gamma f ps rhs (unDom b) False
 
         reportSDoc "rewriting" 10 $ vcat
@@ -335,7 +364,20 @@ checkRewriteRule q = do
     checkAxFunOrCon :: QName -> Definition -> TCM ()
     checkAxFunOrCon f def = case theDef def of
       Axiom{}        -> return ()
-      Function{}     -> return ()
+      def@Function{} -> whenJust (maybeRight (funProjection def)) $ \proj ->
+        case projProper proj of
+          Just{} -> typeError . GenericDocError =<< hsep
+            [ prettyTCM q , " is not a legal rewrite rule, since the head symbol"
+            , prettyTCM f , "is a projection"
+            ]
+          Nothing -> typeError . GenericDocError =<< hsep
+            [ prettyTCM q , " is not a legal rewrite rule, since the head symbol"
+            , hd , "is a projection-like function."
+            , "You can turn off the projection-like optimization for", hd
+            , "with the pragma {-# NOT_PROJECTION_LIKE", hd, "#-}"
+            , "or globally with the flag --no-projection-like"
+            ]
+            where hd = prettyTCM f
       Constructor{}  -> return ()
       AbstractDefn{} -> return ()
       Primitive{}    -> return () -- TODO: is this fine?
