@@ -6,6 +6,8 @@
 
 module Agda.TypeChecking.Rewriting.NonLinPattern where
 
+import Prelude hiding ( null )
+
 import Control.Monad        ( (>=>), forM )
 import Control.Monad.Reader ( asks )
 
@@ -15,6 +17,7 @@ import qualified Data.IntSet as IntSet
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Defs
+import Agda.Syntax.Internal.MetaVars ( AllMetas, unblockOnAllMetasIn )
 
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Free
@@ -27,7 +30,7 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.Primitive.Cubical
+import Agda.TypeChecking.Primitive.Cubical.Base
 
 import Agda.Utils.Either
 import Agda.Utils.Functor
@@ -57,14 +60,14 @@ instance PatternFrom (Type, Term) Elims [Elim' NLPat] where
   patternFrom r k (t,hd) = \case
     [] -> return []
     (Apply u : es) -> do
-      ~(Pi a b) <- unEl <$> reduce t
+      (a, b) <- assertPi t
       p   <- patternFrom r k a u
-      t'  <- t `piApplyM` u
+      let t'  = absApp b (unArg u)
       let hd' = hd `apply` [ u ]
       ps  <- patternFrom r k (t',hd') es
       return $ Apply p : ps
     (IApply x y i : es) -> do
-      ~(PathType s q l b u v) <- pathView =<< reduce t
+      (s, q, l, b, u, v) <- assertPath t
       let t' = El s $ unArg b `apply` [ defaultArg i ]
       let hd' = hd `applyE` [IApply x y i]
       interval <- primIntervalType
@@ -72,7 +75,7 @@ instance PatternFrom (Type, Term) Elims [Elim' NLPat] where
       ps  <- patternFrom r k (t',hd') es
       return $ IApply (PTerm x) (PTerm y) p : ps
     (Proj o f : es) -> do
-      ~(Just (El _ (Pi a b))) <- getDefType f =<< reduce t
+      (a,b) <- assertProjOf f t
       let t' = b `absApp` hd
       hd' <- applyDef o f (argFromDom a $> hd)
       ps  <- patternFrom r k (t',hd') es
@@ -88,14 +91,15 @@ instance PatternFrom () Type NLPType where
 
 instance PatternFrom () Sort NLPSort where
   patternFrom r k _ s = do
-    s <- reduce s
+    s <- abortIfBlocked s
     case s of
       Type l   -> PType <$> patternFrom r k () l
       Prop l   -> PProp <$> patternFrom r k () l
+      SSet l   -> PSSet <$> patternFrom r k () l
       Inf f n  -> return $ PInf f n
-      SSet l   -> __IMPOSSIBLE__
       SizeUniv -> return PSizeUniv
       LockUniv -> return PLockUniv
+      IntervalUniv -> return PIntervalUniv
       PiSort _ _ _ -> __IMPOSSIBLE__
       FunSort _ _ -> __IMPOSSIBLE__
       UnivSort _ -> __IMPOSSIBLE__
@@ -116,21 +120,26 @@ instance PatternFrom () Level NLPat where
 
 instance PatternFrom Type Term NLPat where
   patternFrom r0 k t v = do
-    t <- reduce t
+    t <- abortIfBlocked t
     etaRecord <- isEtaRecordType t
-    prop <- fromRight __IMPOSSIBLE__ <.> runBlocked $ isPropM t
+    prop <- isPropM t
     let r = if prop then Irrelevant else r0
-    v <- unLevel =<< reduce v
+    v <- unLevel =<< abortIfBlocked v
     reportSDoc "rewriting.build" 60 $ sep
       [ "building a pattern from term v = " <+> prettyTCM v
       , " of type " <+> prettyTCM t
       ]
-    let done = return $ PTerm v
+    pview <- pathViewAsPi'whnf
+    let done = blockOnMetasIn v >> return (PTerm v)
     case (unEl t , stripDontCare v) of
       (Pi a b , _) -> do
         let body = raise 1 v `apply` [ Arg (domInfo a) $ var 0 ]
         p <- addContext a (patternFrom r (k+1) (absBody b) body)
         return $ PLam (domInfo a) $ Abs (absName b) p
+      _ | Left ((a,b),(x,y)) <- pview t -> do
+        let body = raise 1 v `applyE` [ IApply (raise 1 $ x) (raise 1 $ y) $ var 0 ]
+        p <- addContext a (patternFrom r (k+1) (absBody b) body)
+        return $ PLam (domInfo a) $ Abs (absName b) p
       (_ , Var i es)
        | i < k     -> do
            t <- typeOfBV i
@@ -138,10 +147,11 @@ instance PatternFrom Type Term NLPat where
        -- The arguments of `var i` should be distinct bound variables
        -- in order to build a Miller pattern
        | Just vs <- allApplyElims es -> do
-           TelV tel _ <- telView =<< typeOfBV i
-           unless (size tel >= size vs) __IMPOSSIBLE__
+           TelV tel rest <- telViewPath =<< typeOfBV i
+           unless (size tel >= size vs) $ blockOnMetasIn rest >> addContext tel (errNotPi rest)
            let ts = applySubst (parallelS $ reverse $ map unArg vs) $ map unDom $ flattenTel tel
            mbvs <- forM (zip ts vs) $ \(t , v) -> do
+             blockOnMetasIn (v,t)
              isEtaVar (unArg v) t >>= \case
                Just j | j < k -> return $ Just $ v $> j
                _              -> return Nothing
@@ -156,9 +166,9 @@ instance PatternFrom Type Term NLPat where
       (_ , _ ) | Just (d, pars) <- etaRecord -> do
         def <- theDef <$> getConstInfo d
         (tel, c, ci, vs) <- etaExpandRecord_ d pars def v
-        caseMaybeM (getFullyAppliedConType c t) __IMPOSSIBLE__ $ \ (_ , ct) -> do
+        ct <- assertConOf c t
         PDef (conName c) <$> patternFrom r k (ct , Con c ci []) (map Apply vs)
-      (_ , Lam i t) -> __IMPOSSIBLE__
+      (_ , Lam{})   -> errNotPi t
       (_ , Lit{})   -> done
       (_ , Def f es) | isIrrelevant r -> done
       (_ , Def f es) -> do
@@ -171,8 +181,8 @@ instance PatternFrom Type Term NLPat where
             ft <- defType <$> getConstInfo f
             PDef f <$> patternFrom r k (ft , Def f []) es
       (_ , Con c ci vs) | isIrrelevant r -> done
-      (_ , Con c ci vs) ->
-        caseMaybeM (getFullyAppliedConType c t) __IMPOSSIBLE__ $ \ (_ , ct) -> do
+      (_ , Con c ci vs) -> do
+        ct <- assertConOf c t
         PDef (conName c) <$> patternFrom r k (ct , Con c ci []) vs
       (_ , Pi a b) | isIrrelevant r -> done
       (_ , Pi a b) -> do
@@ -182,7 +192,7 @@ instance PatternFrom Type Term NLPat where
       (_ , Sort s)     -> PSort <$> patternFrom r k () s
       (_ , Level l)    -> __IMPOSSIBLE__
       (_ , DontCare{}) -> __IMPOSSIBLE__
-      (_ , MetaV{})    -> __IMPOSSIBLE__
+      (_ , MetaV m _)  -> __IMPOSSIBLE__
       (_ , Dummy s _)  -> __IMPOSSIBLE_VERBOSE__ s
 
 -- | Convert from a non-linear pattern to a term.
@@ -226,9 +236,11 @@ instance NLPatToTerm NLPType Type where
 instance NLPatToTerm NLPSort Sort where
   nlPatToTerm (PType l) = Type <$> nlPatToTerm l
   nlPatToTerm (PProp l) = Prop <$> nlPatToTerm l
+  nlPatToTerm (PSSet l) = SSet <$> nlPatToTerm l
   nlPatToTerm (PInf f n) = return $ Inf f n
   nlPatToTerm PSizeUniv = return SizeUniv
   nlPatToTerm PLockUniv = return LockUniv
+  nlPatToTerm PIntervalUniv = return IntervalUniv
 
 -- | Gather the set of pattern variables of a non-linear pattern
 class NLPatVars a where
@@ -247,9 +259,11 @@ instance NLPatVars NLPSort where
   nlPatVarsUnder k = \case
     PType l   -> nlPatVarsUnder k l
     PProp l   -> nlPatVarsUnder k l
+    PSSet l   -> nlPatVarsUnder k l
     PInf f n  -> empty
     PSizeUniv -> empty
     PLockUniv -> empty
+    PIntervalUniv -> empty
 
 instance NLPatVars NLPat where
   nlPatVarsUnder k = \case
@@ -289,7 +303,7 @@ instance GetMatchables NLPat where
   getMatchables p =
     case p of
       PVar _ _       -> empty
-      PDef f _       -> singleton f
+      PDef f es      -> singleton f ++ getMatchables es
       PLam _ x       -> getMatchables x
       PPi a b        -> getMatchables (a,b)
       PSort s        -> getMatchables s
@@ -303,9 +317,11 @@ instance GetMatchables NLPSort where
   getMatchables = \case
     PType l   -> getMatchables l
     PProp l   -> getMatchables l
+    PSSet l   -> getMatchables l
     PInf f n  -> empty
     PSizeUniv -> empty
     PLockUniv -> empty
+    PIntervalUniv -> empty
 
 instance GetMatchables Term where
   getMatchables = getDefs' __IMPOSSIBLE__ singleton
@@ -336,6 +352,68 @@ instance Free NLPSort where
   freeVars' = \case
     PType l   -> freeVars' l
     PProp l   -> freeVars' l
+    PSSet l   -> freeVars' l
     PInf f n  -> mempty
     PSizeUniv -> mempty
     PLockUniv -> mempty
+    PIntervalUniv -> mempty
+
+-- Throws a pattern violation if the given term contains any
+-- metavariables.
+blockOnMetasIn :: (MonadBlock m, AllMetas t) => t -> m ()
+blockOnMetasIn t = case unblockOnAllMetasIn t of
+  UnblockOnAll ms | null ms -> return ()
+  b -> patternViolation b
+
+-- Helper functions
+
+
+assertPi :: Type -> TCM (Dom Type, Abs Type)
+assertPi t = abortIfBlocked t >>= \case
+  El _ (Pi a b) -> return (a,b)
+  t             -> errNotPi t
+
+errNotPi :: Type -> TCM a
+errNotPi t = typeError . GenericDocError =<< fsep
+    [ prettyTCM t
+    , "should be a function type, but it isn't."
+    , "Do you have any non-confluent rewrite rules?"
+    ]
+
+assertPath :: Type -> TCM (Sort, QName, Arg Term, Arg Term, Arg Term, Arg Term)
+assertPath t = abortIfBlocked t >>= pathView >>= \case
+  PathType s q l b u v -> return (s,q,l,b,u,v)
+  OType t -> errNotPath t
+
+errNotPath :: Type -> TCM a
+errNotPath t = typeError . GenericDocError =<< fsep
+    [ prettyTCM t
+    , "should be a path type, but it isn't."
+    , "Do you have any non-confluent rewrite rules?"
+    ]
+
+assertProjOf :: QName -> Type -> TCM (Dom Type, Abs Type)
+assertProjOf f t = do
+  t <- abortIfBlocked t
+  getDefType f t >>= \case
+    Just (El _ (Pi a b)) -> return (a,b)
+    _ -> errNotProjOf f t
+
+errNotProjOf :: QName -> Type -> TCM a
+errNotProjOf f t = typeError . GenericDocError =<< fsep
+      [ prettyTCM f , "should be a projection from type"
+      , prettyTCM t , "but it isn't."
+      , "Do you have any non-confluent rewrite rules?"
+      ]
+
+assertConOf :: ConHead -> Type -> TCM Type
+assertConOf c t = getFullyAppliedConType c t >>= \case
+    Just (_ , ct) -> return ct
+    Nothing -> errNotConOf c t
+
+errNotConOf :: ConHead -> Type -> TCM a
+errNotConOf c t = typeError . GenericDocError =<< fsep
+      [ prettyTCM c , "should be a constructor of type"
+      , prettyTCM t , "but it isn't."
+      , "Do you have any non-confluent rewrite rules?"
+      ]
