@@ -118,6 +118,8 @@
 module Agda.TypeChecking.Rules.LHS.Unify
   ( UnificationResult
   , UnificationResult'(..)
+  , NoLeftInv(..)
+  , unifyIndices'
   , unifyIndices ) where
 
 import Prelude hiding (null)
@@ -135,7 +137,7 @@ import Data.IntMap (IntMap)
 
 import qualified Agda.Benchmarking as Bench
 
-import Agda.Interaction.Options (optInjectiveTypeConstructors)
+import Agda.Interaction.Options (optInjectiveTypeConstructors, optCubical, optWithoutK)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -143,6 +145,11 @@ import Agda.Syntax.Literal
 
 import Agda.TypeChecking.Monad
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
+import Agda.TypeChecking.Monad.Builtin -- (constructorForm, getTerm, builtinPathP)
+import Agda.TypeChecking.Primitive hiding (Nat)
+import Agda.TypeChecking.Primitive.Cubical
+import Agda.TypeChecking.Names
+import Agda.TypeChecking.Conversion -- equalTerm
 import Agda.TypeChecking.Conversion.Pure
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Datatypes
@@ -159,7 +166,11 @@ import Agda.TypeChecking.Free.Reduce
 import Agda.TypeChecking.Records
 
 import Agda.TypeChecking.Rules.LHS.Problem
+import Agda.TypeChecking.Rules.LHS.Unify.Types
+import Agda.TypeChecking.Rules.LHS.Unify.LeftInverse
 
+import Agda.Utils.Empty
+import Agda.Utils.Either
 import Agda.Utils.Benchmark
 import Agda.Utils.Either
 import Agda.Utils.Function
@@ -174,10 +185,12 @@ import Agda.Utils.PartialOrd
 import Agda.Utils.Permutation
 import Agda.Utils.Singleton
 import Agda.Utils.Size
+import Agda.Utils.WithDefault
 import Agda.Utils.Tuple
 
 import Agda.Utils.Impossible
 
+
 -- | Result of 'unifyIndices'.
 type UnificationResult = UnificationResult'
   ( Telescope                  -- @tel@
@@ -185,6 +198,13 @@ type UnificationResult = UnificationResult'
   , [NamedArg DeBruijnPattern] -- @ps@    s.t. @tel ⊢ ps    : eqTel @
   )
 
+type FullUnificationResult = UnificationResult'
+  ( Telescope                  -- @tel@
+  , PatternSubstitution        -- @sigma@ s.t. @tel ⊢ sigma : varTel@
+  , [NamedArg DeBruijnPattern] -- @ps@    s.t. @tel ⊢ ps    : eqTel @
+  , Either NoLeftInv (Substitution, Substitution) -- (τ,leftInv)
+  )
+
 data UnificationResult' a
   = Unifies  a                        -- ^ Unification succeeded.
   | NoUnify  NegativeUnification      -- ^ Terms are not unifiable.
@@ -204,27 +224,29 @@ data UnificationResult' a
 --
 --   The result is the most general unifier of @us@ and @vs@.
 unifyIndices
-  :: (PureTCM m, MonadBench m, BenchPhase m ~ Bench.Phase)
-  => Telescope     -- ^ @gamma@
-  -> FlexibleVars  -- ^ @flex@
-  -> Type          -- ^ @a@
-  -> Args          -- ^ @us@
-  -> Args          -- ^ @vs@
+  :: (PureTCM m, MonadBench m, BenchPhase m ~ Bench.Phase, MonadError TCErr m)
+  => Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
+  -> Telescope       -- ^ @gamma@
+  -> FlexibleVars    -- ^ @flex@
+  -> Type            -- ^ @a@
+  -> Args            -- ^ @us@
+  -> Args            -- ^ @vs@
   -> m UnificationResult
-unifyIndices tel flex a us vs =
+unifyIndices linv tel flex a us vs =
   Bench.billTo [Bench.Typing, Bench.CheckLHS, Bench.UnifyIndices] $
-    unifyIndices' tel flex a us vs
+    fmap (\(a,b,c,_) -> (a,b,c)) <$> unifyIndices' linv tel flex a us vs
 
 unifyIndices'
-  :: (PureTCM m)
-  => Telescope     -- ^ @gamma@
+  :: (PureTCM m, MonadError TCErr m)
+  => Maybe NoLeftInv -- ^ Do we have a reason for not computing a left inverse?
+  -> Telescope     -- ^ @gamma@
   -> FlexibleVars  -- ^ @flex@
   -> Type          -- ^ @a@
   -> Args          -- ^ @us@
   -> Args          -- ^ @vs@
-  -> m UnificationResult
-unifyIndices' tel flex a [] [] = return $ Unifies (tel, idS, [])
-unifyIndices' tel flex a us vs = do
+  -> m FullUnificationResult
+unifyIndices' linv tel flex a [] [] = return $ Unifies (tel, idS, [], Right (idS, raiseS 1))
+unifyIndices' linv tel flex a us vs = do
     reportSDoc "tc.lhs.unify" 10 $
       sep [ "unifyIndices"
           , ("tel  =" <+>) $ nest 2 $ prettyTCM tel
@@ -235,376 +257,29 @@ unifyIndices' tel flex a us vs = do
           ]
     initialState    <- initUnifyState tel flex a us vs
     reportSDoc "tc.lhs.unify" 20 $ "initial unifyState:" <+> prettyTCM initialState
-    reportSDoc "tc.lhs.unify" 70 $ "initial unifyState:" <+> text (show initialState)
-    (result,output) <- runUnifyLogT $ unify initialState rightToLeftStrategy
-    let ps = applySubst (unifyProof output) $ teleNamedArgs (eqTel initialState)
-    return $ fmap (\s -> (varTel s , unifySubst output , ps)) result
-
-----------------------------------------------------
--- Equalities
-----------------------------------------------------
-
-data Equality = Equal
-  { _eqType  :: Dom Type
-  , _eqLeft  :: Term
-  , _eqRight :: Term
-  }
-
-instance Reduce Equality where
-  reduce' (Equal a u v) = Equal <$> reduce' a <*> reduce' u <*> reduce' v
-
-eqConstructorForm :: HasBuiltins m => Equality -> m Equality
-eqConstructorForm (Equal a u v) = Equal a <$> constructorForm u <*> constructorForm v
-
-eqUnLevel :: HasBuiltins m => Equality -> m Equality
-eqUnLevel (Equal a u v) = Equal a <$> unLevel u <*> unLevel v
-  where
-    unLevel (Level l) = reallyUnLevelView l
-    unLevel u         = return u
-
-----------------------------------------------------
--- Unify state
-----------------------------------------------------
-
-data UnifyState = UState
-  { varTel   :: Telescope     -- ^ Don't reduce!
-  , flexVars :: FlexibleVars
-  , eqTel    :: Telescope     -- ^ Can be reduced eagerly.
-  , eqLHS    :: [Arg Term]    -- ^ Ends up in dot patterns (should not be reduced eagerly).
-  , eqRHS    :: [Arg Term]    -- ^ Ends up in dot patterns (should not be reduced eagerly).
-  } deriving (Show)
--- Issues #3578 and #4125: avoid unnecessary reduction in unifier.
-
-lensVarTel   :: Lens' Telescope UnifyState
-lensVarTel   f s = f (varTel s) <&> \ tel -> s { varTel = tel }
---UNUSED Liang-Ting Chen 2019-07-16
---lensFlexVars :: Lens' FlexibleVars UnifyState
---lensFlexVars f s = f (flexVars s) <&> \ flex -> s { flexVars = flex }
-
-lensEqTel    :: Lens' Telescope UnifyState
-lensEqTel    f s = f (eqTel s) <&> \ x -> s { eqTel = x }
-
---UNUSED Liang-Ting Chen 2019-07-16
---lensEqLHS    :: Lens' Args UnifyState
---lensEqLHS    f s = f (eqLHS s) <&> \ x -> s { eqLHS = x }
-
---UNUSED Liang-Ting Chen 2019-07-16
---lensEqRHS    :: Lens' Args UnifyState
---lensEqRHS    f s = f (eqRHS s) <&> \ x -> s { eqRHS = x }
-
--- UNUSED Andreas, 2019-10-14
--- instance Reduce UnifyState where
---   reduce' (UState var flex eq lhs rhs) =
---     UState <$> reduce' var
---            <*> pure flex
---            <*> reduce' eq
---            <*> reduce' lhs
---            <*> reduce' rhs
-
--- Andreas, 2019-10-14, issues #3578 and #4125:
--- | Don't ever reduce the whole 'varTel', as it will destroy
--- readability of the context in interactive editing!
--- To make sure this insight is not lost, the following
--- dummy instance should prevent a proper 'Reduce' instance for 'UnifyState'.
-instance Reduce UnifyState where
-  reduce' = __IMPOSSIBLE__
-
---UNUSED Liang-Ting Chen 2019-07-16
---reduceEqTel :: UnifyState -> TCM UnifyState
---reduceEqTel = lensEqTel reduce
-
--- UNUSED Andreas, 2019-10-14
--- instance Normalise UnifyState where
---   normalise' (UState var flex eq lhs rhs) =
---     UState <$> normalise' var
---            <*> pure flex
---            <*> normalise' eq
---            <*> normalise' lhs
---            <*> normalise' rhs
-
-instance PrettyTCM UnifyState where
-  prettyTCM state = "UnifyState" $$ nest 2 (vcat $
-    [ "variable tel:  " <+> prettyTCM gamma
-    , "flexible vars: " <+> pshow (map flexVarF $ flexVars state)
-    , "equation tel:  " <+> addContext gamma (prettyTCM delta)
-    , "equations:     " <+> addContext gamma (prettyList_ (zipWith prettyEquality (eqLHS state) (eqRHS state)))
-    ])
-    where
-      flexVarF fi = (flexVar fi, flexForced fi)
-      gamma = varTel state
-      delta = eqTel state
-      prettyEquality x y = prettyTCM x <+> "=?=" <+> prettyTCM y
-
-initUnifyState
-  :: PureTCM m
-  => Telescope -> FlexibleVars -> Type -> Args -> Args -> m UnifyState
-initUnifyState tel flex a lhs rhs = do
-  (tel, a, lhs, rhs) <- instantiateFull (tel, a, lhs, rhs)
-  let n = size lhs
-  unless (n == size rhs) __IMPOSSIBLE__
-  TelV eqTel _ <- telView a
-  unless (n == size eqTel) __IMPOSSIBLE__
-  return $ UState tel flex eqTel lhs rhs
-  -- Andreas, 2019-02-23, issue #3578: do not eagerly reduce
-  -- reduce $ UState tel flex eqTel lhs rhs
-
-isUnifyStateSolved :: UnifyState -> Bool
-isUnifyStateSolved = null . eqTel
-
-varCount :: UnifyState -> Int
-varCount = size . varTel
-
--- | Get the type of the i'th variable in the given state
-getVarType :: Int -> UnifyState -> Dom Type
-getVarType i s = indexWithDefault __IMPOSSIBLE__ (flattenTel $ varTel s) i
-
-getVarTypeUnraised :: Int -> UnifyState -> Dom Type
-getVarTypeUnraised i s = snd <$> indexWithDefault __IMPOSSIBLE__ (telToList $ varTel s) i
-
-eqCount :: UnifyState -> Int
-eqCount = size . eqTel
-
--- | Get the k'th equality in the given state. The left- and right-hand sides
---   of the equality live in the varTel telescope, and the type of the equality
---   lives in the varTel++eqTel telescope
-getEquality :: Int -> UnifyState -> Equality
-getEquality k UState { eqTel = eqs, eqLHS = lhs, eqRHS = rhs } =
-    Equal (indexWithDefault __IMPOSSIBLE__ (flattenTel eqs) k)
-          (unArg $ indexWithDefault __IMPOSSIBLE__ lhs k)
-          (unArg $ indexWithDefault __IMPOSSIBLE__ rhs k)
-
--- | As getEquality, but with the unraised type
-getEqualityUnraised :: Int -> UnifyState -> Equality
-getEqualityUnraised k UState { eqTel = eqs, eqLHS = lhs, eqRHS = rhs } =
-    Equal (snd <$> indexWithDefault __IMPOSSIBLE__ (telToList eqs) k)
-          (unArg $ indexWithDefault __IMPOSSIBLE__ lhs k)
-          (unArg $ indexWithDefault __IMPOSSIBLE__ rhs k)
-
---UNUSED Liang-Ting Chen 2019-07-16
---getEqInfo :: Int -> UnifyState -> ArgInfo
---getEqInfo k UState { eqTel = eqs } =
---  domInfo $ indexWithDefault __IMPOSSIBLE__ (telToList eqs) k
---
----- | Add a list of equations to the front of the equation telescope
---addEqs :: Telescope -> [Arg Term] -> [Arg Term] -> UnifyState -> UnifyState
---addEqs tel us vs s =
---  s { eqTel = tel `abstract` eqTel s
---    , eqLHS = us ++ eqLHS s
---    , eqRHS = vs ++ eqRHS s
---    }
---  where k = size tel
---
---addEq :: Type -> Arg Term -> Arg Term -> UnifyState -> UnifyState
---addEq a u v = addEqs (ExtendTel (defaultDom a) (Abs underscore EmptyTel)) [u] [v]
+    (result,log) <- runUnifyLogT $ unify initialState rightToLeftStrategy
+    forM result $ \ s -> do -- Unifies case
+        let output = mconcat [output | (UnificationStep _ _ output,_) <- log ]
+        let ps = applySubst (unifyProof output) $ teleNamedArgs (eqTel initialState)
+        tauInv <- do
+          strict     <- asksTC envSplitOnStrict
+          cubicalCompatible <- cubicalCompatibleOption
+          withoutK <- withoutKOption
+          case linv of
+            Just reason -> pure (Left reason)
+            Nothing
+              | strict            -> pure (Left SplitOnStrict)
+              | cubicalCompatible -> buildLeftInverse initialState log
+              | withoutK          -> pure (Left NoCubical)
+              | otherwise         -> pure (Left WithKEnabled)
+        reportSDoc "tc.lhs.unify" 20 $ "ps:" <+> pretty ps
+        return $ (varTel s, unifySubst output, ps, tauInv)
 
 
 
--- | Instantiate the k'th variable with the given value.
---   Returns Nothing if there is a cycle.
-solveVar :: Int             -- ^ Index @k@
-         -> DeBruijnPattern -- ^ Solution @u@
-         -> UnifyState -> Maybe (UnifyState, PatternSubstitution)
-solveVar k u s = case instantiateTelescope (varTel s) k u of
-  Nothing -> Nothing
-  Just (tel' , sigma , rho) -> Just $ (,sigma) $ UState
-      { varTel   = tel'
-      , flexVars = permuteFlex (reverseP rho) $ flexVars s
-      , eqTel    = applyPatSubst sigma $ eqTel s
-      , eqLHS    = applyPatSubst sigma $ eqLHS s
-      , eqRHS    = applyPatSubst sigma $ eqRHS s
-      }
-  where
-    permuteFlex :: Permutation -> FlexibleVars -> FlexibleVars
-    permuteFlex perm =
-      mapMaybe $ \(FlexibleVar ai fc k p x) ->
-        FlexibleVar ai fc k p <$> List.elemIndex x (permPicks perm)
-
-applyUnder :: Int -> Telescope -> Term -> Telescope
-applyUnder k tel u
- | k < 0     = __IMPOSSIBLE__
- | k == 0    = tel `apply1` u
- | otherwise = case tel of
-    EmptyTel         -> __IMPOSSIBLE__
-    ExtendTel a tel' -> ExtendTel a $
-      Abs (absName tel') $ applyUnder (k-1) (absBody tel') u
-
-dropAt :: Int -> [a] -> [a]
-dropAt _ [] = __IMPOSSIBLE__
-dropAt k (x:xs)
- | k < 0     = __IMPOSSIBLE__
- | k == 0    = xs
- | otherwise = x : dropAt (k-1) xs
-
--- | Solve the k'th equation with the given value, which can depend on
---   regular variables but not on other equation variables.
-solveEq :: Int -> Term -> UnifyState -> (UnifyState, PatternSubstitution)
-solveEq k u s = (,sigma) $ s
-    { eqTel    = applyUnder k (eqTel s) u'
-    , eqLHS    = dropAt k $ eqLHS s
-    , eqRHS    = dropAt k $ eqRHS s
-    }
-  where
-    u'    = raise k u
-    n     = eqCount s
-    sigma = liftS (n-k-1) $ consS (dotP u') idS
-
---UNUSED Liang-Ting Chen 2019-07-16
----- | Simplify the k'th equation with the given value (which can depend on other
-----   equation variables). Returns Nothing if there is a cycle.
---simplifyEq :: Int -> Term -> UnifyState -> Maybe (UnifyState, PatternSubstitution)
---simplifyEq k u s = case instantiateTelescope (eqTel s) k u of
---  Nothing -> Nothing
---  Just (tel' , sigma , rho) -> Just $ (,sigma) $ UState
---    { varTel   = varTel s
---    , flexVars = flexVars s
---    , eqTel    = tel'
---    , eqLHS    = permute rho $ eqLHS s
---    , eqRHS    = permute rho $ eqRHS s
---    }
---
-----------------------------------------------------
--- Unification strategies
-----------------------------------------------------
-
-data UnifyStep
-  = Deletion
-    { deleteAt           :: Int
-    , deleteType         :: Type
-    , deleteLeft         :: Term
-    , deleteRight        :: Term
-    }
-  | Solution
-    { solutionAt         :: Int
-    , solutionType       :: Dom Type
-    , solutionVar        :: FlexibleVar Int
-    , solutionTerm       :: Term
-    }
-  | Injectivity
-    { injectAt           :: Int
-    , injectType         :: Type
-    , injectDatatype     :: QName
-    , injectParameters   :: Args
-    , injectIndices      :: Args
-    , injectConstructor  :: ConHead
-    }
-  | Conflict
-    { conflictAt         :: Int
-    , conflictType       :: Type
-    , conflictDatatype   :: QName
-    , conflictParameters :: Args
-    , conflictLeft       :: Term
-    , conflictRight      :: Term
-    }
-  | Cycle
-    { cycleAt            :: Int
-    , cycleType          :: Type
-    , cycleDatatype      :: QName
-    , cycleParameters    :: Args
-    , cycleVar           :: Int
-    , cycleOccursIn      :: Term
-    }
-  | EtaExpandVar
-    { expandVar           :: FlexibleVar Int
-    , expandVarRecordType :: QName
-    , expandVarParameters :: Args
-    }
-  | EtaExpandEquation
-    { expandAt           :: Int
-    , expandRecordType   :: QName
-    , expandParameters   :: Args
-    }
-  | LitConflict
-    { litConflictAt      :: Int
-    , litType            :: Type
-    , litConflictLeft    :: Literal
-    , litConflictRight   :: Literal
-    }
-  | StripSizeSuc
-    { stripAt            :: Int
-    , stripArgLeft       :: Term
-    , stripArgRight      :: Term
-    }
-  | SkipIrrelevantEquation
-    { skipIrrelevantAt   :: Int
-    }
-  | TypeConInjectivity
-    { typeConInjectAt    :: Int
-    , typeConstructor    :: QName
-    , typeConArgsLeft    :: Args
-    , typeConArgsRight   :: Args
-    } deriving (Show)
-
-instance PrettyTCM UnifyStep where
-  prettyTCM step = case step of
-    Deletion k a u v -> "Deletion" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "lhs:        " <+> prettyTCM u
-      , "rhs:        " <+> prettyTCM v
-      ])
-    Solution k a i u -> "Solution" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "variable:   " <+> text (show (flexVar i, flexPos i, flexForced i, flexKind i))
-      , "term:       " <+> prettyTCM u
-      ])
-    Injectivity k a d pars ixs c -> "Injectivity" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "datatype:   " <+> prettyTCM d
-      , "parameters: " <+> prettyList_ (map prettyTCM pars)
-      , "indices:    " <+> prettyList_ (map prettyTCM ixs)
-      , "constructor:" <+> prettyTCM c
-      ])
-    Conflict k a d pars u v -> "Conflict" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "datatype:   " <+> prettyTCM d
-      , "parameters: " <+> prettyList_ (map prettyTCM pars)
-      , "lhs:        " <+> prettyTCM u
-      , "rhs:        " <+> prettyTCM v
-      ])
-    Cycle k a d pars i u -> "Cycle" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "datatype:   " <+> prettyTCM d
-      , "parameters: " <+> prettyList_ (map prettyTCM pars)
-      , "variable:   " <+> text (show i)
-      , "term:       " <+> prettyTCM u
-      ])
-    EtaExpandVar fi r pars -> "EtaExpandVar" $$ nest 2 (vcat $
-      [ "variable:   " <+> text (show fi)
-      , "record type:" <+> prettyTCM r
-      , "parameters: " <+> prettyTCM pars
-      ])
-    EtaExpandEquation k r pars -> "EtaExpandEquation" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "record type:" <+> prettyTCM r
-      , "parameters: " <+> prettyTCM pars
-      ])
-    LitConflict k a u v -> "LitConflict" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "type:       " <+> prettyTCM a
-      , "lhs:        " <+> prettyTCM u
-      , "rhs:        " <+> prettyTCM v
-      ])
-    StripSizeSuc k u v -> "StripSizeSuc" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "lhs:        " <+> prettyTCM u
-      , "rhs:        " <+> prettyTCM v
-      ])
-    SkipIrrelevantEquation k -> "SkipIrrelevantEquation" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      ])
-    TypeConInjectivity k d us vs -> "TypeConInjectivity" $$ nest 2 (vcat $
-      [ "position:   " <+> text (show k)
-      , "datatype:   " <+> prettyTCM d
-      , "lhs:        " <+> prettyList_ (map prettyTCM us)
-      , "rhs:        " <+> prettyList_ (map prettyTCM vs)
-      ])
-
 type UnifyStrategy = forall m. (PureTCM m, MonadPlus m) => UnifyState -> m UnifyStep
 
+
 --UNUSED Liang-Ting Chen 2019-07-16
 --leftToRightStrategy :: UnifyStrategy
 --leftToRightStrategy s =
@@ -657,10 +332,10 @@ basicUnifyStrategy k s = do
      | Just fi <- findFlexible i flex
      , Just fj <- findFlexible j flex -> do
        let choice = chooseFlex fi fj
-           firstTryLeft  = msum [ return (Solution k dom{unDom = ha} fi v)
-                                , return (Solution k dom{unDom = ha} fj u)]
-           firstTryRight = msum [ return (Solution k dom{unDom = ha} fj u)
-                                , return (Solution k dom{unDom = ha} fi v)]
+           firstTryLeft  = msum [ return (Solution k dom{unDom = ha} fi v left)
+                                , return (Solution k dom{unDom = ha} fj u right)]
+           firstTryRight = msum [ return (Solution k dom{unDom = ha} fj u right)
+                                , return (Solution k dom{unDom = ha} fi v left)]
        reportSDoc "tc.lhs.unify" 40 $ "fi = " <+> text (show fi)
        reportSDoc "tc.lhs.unify" 40 $ "fj = " <+> text (show fj)
        reportSDoc "tc.lhs.unify" 40 $ "chooseFlex: " <+> text (show choice)
@@ -670,20 +345,25 @@ basicUnifyStrategy k s = do
          ExpandBoth   -> mzero -- This should be taken care of by etaExpandEquationStrategy
          ChooseEither -> firstTryRight
     (Just i, _)
-     | Just fi <- findFlexible i flex -> return $ Solution k dom{unDom = ha} fi v
+     | Just fi <- findFlexible i flex -> return $ Solution k dom{unDom = ha} fi v left
     (_, Just j)
-     | Just fj <- findFlexible j flex -> return $ Solution k dom{unDom = ha} fj u
+     | Just fj <- findFlexible j flex -> return $ Solution k dom{unDom = ha} fj u right
     _ -> mzero
   where
     flex = flexVars s
     n = eqCount s
+    left = Left (); right = Right ()
 
 dataStrategy :: Int -> UnifyStrategy
 dataStrategy k s = do
-  Equal Dom{unDom = a} u v <- eqConstructorForm =<< eqUnLevel =<< reduce (getEqualityUnraised k s)
-  sa <- reduce $ getSort a
+  Equal Dom{unDom = a} u v <- eqConstructorForm =<< eqUnLevel =<< getReducedEqualityUnraised k s
+  sortOk <- reduce (getSort a) <&> \case
+    Type{} -> True
+    Inf{}  -> True
+    SSet{} -> True
+    _      -> False
   case unEl a of
-    Def d es | Type{} <- sa -> do
+    Def d es | sortOk -> do
       npars <- catMaybesMP $ getNumberOfParameters d
       let (pars,ixs) = splitAt npars $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       reportSDoc "tc.lhs.unify" 40 $ addContext (varTel s `abstract` eqTel s) $
@@ -726,7 +406,7 @@ literalStrategy k s = do
 
 etaExpandVarStrategy :: Int -> UnifyStrategy
 etaExpandVarStrategy k s = do
-  Equal Dom{unDom = a} u v <- eqUnLevel <=< reduce $ getEquality k s
+  Equal Dom{unDom = a} u v <- eqUnLevel =<< getReducedEquality k s
   shouldEtaExpand u v a s `mplus` shouldEtaExpand v u a s
   where
     -- TODO: use IsEtaVar to check if the term is a variable
@@ -739,7 +419,9 @@ etaExpandVarStrategy k s = do
       -- record or if it's unified against a record constructor term. Basically
       -- we need to avoid EtaExpandEquation if EtaExpandVar is possible, or the
       -- forcing translation is unhappy.
-      b         <- reduce $ unDom $ getVarTypeUnraised (varCount s - 1 - i) s
+      let k  = varCount s - 1 - i -- position of var i in telescope
+          b0 = unDom $ getVarTypeUnraised k s
+      b         <- addContext (telFromList $ take k $ telToList $ varTel s) $ reduce b0
       (d, pars) <- catMaybesMP $ isEtaRecordType b
       ps        <- fromMaybeMP $ allProjElims es
       guard =<< orM
@@ -760,7 +442,7 @@ etaExpandVarStrategy k s = do
 etaExpandEquationStrategy :: Int -> UnifyStrategy
 etaExpandEquationStrategy k s = do
   -- Andreas, 2019-02-23, re #3578, is the following reduce redundant?
-  Equal Dom{unDom = a} u v <- reduce $ getEqualityUnraised k s
+  Equal Dom{unDom = a} u v <- getReducedEqualityUnraised k s
   (d, pars) <- catMaybesMP $ addContext tel $ isEtaRecordType a
   guard =<< orM
     [ (Right True ==) <$> runBlocked (isSingletonRecord d pars)
@@ -789,7 +471,7 @@ etaExpandEquationStrategy k s = do
 simplifySizesStrategy :: Int -> UnifyStrategy
 simplifySizesStrategy k s = do
   isSizeName <- isSizeNameTest
-  Equal Dom{unDom = a} u v <- reduce $ getEquality k s
+  Equal Dom{unDom = a} u v <- getReducedEquality k s
   case unEl a of
     Def d _ -> do
       guard $ isSizeName d
@@ -806,7 +488,7 @@ injectiveTypeConStrategy :: Int -> UnifyStrategy
 injectiveTypeConStrategy k s = do
   injTyCon <- optInjectiveTypeConstructors <$> pragmaOptions
   guard injTyCon
-  eq <- eqUnLevel <=< reduce $ getEquality k s
+  eq <- eqUnLevel =<< getReducedEquality k s
   case eq of
     Equal a u@(Def d es) v@(Def d' es') | d == d' -> do
       -- d must be a data, record or axiom
@@ -829,7 +511,7 @@ injectiveTypeConStrategy k s = do
 
 injectivePragmaStrategy :: Int -> UnifyStrategy
 injectivePragmaStrategy k s = do
-  eq <- eqUnLevel <=< reduce $ getEquality k s
+  eq <- eqUnLevel =<< getReducedEquality k s
   case eq of
     Equal a u@(Def d es) v@(Def d' es') | d == d' -> do
       -- d must have an injective pragma
@@ -842,8 +524,9 @@ injectivePragmaStrategy k s = do
 
 skipIrrelevantStrategy :: Int -> UnifyStrategy
 skipIrrelevantStrategy k s = do
-  let Equal a _ _ = getEquality k s                               -- reduce not necessary
-  guard . (== Right True) =<< runBlocked (isIrrelevantOrPropM a)  -- reduction takes place here
+  let Equal a _ _ = getEquality k s                                 -- reduce not necessary
+  addContext (varTel s `abstract` eqTel s) $
+    guard . (== Right True) =<< runBlocked (isIrrelevantOrPropM a)  -- reduction takes place here
   -- TODO: do something in case the above is blocked (i.e. `Left b`)
   return $ SkipIrrelevantEquation k
 
@@ -852,47 +535,9 @@ skipIrrelevantStrategy k s = do
 -- Actually doing the unification
 ----------------------------------------------------
 
-data UnifyLogEntry
-  = UnificationStep  UnifyState UnifyStep
---  | UnificationDone  UnifyState -- unused?
-
-type UnifyLog = [UnifyLogEntry]
-
-data UnifyOutput = UnifyOutput
-  { unifySubst :: PatternSubstitution
-  , unifyProof :: PatternSubstitution
-  , unifyLog   :: UnifyLog
-  }
-
-instance Semigroup UnifyOutput where
-  x <> y = UnifyOutput
-    { unifySubst = unifySubst y `composeS` unifySubst x
-    , unifyProof = unifyProof y `composeS` unifyProof x
-    , unifyLog   = unifyLog x ++ unifyLog y
-    }
-
-instance Monoid UnifyOutput where
-  mempty  = UnifyOutput IdS IdS []
-  mappend = (<>)
-
-type UnifyLogT m a = WriterT UnifyOutput m a
-
-tellUnifySubst :: MonadWriter UnifyOutput m => PatternSubstitution -> m ()
-tellUnifySubst sub = tell $ UnifyOutput sub IdS []
-
-tellUnifyProof :: MonadWriter UnifyOutput m => PatternSubstitution -> m ()
-tellUnifyProof sub = tell $ UnifyOutput IdS sub []
-
-writeUnifyLog :: MonadWriter UnifyOutput m => UnifyLogEntry -> m ()
-writeUnifyLog x = tell $ UnifyOutput IdS IdS [x]
-
-runUnifyLogT :: UnifyLogT m a -> m (a,UnifyOutput)
-runUnifyLogT = runWriterT
-
 unifyStep
-  :: (PureTCM m, MonadWriter UnifyOutput m)
+  :: (PureTCM m, MonadWriter UnifyOutput m, MonadError TCErr m)
   => UnifyState -> UnifyStep -> m (UnificationResult' UnifyState)
-
 unifyStep s Deletion{ deleteAt = k , deleteType = a , deleteLeft = u , deleteRight = v } = do
     -- Check definitional equality of u and v
     isReflexive <- addContext (varTel s) $ runBlocked $ pureEqualTerm a u v
@@ -923,7 +568,7 @@ unifyStep s (Injectivity k a d pars ixs c) = do
   let ctype  = defType cdef `piApply` pars
   addContext (varTel s `abstract` eqTel1) $ reportSDoc "tc.lhs.unify" 40 $
     "Constructor type: " <+> prettyTCM ctype
-  TelV ctel ctarget <- telView ctype
+  TelV ctel ctarget <- addContext (varTel s `abstract` eqTel1) $ telView ctype
   let cixs = case unEl ctarget of
                Def d' es | d == d' ->
                  let args = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -943,7 +588,12 @@ unifyStep s (Injectivity k a d pars ixs c) = do
   -- computeNeighbourhood function in Agda.TypeChecking.Coverage.
   let hduTel = eqTel1 `abstract` ctel
       notforced = replicate (size hduTel) NotForced
-  res <- addContext (varTel s) $ unifyIndices'
+
+  -- The left inverse computed here is not actually used when computing
+  -- a left inverse for the overall match, so as a slight optimisation
+  -- we just don't bother computing it. __IMPOSSIBLE__ because that
+  -- field in the result is never evaluated.
+  res <- addContext (varTel s) $ unifyIndices' (Just __IMPOSSIBLE__)
            hduTel
            (allFlexVars notforced hduTel)
            (raise (size ctel) dtype)
@@ -973,10 +623,10 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
       tellUnifyProof rho
 
-      eqTel' <- reduce eqTel'
+      eqTel' <- addContext (varTel s) $ reduce eqTel'
 
       -- Compute new lhs and rhs by matching the old ones against rho
-      (lhs', rhs') <- do
+      (lhs', rhs') <- addContext (varTel s) $ do
         let ps = applySubst rho $ teleNamedArgs $ eqTel s
         (lhsMatch, _) <- Match.matchPatterns ps $ eqLHS s
         (rhsMatch, _) <- Match.matchPatterns ps $ eqRHS s
@@ -995,7 +645,7 @@ unifyStep s (Injectivity k a d pars ixs c) = do
                          (varTel s `abstract` eqTel s) a
                          (raise n u) (raise n v) (raise (n-k) ixs)]
 
-    Unifies (eqTel1', rho0, _) -> do
+    Unifies (eqTel1', rho0, _, _) -> do
       -- Split ps0 into parts for eqTel1 and ctel
       let (rho1, rho2) = splitS (size ctel) rho0
 
@@ -1008,10 +658,10 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
       tellUnifyProof rho
 
-      eqTel' <- reduce eqTel'
+      eqTel' <- addContext (varTel s) $ reduce eqTel'
 
       -- Compute new lhs and rhs by matching the old ones against rho
-      (lhs', rhs') <- do
+      (lhs', rhs') <- addContext (varTel s) $ do
         let ps = applySubst rho $ teleNamedArgs $ eqTel s
         (lhsMatch, _) <- Match.matchPatterns ps $ eqLHS s
         (rhsMatch, _) <- Match.matchPatterns ps $ eqRHS s
@@ -1226,7 +876,7 @@ solutionStep retry s
   caseEitherM (return eusable) (return . UnifyBlocked) $ \ usable -> do
 
   reportSDoc "tc.lhs.unify" 45 $ "Modality ok: " <+> prettyTCM usable
-  unless usable $ reportSLn "tc.lhs.unify" 65 $ "Rejected solution: " ++ show u
+  unless usable $ reportSDoc "tc.lhs.unify" 65 $ "Rejected solution: " <+> prettyTCM u
 
   -- We need a Flat equality to solve a Flat variable.
   -- This also ought to take care of the need for a usableCohesion check.
@@ -1255,14 +905,14 @@ solutionStep retry s
 solutionStep _ _ _ = __IMPOSSIBLE__
 
 unify
-  :: (PureTCM m, MonadWriter UnifyOutput m)
+  :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
   => UnifyState -> UnifyStrategy -> m (UnificationResult' UnifyState)
 unify s strategy = if isUnifyStateSolved s
                    then return $ Unifies s
                    else tryUnifyStepsAndContinue (strategy s)
   where
     tryUnifyStepsAndContinue
-      :: (PureTCM m, MonadWriter UnifyOutput m)
+      :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
       => ListT m UnifyStep -> m (UnificationResult' UnifyState)
     tryUnifyStepsAndContinue steps = do
       x <- foldListT tryUnifyStep failure steps
@@ -1272,19 +922,20 @@ unify s strategy = if isUnifyStateSolved s
         UnifyBlocked b -> return $ UnifyBlocked b
         UnifyStuck err -> return $ UnifyStuck err
 
-    tryUnifyStep :: (PureTCM m, MonadWriter UnifyOutput m)
+    tryUnifyStep :: (PureTCM m, MonadWriter UnifyLog' m, MonadError TCErr m)
                  => UnifyStep
                  -> m (UnificationResult' UnifyState)
                  -> m (UnificationResult' UnifyState)
     tryUnifyStep step fallback = do
       addContext (varTel s) $
         reportSDoc "tc.lhs.unify" 20 $ "trying unifyStep" <+> prettyTCM step
-      x <- unifyStep s step
+      (x, output) <- runWriterT $ unifyStep s step
       case x of
         Unifies s'   -> do
           reportSDoc "tc.lhs.unify" 20 $ "unifyStep successful."
           reportSDoc "tc.lhs.unify" 20 $ "new unifyState:" <+> prettyTCM s'
-          writeUnifyLog $ UnificationStep s step
+          -- tell output
+          writeUnifyLog $ (UnificationStep s step output,s')
           return x
         NoUnify{}       -> return x
         UnifyBlocked b1 -> do
