@@ -11,6 +11,7 @@ import qualified Data.Set as Set
 import Agda.Interaction.Options
 
 import qualified Agda.Syntax.Abstract as A
+import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
@@ -31,11 +32,15 @@ import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.CompiledClause (hasProjectionPatterns)
 import Agda.TypeChecking.CompiledClause.Compile
 
-import Agda.TypeChecking.Rules.Data ( getGeneralizedParameters, bindGeneralizedParameters, bindParameters, fitsIn, forceSort,
-                                      defineCompData, defineTranspOrHCompForFields )
+import Agda.TypeChecking.Rules.Data
+  ( getGeneralizedParameters, bindGeneralizedParameters, bindParameters
+  , checkDataSort, fitsIn, forceSort
+  , defineCompData, defineKanOperationForFields
+  )
 import Agda.TypeChecking.Rules.Term ( isType_ )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 
+import Agda.Utils.List (headWithDefault)
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
@@ -60,7 +65,7 @@ import Agda.Utils.Impossible
 --
 --     [@ps@]      Record parameters.
 --
---     [@contel@]  Approximate type of constructor (@fields@ -> Set).
+--     [@contel@]  Approximate type of constructor (@fields@ -> dummy).
 --                 Does not include record parameters.
 --
 --     [@fields@]  List of field signatures.
@@ -71,11 +76,19 @@ checkRecDef
   -> UniverseCheck             -- ^ Check universes?
   -> A.RecordDirectives        -- ^ (Co)Inductive, (No)Eta, (Co)Pattern, Constructor?
   -> A.DataDefParams           -- ^ Record parameters.
-  -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> Set).
+  -> A.Expr                    -- ^ Approximate type of constructor (@fields@ -> dummy).
                                --   Does not include record parameters.
   -> [A.Field]                 -- ^ Field signatures.
   -> TCM ()
-checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars ps) contel fields =
+checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars ps) contel0 fields = do
+
+  -- Andreas, 2022-10-06, issue #6165:
+  -- The target type of the constructor is a meaningless dummy expression which does not type-check.
+  -- We replace it by Set/Type (builtinSet) which is still incorrect but type-checks.
+  -- It will be fixed after type-checking.
+  aType <- A.Def . fromMaybe __IMPOSSIBLE__ <$> getBuiltinName' builtinSet
+  let contel = A.unPiView . (\ (A.PiView tels _) -> A.PiView tels aType) . A.piView $ contel0
+
   traceCall (CheckRecDef (getRange name) name ps fields) $ do
     reportSDoc "tc.rec" 10 $ vcat
       [ "checking record def" <+> prettyTCM name
@@ -226,6 +239,8 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
                   -- in case the record turns out to be recursive.
               -- Determined by positivity checker:
               , recMutual         = Nothing
+              -- Determined by the termination checker:
+              , recTerminates     = Nothing
               , recComp           = emptyCompKit -- filled in later
               }
 
@@ -258,6 +273,10 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
       -- Check that the fields fit inside the sort
       _ <- fitsIn uc [] contype s
 
+      -- Check that the sort admits record declarations.
+      checkDataSort name s
+
+
       {- Andreas, 2011-04-27 WRONG because field types are checked again
          and then non-stricts should not yet be irrelevant
 
@@ -298,11 +317,16 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
 
       let m = qnameToMName name  -- Name of record module.
 
+      eraseRecordParameters <- optEraseRecordParameters <$> pragmaOptions
+      let maybeErase :: forall a. LensQuantity a => a -> a
+          maybeErase | eraseRecordParameters = setQuantity zeroQuantity
+                     | otherwise             = id
+
       -- Andreas, 2016-02-09 setting all parameters hidden in the record
       -- section telescope changes the semantics, see e.g.
       -- test/Succeed/RecordInParModule.
       -- Ulf, 2016-03-02 but it's the right thing to do (#1759)
-      modifyContextInfo hideOrKeepInstance $ addRecordVar $ do
+      modifyContextInfo (hideOrKeepInstance . maybeErase) $ addRecordVar $ do
 
         -- Add the record section.
 
@@ -321,17 +345,20 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
       -- Andreas, 2016-02-09, Issue 1815 (see also issue 1759).
       -- For checking the record declarations, hide the record parameters
       -- and the parameters of the parent modules.
-      modifyContextInfo hideOrKeepInstance $ addRecordVar $ do
+      modifyContextInfo (hideOrKeepInstance . maybeErase) $ do
+        params <- getContext
 
         -- Check the types of the fields and the other record declarations.
-        withCurrentModule m $ do
+        addRecordVar $ withCurrentModule m $ do
 
           -- Andreas, 2013-09-13, 2016-01-06.
           -- Argument telescope for the projections: all parameters are hidden.
           -- This means parameters of the parent modules and of the current
           -- record type.
           -- See test/Succeed/ProjectionsTakeModuleTelAsParameters.agda.
-          tel' <- getContextTelescope
+          tel' <- do
+            r <- headWithDefault __IMPOSSIBLE__ <$> getContext
+            return $ telFromList' nameToArgName $ reverse $ r : params
           setModuleCheckpoint m
           checkRecordProjections m name hasNamedCon con tel' ftel fields
 
@@ -345,7 +372,6 @@ checkRecDef i name uc (RecordDirectives ind eta0 pat con) (A.DataDefParams gpars
       modifySignature $ updateDefinition conName $ \def ->
         def { defMatchable = Set.fromList $ map unDom fs }
 
-      return ()
   where
   -- Andreas, 2020-04-19, issue #4560
   -- If the user declared the record constructor as @pattern@,
@@ -408,8 +434,8 @@ defineCompKitR name params fsT fns rect = do
   reportSDoc "tc.rec.cxt" 30 $ prettyTCM fsT
   reportSDoc "tc.rec.cxt" 30 $ pretty rect
   if not $ all isJust required then return $ emptyCompKit else do
-    transp <- whenDefined [builtinTrans]              (defineTranspOrHCompR DoTransp name params fsT fns rect)
-    hcomp  <- whenDefined [builtinTrans,builtinHComp] (defineTranspOrHCompR DoHComp name params fsT fns rect)
+    transp <- whenDefined [builtinTrans]              (defineKanOperationR DoTransp name params fsT fns rect)
+    hcomp  <- whenDefined [builtinTrans,builtinHComp] (defineKanOperationR DoHComp name params fsT fns rect)
     return $ CompKit
       { nameOfTransp = transp
       , nameOfHComp  = hcomp
@@ -420,17 +446,17 @@ defineCompKitR name params fsT fns rect = do
       if all isJust xs then m else return Nothing
 
 
-defineTranspOrHCompR ::
-  TranspOrHComp
+defineKanOperationR
+  :: Command
   -> QName       -- ^ some name, e.g. record name
   -> Telescope   -- ^ param types Δ
   -> Telescope   -- ^ fields' types Δ ⊢ Φ
   -> [Arg QName] -- ^ fields' names
   -> Type        -- ^ record type Δ ⊢ T
   -> TCM (Maybe QName)
-defineTranspOrHCompR cmd name params fsT fns rect = do
+defineKanOperationR cmd name params fsT fns rect = do
   let project = (\ t fn -> t `applyE` [Proj ProjSystem fn])
-  stuff <- fmap fst <$> defineTranspOrHCompForFields cmd Nothing project name params fsT fns rect
+  stuff <- fmap fst <$> defineKanOperationForFields cmd Nothing project name params fsT fns rect
 
   caseMaybe stuff (return Nothing) $ \ (theName, gamma, rtype, clause_types, bodies) -> do
   -- phi = 1 clause
@@ -474,36 +500,38 @@ defineTranspOrHCompR cmd name params fsT fns rect = do
                 (g0,_:g1) = splitAt (size gamma - 1 - ix) $ flattenTel gamma
                 (ns0,_:ns1) = splitAt (size gamma - 1 - ix) $ teleNames gamma
 
-              c = Clause { clauseTel       = gamma'
-                         , clauseType      = Just $ argN t
-                         , namedClausePats = pats
-                         , clauseFullRange = noRange
+              c = Clause { clauseFullRange = noRange
                          , clauseLHSRange  = noRange
-                         , clauseCatchall  = False
+                         , clauseTel       = gamma'
+                         , namedClausePats = pats
                          , clauseBody      = Just $ rhs
+                         , clauseType      = Just $ argN t
+                         , clauseCatchall    = False
                          , clauseExact       = Just True
                          , clauseRecursive   = Just False  -- definitely non-recursive!
                          , clauseUnreachable = Just False
-                         , clauseEllipsis  = NoEllipsis
+                         , clauseEllipsis    = NoEllipsis
+                         , clauseWhereModule = Nothing
                          }
            reportSDoc "trans.rec.face" 17 $ text $ show c
            return c
   cs <- forM (zip3 fns clause_types bodies) $ \ (fname, clause_ty, body) -> do
           let
               pats = teleNamedArgs gamma ++ [defaultNamedArg $ ProjP ProjSystem $ unArg fname]
-              c = Clause { clauseTel       = gamma
-                         , clauseType      = Just $ argN (unDom clause_ty)
-                         , namedClausePats = pats
-                         , clauseFullRange = noRange
+              c = Clause { clauseFullRange = noRange
                          , clauseLHSRange  = noRange
-                         , clauseCatchall  = False
+                         , clauseTel       = gamma
+                         , namedClausePats = pats
                          , clauseBody      = Just body
+                         , clauseType      = Just $ argN (unDom clause_ty)
+                         , clauseCatchall    = False
                          , clauseExact       = Just True
                          , clauseRecursive   = Nothing
                              -- Andreas 2020-02-06 TODO
                              -- Or: Just False;  is it known to be non-recursive?
                          , clauseUnreachable = Just False
-                         , clauseEllipsis  = NoEllipsis
+                         , clauseEllipsis    = NoEllipsis
+                         , clauseWhereModule = Nothing
                          }
           reportSDoc "trans.rec" 17 $ text $ show c
           reportSDoc "trans.rec" 16 $ text "type =" <+> text (show (clauseType c))
@@ -680,7 +708,8 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
                             , clauseExact       = Just True
                             , clauseRecursive   = Just False
                             , clauseUnreachable = Just False
-                            , clauseEllipsis  = NoEllipsis
+                            , clauseEllipsis    = NoEllipsis
+                            , clauseWhereModule = Nothing
                             }
 
         let projection = Projection
@@ -717,14 +746,14 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
         escapeContext impossible (size tel) $ do
           lang <- getLanguage
           addConstant projname $
-            (defaultDefn ai projname (killRange finalt) lang
-              emptyFunction
-                { funClauses        = [clause]
-                , funCompiled       = Just cc
-                , funSplitTree      = mst
-                , funProjection     = Just projection
-                , funMutual         = Just []  -- Projections are not mutually recursive with anything
-                , funTerminates     = Just True
+            (defaultDefn ai projname (killRange finalt) lang $ FunctionDefn
+              emptyFunctionData
+                { _funClauses        = [clause]
+                , _funCompiled       = Just cc
+                , _funSplitTree      = mst
+                , _funProjection     = Right projection
+                , _funMutual         = Just []  -- Projections are not mutually recursive with anything
+                , _funTerminates     = Just True
                 })
               { defArgOccurrences = [StrictPos]
               , defCopatternLHS   = hasProjectionPatterns cc
