@@ -17,7 +17,7 @@ import qualified Data.Set as Set
 import Agda.Interaction.Options
 import Agda.Interaction.Highlighting.Generate (disambiguateRecordFields)
 
-import Agda.Syntax.Abstract (Binder)
+import Agda.Syntax.Abstract (Binder, TypedBindingInfo (tbTacticAttr))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Views as A
 import qualified Agda.Syntax.Info as A
@@ -207,7 +207,7 @@ isType_ e = traceCall (IsType_ e) $ do
         , prettyTCM x
         , text $ " for interaction point " ++ show ii
         ]
-      mv <- lookupMeta x
+      mv <- lookupLocalMeta x
       let s0 = jMetaType . mvJudgement $ mv
       -- Andreas, 2016-10-14, issue #2257
       -- The meta was created in a context of length @n@.
@@ -357,7 +357,7 @@ checkPiDomain = checkDomain PiNotLam
 checkTypedBindings :: LamOrPi -> A.TypedBinding -> (Telescope -> TCM a) -> TCM a
 checkTypedBindings lamOrPi (A.TBind r tac xps e) ret = do
     let xs = fmap (updateNamedArg $ A.unBind . A.binderName) xps
-    tac <- traverse (checkTacticAttribute lamOrPi) tac
+    tac <- traverse (checkTacticAttribute lamOrPi) (tbTacticAttr tac)
     whenJust tac $ \ t -> reportSDoc "tc.term.tactic" 30 $ "Checked tactic attribute:" <?> prettyTCM t
     -- Andreas, 2011-04-26 irrelevant function arguments may appear
     -- non-strictly in the codomain type
@@ -415,14 +415,13 @@ checkTacticAttribute PiNotLam e = do
   expectedType <- el primAgdaTerm --> el (primAgdaTCM <#> primLevelZero <@> primUnit)
   checkExpr e expectedType
 
-ifPath :: Type -> TCM a -> TCM a -> TCM a
-ifPath ty fallback work = do
-  pv <- pathView ty
-  if isPathType pv then work else fallback
-
 checkPath :: A.TypedBinding -> A.Expr -> Type -> TCM Term
-checkPath b@(A.TBind _ _ (x':|[]) typ) body ty = do
-    let x    = updateNamedArg (A.unBind . A.binderName) x'
+checkPath b@(A.TBind _r _tac (xp :| []) typ) body ty = do
+ reportSDoc "tc.term.lambda" 30 $ hsep [ "checking path lambda", prettyA xp ]
+ case (A.extractPattern $ namedArg xp) of
+  Just{}  -> setCurrentRange xp $ genericError $ "Patterns are not allowed in Path-lambdas"
+  Nothing -> do
+    let x    = updateNamedArg (A.unBind . A.binderName) xp
         info = getArgInfo x
     PathType s path level typ lhs rhs <- pathView ty
     interval <- primIntervalType
@@ -434,11 +433,12 @@ checkPath b@(A.TBind _ _ (x':|[]) typ) body ty = do
         rhs' = subst 0 iOne  v
     let t = Lam info $ Abs (namedArgName x) v
     let btyp i = El s (unArg typ `apply` [argN i])
-    locallyTC eRange (const noRange) $ blockTerm ty $ traceCall (SetRange $ getRange body) $ do
+    locallyTC eRange (const noRange) $ blockTerm ty $ setCurrentRange body $ do
       equalTerm (btyp iZero) lhs' (unArg lhs)
       equalTerm (btyp iOne) rhs' (unArg rhs)
       return t
 checkPath b body ty = __IMPOSSIBLE__
+
 ---------------------------------------------------------------------------
 -- * Lambda abstractions
 ---------------------------------------------------------------------------
@@ -490,11 +490,11 @@ checkLambda' cmp b xps typ body target = do
     info = getArgInfo $ List1.head xs
 
     trySeeingIfPath = do
-      cubical <- optCubical <$> pragmaOptions
+      cubical <- isJust . optCubical <$> pragmaOptions
       reportSLn "tc.term.lambda" 60 $ "trySeeingIfPath for " ++ show xps
-      let postpone' = if isJust cubical then postpone else \ _ _ -> dontUseTargetType
+      let postpone' = if cubical then postpone else \ _ _ -> dontUseTargetType
       ifBlocked target postpone' $ \ _ t -> do
-        ifPath t dontUseTargetType $ if isJust cubical
+        ifNotM (isPathType <$> pathView t) dontUseTargetType {-else-} $ if cubical
           then checkPath b body t
           else genericError $ unwords
                  [ "Option --cubical/--erased-cubical needed to build"
@@ -595,15 +595,7 @@ lambdaIrrelevanceCheck dom info
   | otherwise = do
       let rPi  = getRelevance dom  -- relevance of function type
       let rLam = getRelevance info -- relevance of lambda
-        -- Andreas, 2017-01-24, issue #2429
-        -- we should report an error if we try to check a relevant function
-        -- against an irrelevant function type (subtyping violation)
-      unless (moreRelevant rPi rLam) $ do
-        -- @rLam == Relevant@ is impossible here
-        -- @rLam == Irrelevant@ is impossible here (least relevant)
-        -- this error can only happen if @rLam == NonStrict@ and @rPi == Irrelevant@
-        unless (rLam == NonStrict) __IMPOSSIBLE__  -- separate tests for separate line nums
-        unless (rPi == Irrelevant) __IMPOSSIBLE__
+      unless (sameRelevance rPi rLam) $
         typeError WrongIrrelevanceInLambda
       return info
 
@@ -754,8 +746,8 @@ checkAbsurdLambda cmp i h e t = localTC (set eQuantity topQuantity) $ do
             (\ d -> (defaultDefn (setModality mod info') aux t' lang d)
                     { defPolarity       = [Nonvariant]
                     , defArgOccurrences = [Unused] })
-            $ emptyFunction
-              { funClauses        =
+            $ FunctionDefn emptyFunctionData
+              { _funClauses        =
                   [ Clause
                     { clauseLHSRange  = getRange e
                     , clauseFullRange = getRange e
@@ -767,14 +759,15 @@ checkAbsurdLambda cmp i h e t = localTC (set eQuantity topQuantity) $ do
                     , clauseExact       = Just False
                     , clauseRecursive   = Just False
                     , clauseUnreachable = Just True -- absurd clauses are unreachable
-                    , clauseEllipsis  = NoEllipsis
+                    , clauseEllipsis    = NoEllipsis
+                    , clauseWhereModule = Nothing
                     }
                   ]
-              , funCompiled       = Just $ Fail [Arg info' "()"]
-              , funSplitTree      = Just $ SplittingDone 0
-              , funMutual         = Just []
-              , funTerminates     = Just True
-              , funExtLam         = Just $ ExtLamInfo top True empty
+              , _funCompiled       = Just $ Fail [Arg info' "()"]
+              , _funSplitTree      = Just $ SplittingDone 0
+              , _funMutual         = Just []
+              , _funTerminates     = Just True
+              , _funExtLam         = Just $ ExtLamInfo top True empty
               }
           -- Andreas 2012-01-30: since aux is lifted to toplevel
           -- it needs to be applied to the current telescope (issue 557)
@@ -894,7 +887,7 @@ catchIlltypedPatternBlockedOnMeta m handle = do
     -- There might be metas in the blocker not known in the reset state, as they could have been
     -- created somewhere on the way to the type error.
     blocker <- (`onBlockingMetasM` blocker) $ \ x ->
-                lookupMeta' x >>= \ case
+                lookupMeta x >>= \ case
       -- Case: we do not know the meta, so cannot unblock.
       Nothing -> return neverUnblock
       -- Case: we know the meta here.
@@ -908,8 +901,10 @@ catchIlltypedPatternBlockedOnMeta m handle = do
       -- fact not very helpful. Yes there is no hope of solving the problem, but throwing a hard
       -- error means we rob the user of the tools needed to figure out why the meta has not been
       -- solved. Better to leave the constraint.
-      Just m | InstV{} <- mvInstantiation m -> return alwaysUnblock
-             | otherwise -> return $ unblockOnMeta x
+      Just Left{} -> return alwaysUnblock
+      Just (Right m)
+        | InstV{} <- mvInstantiation m -> return alwaysUnblock
+        | otherwise                    -> return $ unblockOnMeta x
 
     -- If it's not blocked or we can't ever unblock reraise the error.
     if blocker `elem` [neverUnblock, alwaysUnblock] then reraise else handle (err, blocker)
@@ -993,6 +988,15 @@ checkRecordExpression cmp mfs e t = do
         , "  con = " <> return (P.pretty con)
         ]
 
+      -- Record expressions corresponding to erased record
+      -- constructors can only be used in compile-time mode.
+      constructorQ <- getQuantity <$> getConstInfo (conName con)
+      currentQ     <- viewTC eQuantity
+      unless (constructorQ `moreQuantity` currentQ) $
+        typeError $ GenericError $
+        "A record expression corresponding to an erased record " ++
+        "constructor must only be used in erased settings"
+
       -- Andreas, 2018-09-06, issue #3122.
       -- Associate the concrete record field names used in the record expression
       -- to their counterpart in the record type definition.
@@ -1245,8 +1249,6 @@ checkExpr' cmp e t =
             (dontCare <$> do applyRelevanceToContext Irrelevant $ checkExpr' cmp e t)
             (internalError "DontCare may only appear in irrelevant contexts")
 
-        A.ETel _   -> __IMPOSSIBLE__
-
         A.Dot{} -> genericError "Invalid dotted expression"
 
         -- Application
@@ -1327,7 +1329,6 @@ checkExpr' cmp e t =
       A.Rec{}        -> True
       A.RecUpdate{}  -> True
       A.ScopedExpr{} -> __IMPOSSIBLE__
-      A.ETel{}       -> __IMPOSSIBLE__
       _ -> False
 
 ---------------------------------------------------------------------------
@@ -1362,10 +1363,10 @@ unquoteTactic tac hole goal = do
   case ok of
     Left (BlockedOnMeta oldState blocker) -> do
       putTC oldState
-      let stripFreshMeta x = maybe neverUnblock (const $ unblockOnMeta x) <$> lookupMeta' x
+      let stripFreshMeta x = maybe neverUnblock (const $ unblockOnMeta x) <$> lookupLocalMeta' x
       blocker' <- onBlockingMetasM stripFreshMeta blocker
       r <- case Set.toList $ allBlockingMetas blocker' of
-            x : _ -> getRange <$> lookupMeta' x
+            x : _ -> getRange <$> lookupLocalMeta' x
             []    -> return noRange
       setCurrentRange r $
         addConstraint blocker' (UnquoteTactic tac hole goal)
@@ -1470,7 +1471,7 @@ checkKnownArguments
   -> TCM (Args, Type)   -- ^ Remaining inferred arguments, remaining type.
 checkKnownArguments []           vs t = return (vs, t)
 checkKnownArguments (arg : args) vs t = do
-  (vs', t') <- traceCall (SetRange $ getRange arg) $ checkKnownArgument arg vs t
+  (vs', t') <- setCurrentRange arg $ checkKnownArgument arg vs t
   checkKnownArguments args vs' t'
 
 -- | Check an argument whose value we already know.
@@ -1647,7 +1648,7 @@ checkLetBinding b@(A.LetPatBind i p e) ret =
         ]
       ]
     fvs <- getContextSize
-    checkLeftHandSide (CheckPattern p EmptyTel t) Nothing [p0] t0 Nothing [] $ \ (LHSResult _ delta0 ps _ _t _ asb _) -> bindAsPatterns asb $ do
+    checkLeftHandSide (CheckPattern p EmptyTel t) Nothing [p0] t0 Nothing [] $ \ (LHSResult _ delta0 ps _ _t _ asb _ _) -> bindAsPatterns asb $ do
           -- After dropping the free variable patterns there should be a single pattern left.
       let p = case drop fvs ps of [p] -> namedArg p; _ -> __IMPOSSIBLE__
           -- Also strip the context variables from the telescope
