@@ -13,7 +13,6 @@ import Data.Array.IArray
 import Data.Word
 import qualified Data.Foldable as Fold
 import Data.Hashable
-import qualified Data.HashTable.IO as H
 import Data.Int (Int32)
 
 import Data.Map (Map)
@@ -21,8 +20,6 @@ import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.IntSet as IntSet
 import Data.IntSet (IntSet)
-import Data.List.NonEmpty (NonEmpty(..), nonEmpty)
-import qualified Data.List.NonEmpty as NonEmpty
 import qualified Data.Set as Set
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
@@ -41,6 +38,7 @@ import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Position as P
 import Agda.Syntax.Literal
+import Agda.Syntax.TopLevelModuleName
 import Agda.Interaction.FindFile
 
 import Agda.TypeChecking.Serialise.Base
@@ -49,6 +47,9 @@ import Agda.Utils.BiMap (BiMap)
 import qualified Agda.Utils.BiMap as BiMap
 import qualified Agda.Utils.Empty as Empty
 import Agda.Utils.FileName
+import qualified Agda.Utils.HashTable as H
+import Agda.Utils.List1 (List1)
+import qualified Agda.Utils.List1 as List1
 import Agda.Utils.List2 (List2(List2))
 import qualified Agda.Utils.List2 as List2
 import Agda.Utils.Maybe
@@ -194,36 +195,6 @@ instance EmbPrj Language where
     [1, a] -> valuN Cubical a
     _      -> malformed
 
-instance EmbPrj AbsolutePath where
-  icod_ file = do
-    d <- asks absPathD
-    -- Andreas, 2020-08-11, issue #4828
-    -- AbsolutePath is no longer canonical (can contain symlinks).
-    -- The dictonary contains canonical pathes, though.
-    file <- liftIO $ canonicalizeAbsolutePath file
-    liftIO $ flip fromMaybeM (H.lookup d file) $ do
-      -- The path @file@ should be cached in the dictionary @d@.
-      -- This seems not to be the case, thus, crash here.
-      -- But leave some hints for the posterity why things could go so wrong.
-      -- reportSLn "impossible" 10 -- does not work here
-      putStrLn $ unlines $
-        [ "Panic while serializing absolute path: " ++ show file
-        , "The path could not be found in the dictionary:"
-        ]
-      print =<< H.toList d
-      __IMPOSSIBLE__
-
-  value m = do
-    m :: TopLevelModuleName
-            <- value m
-    mf      <- gets modFile
-    incs    <- gets includes
-    (r, mf) <- liftIO $ findFile'' incs m mf
-    modify $ \s -> s { modFile = mf }
-    case r of
-      Left err -> throwError $ findErrorToTypeError m err
-      Right f  -> return (srcFilePath f)
-
 instance EmbPrj a => EmbPrj (Position' a) where
   icod_ (P.Pn file pos line col) = icodeN' P.Pn file pos line col
 
@@ -240,7 +211,7 @@ instance Typeable b => EmbPrj (WithDefault b) where
     _ -> malformed
 
 instance EmbPrj TopLevelModuleName where
-  icod_ (TopLevelModuleName a b) = icodeN' TopLevelModuleName a b
+  icod_ (TopLevelModuleName a b c) = icodeN' TopLevelModuleName a b c
 
   value = valueN TopLevelModuleName
 
@@ -253,9 +224,9 @@ instance {-# OVERLAPPABLE #-} EmbPrj a => EmbPrj [a] where
 --                            valu [x, xs] = valu2 (:) x xs
 --                            valu _       = malformed
 
-instance EmbPrj a => EmbPrj (NonEmpty a) where
-  icod_ = icod_ . NonEmpty.toList
-  value = maybe malformed return . nonEmpty <=< value
+instance EmbPrj a => EmbPrj (List1 a) where
+  icod_ = icod_ . List1.toList
+  value = maybe malformed return . List1.nonEmpty <=< value
 
 instance EmbPrj a => EmbPrj (List2 a) where
   icod_ = icod_ . List2.toList
@@ -266,9 +237,31 @@ instance (EmbPrj k, EmbPrj v, EmbPrj (BiMap.Tag v)) =>
   icod_ m = icode (BiMap.toDistinctAscendingLists m)
   value m = BiMap.fromDistinctAscendingLists <$> value m
 
+
+-- | Encode a list of key-value pairs as a flat list.
+mapPairsIcode :: (EmbPrj k, EmbPrj v) => [(k, v)] -> S Int32
+mapPairsIcode xs = icodeNode =<< convert [] xs where
+  -- As we need to call `convert' in the tail position, the resulting list is
+  -- written (and read) in reverse order, with the highest pair first in the
+  -- resulting list.
+  convert ys [] = return ys
+  convert ys ((start, entry):xs) = do
+    start <- icode start
+    entry <- icode entry
+    convert (start:entry:ys) xs
+
+mapPairsValue :: (EmbPrj k, EmbPrj v) => [Int32] -> R [(k, v)]
+mapPairsValue = convert [] where
+  convert ys [] = return ys
+  convert ys (start:entry:xs) = do
+    start <- value start
+    entry <- value entry
+    convert ((start, entry):ys) xs
+  convert _ _ = malformed
+
 instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Map a b) where
-  icod_ m = icode (Map.toAscList m)
-  value m = Map.fromDistinctAscList <$> value m
+  icod_ m = mapPairsIcode (Map.toAscList m)
+  value = vcase (fmap Map.fromDistinctAscList . mapPairsValue)
 
 instance (Ord a, EmbPrj a) => EmbPrj (Set a) where
   icod_ s = icode (Set.toAscList s)
@@ -292,6 +285,21 @@ instance EmbPrj a => EmbPrj (P.Interval' a) where
 
   value = valueN P.Interval
 
+instance EmbPrj RangeFile where
+  icod_ (RangeFile _ Nothing)  = __IMPOSSIBLE__
+  icod_ (RangeFile _ (Just a)) = icode a
+
+  value r = do
+    m :: TopLevelModuleName
+            <- value r
+    mf      <- gets modFile
+    incs    <- gets includes
+    (r, mf) <- liftIO $ findFile'' incs m mf
+    modify $ \s -> s { modFile = mf }
+    case r of
+      Left err -> throwError $ findErrorToTypeError m err
+      Right f  -> return $ RangeFile (srcFilePath f) (Just m)
+
 -- | Ranges are always deserialised as 'noRange'.
 
 instance EmbPrj Range where
@@ -386,22 +394,28 @@ instance EmbPrj Fixity' where
 
   value = valueN (\ f n -> Fixity' f n noRange)
 
-instance EmbPrj GenPart where
-  icod_ (BindHole a b)   = icodeN 0 BindHole a b
-  icod_ (NormalHole a b) = icodeN 1 NormalHole a b
-  icod_ (WildHole a)     = icodeN 2 WildHole a
-  icod_ (IdPart a)       = icodeN' IdPart a
+instance EmbPrj BoundVariablePosition where
+  icod_ (BoundVariablePosition a b) = icodeN' BoundVariablePosition a b
+
+  value = valueN BoundVariablePosition
+
+instance EmbPrj NotationPart where
+  icod_ (VarPart a b)  = icodeN 0 VarPart a b
+  icod_ (HolePart a b) = icodeN 1 HolePart a b
+  icod_ (WildPart a)   = icodeN 2 WildPart a
+  icod_ (IdPart a)     = icodeN' IdPart a
 
   value = vcase valu where
-    valu [0, a, b] = valuN BindHole a b
-    valu [1, a, b] = valuN NormalHole a b
-    valu [2, a]    = valuN WildHole a
+    valu [0, a, b] = valuN VarPart a b
+    valu [1, a, b] = valuN HolePart a b
+    valu [2, a]    = valuN WildPart a
     valu [a]       = valuN IdPart a
     valu _         = malformed
 
 instance EmbPrj MetaId where
-  icod_ (MetaId n) = icod_ n
-  value i = MetaId <$> value i
+  icod_ (MetaId a b) = icode (a, b)
+
+  value m = uncurry MetaId <$> value m
 
 instance EmbPrj A.QName where
   icod_ n@(A.QName a b) = icodeMemo qnameD qnameC (qnameId n) $ icodeN' A.QName a b
@@ -453,8 +467,8 @@ instance EmbPrj NameId where
   value = valueN NameId
 
 instance (Eq k, Hashable k, EmbPrj k, EmbPrj v) => EmbPrj (HashMap k v) where
-  icod_ m = icode (HMap.toList m)
-  value m = HMap.fromList `fmap` value m
+  icod_ m = mapPairsIcode (HMap.toList m)
+  value = vcase (fmap HMap.fromList . mapPairsValue)
 
 instance EmbPrj a => EmbPrj (WithHiding a) where
   icod_ (WithHiding a b) = icodeN' WithHiding a b
