@@ -11,6 +11,8 @@ import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 
+import Agda.TypeChecking.Monad.Base (TCM)
+import qualified Agda.TypeChecking.Pretty as P
 import Agda.TypeChecking.SizedTypes.Utils
 
 import Agda.Utils.Functor
@@ -58,6 +60,9 @@ instance Show Flex where
 instance Pretty Flex where
   pretty = text . flexId
 
+instance P.PrettyTCM Flex where
+  prettyTCM = return . pretty
+
 -- | Size expressions appearing in constraints.
 data SizeExpr' rigid flex
   = Const { offset :: Offset }                   -- ^ Constant number @n@.
@@ -170,14 +175,19 @@ instance Plus (SizeExpr' r f) Offset (SizeExpr' r f) where
       Flex x  n -> Flex x  $ n + m
       Infty     -> Infty
 
+-- | Error messages produced by the constraint simplification monad.
+
+type Error = TCM Doc
+
 -- * Constraint simplification
 
-type CTrans r f = Constraint' r f -> Either String [Constraint' r f]
+type CTrans r f = Constraint' r f -> Either Error [Constraint' r f]
 
 -- | Returns an error message if we have a contradictory constraint.
 simplify1 :: (Pretty f, Pretty r, Eq r) => CTrans r f -> CTrans r f
 simplify1 test c = do
-  let err = Left $ "size constraint " ++ prettyShow c ++ " is inconsistent"
+  let err = Left $ "size constraint" P.<+> P.pretty c P.<+>
+                   "is inconsistent"
   case c of
     -- rhs is Infty
     Constraint a           Le  Infty -> return []
