@@ -14,6 +14,8 @@ import qualified Data.Set as Set
 import Data.Map (Map)
 import qualified Data.Map as Map
 
+import Agda.TypeChecking.Pretty (PrettyTCM)
+import qualified Agda.TypeChecking.Pretty as P
 import Agda.TypeChecking.SizedTypes.Syntax
 import Agda.TypeChecking.SizedTypes.Utils
 
@@ -428,7 +430,7 @@ type Hyp' = Constraint'
 type HypGraph r f = Graph r f Label
 
 hypGraph :: (Ord rigid, Ord flex, Pretty rigid, Pretty flex) =>
-  Set rigid -> [Hyp' rigid flex] -> Either String (HypGraph rigid flex)
+  Set rigid -> [Hyp' rigid flex] -> Either Error (HypGraph rigid flex)
 hypGraph is hyps0 = do
   -- get a list of hypothesis from a list of constraints
   hyps <- concat <$> mapM (simplify1 $ \ c -> return [c]) hyps0
@@ -446,8 +448,10 @@ hypConn hg n1 n2
   | Just l <- lookupEdge hg n1 n2 = l
   | otherwise                               = top
 
-simplifyWithHypotheses :: (Ord rigid, Ord flex, Pretty rigid, Pretty flex) =>
-  HypGraph rigid flex -> [Constraint' rigid flex] -> Either String [Constraint' rigid flex]
+simplifyWithHypotheses ::
+  (Ord rigid, Ord flex, Pretty rigid, Pretty flex) =>
+  HypGraph rigid flex -> [Constraint' rigid flex] ->
+  Either Error [Constraint' rigid flex]
 simplifyWithHypotheses hg cons = concat <$> mapM (simplify1 test) cons
   where
     -- Test whether a constraint is compatible with the hypotheses:
@@ -458,7 +462,8 @@ simplifyWithHypotheses hg cons = concat <$> mapM (simplify1 test) cons
           l' = hypConn hg n1 n2
       -- l' <- lookupEdge hg n1 n2
       unless (l' <= l) $ Left $
-        "size constraint " ++ prettyShow c ++ " not consistent with size hypotheses"
+        "size constraint" P.<+> P.pretty c P.<+>
+        "not consistent with size hypotheses"
       return [c]
       -- if (l' <= l) then Just [c] else Nothing
 
@@ -467,7 +472,9 @@ simplifyWithHypotheses hg cons = concat <$> mapM (simplify1 test) cons
 
 type ConGraph r f = Graph r f Label
 
-constraintGraph :: (Ord r, Ord f, Pretty r, Pretty f) => [Constraint' r f] -> HypGraph r f -> Either String (ConGraph r f)
+constraintGraph ::
+  (Ord r, Ord f, Pretty r, Pretty f) =>
+  [Constraint' r f] -> HypGraph r f -> Either Error (ConGraph r f)
 constraintGraph cons0 hg = do
   traceM $ "original constraints cons0 = " ++ prettyShow cons0
   -- Simplify constraints, ensure they are locally consistent with
@@ -487,7 +494,9 @@ constraintGraph cons0 hg = do
 
 type ConGraphs r f = Graphs r f Label
 
-constraintGraphs :: (Ord r, Ord f, Pretty r, Pretty f) => [Constraint' r f] -> HypGraph r f -> Either String ([f], ConGraphs r f)
+constraintGraphs ::
+  (Ord r, Ord f, Pretty r, Pretty f) =>
+  [Constraint' r f] -> HypGraph r f -> Either Error ([f], ConGraphs r f)
 constraintGraphs cons0 hg = do
   traceM $ "original constraints cons0 = " ++ prettyShow cons0
   -- Simplify constraints, ensure they are locally consistent with
@@ -858,11 +867,11 @@ findRigidBelow hg e = __IMPOSSIBLE__
 
 
 solveGraph
-  :: (Ord r, Ord f, Pretty r, Pretty f, Show r, Show f)
+  :: (Ord r, Ord f, Pretty r, Pretty f, PrettyTCM f, Show r, Show f)
   => Polarities f
   -> HypGraph r f
   -> ConGraph r f
-  -> Either String (Solution r f)
+  -> Either Error (Solution r f)
 solveGraph pols hg g = do
   let (Bounds lbs ubs fs) = bounds g
       -- flexibles to solve for
@@ -881,8 +890,9 @@ solveGraph pols hg g = do
           []     -> return $ Nothing
           (a:as) -> do
             case foldM (lub hg) a as of
-              Nothing -> Left $ "inconsistent lower bound for " ++ prettyShow x
               Just l  -> return $ Just $ truncateOffset l
+              Nothing -> Left $
+                "inconsistent lower bound for" P.<+> P.prettyTCM x
       -- compute minimum of upper bounds
       ub <- do
         case ux of
@@ -891,7 +901,8 @@ solveGraph pols hg g = do
             case foldM (glb hg) a as of
               Just l | validOffset l                  -> return $ Just l
                      | Just l' <- findRigidBelow hg l -> return $ Just l'
-              _ -> Left $ "inconsistent upper bound for " ++ prettyShow x
+              _ -> Left $
+                "inconsistent upper bound for" P.<+> P.prettyTCM x
       case (lb, ub) of
         (Just l, Nothing) -> return $ Just l  -- solve x = lower bound
         (Nothing, Just u) -> return $ Just u  -- solve x = upper bound
@@ -907,11 +918,11 @@ solveGraph pols hg g = do
 -- | Solve a forest of constraint graphs relative to a hypotheses graph.
 --   Concatenate individual solutions.
 solveGraphs
-  :: (Ord r, Ord f, Pretty r, Pretty f, Show r, Show f)
+  :: (Ord r, Ord f, Pretty r, Pretty f, PrettyTCM f, Show r, Show f)
   => Polarities f
   -> HypGraph r f
   -> ConGraphs r f
-  -> Either String (Solution r f)
+  -> Either Error (Solution r f)
 solveGraphs pols hg gs =
   Solution . Map.unions <$> mapM (theSolution <.> solveGraph pols hg) gs
 
@@ -924,7 +935,7 @@ verifySolution
   => HypGraph r f
   -> [Constraint' r f]
   -> Solution r f
-  -> Either String ()
+  -> Either Error ()
 verifySolution hg cs sol = do
   cs <- return $ subst sol cs
   traceM $ "substituted constraints " ++ prettyShow cs
@@ -949,7 +960,7 @@ verifySolution hg cs sol = do
 --   which would otherwise go unnoticed.
 
 iterateSolver
-  :: (Ord r, Ord f, Pretty r, Pretty f, Show r, Show f)
+  :: (Ord r, Ord f, Pretty r, Pretty f, PrettyTCM f, Show r, Show f)
   => Polarities f
      -- ^ Meta variable polarities (prefer lower or upper solution?).
   -> HypGraph r f
@@ -958,7 +969,7 @@ iterateSolver
      -- ^ Constraints to solve.
   -> Solution r f
      -- ^ Previous substitution (already applied to constraints).
-  -> Either String (Solution r f)
+  -> Either Error (Solution r f)
      -- ^ Accumulated substition.
 
 iterateSolver pols hg cs sol0 = do
