@@ -64,8 +64,8 @@ inferUnivSort
 inferUnivSort s = do
   s <- reduce s
   case univSort' s of
-    Just s' -> return s'
-    Nothing -> do
+    Right s' -> return s'
+    Left _ -> do
       -- Jesper, 2020-04-19: With the addition of Setωᵢ and the PTS
       -- rule SizeUniv : Setω, every sort (with no metas) now has a
       -- bigger sort, so we do not need to add a constraint.
@@ -165,7 +165,7 @@ sortOf
   :: forall m. (PureTCM m, MonadBlock m)
   => Term -> m Sort
 sortOf t = do
-  reportSDoc "tc.sort" 40 $ "sortOf" <+> prettyTCM t
+  reportSDoc "tc.sort" 60 $ "sortOf" <+> prettyTCM t
   sortOfT =<< elimView EvenLone t
 
   where
@@ -175,7 +175,7 @@ sortOf t = do
         let a = unEl $ unDom adom
         sa <- sortOf a
         sb <- mapAbstraction adom (sortOf . unEl) b
-        return $ piSort (unEl <$> adom) sa sb
+        inferPiSort (adom $> El sa a) sb
       Sort s     -> return $ univSort s
       Var i es   -> do
         a <- typeOfBV i
