@@ -19,15 +19,17 @@ module Agda.TypeChecking.Substitute
 
 import Control.Arrow (first, second)
 import Control.Monad (guard)
+import Control.Monad.Except (throwError)
+
 import Data.Coerce
-import Data.Function
+import Data.Function (on)
 import qualified Data.List as List
 import Data.Map (Map)
+import qualified Data.Map.Strict as MapS
 import Data.Maybe
 import Data.HashMap.Strict (HashMap)
 
 import Debug.Trace (trace)
-import Language.Haskell.TH.Syntax (thenCmp) -- lexicographic combination of Ordering
 
 import Agda.Interaction.Options
 
@@ -46,6 +48,7 @@ import Agda.TypeChecking.Positivity.Occurrence as Occ
 import Agda.TypeChecking.Substitute.Class
 import Agda.TypeChecking.Substitute.DeBruijn
 
+import Agda.Utils.Either
 import Agda.Utils.Empty
 import Agda.Utils.Functor
 import Agda.Utils.List
@@ -79,7 +82,7 @@ applyTermE err' m es = coerce $
         case es of
           Apply a : es0      -> lazyAbsApp (coerce b :: Abs t) (coerce $ unArg a) `app` es0
           IApply _ _ a : es0 -> lazyAbsApp (coerce b :: Abs t) (coerce a)         `app` es0
-          _             -> err __IMPOSSIBLE__
+          _                  -> err __IMPOSSIBLE__
       MetaV x es' -> MetaV x (es' ++ es)
       Lit{}       -> err __IMPOSSIBLE__
       Level{}     -> err __IMPOSSIBLE__
@@ -89,8 +92,8 @@ applyTermE err' m es = coerce $
       DontCare mv -> dontCare $ mv `app` es  -- Andreas, 2011-10-02
         -- need to go under DontCare, since "with" might resurrect irrelevant term
    where
-     app :: Coercible t x => x -> Elims -> Term
-     app t es = coerce $ (coerce t :: t) `applyE` es
+     app :: Coercible t a => a -> Elims -> Term
+     app u es = coerce $ (coerce u :: t) `applyE` es
      err e = err' e (coerce m) es
 
 instance Apply Term where
@@ -99,12 +102,14 @@ instance Apply Term where
 instance Apply BraveTerm where
   applyE = applyTermE (\ _ t es ->  Dummy "applyE" (Apply (defaultArg t) : es))
 
--- | If $v$ is a record value, @canProject f v@
+-- | If @v@ is a record or constructed value, @canProject f v@
 --   returns its field @f@.
 canProject :: QName -> Term -> Maybe (Arg Term)
 canProject f v =
   case v of
-    (Con (ConHead _ IsRecord{} _ fs) _ vs) -> do
+    -- Andreas, 2022-06-10, issue #5922: also unfold data projections
+    -- (not just record projections).
+    (Con (ConHead _ _ _ fs) _ vs) -> do
       (fld, i) <- findWithIndex ((f==) . unArg) fs
       -- Jesper, 2019-10-17: dont unfold irrelevant projections
       guard $ not $ isIrrelevant fld
@@ -115,10 +120,10 @@ canProject f v =
 
 -- | Eliminate a constructed term.
 conApp :: forall t. (Coercible t Term, Apply t) => (Empty -> Term -> Elims -> Term) -> ConHead -> ConInfo -> Elims -> Elims -> Term
-conApp fk ch                  ci args []             = Con ch ci args
-conApp fk ch                  ci args (a@Apply{} : es) = conApp @t fk ch ci (args ++ [a]) es
-conApp fk ch                  ci args (a@IApply{} : es) = conApp @t fk ch ci (args ++ [a]) es
-conApp fk ch@(ConHead c _ _ fs) ci args ees@(Proj o f : es) =
+conApp fallback ch                    ci args []                  = Con ch ci args
+conApp fallback ch                    ci args    (a@Apply{} : es) = conApp @t fallback ch ci (args ++ [a]) es
+conApp fallback ch                    ci args   (a@IApply{} : es) = conApp @t fallback ch ci (args ++ [a]) es
+conApp fallback ch@(ConHead c _ _ fs) ci args ees@(Proj o f : es) =
   let failure :: forall a. a -> a
       failure err = flip trace err $ concat
         [ "conApp: constructor ", prettyShow c
@@ -127,7 +132,7 @@ conApp fk ch@(ConHead c _ _ fs) ci args ees@(Proj o f : es) =
         , " projected by ", prettyShow f
         ]
       isApply e = fromMaybe (failure __IMPOSSIBLE__) $ isApplyElim e
-      stuck err = fk err (Con ch ci args) [Proj o f]
+      stuck err = fallback err (Con ch ci args) [Proj o f]
       -- Recurse using the instance for 't', see @applyTermE@
       app :: Term -> Elims -> Term
       app v es = coerce $ applyE (coerce v :: t) es
@@ -303,7 +308,7 @@ instance Apply Defn where
     AbstractDefn d -> AbstractDefn $ apply d args
     Function{ funClauses = cs, funCompiled = cc, funCovering = cov, funInv = inv
             , funExtLam = extLam
-            , funProjection = Nothing } ->
+            , funProjection = Left _ } ->
       d { funClauses    = apply cs args
         , funCompiled   = apply cc args
         , funCovering   = apply cov args
@@ -313,19 +318,19 @@ instance Apply Defn where
 
     Function{ funClauses = cs, funCompiled = cc, funCovering = cov, funInv = inv
             , funExtLam = extLam
-            , funProjection = Just p0} ->
+            , funProjection = Right p0 } ->
       case p0 `apply` args of
         p@Projection{ projIndex = n }
           | n < 0     -> d { funProjection = __IMPOSSIBLE__ } -- TODO (#3123): we actually get here!
           -- case: applied only to parameters
-          | n > 0     -> d { funProjection = Just p }
+          | n > 0     -> d { funProjection = Right p }
           -- case: applied also to record value (n == 0)
           | otherwise ->
               d { funClauses        = apply cs args'
                 , funCompiled       = apply cc args'
                 , funCovering       = apply cov args'
                 , funInv            = apply inv args'
-                , funProjection     = if isVar0 then Just p{ projIndex = 0 } else Nothing
+                , funProjection     = if isVar0 then Right p{ projIndex = 0 } else Left MaybeProjection
                 , funExtLam         = modifySystem (\ _ -> __IMPOSSIBLE__) <$> extLam
                 }
               where
@@ -360,7 +365,7 @@ instance Apply Clause where
     -- It is assumed that we only apply a clause to "parameters", i.e.
     -- arguments introduced by lambda lifting. The problem is that these aren't
     -- necessarily the first elements of the clause telescope.
-    apply cls@(Clause rl rf tel ps b t catchall exact recursive unreachable ell) args
+    apply cls@(Clause rl rf tel ps b t catchall exact recursive unreachable ell wm) args
       | length args > length ps = __IMPOSSIBLE__
       | otherwise =
       Clause rl rf
@@ -373,6 +378,7 @@ instance Apply Clause where
              recursive
              unreachable
              ell
+             wm
       where
         -- We have
         --  Γ ⊢ args, for some outer context Γ
@@ -443,13 +449,19 @@ instance Apply Clause where
             IApplyP _ _ _ (DBPatVar _ i) -> newTel (n - 1) (subTel (size tel - 1 - i) v tel) (substP i (raise (n - 1) v) ps) vs
         newTel _ tel _ _ = __IMPOSSIBLE__
 
+        projections :: ConHead -> Term -> [Term]
         projections c v = [ relToDontCare ai $
                             -- #4528: We might have bogus terms here when printing a clause that
                             --        cannot be taken. To mitigate the problem we use a Def instead
                             --        a Proj elim for data constructors, which at least stops conApp
                             --        from crashing. See #4989 for not printing bogus terms at all.
                             case conDataRecord c of
-                              IsData     -> Def f [Apply (Arg ai v)]
+                              IsData     -> defApp f [] [Apply (Arg ai v)]
+                                              -- Andreas, 2022-06-10, issue #5922.
+                                              -- This was @Def f [Apply (Arg ai v)]@, but are we sure
+                                              -- that @v@ isn't a matching @Con@?  The testcase for
+                                              -- #5922 does not require this precaution,
+                                              -- but I sleep better this way...
                               IsRecord{} -> applyE v [Proj ProjSystem f]
                           | Arg ai f <- conFields c ]
 
@@ -655,7 +667,7 @@ instance Abstract Defn where
     AbstractDefn d -> AbstractDefn $ abstract tel d
     Function{ funClauses = cs, funCompiled = cc, funCovering = cov, funInv = inv
             , funExtLam = extLam
-            , funProjection = Nothing  } ->
+            , funProjection = Left _  } ->
       d { funClauses  = abstract tel cs
         , funCompiled = abstract tel cc
         , funCovering = abstract tel cov
@@ -664,19 +676,22 @@ instance Abstract Defn where
         }
     Function{ funClauses = cs, funCompiled = cc, funCovering = cov, funInv = inv
             , funExtLam = extLam
-            , funProjection = Just p } ->
+            , funProjection = Right p } ->
       -- Andreas, 2015-05-11 if projection was applied to Var 0
       -- then abstract over last element of tel (the others are params).
-      if projIndex p > 0 then d' else
-        d' { funClauses  = map (abstractClause tel1) cs
-           , funCompiled = abstract tel1 cc
-           , funCovering = abstract tel1 cov
-           , funInv      = abstract tel1 inv
-           , funExtLam   = modifySystem (\ _ -> __IMPOSSIBLE__) <$> extLam
-           }
+      if projIndex p > 0 then
+        d { funProjection = Right $ abstract tel p
+          , funClauses    = map (abstractClause EmptyTel) cs
+          }
+      else
+        d { funProjection = Right $ abstract tel p
+          , funClauses    = map (abstractClause tel1) cs
+          , funCompiled   = abstract tel1 cc
+          , funCovering   = abstract tel1 cov
+          , funInv        = abstract tel1 inv
+          , funExtLam     = modifySystem (\ _ -> __IMPOSSIBLE__) <$> extLam
+          }
         where
-          d' = d { funProjection = Just $ abstract tel p
-                 , funClauses    = map (abstractClause EmptyTel) cs }
           tel1 = telFromList $ drop (size tel - 1) $ telToList tel
           -- #5128: clause telescopes should be abstracted over the full telescope, regardless of
           --        projection shenanigans.
@@ -702,7 +717,7 @@ instance Abstract PrimFun where
         where n = size tel
 
 instance Abstract Clause where
-  abstract tel (Clause rl rf tel' ps b t catchall exact recursive unreachable ell) =
+  abstract tel (Clause rl rf tel' ps b t catchall exact recursive unreachable ell wm) =
     Clause rl rf (abstract tel tel')
            (namedTelVars m tel ++ ps)
            b
@@ -712,6 +727,7 @@ instance Abstract Clause where
            recursive
            unreachable
            ell
+           wm
       where m = size tel + size tel'
 
 instance Abstract CompiledClauses where
@@ -777,7 +793,27 @@ renaming err p = prependS err gamma $ raiseS $ size p
 
 -- | If @permute π : [a]Γ -> [a]Δ@, then @applySubst (renamingR π) : Term Δ -> Term Γ@
 renamingR :: DeBruijn a => Permutation -> Substitution' a
-renamingR p@(Perm n _) = permute (reverseP p) (map deBruijnVar [0..]) ++# raiseS n
+renamingR p@(Perm n is) = xs ++# raiseS n
+  where
+  xs = map (\i -> deBruijnVar (n - 1 - i)) (reverse is)
+
+  -- The list xs used to be defined in the following way:
+  --
+  --   permute (reverseP p) (map deBruijnVar [0..])
+  --
+  -- We have that
+  --
+  --     permute (reverseP p) (map deBruijnVar [0..])
+  --   = permute (Perm n $ map ((n - 1) -) $ reverse is)
+  --       (map deBruijnVar [0..])
+  --   = map (map deBruijnVar [0..] !!)
+  --       (map ((n - 1) -) $ reverse is)
+  --   = map deBruijnVar (map ((n - 1) -) $ reverse is)
+  --   = map (\i -> deBruijnVar (n - 1 - i)) (reverse is).
+  --
+  -- The latter code is linear in the length of is (if deBruijnVar
+  -- takes constant time), while the time complexity of the former
+  -- code depends on the value of the largest index in is.
 
 -- | The permutation should permute the corresponding context. (right-to-left list)
 renameP :: Subst a => Impossible -> Permutation -> a -> a
@@ -832,6 +868,7 @@ instance (Coercible a Term, Subst a) => Subst (Sort' a) where
     SSet n     -> SSet $ sub n
     SizeUniv   -> SizeUniv
     LockUniv   -> LockUniv
+    IntervalUniv -> IntervalUniv
     PiSort a s1 s2 -> coerce $ piSort (coerce $ sub a) (coerce $ sub s1) (coerce $ sub s2)
     FunSort s1 s2 -> coerce $ funSort (coerce $ sub s1) (coerce $ sub s2)
     UnivSort s -> coerce $ univSort $ coerce $ sub s
@@ -937,9 +974,11 @@ instance Subst NLPSort where
   applySubst rho = \case
     PType l   -> PType $ applySubst rho l
     PProp l   -> PProp $ applySubst rho l
+    PSSet l   -> PSSet $ applySubst rho l
     PInf f n  -> PInf f n
     PSizeUniv -> PSizeUniv
     PLockUniv -> PLockUniv
+    PIntervalUniv -> PIntervalUniv
 
 instance Subst RewriteRule where
   type SubstArg RewriteRule = NLPat
@@ -992,9 +1031,10 @@ instance Subst Constraint where
     HasPTSRule a s           -> HasPTSRule (rf a) (rf s)
     CheckLockedVars a b c d  -> CheckLockedVars (rf a) (rf b) (rf c) (rf d)
     UnquoteTactic t h g      -> UnquoteTactic (rf t) (rf h) (rf g)
+    CheckDataSort q s        -> CheckDataSort q (rf s)
     CheckMetaInst m          -> CheckMetaInst m
     CheckType t              -> CheckType (rf t)
-    UsableAtModality mod m   -> UsableAtModality mod (rf m)
+    UsableAtModality cc ms mod m -> UsableAtModality cc (rf ms) mod (rf m)
     where
       rf :: forall a. TermSubst a => a -> a
       rf x = applySubst rho x
@@ -1072,11 +1112,14 @@ instance Subst Candidate where
 
 instance Subst EqualityView where
   type SubstArg EqualityView = Term
-  applySubst rho (OtherType t) = OtherType
-    (applySubst rho t)
-  applySubst rho (IdiomType t) = IdiomType
-    (applySubst rho t)
-  applySubst rho (EqualityType s eq l t a b) = EqualityType
+  applySubst rho = \case
+    OtherType t          -> OtherType $ applySubst rho t
+    IdiomType t          -> IdiomType $ applySubst rho t
+    EqualityViewType eqt -> EqualityViewType $ applySubst rho eqt
+
+instance Subst EqualityTypeData where
+  type SubstArg EqualityTypeData = Term
+  applySubst rho (EqualityTypeData s eq l t a b) = EqualityTypeData
     (applySubst rho s)
     eq
     (map (applySubst rho) l)
@@ -1198,8 +1241,10 @@ telView'UpTo 0 t = TelV EmptyTel t
 telView'UpTo n t = case unEl t of
   Pi a b  -> absV a (absName b) $ telView'UpTo (n - 1) (absBody b)
   _       -> TelV EmptyTel t
-  where
-    absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
+
+-- | Add given binding to the front of the telescope.
+absV :: Dom a -> ArgName -> TelV a -> TelV a
+absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
 
 
 -- ** Creating telescopes from lists of types
@@ -1403,7 +1448,8 @@ instance Eq a => Eq (Pattern' a) where
   _               == _                 = False
 
 instance Ord Term where
-  Var a b    `compare` Var x y    = compare x a `thenCmp` compare b y -- sort de Bruijn indices down (#2765)
+  Var a b    `compare` Var x y    = compare (x, b) (a, y)
+                                    -- sort de Bruijn indices down (#2765)
   Var{}      `compare` _          = LT
   _          `compare` Var{}      = GT
   Def a b    `compare` Def x y    = compare (a, b) (x, y)
@@ -1492,17 +1538,23 @@ instance (Subst a, Ord a) => Ord (Elim' a) where
 --   known (i.e. it is not just @UnivSort s@).
 --
 --   Precondition: @s@ is reduced
-univSort' :: Sort -> Maybe Sort
-univSort' (Type l) = Just $ Type $ levelSuc l
-univSort' (Prop l) = Just $ Type $ levelSuc l
-univSort' (Inf f n) = Just $ Inf f $ 1 + n
-univSort' (SSet l) = Just $ SSet $ levelSuc l
-univSort' SizeUniv = Just $ Inf IsFibrant 0
-univSort' LockUniv = Just $ Inf IsFibrant 0 -- lock polymorphism is not actually supported
-univSort' s        = Nothing
+univSort' :: Sort -> Either Blocker Sort
+univSort' (Type l)     = Right $ Type $ levelSuc l
+univSort' (Prop l)     = Right $ Type $ levelSuc l
+univSort' (Inf f n)    = Right $ Inf f $ 1 + n
+univSort' (SSet l)     = Right $ SSet $ levelSuc l
+univSort' SizeUniv     = Right $ Inf IsFibrant 0
+univSort' LockUniv     = Right $ Inf IsFibrant 0 -- lock polymorphism is not actually supported
+univSort' IntervalUniv = Right $ SSet $ ClosedLevel 1
+univSort' (MetaS m _)  = Left neverUnblock
+univSort' FunSort{}    = Left neverUnblock
+univSort' PiSort{}     = Left neverUnblock
+univSort' UnivSort{}   = Left neverUnblock
+univSort' DefS{}       = Left neverUnblock
+univSort' DummyS{}     = Left neverUnblock
 
 univSort :: Sort -> Sort
-univSort s = fromMaybe (UnivSort s) $ univSort' s
+univSort s = fromRight (const $ UnivSort s) $ univSort' s
 
 sort :: Sort -> Type
 sort s = El (univSort s) $ Sort s
@@ -1510,23 +1562,33 @@ sort s = El (univSort s) $ Sort s
 ssort :: Level -> Type
 ssort l = sort (SSet l)
 
--- | Returns @Nothing@ for unknown (meta) sorts, and otherwise returns
---   @Just (b,f)@ where @b@ indicates smallness and @f@ fibrancy.
---   I.e., @b@ is @True@ for (relatively) small sorts like @Set l@ and
---   @Prop l@, and instead @b@ is @False@ for large sorts such as @Setω@.
-isSmallSort :: Sort -> Maybe (Bool,IsFibrant)
-isSmallSort Type{}     = Just (True,IsFibrant)
-isSmallSort Prop{}     = Just (True,IsFibrant)
-isSmallSort SizeUniv   = Just (True,IsFibrant)
-isSmallSort LockUniv   = Just (True,IsFibrant)
-isSmallSort (Inf f _)  = Just (False,f)
-isSmallSort SSet{}     = Just (True,IsStrict)
-isSmallSort MetaS{}    = Nothing
-isSmallSort FunSort{}  = Nothing
-isSmallSort PiSort{}   = Nothing
-isSmallSort UnivSort{} = Nothing
-isSmallSort DefS{}     = Nothing
-isSmallSort DummyS{}   = Nothing
+-- | A sort can either be small (Set l, Prop l, Size, ...)  or large
+--   (Setω n).
+data SizeOfSort
+  = SmallSort IsFibrant
+  | LargeSort IsFibrant Integer
+
+-- | Returns @Left blocker@ for unknown (blocked) sorts, and otherwise
+--   returns @Right s@ where @s@ indicates the size and fibrancy.
+sizeOfSort :: Sort -> Either Blocker SizeOfSort
+sizeOfSort Type{}       = Right $ SmallSort IsFibrant
+sizeOfSort Prop{}       = Right $ SmallSort IsFibrant
+sizeOfSort SizeUniv     = Right $ SmallSort IsFibrant
+sizeOfSort LockUniv     = Right $ SmallSort IsFibrant
+sizeOfSort IntervalUniv = Right $ SmallSort IsStrict
+sizeOfSort (Inf f n)    = Right $ LargeSort f n
+sizeOfSort SSet{}       = Right $ SmallSort IsStrict
+sizeOfSort (MetaS m _)  = Left $ unblockOnMeta m
+sizeOfSort FunSort{}    = Left neverUnblock
+sizeOfSort PiSort{}     = Left neverUnblock
+sizeOfSort UnivSort{}   = Left neverUnblock
+sizeOfSort DefS{}       = Left neverUnblock
+sizeOfSort DummyS{}     = Left neverUnblock
+
+isSmallSort :: Sort -> Bool
+isSmallSort s = case sizeOfSort s of
+  Right SmallSort{} -> True
+  _                 -> False
 
 fibrantLub :: IsFibrant -> IsFibrant -> IsFibrant
 fibrantLub IsStrict a = IsStrict
@@ -1535,41 +1597,71 @@ fibrantLub a b = a
 
 -- | Compute the sort of a function type from the sorts of its
 --   domain and codomain.
-funSort' :: Sort -> Sort -> Maybe Sort
+funSort' :: Sort -> Sort -> Either Blocker Sort
 funSort' a b = case (a, b) of
-  (Inf af m      , Inf bf n     ) -> Just $ Inf (fibrantLub af bf) $ max m n
-  (Inf af m      , b            ) | Just (True,bf) <- isSmallSort b -> Just $ Inf (fibrantLub af bf) m
-  (a             , Inf bf n     ) | Just (True,af) <- isSmallSort a -> Just $ Inf (fibrantLub af bf) n
-  (Type a        , Type b       ) -> Just $ Type $ levelLub a b
-  (LockUniv      , b            ) -> Just b
+  (Type a        , Type b       ) -> Right $ Type $ levelLub a b
+  (Prop a        , Type b       ) -> Right $ Type $ levelLub a b
+  (Type a        , Prop b       ) -> Right $ Prop $ levelLub a b
+  (Prop a        , Prop b       ) -> Right $ Prop $ levelLub a b
+  (SSet a        , SSet b       ) -> Right $ SSet $ levelLub a b
+  (Type a        , SSet b       ) -> Right $ SSet $ levelLub a b
+  (SSet a        , Type b       ) -> Right $ SSet $ levelLub a b
+  (SSet a        , Prop b       ) -> Right $ SSet $ levelLub a b
+  (Prop a        , SSet b       ) -> Right $ SSet $ levelLub a b
+  (Inf af m      , b            ) -> sizeOfSort b >>= \case
+    SmallSort bf   -> Right $ Inf (fibrantLub af bf) m
+    LargeSort bf n -> Right $ Inf (fibrantLub af bf) $ max m n
+  (a             , Inf bf n     ) -> sizeOfSort a >>= \case
+    SmallSort af   -> Right $ Inf (fibrantLub af bf) n
+    LargeSort af m -> Right $ Inf (fibrantLub af bf) $ max m n
+  (LockUniv      , b            ) -> Right b
   -- No functions into lock types
-  (a             , LockUniv     ) -> Nothing
-  (SizeUniv      , b            ) -> Just b
-  (a             , SizeUniv     ) | Just (True,_) <- isSmallSort a -> Just SizeUniv
-  (Prop a        , Type b       ) -> Just $ Type $ levelLub a b
-  (Type a        , Prop b       ) -> Just $ Prop $ levelLub a b
-  (Prop a        , Prop b       ) -> Just $ Prop $ levelLub a b
-  (SSet a        , SSet b       ) -> Just $ SSet $ levelLub a b
-  (Type a        , SSet b       ) -> Just $ SSet $ levelLub a b
-  (SSet a        , Type b       ) -> Just $ SSet $ levelLub a b
-  (a             , b            ) -> Nothing
+  (a             , LockUniv     ) -> Left neverUnblock
+  -- @IntervalUniv@ behaves like @SSet@, but functions into @Type@ land in @Type@
+  (IntervalUniv  , IntervalUniv ) -> Right $ SSet $ ClosedLevel 0
+  (IntervalUniv  , SSet b       ) -> Right $ SSet $ b
+  (IntervalUniv  , Type b       ) -> Right $ Type $ b
+  (IntervalUniv  , _            ) -> Left neverUnblock
+  (Type a        , IntervalUniv ) -> Right $ SSet $ a
+  (SSet a        , IntervalUniv ) -> Right $ SSet $ a
+  (_             , IntervalUniv ) -> Left neverUnblock
+  (SizeUniv      , b            ) -> Right b
+  (a             , SizeUniv     ) -> sizeOfSort a >>= \case
+    SmallSort{} -> Right SizeUniv
+    LargeSort{} -> Left neverUnblock
+  (MetaS m _     , _            ) -> Left $ unblockOnMeta m
+  (_             , MetaS m _    ) -> Left $ unblockOnMeta m
+  (FunSort{}     , _            ) -> Left neverUnblock
+  (_             , FunSort{}    ) -> Left neverUnblock
+  (PiSort{}      , _            ) -> Left neverUnblock
+  (_             , PiSort{}     ) -> Left neverUnblock
+  (UnivSort{}    , _            ) -> Left neverUnblock
+  (_             , UnivSort{}   ) -> Left neverUnblock
+  (DefS{}        , _            ) -> Left neverUnblock
+  (_             , DefS{}       ) -> Left neverUnblock
+  (DummyS{}      , _            ) -> Left neverUnblock
+  (_             , DummyS{}     ) -> Left neverUnblock
 
 funSort :: Sort -> Sort -> Sort
-funSort a b = fromMaybe (FunSort a b) $ funSort' a b
+funSort a b = fromRight (const $ FunSort a b) $ funSort' a b
 
 -- | Compute the sort of a pi type from the sorts of its domain
 --   and codomain.
-piSort' :: Dom Term -> Sort -> Abs Sort -> Maybe Sort
-piSort' a s1       (NoAbs _ s2) = Just $ FunSort s1 s2
+piSort' :: Dom Term -> Sort -> Abs Sort -> Either Blocker Sort
+piSort' a s1       (NoAbs _ s2) = Right $ FunSort s1 s2
 piSort' a s1 s2Abs@(Abs   _ s2) = case flexRigOccurrenceIn 0 s2 of
-  Nothing -> Just $ FunSort s1 $ noabsApp __IMPOSSIBLE__ s2Abs
-  Just o | Just (True, f1) <- isSmallSort s1, Just (True, f2) <- isSmallSort s2 -> case o of
-    StronglyRigid -> Just $ Inf (fibrantLub f1 f2) 0
-    Unguarded     -> Just $ Inf (fibrantLub f1 f2) 0
-    WeaklyRigid   -> Just $ Inf (fibrantLub f1 f2) 0
-    Flexible _    -> Nothing
-  Just o | Inf f1 n <- s1 , Just (True, f2) <- isSmallSort s2 -> Just $ Inf (fibrantLub f1 f2) n
-  Just _ -> Nothing
+  Nothing -> Right $ FunSort s1 $ noabsApp __IMPOSSIBLE__ s2Abs
+  Just o  -> case (sizeOfSort s1 , sizeOfSort s2) of
+    (Right (SmallSort f1) , Right (SmallSort f2)) -> case o of
+      StronglyRigid -> Right $ Inf (fibrantLub f1 f2) 0
+      Unguarded     -> Right $ Inf (fibrantLub f1 f2) 0
+      WeaklyRigid   -> Right $ Inf (fibrantLub f1 f2) 0
+      Flexible ms   -> Left $ metaSetToBlocker ms
+    (Right (LargeSort f1 n) , Right (SmallSort f2)) -> Right $ Inf (fibrantLub f1 f2) n
+    (_                     , Right LargeSort{}    ) -> __IMPOSSIBLE__ -- large sorts cannot depend on variables
+    (Left blocker          , Right _              ) -> Left blocker
+    (Right _               , Left blocker         ) -> Left blocker
+    (Left blocker1         , Left blocker2        ) -> Left $ unblockOnBoth blocker1 blocker2
 
 -- Andreas, 2019-06-20
 -- KEEP the following commented out code for the sake of the discussion on irrelevance.
@@ -1601,9 +1693,7 @@ piSort' a s1 s2Abs@(Abs   _ s2) = case flexRigOccurrenceIn 0 s2 of
 --     Flexible _    -> Nothing
 
 piSort :: Dom Term -> Sort -> Abs Sort -> Sort
-piSort a s1 s2 = case piSort' a s1 s2 of
-  Just s  -> s
-  Nothing -> PiSort a s1 s2
+piSort a s1 s2 = fromRight (const $ PiSort a s1 s2) $ piSort' a s1 s2
 
 ---------------------------------------------------------------------------
 -- * Level stuff
@@ -1611,15 +1701,13 @@ piSort a s1 s2 = case piSort' a s1 s2 of
 
 -- ^ Computes @n0 ⊔ a₁ ⊔ a₂ ⊔ ... ⊔ aₙ@ and return its canonical form.
 levelMax :: Integer -> [PlusLevel] -> Level
-levelMax n0 as0 = Max n as
+levelMax !n0 as0 = Max n as
   where
     -- step 1: flatten nested @Level@ expressions in @PlusLevel@s
     Max n1 as1 = expandLevel $ Max n0 as0
-    -- step 2: remove subsumed @PlusLevel@s
-    as2       = removeSubsumed as1
-    -- step 3: sort remaining @PlusLevel@s
-    as        = List.sort as2
-    -- step 4: set constant to 0 if it is subsumed by one of the @PlusLevel@s
+    -- step 2: remove subsumed @PlusLevel@s and sort what remains
+    as        = removeSubsumed as1
+    -- step 3: set constant to 0 if it is subsumed by one of the @PlusLevel@s
     greatestB = Prelude.maximum $ 0 : [ n | Plus n _ <- as ]
     n | n1 > greatestB = n1
       | otherwise      = 0
@@ -1637,12 +1725,11 @@ levelMax n0 as0 = Max n as
     expandTm (Level l)       = expandLevel l
     expandTm l               = atomicLevel l
 
-    removeSubsumed [] = []
-    removeSubsumed (Plus n a : bs)
-      | not $ null ns = removeSubsumed bs
-      | otherwise     = Plus n a : removeSubsumed [ b | b@(Plus _ a') <- bs, a /= a' ]
-      where
-        ns = [ m | Plus m a' <- bs, a == a', m > n ]
+    removeSubsumed =
+      map (\(a, n) -> Plus n a) .
+      MapS.toAscList .
+      MapS.fromListWith max .
+      map (\(Plus n a) -> (a, n))
 
 -- | Given two levels @a@ and @b@, compute @a ⊔ b@ and return its
 --   canonical form.
