@@ -20,12 +20,17 @@ module Agda.Utils.Either
   , allRight
   , groupByEither
   , maybeToEither
+  , swapEither
   ) where
 
 import Data.Bifunctor
 import Data.Either (isLeft, isRight)
+import Data.List   (unfoldr)
 
-import Agda.Utils.List ( listCase )
+import Agda.Utils.List ( spanJust )
+import Agda.Utils.List1 ( List1, pattern (:|), (<|) )
+import qualified Agda.Utils.List1 as List1
+import Agda.Utils.Singleton
 
 -- | Loop while we have an exception.
 
@@ -58,11 +63,11 @@ traverseEither f g = either (fmap Left . f) (fmap Right . g)
 
 -- | Analogue of 'Data.Maybe.fromMaybe'.
 fromLeft :: (b -> a) -> Either a b -> a
-fromLeft = either id
+fromLeft = (id `either`)
 
 -- | Analogue of 'Data.Maybe.fromMaybe'.
 fromRight :: (a -> b) -> Either a b -> b
-fromRight f = either f id
+fromRight = (`either` id)
 
 -- | Analogue of 'Agda.Utils.Maybe.fromMaybeM'.
 fromLeftM :: Monad m => (b -> m a) -> m (Either a b) -> m a
@@ -108,20 +113,18 @@ allRight :: [Either a b] -> Maybe [b]
 allRight = mapM maybeRight
 
 -- | Groups a list into alternating chunks of 'Left' and 'Right' values
-groupByEither :: forall a b. [Either a b] -> [Either [a] [b]]
-groupByEither = listCase [] (go . init) where
+groupByEither :: forall a b. [Either a b] -> [Either (List1 a) (List1 b)]
+groupByEither = unfoldr c
+  where
+  c :: [Either a b] -> Maybe (Either (List1 a) (List1 b), [Either a b])
+  c []             = Nothing
+  c (Left  a : xs) = Just $ first (Left  . (a :|)) $ spanJust maybeLeft  xs
+  c (Right b : xs) = Just $ first (Right . (b :|)) $ spanJust maybeRight xs
 
-  go :: Either [a] [b] -> [Either a b] -> [Either [a] [b]]
-  go acc         []              = adjust acc : []
-  -- match: next value can be tacked onto the accumulator
-  go (Left  acc) (Left  a : abs) = go (Left  $ a : acc) abs
-  go (Right acc) (Right b : abs) = go (Right $ b : acc) abs
-  -- mismatch: switch the accumulator to the other mode
-  go acc         (ab      : abs) = adjust acc : go (init ab) abs
+-- | Convert 'Maybe' to @'Either' e@, given an error @e@ for the 'Nothing' case.
+maybeToEither :: e -> Maybe a -> Either e a
+maybeToEither e = maybe (Left e) Right
 
-  adjust = bimap reverse reverse
-  init   = bimap pure pure
-
--- | Convert 'Maybe' to @'Either' ()@.
-maybeToEither :: Maybe a -> Either () a
-maybeToEither = maybe (Left ()) Right
+-- | Swap tags 'Left' and 'Right'.
+swapEither :: Either a b -> Either b a
+swapEither = either Right Left
