@@ -4,15 +4,11 @@ module Agda.Utils.Empty where
 import Control.DeepSeq
 import Control.Exception (evaluate)
 
-import Data.Data (Data)
-
 import Agda.Utils.Impossible
 
 
 data Empty
 
-deriving instance Data Empty
-
 -- | Values of type 'Empty' are not forced, because 'Empty' is used as
 -- a constructor argument in 'Agda.Syntax.Internal.Substitution''.
 
