@@ -3,6 +3,7 @@
 module Agda.Utils.IO where
 
 import Control.Exception
+import Control.Monad.State
 import Control.Monad.Writer
 
 -- | Catch 'IOException's.
@@ -19,3 +20,8 @@ instance CatchIO IO where
 --
 instance CatchIO m => CatchIO (WriterT w m) where
   catchIO m h = WriterT $ runWriterT m `catchIO` \ e -> runWriterT (h e)
+
+-- | Upon exception, the state is reset.
+--
+instance CatchIO m => CatchIO (StateT s m) where
+  catchIO m h = StateT $ \s -> runStateT m s `catchIO` \ e -> runStateT (h e) s
