@@ -1,25 +1,31 @@
 -- | Text IO using the UTF8 character encoding.
 
 module Agda.Utils.IO.UTF8
-  ( readTextFile
+  ( ReadException
+  , readTextFile
   , Agda.Utils.IO.UTF8.readFile
   , Agda.Utils.IO.UTF8.writeFile
   , writeTextToFile
   ) where
 
+import Control.Exception
 import Data.Text.Lazy (Text)
 import qualified Data.Text.Lazy as T
+import qualified Data.Text.Lazy.Encoding as T
 import qualified Data.Text.Lazy.IO as T
+import qualified Data.ByteString.Lazy as B
 import qualified System.IO as IO
+import qualified System.IO.Error as E
 
 -- | Converts many character sequences which may be interpreted as
 -- line or paragraph separators into '\n'.
---
--- Note that '\r\n' is assumed to have already been converted to '\n'.
 
 convertLineEndings :: Text -> Text
-convertLineEndings = T.map convert
+convertLineEndings = T.map convert . convertCRLF
   where
+  -- Replaces CR LF with LF.
+  convertCRLF = T.replace "\x000D\n" "\n"
+
   -- ASCII:
   convert '\x000D' = '\n'  -- CR  (Carriage return)
   convert '\x000C' = '\n'  -- FF  (Form feed)
@@ -30,21 +36,36 @@ convertLineEndings = T.map convert
   -- Not a line ending (or '\x000A'):
   convert c        = c
 
+-- | A kind of exception that can be thrown by 'readTextFile' and
+-- 'readFile'.
+newtype ReadException
+  = DecodingError FilePath
+    -- ^ Decoding failed for the given file.
+  deriving Show
+
+instance Exception ReadException where
+  displayException (DecodingError file) =
+    "Failed to read " ++ file ++ ".\n" ++
+    "Please ensure that this file uses the UTF-8 character encoding."
+
 -- | Reads a UTF8-encoded text file and converts many character
 -- sequences which may be interpreted as line or paragraph separators
 -- into '\n'.
+--
+-- If the file cannot be decoded, then a 'ReadException' is raised.
 
 readTextFile :: FilePath -> IO Text
-readTextFile file = convertLineEndings <$> do
-  h <- IO.openFile file IO.ReadMode
-  IO.hSetNewlineMode h $
-    IO.NewlineMode { IO.inputNL = IO.CRLF, IO.outputNL = IO.LF }
-  IO.hSetEncoding h IO.utf8
-  T.hGetContents h
+readTextFile file = do
+  s <- T.decodeUtf8' <$> B.readFile file
+  case s of
+    Right s -> return $ convertLineEndings s
+    Left _  -> throw $ DecodingError file
 
 -- | Reads a UTF8-encoded text file and converts many character
 -- sequences which may be interpreted as line or paragraph separators
 -- into '\n'.
+--
+-- If the file cannot be decoded, then a 'ReadException' is raised.
 
 readFile :: FilePath -> IO String
 readFile f = do
