@@ -7,26 +7,16 @@ module Agda.Utils.Pointer
 import Control.DeepSeq
 import Control.Concurrent.MVar
 
-import Data.Function
+import Data.Function (on)
 import Data.Hashable
 import Data.IORef
 
 import System.IO.Unsafe
 
-import Data.Data (Data (..))
-import Data.Typeable (Typeable)
-
 import Agda.Utils.Impossible
 
 data Ptr a = Ptr { ptrTag :: !Integer
                  , ptrRef :: !(IORef a) }
-  deriving Data
-
--- cheating because you shouldn't be digging this far anyway
-instance Typeable a => Data (IORef a) where
-  gunfold _ _ _ = __IMPOSSIBLE__
-  toConstr      = __IMPOSSIBLE__
-  dataTypeOf    = __IMPOSSIBLE__
 
 {-# NOINLINE freshVar #-}
 freshVar :: MVar Integer
