@@ -1,6 +1,7 @@
 module Agda.Version
   ( version
   , package
+  , docsUrl
   ) where
 
 import GHC.Generics ( Generic, Rep, packageName )
@@ -22,4 +23,9 @@ version = intercalate "." $ map show $
 package :: String
 package = packageName (undefined :: Rep AnArbitrarySymbolInThisPackage p)
 
+-- | Returns a URL corresponding to the given section in the documentation for
+-- the current version.
+docsUrl :: String -> String
+docsUrl section = "https://agda.readthedocs.io/en/v" ++ version ++ "/" ++ section
+
 data AnArbitrarySymbolInThisPackage deriving Generic
