@@ -1,5 +1,10 @@
+{-# LANGUAGE CPP             #-}
 {-# LANGUAGE TemplateHaskell #-}
 
+#if __GLASGOW_HASKELL__ >= 900
+{-# OPTIONS_GHC -Wno-overlapping-patterns #-}
+#endif
+
 module Agda.VersionCommit where
 
 import Development.GitRev
@@ -7,9 +12,7 @@ import Development.GitRev
 import Agda.Version
 
 versionWithCommitInfo :: String
-versionWithCommitInfo = version ++ case commitInfo of
-                                     Nothing   -> ""
-                                     Just info -> "-" ++ info
+versionWithCommitInfo = version ++ maybe "" ("-" ++) commitInfo
 
 -- | Information about current git commit, generated at compile time
 commitInfo :: Maybe String
@@ -19,9 +22,9 @@ commitInfo
   where
     hash = $(gitHash)
 
-    -- | Check if any tracked files have uncommitted changes
+    -- Check if any tracked files have uncommitted changes
     dirty | $(gitDirtyTracked) = "-dirty"
           | otherwise          = ""
 
-    -- | Abbreviate a commit hash while keeping it unambiguous
+    -- Abbreviate a commit hash while keeping it unambiguous
     abbrev = take 7
